//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_ANP_AnalysisBase.h,v $
//
// Purpose        : Base class for Analysis types
//
// Special Notes  : Specify any "hidden" or subtle details of the class here.
//                  Portability details, error handling information, etc.
//
// Creator        : Richard Schiek, SNL, Electrical and Microsystem Modeling
//
// Creation Date  : 01/24/08
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.63.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:06 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#ifndef Xyce_N_ANP_AnalysisBase_h
#define Xyce_N_ANP_AnalysisBase_h

#include <N_ANP_fwd.h>
#include <N_IO_fwd.h>
#include <N_LOA_fwd.h>
#include <N_NLS_fwd.h>
#include <N_TIA_fwd.h>
#include <N_UTL_fwd.h>

#include <N_UTL_JSON.h>
#include <N_UTL_Stats.h>

namespace Xyce {
namespace Analysis {

struct StatCounts
{
  StatCounts();
  StatCounts &operator+=(const StatCounts &stats);

  unsigned int        successfulStepsTaken_;        ///< Number of consecutive successful time-integration steps.
  unsigned int        successStepsThisParameter_;
  unsigned int        failedStepsAttempted_;        ///< Total number of failed time-integration steps.
  unsigned int        jacobiansEvaluated_;
  unsigned int        iterationMatrixFactorizations_;
  unsigned int        linearSolves_;
  unsigned int        failedLinearSolves_;
  unsigned int        linearIters_;
  unsigned int        residualEvaluations_;
  unsigned int        nonlinearConvergenceFailures_;
  double              linearSolutionTime_;
  double              residualLoadTime_;
  double              jacobianLoadTime_;
};

Util::JSON &operator<<(Util::JSON &json, const StatCounts &s);

//-------------------------------------------------------------------------
// Class         : AnalysisBase
// Purpose       : Base class for common analysis functions
// Special Notes :
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 01/24/08
//-------------------------------------------------------------------------
class AnalysisBase
{
public:
  AnalysisBase(AnalysisManager &analysis_manager, const char *name);
  virtual ~AnalysisBase();

private:
  AnalysisBase(const AnalysisBase &);                   ///< No copying
  AnalysisBase &operator=(const AnalysisBase &);        ///< No assignment

public:
  void setTranStepNumber(int step) {
    tranStepNumber = step;
  }

  int getTranStepNumber()
  {
    return tranStepNumber;
  }

  int getStepNumber() {
    return stepNumber;
  }

  void setStepNumber(int step) {
    stepNumber=step;
  }

  virtual const TimeIntg::TIAParams &getTIAParams() const = 0;
  virtual TimeIntg::TIAParams &getTIAParams() = 0;

  virtual bool outputFailureStats(std::ostream &os) {return true;}

  bool run();
  bool init();
  bool loopProcess();
  bool processSuccessfulStep();
  bool processFailedStep();
  bool finish();
  bool handlePredictor();

  virtual bool doRun() = 0;
  virtual bool doInit() = 0;
  virtual bool doLoopProcess() = 0;
  virtual bool doProcessSuccessfulStep() = 0;
  virtual bool doProcessFailedStep() = 0;
  virtual bool doFinish() = 0;
  virtual bool doHandlePredictor() = 0;

  //-----------------------------------------------------------------------------
  // Function      : AnalysisBase::printStepHeader()
  // Purpose       : Prints out step information.
  // Special Notes :
  // Scope         : public
  // Creator       : Eric Keiter, SNL, Parallel Computational Sciences
  // Creation Date : 6/26/00
  //-----------------------------------------------------------------------------
  virtual void printStepHeader(std::ostream &os)
  {}

  //-----------------------------------------------------------------------------
  // Function      : AnalysisBase::printProgress()
  // Purpose       : Outputs run completion percentage and estimated
  //                 time-to-completion.
  // Special Notes :
  // Scope         : public
  // Creator       : Scott A. Hutchinson, SNL, Computational Sciences
  // Creation Date : 06/07/2002
  //-----------------------------------------------------------------------------
  virtual void printProgress(std::ostream &os)
  {}

  // Two Level specific
  virtual bool twoLevelStep() { return true; }

  // Utility function for AnalysisManager to determine if a complex analysis type (like HB)
  // is performing another type of analysis under the hood.  This is necessary for populating
  // the TimeIntg::TimeIntInfo struct.  For straightforward analysis types, this method is not
  // needed because the AnalysisManager already knows which analysis is being performed.
  virtual bool isAnalysis(int analysis_type) const { return false; }

  virtual bool getDCOPFlag() const = 0;

  virtual int getDoubleDCOPStep() const
  {
    return doubleDCOPStep_ ;
  }

  virtual bool printLoopInfo(int start, int finish);

  void setBeginningIntegrationFlag(bool bif)
  {
    beginningIntegration = bif;
  }

  bool getBeginningIntegrationFlag() const
  {
    return beginningIntegration;
  }

  void setIntegrationMethod(int im)
  {
    baseIntegrationMethod_= im;
  }

  int getIntegrationMethod() const
  {
    return baseIntegrationMethod_;
  }

  void setInputOPFlag(bool initial_conditions_loaded) {
    inputOPFlag_ = initial_conditions_loaded;
  }
  
  bool getInputOPFlag() const
  {
    return inputOPFlag_;
  }

  bool resetForStepAnalysis();
  void resetAll();
  int saveLoopInfo ();

  // step statistic functions
  double getTotalLinearSolutionTime() const;
  double getTotalResidualLoadTime() const;
  double getTotalJacobianLoadTime() const;

  // functions related to PDE problems requiring two DCOP solves
  bool getDoubleDCOPEnabled() const
  {
    return doubleDCOPFlag_;
  }

  bool firstDoubleDCOPStep();

  const StatCounts &getStatCounts(int index = -1) const
  {
    if (index == -1)
      return saveStatCountsVector_.back();
    else
      return saveStatCountsVector_[index];
  }

  bool setDCOPOption(const Util::Param &param);

  int getLastDCOPStep() const {
    return lastDCOPStep_;
  }

  int getFirstDCOPStep() const
  {
    return firstDCOPStep_;
  }

  bool getNOOP() const
  {
    return NOOP_;
  }

  void setNOOP(bool noop)
  {
    NOOP_ = noop;
  }

  const char *getName() const
  {
    return name_;
  }
  
  
  
private:
  const char *          name_;

  bool                  inputOPFlag_;                   ///< true if starting from an initial condition.

protected:
  bool                  beginningIntegration;
  unsigned int          baseIntegrationMethod_;         ///< Current time-integration method flag.

    // NOTE: For now tranStepNumber is the same as stepNumber, but later I will
    //       change it.  stepNumber will later include both dcop and tran steps.
    //       I haven't changed it yet b/c I need to check what devices call
    //       getStepNumber, and what they expect to get.
  unsigned int          stepNumber;                     ///< Time-integration step number counter.

protected:
  unsigned int          tranStepNumber;

  bool                  NOOP_;
  bool                  doubleDCOPFlag_;                ///< true if doing a double-DCOP is possible.
  int                   doubleDCOPStep_;                ///< current step in the DCOP loop.
  int                   firstDCOPStep_;
  int                   lastDCOPStep_;

  std::vector<StatCounts>     saveStatCountsVector_;

public:
  StatCounts                    stats_;
};

StatCounts operator-(const StatCounts &s0, const StatCounts &s1);

//-----------------------------------------------------------------------------
inline double AnalysisBase::getTotalLinearSolutionTime() const
{
  return stats_.linearSolutionTime_;
}

//-----------------------------------------------------------------------------
inline double AnalysisBase::getTotalResidualLoadTime() const
{
  return stats_.residualLoadTime_;
}

//-----------------------------------------------------------------------------
inline double AnalysisBase::getTotalJacobianLoadTime() const
{
  return stats_.jacobianLoadTime_;
}

void gatherStepStatistics(StatCounts &stats, Nonlinear::NonLinearSolver &nonlinear_solver, int newton_convergence_status);

} // namespace Analysis
} // namespace Xyce


#endif // Xyce_N_ANP_AnalysisBase_h
