//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_ANP_AnalysisManager.h,v $
//
// Purpose       :
//
// Special Notes :
//
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
//
// Creation Date : 01/24/08
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.150.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:06 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#ifndef Xyce_N_ANP_AnalysisManager_h
#define Xyce_N_ANP_AnalysisManager_h

#include <list>

#include <N_ANP_fwd.h>
#include <N_DEV_fwd.h>
#include <N_IO_fwd.h>
#include <N_LAS_fwd.h>
#include <N_LOA_fwd.h>
#include <N_NLS_fwd.h>
#include <N_PDS_fwd.h>
#include <N_TIA_fwd.h>
#include <N_TOP_fwd.h>
#include <N_UTL_fwd.h>

#include <N_ANP_AnalysisBase.h>
#include <N_ANP_AnalysisEvent.h>
#include <N_ANP_Factory.h>
#include <N_ANP_StepEvent.h>
#include <N_TIA_TIAParams.h>
#include <N_UTL_Listener.h>
#include <N_UTL_Stats.h>
#include <N_UTL_Timer.h>

namespace Xyce {
namespace Analysis {

typedef std::vector<Factory<void> *> CreatorVector;
typedef Util::Notifier<StepEvent> StepEventNotifier;
typedef Util::Notifier<AnalysisEvent> AnalysisEventNotifier;
typedef Util::ListenerAutoSubscribe<StepEvent> StepEventListener;
typedef Util::ListenerAutoSubscribe<AnalysisEvent> AnalysisEventListener;

//-----------------------------------------------------------------------------
// Class         :
// Purpose       : This function converts between Nonlinear::AnalysisMode
//               : and AnalysisManager.h ANP_Analysis_Mode
// Special Notes :
// Creator       : Todd Coffey, 1414
// Creation Date : 07/23/08
//-----------------------------------------------------------------------------
Nonlinear::AnalysisMode nonlinearAnalysisMode(Mode mode);

//-----------------------------------------------------------------------------
// Function      : analysisModeName
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Mon Mar 16 13:56:09 2015
//-----------------------------------------------------------------------------
///
/// Returns the name of the analysis mode given by mode.
///
/// @param mode         Analysis mode
///
/// @return name of the mode
///
///
const char *analysisModeName(Mode mode);

//-----------------------------------------------------------------------------
// Class         : AnalysisManager
//
// Purpose       : This class manages, allocates, and sets up the
//                 various analysis types, such as DC, Tran, HB, etc.
//
// Special Notes : Some of this class was once in the TimeIntg::ControlAlgorithm
//                 class, which was set up back when Xyce only did transient
//                 simulations.  As we added analysis types it became necessary
//                 to refactor the code so that each analysis type had its own
//                 set of classes.
//
// Creator       : Eric Keiter, SNL
// Creation Date : 6/01/00 (TimeIntg::ControlAlgorithm, now deprecated)
// Creation Date : 1/24/08 (for this version of the class. Date is approximate)
//-----------------------------------------------------------------------------
class AnalysisManager : public StepEventNotifier,
                        public AnalysisEventNotifier,
                        public StepEventListener,
                        public AnalysisEventListener
{
public:
  AnalysisManager(
    const IO::CmdParse &        command_line,
    IO::RestartMgr &            restart_manager,
    OutputMgrAdapter &          output_manager_adapter,
    Stats::Stat                 analysis_stat);
  virtual ~AnalysisManager();

private:
  AnalysisManager(const AnalysisManager &);
  AnalysisManager &operator=(const AnalysisManager &);

public:
  void notify(const StepEvent &step_event);
  void notify(const AnalysisEvent &analysis_event);

  Loader::Loader &createLoader(Device::DeviceMgr &device_manager);

  void allocateAnalysisObject();

  bool initializeSolverSystem(
    const TimeIntg::TIAParams & tia_params,
    Loader::Loader &            loader,
    Linear::System &            linear_system,
    Nonlinear::Manager &        nonlinear_manager,
    Device::DeviceMgr &         device_manager);

  void resetSolverSystem();

  // Execute the control loop for the set analysis type.
  bool run();

  // Gets the next time-step value.
  double getTime() const;

  // Gets the final time-step value.
  double getFinalTime() const;

  // Gets the initial time-step value.
  double getInitialTime() const;

  // Updates the divided difference values.
  bool updateDivDiffs();

    // Calls the time int. method to update the corrector derivatives.
    bool updateDerivs();

    // updates state vectors following initial solve with previous operating point constraints
    bool completeOPStartStep();

  // updates the State vectors. This function is called from the LOCA interface.
  bool completeHomotopyStep(
    Loader::NonlinearEquationLoader &   loader,
    const std::vector<std::string> &    paramNames,
    const std::vector<double> &         paramVals,
    Linear::Vector *                    solnVecPtr );

  bool failHomotopyStep(Loader::NonlinearEquationLoader &loader);

    // Compute an estimate of the error in the integration step.
    bool updateDerivsBlock(const std::list< IndexPair > & solGIDList,
                           const std::list< IndexPair > & staGIDList);

    // Prints out time loop information.
    bool printLoopInfo(int start, int finish);

  // Get the steady-state flag (true if time int mode is none)
  bool getDCOPFlag() const;

  // Get the dcop flag for transient.(true if doing DCOP for transient initialization)
  bool getTranOPFlag() const;

  // Get the dcop flag for AC.(true if doing DCOP for AC initialization)
  bool getACOPFlag() const;

  bool getDCSweepFlag() const;

  bool getDotOpSpecified()
  {
    return dotOpSpecified_;
  }

  bool getSweepSourceResetFlag() const
  {
    return sweepSourceResetFlag_;
  }

  void setSweepSourceResetFlag (bool ssrf) {
    sweepSourceResetFlag_ = ssrf;
  }

  bool getTransientFlag () const;

  // Is the doubleDCOP algorithm enabled?
  bool getDoubleDCOPEnabled() const;

  void setTwoLevelMode(TwoLevelMode current_mode) {
    twoLevelMode_ = current_mode;
  }

  TwoLevelMode getTwoLevelMode() const {
    return twoLevelMode_;
  }

  // Get block analysis information for HB
  bool getBlockAnalysisFlag() const;

  // gets the index of the DCOP step.
  // 0 = nonlinear poisson, 1=full TCAD
  int getDoubleDCOPStep() const;

  // Gets/sets the step number.
  int getStepNumber() const;
  int getTranStepNumber();

  void setStepNumber(int step);
  void setTranStepNumber(int step);

    // This is true only at the beginning of integration, not at a breakpoint.
  bool getInitTranFlag() const;

  const IO::CmdParse &getCommandLine() const {
    return commandLine_;
  }

  const std::string &getNetlistFilename() const
  {
    return netlistFilename_;
  }

  // returns whether transient analysis is completed
    bool isSimulationComplete();

    // Gets the size of the restart data (bytes?).
    int getRestartDataSize( bool pack ) const;

    // Sets the DC sweep calculation parameters.
    bool setDCAnalysisParams(const Util::OptionBlock & paramsBlock);

    // Method to handle OP statements.
    bool setOPAnalysisParams(const Util::OptionBlock & paramsBlock);

    // Sets the SAVE parameters.
    bool setSaveOptions(const Util::OptionBlock & OB);

    // Sets the DCOP restart parameters.
    bool setDCOPRestartParams(const Util::OptionBlock & OB);

    // Sets the AC Analysis options.
    bool setACAnalysisParams(const Util::OptionBlock & OB);

    // Sets the NOISE Analysis options.
    bool setNOISEAnalysisParams(const Util::OptionBlock & OB);

    // sets a time at which to pause the simulation
    void setPauseTime(double pauseTime, double initial_time);

    // returns time at which to pause the simulation
    double getPauseTime() const;

    // returns true if the simulation is currently paused
    bool isPaused() const;

  //-----------------------------------------------------------------------------
  // Function      : AnalysisManager::resumeSimulation
  //
  // Purpose       : set flag to signify that simulation is continuation of
  //                 previously paused simulation.
  //
  // Special Notes : This is a temporary implementation that I'm using to
  //                 begin the process of hiding time integrator internals from
  //                 the "N_CIR_Xyce::simulateUntil" method, so that I can
  //                 ultimately change those internals without the simulateUntil
  //                 method
  //                 In the zero order version, just sets tiaParams_.resume=true
  // Scope         : public
  // Creator       : Tom Russo, SNL, Component Information and Models
  // Creation Date : 04/29/2004
  //-----------------------------------------------------------------------------
  void setResumeSimulation(bool resume)
  {
    resumeSimulation_ = resume;
  }

    // Sets the sensitivity options.
    bool setSensOptions(const Util::OptionBlock & OB);

    // sets the nonlinear loader pointer.
  bool setLoader(Loader::Loader &loader_);

    // Registers the parallel services manager pointer.
  bool registerParallelServices(Parallel::Manager * pds_tmp);

    // Registers the elapsed time timer
    bool registerElapsedTimer(Util::Timer *);

    // Writes-out the restart data.
    bool dumpRestartData(char * buf, int bsize, int & pos, Parallel::Communicator * comm, bool pack);

    // Restores the restart data.
    bool restoreRestartData(char * buf, int bsize, int & pos, Parallel::Communicator * comm, bool pack );

    // Gets the solution variable data.
    bool getSolnVarData(const int & gid, std::vector< double > & varData) const;

    // Gets the state variable data.
    bool getStateVarData(const int & gid, std::vector< double > & varData) const;

    // Gets the store variable data.
    bool getStoreVarData(const int & gid, std::vector< double > & varData) const;

    // Sets the solution variable data.
    bool setSolnVarData(const int & gid, const std::vector< double > & varData);

    // Sets the state variable data.
    bool setStateVarData(const int & gid, const std::vector< double > & varData);

    // Sets the store variable data.
    bool setStoreVarData(const int & gid, const std::vector< double > & varData);

    // set/get beginning integration flag
    void setBeginningIntegrationFlag(bool bif);
    bool getBeginningIntegrationFlag() const;

    // set/get integration method
    void setIntegrationMethod(int im);
    int getIntegrationMethod();

    // Gets the total time spent in the linear solvers.
    double getTotalLinearSolutionTime() const;

    // Gets the total time spent in the residual load calculations.
    double getTotalResidualLoadTime() const;

    // Gets the total time spent in the Jacobian load calculations.
    double getTotalJacobianLoadTime() const;

    // set the next solution vector pointer.  Needed for NOX...
    bool setNextSolVectorPtr (Linear::Vector * solVecPtr);

  void initializeTransientModel(const TimeIntg::TIAParams &tia_params);

  void createTimeIntegratorMethod(const TimeIntg::TIAParams &tia_params, const unsigned int integration_method);

  //-----------------------------------------------------------------------------
  // Function      : AnalysisManager::getTIAParams
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : Eric R. Keiter, SNL, Computational Sciences
  // Creation Date : 11/05/04
  //-----------------------------------------------------------------------------
  const TimeIntg::TIAParams &getTIAParams() const;
  TimeIntg::TIAParams &getTIAParams();

  bool getSensFlag() const {
    return sensFlag_;
  }

  void addAnalysisFactory(Factory<void> *factory) {
    registry_.push_back(factory);
  }

  void addAnalysis(Factory<void> *factory) {
    creatorVector_.push_back(factory);
  }


  void setDAEStateDerivFlag(bool state) {
    daeStateDerivFlag_ = state;
  }

public:
  // Queries about output and restart times
  bool testRestartSaveTime();

  // For DCOP restart and/or .SAVE files.
  bool testDCOPOutputTime();
  bool testSaveOutputTime();

public:
  TimeIntg::DataStore *getDataStore() {
    return dataStore_;
  }

  const AnalysisBase &getAnalysisObject() const {
    return *primaryAnalysisObject_;
  }

  AnalysisBase &getAnalysisObject() {
    return *primaryAnalysisObject_;
  }

  void setPrimaryAnalysisObject(AnalysisBase *primary) {
    primaryAnalysisObject_ = primary;
  }

  CreatorVector &getCreatorVector() {
    return creatorVector_;
  }

  Parallel::Manager *getPDSManager() const {
    return parallelManager_;
  }

  bool getSwitchIntegrator() const {
    return switchIntegrator_;
  }

  void setSwitchIntegrator(bool switch_itegrator) {
    switchIntegrator_ = switch_itegrator;
  }

  void setNextOutputTime(double next_output_time) {
    nextOutputTime_ = next_output_time;
  }

  double getNextOutputTime() const {
    return nextOutputTime_;
  }

  Util::Timer &getXyceTranTimer() {
    return xyceTranTimerPtr_;
  }

  OutputMgrAdapter &getOutputManagerAdapter() const {
    return outputManagerAdapter_;
  }

  TimeIntg::WorkingIntegrationMethod &getWorkingIntegrationMethod() {
    return *workingIntgMethod_;
  }

  const TimeIntg::WorkingIntegrationMethod &getWorkingIntegrationMethod() const {
    return *workingIntgMethod_;
  }

  TimeIntg::StepErrorControl &getStepErrorControl() {
    return *stepErrorControl_;
  }

  const TimeIntg::StepErrorControl &getStepErrorControl() const {
    return *stepErrorControl_;
  }

  Loader::NonlinearEquationLoader &getNonlinearEquationLoader() {
    return *nonlinearEquationLoader_;
  }

  Loader::Loader &getLoader() {
    return *loader_;
  }

  const Loader::Loader &getLoader() const {
    return *loader_;
  }

  void setAnalysisMode(Mode analysis_mode) {
    analysisMode_ = analysis_mode;
  }

  Mode getAnalysisMode() const {
    return analysisMode_;
  }

  double getSolverStartTime() const {
    return solverStartTime_;
  }

  void silenceProgress()
  {
    progressFlag_ = false;
  }

  bool getProgressFlag() const {
    return progressFlag_;
  }

  double getSaveTime() const {
    return saveTime_;
  }

  void pushActiveAnalysis(AnalysisBase *analysis) {
    currentAnalysisStack_.push_back(analysis);
  }

  void popActiveAnalysis() {
    currentAnalysisStack_.pop_back();
  }

  const AnalysisBase *getActiveAnalysis() const {
    return currentAnalysisStack_.front();
  }

  bool getResumingSimulation() const {
    return resumeSimulation_;
  }

private:
  const IO::CmdParse &                  commandLine_;                   ///< Command line object
  const std::string                     netlistFilename_;               ///< Netlist file name

  IO::RestartMgr &                      restartManager_;                ///< Restart manager
  OutputMgrAdapter &                    outputManagerAdapter_;          ///< Output manager adapter

  TimeIntg::WorkingIntegrationMethod *  workingIntgMethod_;             ///< Working integration method
  TimeIntg::StepErrorControl *          stepErrorControl_;              ///< Pointer to the TIA step-error control object
  Loader::NonlinearEquationLoader *     nonlinearEquationLoader_;    ///< Pointer to the nonlinear equation loader

  Loader::Loader *                      loader_;                        ///< Pointer to the nonlinear loader object
  Loader::CktLoader *                   cktLoaderPtr_;                  ///< 'real' pointer to the ckt-loader
  Parallel::Manager *                   parallelManager_;               ///< Pointer to the parallel services manager
  TimeIntg::DataStore *                 dataStore_;                     ///< Data store object
  IO::ActiveOutput *                    activeOutput_;


  Mode                  analysisMode_;
  TwoLevelMode          twoLevelMode_;

  bool                  resumeSimulation_;              ///< Resume simulation from a paused transient
  bool                  blockAnalysisFlag_;             ///< HB Analysis (maybe something with MPDE too)
  bool                  daeStateDerivFlag_;             ///< .OPTIONS TIMEINT DAESTATEDERIV=
  bool                  dcopRestartFlag_;               ///< Set if .OPTIONS OP_IO
  bool                  saveFlag_;                      ///< Set if .SAVE
  bool                  dotOpSpecified_;                ///< Set if .OP
  bool                  gui_;                           ///< Set if -giu appears on command line
  bool                  progressFlag_;
  bool                  saveTimeGiven_;
  bool                  savedAlready_;
  bool                  sensFlag_;
  bool                  sweepSourceResetFlag_;
  bool                  switchIntegrator_;              ///< Set to true when Transient::integrationMethod_ is changed

  Util::Timer           xyceTranTimerPtr_;              /// Xyce timing utility for timing the transient simulation CPU time.
  Util::Timer *         elapsedTimerPtr_;               /// Xyce timing utility for timing elapsed run time

  double                solverStartTime_;
  double                saveTime_;                      /// for .SAVE and/or DCOP restart.

  double                nextOutputTime_;
  double                nextRestartSaveTime_;

  AnalysisBase *        analysisObject_;                ///< .STEP, Dakota
  AnalysisBase *        primaryAnalysisObject_;         ///< .TRAN, .AC, .HB, ...

  std::vector<AnalysisBase *> analysisVector_;
  std::vector<AnalysisBase *> currentAnalysisStack_;

  CreatorVector         creatorVector_;
  Registry              registry_;

  Stats::Stat           analysisStat_;
  
public:
    unsigned int breakPointRestartStep;
};

bool registerPkgOptionsMgr(AnalysisManager &analysis_manager, const std::string &netList_file, IO::PkgOptionsMgr &options_manager);

void registerAnalysisFactory(
  const std::string &           netlist_filename,
  IO::PkgOptionsMgr &           options_manager,
  AnalysisManager &             analysis_manager,
  Linear::System &              linear_system,
  Nonlinear::Manager &          nonlinear_manager,
  Device::DeviceMgr &           device_manager,
  Linear::Builder &             builder,
  Topo::Topology &              topology);

void getTimeIntInfo(const AnalysisManager &analysis_manager, TimeIntg::TimeIntInfo & tiInfo);

bool evalTransientModel(
  Loader::Loader &      loader,
  double                t,
  Linear::Vector *      SolVectorPtr,
  Linear::Vector *      CurrSolVectorPtr,
  Linear::Vector *      LasSolVectorPtr,
  Linear::Vector *      StaVectorPtr,
  Linear::Vector *      CurrStaVectorPtr,
  Linear::Vector *      LasStaVectorPtr,
  Linear::Vector *      StaDerivVectorPtr,
  Linear::Vector *      StoVectorPtr,
  Linear::Vector *      CurrStoVectorPtr,
  Linear::Vector *      LasStoVectorPtr,
  Linear::Vector *      stoLeadCurrQVectorPtr,
  Linear::Vector *      NextLeadFVectorPtr,
  Linear::Vector *      CurrLeadFVectorPtr,
  Linear::Vector *      LastLeadFVectorPtr,
  Linear::Vector *      LeadQVectorPtr,
  Linear::Vector *      NextJunctionVVectorPtr,
  Linear::Vector *      CurrentJunctionVVectorPtr,
  Linear::Vector *      LastJunctionVVectorPtr,
  Linear::Vector *      QVectorPtr,
  Linear::Vector *      FVectorPtr,
  Linear::Vector *      BVectorPtr,
  Linear::Vector *      dFdxdVpVectorPtr,
  Linear::Vector *      dQdxdVpVectorPtr,
  Linear::Matrix *      dQdxMatrixPtr,
  Linear::Matrix *      dFdxMatrixPtr);

bool evalTransientModelState(
  TimeIntg::StepErrorControl &  step_error_control,
  Loader::Loader &              loader,
  double                        t,
  Linear::Vector *              SolVectorPtr,
  Linear::Vector *              StaVectorPtr,
  Linear::Vector *              StoVectorPtr);

void writeConductanceFile(const std::vector<std::string> &device_names, Nonlinear::ConductanceExtractor &conductance_extractor, const std::string &filename);

} // namespace Analysis
} // namespace Xyce

#endif // Xyce_N_ANP_AnalysisManager_h
