//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_ANP_Factory.h,v $
//
// Purpose       :
//
// Special Notes :
//
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
//
// Creation Date : 01/24/08
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.6.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:06 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

///
/// @file   N_ANP_Factory.h
/// @author David G. Baur  Raytheon  Sandia National Laboratories 1355 <dgbaur@sandia.gov>
/// @date   Thu Jan 29 12:51:28 2015
///
/// @brief  Factory for creating analysis objects
///
///
#ifndef Xyce_N_ANP_Factory_h
#define Xyce_N_ANP_Factory_h

#include <typeinfo>

#include <N_LAS_fwd.h>

namespace Xyce {
namespace Analysis {

template<class T>
class Factory;

//-----------------------------------------------------------------------------
// Template      : Factory
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Thu Jan 29 12:40:53 2015
//-----------------------------------------------------------------------------
///
/// The analysis factory template defines an interface for analysis type testing and
/// analysis creation.
///
template<>
class Factory<void>
{
public:
  //-----------------------------------------------------------------------------
  // Function      : Factory
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:41:54 2015
  //-----------------------------------------------------------------------------
  ///
  /// Constructor
  ///
  Factory()
  {}

  //-----------------------------------------------------------------------------
  // Function      : ~Factory
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:43:11 2015
  //-----------------------------------------------------------------------------
  ///
  /// Destructor
  ///
  virtual ~Factory()
  {}

  //-----------------------------------------------------------------------------
  // Function      : type
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:43:43 2015
  //-----------------------------------------------------------------------------
  ///
  /// Defines the interface to get the type info of the analysis created by the factory.
  ///
  /// @return type info of the analysis that would be created.
  ///
  ///
  virtual const std::type_info &type() const = 0;

  //-----------------------------------------------------------------------------
  // Function      : isType
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:45:04 2015
  //-----------------------------------------------------------------------------
  ///
  /// Returns true if the analysis type of the factory matches class U
  ///
  /// @param U Analysis class type to test
  ///
  /// @return true if the analysis type of the factory matches class U
  ///
  ///
  template <class U>
  bool isType() const {
    return type() == typeid(U);
  }
  
private:
  Factory(const Factory &);                     ///< not copyable
  Factory &operator=(const Factory &);          ///< not assignable

public:
  //-----------------------------------------------------------------------------
  // Function      : create
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:47:13 2015
  //-----------------------------------------------------------------------------
  ///
  /// Creates the analysis object.
  ///
  /// @return the new analysis object
  ///
  ///
  virtual AnalysisBase *create() const = 0;
};

//-----------------------------------------------------------------------------
// Template      : Factory
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Thu Jan 29 12:40:53 2015
//-----------------------------------------------------------------------------
///
/// The analysis factory template defines an interface for analysis type testing and
/// analysis creation.  This template implements the type checking.
///
/// @param T analysis type to be created
///
template<class T>
class Factory : public Factory<void>
{
public:
  //-----------------------------------------------------------------------------
  // Function      : Factory
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:49:44 2015
  //-----------------------------------------------------------------------------
  ///
  /// Constructor
  ///
  Factory()
    : Factory<void>()
  {}

  //-----------------------------------------------------------------------------
  // Function      : ~Factory
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:50:03 2015
  //-----------------------------------------------------------------------------
  ///
  /// Destructor
  ///
  virtual ~Factory()
  {}

  //-----------------------------------------------------------------------------
  // Function      : type
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:50:21 2015
  //-----------------------------------------------------------------------------
  ///
  /// Returns the type info the analysis type.
  ///
  /// @return the type info the analysis type.
  ///
  virtual const std::type_info &type() const {
    return typeid(T);
  }
};

typedef std::vector<Factory<void> *> Registry;                  ///< Registry of factories

} // namespace Analysis
} // namespace Xyce

#endif // Xyce_N_ANP_Factory_h
