//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-NOISE04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_ANP_NOISE.h,v $
//
// Purpose        : NOISE analysis class
//
// Special Notes  : Specify any "hidden" or subtle details of the class here.
//                  Portability details, error handling information, etc.
//
// Creator        : Eric Keiter
//
// Creation Date  : 01/11
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.9.2.1 $
// Revision Date  : $Date: 2015/04/02 18:20:06 $
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#ifndef Xyce_N_ANP_NOISE_h
#define Xyce_N_ANP_NOISE_h

#include <N_ANP_fwd.h>
#include <N_UTL_fwd.h>
#include <N_LAS_fwd.h>
#include <N_TOP_fwd.h>

#include <N_ANP_AnalysisBase.h>
#include <N_ANP_Factory.h>
#include <N_ANP_StepEvent.h>
#include <N_TIA_TIAParams.h>
#include <N_UTL_FixedQueue.h>
#include <N_UTL_Listener.h>
#include <N_UTL_OptionBlock.h>

// ---------- Forward Declarations ----------
class Amesos_BaseSolver;
class Epetra_LinearProblem;

namespace Xyce {
namespace Analysis {

//-------------------------------------------------------------------------
// Class         : NOISE
// Purpose       : NOISE analysis class
// Special Notes :
// Creator       : Eric Keiter
// Creation Date : 12/8/2014
//-------------------------------------------------------------------------
class NOISE: public AnalysisBase, public Util::ListenerAutoSubscribe<StepEvent>
{
  public:
  NOISE(
    AnalysisManager &           analysis_manager,
    Linear::System &            linear_system,
    Nonlinear::Manager &        nonlinear_manager,
    Topo::Topology &            topology);

  ~NOISE();

  void notify(const StepEvent &event);

  const TimeIntg::TIAParams &getTIAParams() const {
    return tiaParams_;
  }

  TimeIntg::TIAParams &getTIAParams() {
    return tiaParams_;
  }

  bool setAnalysisParams(const N_UTL_OptionBlock & paramsBlock);
  bool setTimeIntegratorOptions(const Util::OptionBlock &option_block);

  bool getDCOPFlag() const;

  bool doRun();
  bool doInit();
  bool doLoopProcess();
  bool doProcessSuccessfulStep();
  bool doProcessFailedStep();
  bool doFinish();
  bool doHandlePredictor();

  void printStepHeader(std::ostream &os)
  {}

  void printProgress(std::ostream &os)
  {}

private:
  int setupSweepParam_();

  bool updateCurrentFreq_(int stepNumber);

  bool createACLinearSystem_();

  bool resetACLinearSystem_();

  bool solveACLinearSystem_();

  bool solveDirectNOISE_();

  void resetAdjointNOISELinearSystem_();
  bool solveAdjointNOISE_();
  void setupAdjointRHS_();

  double noiseIntegral( double noizDens, double lnNdens, double lnNlstDens, 
      double delLnFreq, double delFreq, double lnFreq, double lnLastFreq);

  void processOutputNodes ();

  void hackTecplotOutput();

private:
  AnalysisManager &     analysisManager_;
  Loader::Loader &      loader_;
  N_LAS_System &        linearSystem_;
  Nonlinear::Manager &  nonlinearManager_;
  Topo::Topology &      topology_;
  OutputMgrAdapter &    outputManagerAdapter_;
  TimeIntg::TIAParams   tiaParams_;

  // AC B-vectors
  N_LAS_Vector * bVecRealPtr;
  N_LAS_Vector * bVecImagPtr;

  // NOISE B-vectors
  N_LAS_Vector * bNoiseVecRealPtr;
  N_LAS_Vector * bNoiseVecImagPtr;

  bool hackOutputCalledBefore_;
  bool useAdjoint_;
  bool outputNodeSingle_;
  std::string outputNode1_;
  std::string outputNode2_;
  std::string specifiedSource_;
  int noiseLoopSize_;

  std::vector<int>              noiseSweepFailures_;

  bool                          stepFlag_;
  std::string                   type_;
  double                        np_;
  double                        fStart_;
  double                        fStop_;

  double                        stepMult_;
  double                        fstep_;
  int                           pts_per_summary_;
  bool                          pts_per_summary_Given;

  double                        delFreq_;
  double                        lastFreq_;
  double                        currentFreq_;
  double                        lnFreq_;
  double                        lnLastFreq_;
  double                        delLnFreq_;

  double                        GainSqInv_;
  double                        lnGainInv_;

  double                        totalOutputNoise_;
  double                        totalInputNoise_;

  double                        totalOutputNoiseDens_;
  double                        totalInputNoiseDens_;

  N_LAS_Matrix *                C_;
  N_LAS_Matrix *                G_;
  N_LAS_BlockMatrix *           ACMatrix_;
  N_LAS_BlockVector *           B_;
  N_LAS_BlockVector *           X_;
  N_LAS_BlockVector *           saved_AC_X_;

  Amesos_BaseSolver *           blockSolver_;
  Epetra_LinearProblem *        blockProblem_;

  std::vector<std::string> outputVarNames_;
  std::vector<int>    outputVarGIDs_;

  std::vector<Xyce::Analysis::NoiseData*> noiseDataVec_;

  std::vector<double> objectiveVec_; 
  std::vector<double> dOdpVec_; 
  std::vector<double> dOdpAdjVec_;
  std::vector<double> scaled_dOdpVec_;
  std::vector<double> scaled_dOdpAdjVec_;
};

bool registerNOISEFactory(
  const std::string &   netlist_filename,
  IO::PkgOptionsMgr &   options_manager,
  AnalysisManager &     analysis_manager,
  Linear::System &      linear_system,
  Nonlinear::Manager &  nonlinear_manager,
  Topo::Topology &      topology);

} // namespace Analysis
} // namespace Xyce

#endif // Xyce_N_ANP_NOISE_h
