//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_ANP_OutputMgrAdapter.h,v $
//
// Purpose        :
//
// Special Notes  :
//
// Creator        : Richard Schiek, SNL, Electrical and Microsystem Modeling
//
// Creation Date  : 01/24/08
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.75.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:06 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#ifndef Xyce_N_ANP_OutputMgrAdapter_h
#define Xyce_N_ANP_OutputMgrAdapter_h

#include <Teuchos_SerialDenseMatrix.hpp>

#include <N_ANP_fwd.h>
#include <N_DEV_fwd.h>
#include <N_IO_fwd.h>
#include <N_LAS_fwd.h>
#include <N_PDS_fwd.h>
#include <N_TOP_fwd.h>
#include <N_UTL_fwd.h>

#include <N_UTL_Listener.h>
#include <N_ANP_StepEvent.h>
#include <N_LAS_BlockVector.h>

namespace Xyce {
namespace Analysis {

//-------------------------------------------------------------------------
// Class         : OutputMgrAdapter
// Purpose       : Inteface class for the output manager
// Special Notes :
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 01/24/08
//-------------------------------------------------------------------------
class OutputMgrAdapter : public Util::Listener<StepEvent>
{
public:
  OutputMgrAdapter(
    Parallel::Machine                   comm,
    IO::OutputMgr &                     output_manager,
    IO::Measure::Manager &              measure_manager,
    IO::FourierMgr &                    fourier_manager,
    IO::ObjectiveManager &              objective_manager,
    IO::InitialConditionsManager &      initial_conditions_manager,
    Device::DeviceMgr &                 device_manager,
    Topo::Topology &                    topology,
    IO::RestartMgr &                    restart_manager);

  virtual ~OutputMgrAdapter();

  void notify(const StepEvent &event);

  void addOutputResults(const Util::OptionBlock & option_block);
  void addOutputResponse(const Util::OptionBlock & option_block);

  void setStepSweepVector(const Analysis::SweepVector &sweep_vector);

  void setDCSweepVector(const Analysis::SweepVector &sweep_vector);

  const Analysis::SweepVector &getStepSweepVector() const
  {
    return stepSweepVector_;
  }

  const Analysis::SweepVector &getDCSweepVector() const
  {
    return dcSweepVector_;
  }

  // accessor methods
  int getStepAnalysisStepNumber() const
  {
    return stepAnalysisStepNumber_;
  }

  int getDCAnalysisStepNumber() const
  {
    return dcAnalysisStepNumber_;
  }

  int getDCAnalysisMaxSteps() const
  {
    return dcAnalysisMaxSteps_;
  }

  IO::OutputMgr &getOutputManager()
  {
    return outputManager_;
  }

  IO::InitialConditionsManager &getInitialConditionsManager()
  {
    return initialConditionsManager_;
  }

  void setDCAnalysisStepNumber( int num )
  {
    dcAnalysisStepNumber_ = num;
  }

  void setDCAnalysisMaxSteps( int num )
  {
    dcAnalysisMaxSteps_ = num;
  }

  bool isRestart() const;

  void dumpRestart(Parallel::Communicator &parallel_communicator, Analysis::AnalysisManager &analysis_manager, double current_time) const;

  void tranOutput(
    double time, Linear::Vector & currSolutionPtr, Linear::Vector & stateVecPtr, Linear::Vector & storeVecPtr, 
    Linear::Vector &        lead_current_vector,
    Linear::Vector &        junction_voltage_vector,
    std::vector<double> & objectiveVec_, 
    std::vector<double> & dOdpVec_, 
    std::vector<double> & dOdpAdjVec_,
    std::vector<double> & scaled_dOdpVec_, 
    std::vector<double> & scaled_dOdpAdjVec_,
    bool skipPrintLineOutput = false);

  void dcOutput( 
    int dcStepNumber, 
    Linear::Vector & currSolutionPtr, Linear::Vector & stateVecPtr, Linear::Vector & storeVecPtr,
    Linear::Vector &lead_current_vector,
    Linear::Vector &junction_voltage_vector,
    std::vector<double> & objectiveVec_, 
    std::vector<double> & dOdpVec_, 
    std::vector<double> & dOdpAdjVec_, 
    std::vector<double> & scaled_dOdpVec_, 
    std::vector<double> & scaled_dOdpAdjVec_);

  void outputResult(const Linear::Vector &currSolutionPtr, 
    const Linear::Vector &currStatePtr, 
    const Linear::Vector &currStorePtr,
    const Linear::Vector &lead_current_vector,
    const Linear::Vector &junction_voltage_vector );

  void steppingComplete();

  void finishOutput();

  bool setupInitialConditions( Linear::Vector & solnVec, Linear::Vector & flagVec);

  void outputDCOP(const Linear::Vector &solution);

  void outputMPDE(double time, const std::vector<double> &fast_time_points, const Linear::BlockVector &solution_vector);

  void outputHB(
    const std::vector< double > & timePoints, const std::vector< double > & freqPoints,
    const Linear::BlockVector & timeDomainSolnVec, const Linear::BlockVector & freqDomainSolnVecReal,
    const Linear::BlockVector & freqDomainSolnVecImaginary, const Linear::BlockVector & timeDomainStoreVec,
    const Linear::BlockVector & freqDomainStoreVecReal, const Linear::BlockVector & freqDomainStoreVecImaginary,
    const Linear::BlockVector & timeDomainLeadCurrentVec, const Linear::BlockVector & freqDomainLeadCurrentVecReal, 
    const Linear::BlockVector & freqDomainLeadCurrentVecImaginary, const Linear::BlockVector & timeDomainJunctionVoltageVec, 
    const Linear::BlockVector & freqDomainJunctionVoltageVecReal, const Linear::BlockVector & freqDomainJunctionVoltageVecImaginary );

  void outputAC(double freq, const Linear::Vector & solnVecRealPtr, const Linear::Vector & solnVecImaginaryPtr);

  void outputNoise(
      double freq, double totalOutputNoiseDens_, double totalInputNoiseDens_, 
      const std::vector<Xyce::Analysis::NoiseData*> & noiseDataVec_);

  void outputMORTF(bool origSys, double freq, const Teuchos::SerialDenseMatrix<int, std::complex<double> >& H );

  void resetOutputMORTF();

  void outputROM(
    const Teuchos::SerialDenseMatrix<int, double>& Ghat, const Teuchos::SerialDenseMatrix<int, double>& Chat,
    const Teuchos::SerialDenseMatrix<int, double>& Bhat, const Teuchos::SerialDenseMatrix<int, double>& Lhat );

  void outputROM(
    const Linear::Matrix& Ghat, const Linear::Matrix& Chat,
    const Teuchos::SerialDenseMatrix<int, double>& Bhat,
    const Teuchos::SerialDenseMatrix<int, double>& Lhat );

  double getInitialOutputInterval() const;

  const IO::IntervalVector &getOutputIntervals() const;

  void outputHomotopy( const std::vector<std::string> & paramNames, const std::vector<double> & paramVals, Linear::Vector & solnVecPtr );

  const Xyce::NodeNameMap & getAllNodes() const;

private:
  Parallel::Machine                     comm_;
  IO::OutputMgr &                       outputManager_;
  IO::Measure::Manager &                measureManager_;
  IO::FourierMgr &                      fourierManager_;
  IO::ObjectiveManager &                objectiveManager_;
  IO::InitialConditionsManager &        initialConditionsManager_;
  Device::DeviceMgr &                   deviceManager_;
  Topo::Topology &                      topology_;
  IO::RestartMgr &                      restartManager_;

  IO::OutputMOR *                       outputMOR_;
  IO::OutputResults *                   outputResults_;
  IO::OutputResponse *                  outputResponse_;

  Analysis::SweepVector                 stepSweepVector_;
  Analysis::SweepVector                 dcSweepVector_;

  int stepAnalysisStepNumber_;
  int stepAnalysisMaxSteps_;
  int dcAnalysisStepNumber_;
  int dcAnalysisMaxSteps_;
};

class OutputAdapter
{
public:
  OutputAdapter(OutputMgrAdapter &adapter)
    : adapter_(adapter)
  {}

  virtual void outputMPDE(double time, const std::vector<double> &fast_time_points, const Linear::Vector *solution_vector)
  {}

  OutputMgrAdapter &    adapter_;
};

class MPDEOutputAdapter : public OutputAdapter
{
public:
  MPDEOutputAdapter(OutputMgrAdapter &adapter)
    : OutputAdapter(adapter)
  {}
  
  virtual void outputMPDE(double time, const std::vector<double> &fast_time_points, const Linear::Vector *solution_vector)
  {
    if (dynamic_cast<const Linear::BlockVector *>(solution_vector))
      adapter_.outputMPDE(time, fast_time_points, dynamic_cast<const Linear::BlockVector &>(*solution_vector));
  }
};

} // namespace Analysis
} // namespace Xyce

#endif // Xyce_N_ANP_OutputMgrAdapter_h
