//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_ANP_AnalysisManager.C,v $
// Purpose       : This file contains the functions which define the time
//                 domain & integration algorithm classes.
// Special Notes :
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 01/24/08
//
// Revision Information:
// ---------------------
// Revision Number: $Revision: 1.243.2.1 $
// Revision Date  : $Date: 2015/04/02 18:20:07 $
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

// ---------- Standard Includes ----------
#include <ctime>
#include <iostream>
#include <sstream>

#include <N_ANP_AnalysisManager.h>

#include <N_ANP_AC.h>
#include <N_ANP_DCSweep.h>
#include <N_ANP_Dakota.h>
#include <N_ANP_Factory.h>
#include <N_ANP_HB.h>
#include <N_ANP_MOR.h>
#include <N_ANP_MPDE.h>
#include <N_ANP_NOISE.h>
#include <N_ANP_OutputMgrAdapter.h>
#include <N_ANP_Report.h>
#include <N_ANP_Step.h>
#include <N_ANP_Transient.h>
#include <N_DEV_DeviceMgr.h>
#include <N_IO_ActiveOutput.h>
#include <N_IO_CmdParse.h>
#include <N_IO_OutputMgr.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_IO_RestartMgr.h>
#include <N_LAS_LAFactory.h>
#include <N_LAS_Matrix.h>
#include <N_LAS_System.h>
#include <N_LAS_Vector.h>
#include <N_LOA_CktLoader.h>
#include <N_LOA_Loader.h>
#include <N_LOA_NonlinearEquationLoader.h>
#include <N_NLS_Manager.h>
#include <N_NLS_ConductanceExtractor.h>
#include <N_TIA_BackwardDifferentiation15.h>
#include <N_TIA_DataStore.h>
#include <N_TIA_NoTimeIntegration.h>
#include <N_TIA_StepErrorControl.h>
#include <N_TIA_TIAParams.h>
#include <N_TIA_TimeIntInfo.h>
#include <N_TIA_WorkingIntegrationMethod.h>
#include <N_TIA_TwoLevelError.h>
#include <N_UTL_BreakPoint.h>
#include <N_UTL_Diagnostic.h>
#include <N_UTL_ExtendedString.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_OptionBlock.h>
#include <N_UTL_Stats.h>
#include <N_UTL_Timer.h>

namespace Xyce {
namespace Analysis {

//-----------------------------------------------------------------------------
// Function      : analysisModeName
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Tue Oct 21 11:22:31 2014
//-----------------------------------------------------------------------------
///
/// Returns a const char pointer to a name representing the mode.
///
/// @param mode Analysis mode to get name of
///
/// @return Name corresponding to the mode
///
///
const char *
analysisModeName(
  Mode          mode) 
{
  static const char * const mode_names[] = {"Invalid", "DC OP", "DC Sweep", "Transient", "MPDE", "HB", "AC", "NOISE", "MOR"};

  if (mode < sizeof(mode_names)/sizeof(mode_names[0]))
    return mode_names[mode];
  else
    return mode_names[0];
}

//-----------------------------------------------------------------------------
// Function      : nonlinearAnalysisMode
// Purpose       : Convert between Nonlinear::Manager AnalysisMode enum and
//               : expanded AnalysisManager::Manager::Mode enum.
// Special Notes :
// Scope         :
// Creator       : Todd Coffey, 1414
// Creation Date : 07/23/08
//-----------------------------------------------------------------------------
///
/// Returns the nonlinear analysis mode given the analysis mode
///
/// @param mode Analysis mode
///
/// @return Nonlinear analysis mode
///
///
Nonlinear::AnalysisMode
nonlinearAnalysisMode(
  Mode          mode)
{
  Nonlinear::AnalysisMode outMode;
  if (mode == ANP_MODE_TRANSIENT)
  {
    outMode = Nonlinear::TRANSIENT;
  }
  else if (mode == ANP_MODE_DC_OP)
  {
    outMode = Nonlinear::DC_OP;
  }
  else if (mode == ANP_MODE_DC_SWEEP)
  {
    outMode = Nonlinear::DC_SWEEP;
  }
  else if (mode == ANP_MODE_HB)
  {
    outMode = Nonlinear::HB_MODE;
  }
  else
  {
    outMode = Nonlinear::NUM_MODES; // Should be this be TRANSIENT?
  }
  return outMode;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::AnalysisManager
// Purpose       : constructor
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 01/24/08
//-----------------------------------------------------------------------------
///
/// Constructs the analysis manager.  The analysis manager is responsible for the construction and control of the top
/// level analysis.
///
/// @invariant
///
/// @param command_line                 Command line that started Xyce
/// @param restart_manager              
/// @param output_manager_adapter 
/// @param analysis_stat                Base node of the analysis stats collection
///
///
AnalysisManager::AnalysisManager(
  const IO::CmdParse &  command_line,
  IO::RestartMgr &      restart_manager,
  OutputMgrAdapter &    output_manager_adapter,
  Stats::Stat           analysis_stat)
  : StepEventNotifier(),
    AnalysisEventNotifier(),
    StepEventListener(this),
    AnalysisEventListener(this),
    commandLine_(command_line),
    netlistFilename_(command_line.getArgumentValue("netlist")),
    restartManager_(restart_manager),
    outputManagerAdapter_(output_manager_adapter),
    workingIntgMethod_(0),
    stepErrorControl_(0),
    nonlinearEquationLoader_(0),
    loader_(0),
    cktLoaderPtr_(0),
    parallelManager_(0),
    dataStore_(0),
    activeOutput_(0),
    analysisMode_(ANP_MODE_TRANSIENT),
    resumeSimulation_(false),
    blockAnalysisFlag_(false),
    daeStateDerivFlag_(true),
    dcopRestartFlag_(false),
    saveFlag_(false),
    dotOpSpecified_(false),
    gui_(command_line.argExists("-gui")),
    progressFlag_(true),
    saveTimeGiven_(false),
    savedAlready_(false),
    sensFlag_(false),
    sweepSourceResetFlag_(true),
    switchIntegrator_(false),
    xyceTranTimerPtr_(),
    elapsedTimerPtr_(0),
    solverStartTime_(0.0),
    saveTime_(0.0),
    nextOutputTime_(0.0),
    nextRestartSaveTime_(0.0),
    analysisObject_(0),
    primaryAnalysisObject_(0),
    breakPointRestartStep(0),
    analysisStat_(analysis_stat),
    twoLevelMode_(TWO_LEVEL_MODE_TRANSIENT_DCOP)
{
  IO::setTimeIntegratorDebugLevel(command_line, 1);
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::~AnalysisManager
//
// Purpose       : destructor
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 01/24/08
//-----------------------------------------------------------------------------
AnalysisManager::~AnalysisManager()
{
  delete cktLoaderPtr_;
  delete nonlinearEquationLoader_;
  delete workingIntgMethod_;
  delete dataStore_;
  delete stepErrorControl_;

  for (std::vector<AnalysisBase *>::iterator it = analysisVector_.begin(), end = analysisVector_.end(); it !=end; ++it)
    delete (*it);

  for (Registry::iterator it = registry_.begin(), end = registry_.end(); it !=end; ++it)
    delete (*it);
}


//-----------------------------------------------------------------------------
// Function      : AnalysisManager::notify
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Mon Jul 14 12:15:30 2014
//-----------------------------------------------------------------------------
///
/// Notification that there is a StepEvent.
///
/// @param step_event   information about the event
///
void
AnalysisManager::notify(
  const StepEvent &     step_event)
{
  if (step_event.state_ == StepEvent::STEP_STARTED) {
    dataStore_->setZeroHistory();
  }
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::notify
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Mon Jul 14 12:15:30 2014
//-----------------------------------------------------------------------------
///
/// Notification that there is a AnalysisEvent.
///
/// @param time_integrator_event   information about the event
///
void
AnalysisManager::notify(
  const AnalysisEvent &     analysis_event)
{
//  Xyce::dout() << Dump<AnalysisEvent>(analysis_event) << std::endl;
}


void
AnalysisManager::createTimeIntegratorMethod(
  const TimeIntg::TIAParams &   tia_params,
  const unsigned int            integration_method)
{
  workingIntgMethod_->createTimeIntegMethod(integration_method, tia_params, *stepErrorControl_, *dataStore_);
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::resetSolverSystem
//
// Purpose       : just like a destructor without the death
// Special Notes :
// Scope         : public
// Creator       : Todd Coffey, Rich Schiek, Ting Mei
// Creation Date : 7/24/08
//-----------------------------------------------------------------------------
void AnalysisManager::resetSolverSystem()
{
  // dataStore_ is created in initializeAll
  delete dataStore_;
  dataStore_ = 0;

  // stepErrorControl_ is created in initializeAll
  delete stepErrorControl_;
  stepErrorControl_ = 0;

  // workingIntgMethod_ is created in initializeAll
  delete workingIntgMethod_;
  workingIntgMethod_ = 0;

  // Reset step statistics to zero.
  primaryAnalysisObject_->resetAll();
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::isPaused
//
// Purpose       : return the true if the simulation is currently paused
//
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, SNL, Electrical and Microsystems Simulation
// Creation Date : 02/18/2008
//-----------------------------------------------------------------------------
bool AnalysisManager::isPaused() const
{
  return stepErrorControl_->isPauseTime();
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::createLoader
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Mon Mar 16 14:04:17 2015
//-----------------------------------------------------------------------------
///
/// Creates the defaults circuit loader.
///
/// @invariant This circuit loader is owned by the manager and will be destroyed during manager destruction.
///
/// @param device_manager Device manager needed to construct the circuit loader.
///
/// @return default circuit loader
///
///
Loader::Loader &
AnalysisManager::createLoader(Device::DeviceMgr &device_manager)
{
  cktLoaderPtr_ = new Loader::CktLoader(device_manager);

  return *cktLoaderPtr_;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::initializeSolverSystem
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Mon Mar 16 14:06:00 2015
//-----------------------------------------------------------------------------
///
/// Initializes the solver system.  Called on initial analysis assembly and on change of time integrator.
///
/// @invariant There be gremlins here.  The datas tore, step error control, time integrator (managed by the working
/// integrator method) and nonlinear equation loader are destroyed and recreated.  Hopefully none of the subordinate
/// users of those objects are stll hanging on.  The linear system will have much of its information replaced by data
/// from the newly created datastore and vice versa.
///
///
///
/// @param tia_params           Time integrator parameters to construct new time integrator
/// @param loader               Loader to loader linear system
/// @param linear_system        Linear system
/// @param nonlinear_manager    Nonlinear system
/// @param device_manager       Device Manager
///
/// @return true if initialization was successful
///
///
bool AnalysisManager::initializeSolverSystem(
  const TimeIntg::TIAParams &   tia_params,
  Loader::Loader &              loader,
  Linear::System &              linear_system,
  Nonlinear::Manager &          nonlinear_manager,
  Device::DeviceMgr &           device_manager)
{
  Stats::Stat stat(std::string(analysisObject_->getName()), analysisStat_);

  int solutionSize = linear_system.getSolutionSize();
  int stateSize    = linear_system.getStateSize();

  // allocate data store class, which will allocate all the vectors.
  delete dataStore_;
  delete stepErrorControl_;
  delete workingIntgMethod_;
  delete nonlinearEquationLoader_;

  dataStore_ = new TimeIntg::DataStore(solutionSize, stateSize, tia_params.maxOrder, linear_system, linear_system.builder());
  workingIntgMethod_ = new TimeIntg::WorkingIntegrationMethod(stat);
  stepErrorControl_ = new TimeIntg::StepErrorControl(netlistFilename_, *this, *workingIntgMethod_, tia_params);
  nonlinearEquationLoader_ = new Loader::NonlinearEquationLoader(*dataStore_, loader, device_manager, *workingIntgMethod_, daeStateDerivFlag_);

  dataStore_->resetAll(tia_params.absErrorTol, tia_params.relErrorTol);

  nextOutputTime_ = stepErrorControl_->initialTime;
  nextRestartSaveTime_ = stepErrorControl_->initialTime;

  // register current:
  linear_system.registerCurrStaVector(&dataStore_->currStatePtr);
  linear_system.registerCurrSolVector(&dataStore_->currSolutionPtr);

  // register next:
  linear_system.registerNextStaVector(&dataStore_->nextStatePtr);
  linear_system.registerNextSolVector(&dataStore_->nextSolutionPtr);

  // register last:
  linear_system.registerLastStaVector(&dataStore_->lastStatePtr);
  linear_system.registerLastSolVector(&dataStore_->lastSolutionPtr);

  linear_system.registerFlagSolVector(&dataStore_->flagSolutionPtr);

  // register temporaries:
  linear_system.registerTmpSolVector(&dataStore_->tmpSolVectorPtr);
  linear_system.registerTmpStaVector(&dataStore_->tmpStaVectorPtr);
  linear_system.registerTmpStaDerivVector(&dataStore_->tmpStaDerivPtr);
  linear_system.registerTmpStaDivDiffVector(&dataStore_->tmpStaDivDiffPtr);

  // register next derivatives:
  linear_system.registerNextSolDerivVector(&dataStore_->nextSolutionDerivPtr);
  linear_system.registerNextStaDerivVector(&dataStore_->nextStateDerivPtr);

  // register the device mask
  linear_system.registerDeviceMaskVector(&dataStore_->deviceErrorWeightMask_);

  // Get the RHS and the Jacobian
  dataStore_->JMatrixPtr    = linear_system.getJacobianMatrix();
  dataStore_->RHSVectorPtr  = linear_system.getRHSVector();

  // DAE formulation vectors
  linear_system.registerDAEQVector(dataStore_->daeQVectorPtr);
  linear_system.registerDAEFVector(dataStore_->daeFVectorPtr);
  linear_system.registerDAEBVector(dataStore_->daeBVectorPtr);

  // DAE formulation matrices
  linear_system.registerDAEdQdxMatrix(dataStore_->dQdxMatrixPtr);
  linear_system.registerDAEdFdxMatrix(dataStore_->dFdxMatrixPtr);

  // Get the limiter vectors
  dataStore_->dFdxdVpVectorPtr = linear_system.getdFdxdVpVector();
  dataStore_->dQdxdVpVectorPtr = linear_system.getdQdxdVpVector();

  dataStore_->limiterFlag = loader.getLimiterFlag ();

  return true;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getBlockAnalysisFlag
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Mon Mar 16 14:12:17 2015
//-----------------------------------------------------------------------------
///
/// Return true if primary analysis is HB or MPDE.
///
///
/// @return true if primary analysis is HB or MPDE.
///
///
bool
AnalysisManager::getBlockAnalysisFlag() const
{
  return dynamic_cast<Analysis::HB *>(primaryAnalysisObject_) || dynamic_cast<Analysis::MPDE *>(primaryAnalysisObject_);
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::run
// Purpose       : Execute the control loop for the set analysis type.
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 01/24/08
//-----------------------------------------------------------------------------
///
/// Runs the top level analysis.
///
/// @invariant
///
///
/// @return 
///
///
bool
AnalysisManager::run()
{
  Stats::StatTop _analysisStat("Analysis");
  Stats::TimeBlock _analysisTimer(_analysisStat);

  if (!primaryAnalysisObject_)
  {
    Report::UserError0() << "No analysis statement in the netlist";
    return false;
  }

  bool runStatus = false;
  {
    // This prepares the outputters for this analysis mode
    IO::ActiveOutput active(outputManagerAdapter_.getOutputManager());

    active.setStepSweepVector(outputManagerAdapter_.getStepSweepVector());
    active.setDCSweepVector(outputManagerAdapter_.getDCSweepVector());
    active.add(parallelManager_->getPDSComm()->comm(), analysisMode_);

    if (VERBOSE_TIME)
      getTIAParams().printParams(Xyce::lout(), nonlinearAnalysisMode(analysisMode_));

    nonlinearEquationLoader_->getLoader().loadDeviceErrorWeightMask((dataStore_->deviceErrorWeightMask_));

    Report::safeBarrier(parallelManager_->getPDSComm()->comm());

    solverStartTime_ = elapsedTimerPtr_->elapsedTime();

    // Start the solvers timers.
    xyceTranTimerPtr_.resetStartTime();

    runStatus = analysisObject_->run();
  }

  return runStatus;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::allocateAnalysisObject
// Purpose       : Allocate analysis objects, and also setup params.
// Special Notes :
// Scope         : private
// Creator       : Eric Keiter, SNL
// Creation Date : 06/24/10
//-----------------------------------------------------------------------------
///
/// Creates the primary analysis and driving analysis (.STEP, dakota).
///
/// If no analysis was specified and .OP was specified, create a DC analysis as the primary analysis
///
/// @invariant
///
///
void
AnalysisManager::allocateAnalysisObject()
{
  // .OP, but no analysis creates a dcsweep analysis.
  if (dotOpSpecified_ && (creatorVector_.empty() || (creatorVector_.size() == 1 && creatorVector_.front()->isType<Step>())))
  {
    for (CreatorVector::const_iterator it = registry_.begin(), end = registry_.end(); it != end; ++it) {
      if ((*it)->isType<DCSweep>()) {
        analysisMode_ = ANP_MODE_DC_SWEEP;
        primaryAnalysisObject_ = (*it)->create();
        analysisVector_.push_back(primaryAnalysisObject_);
      }
    }
  }

  // CRAZY HACK UNTIL PRIORITY QUEUE OR SOMETHING CLEVER ON ANALYSIS TYPES
  for (CreatorVector::const_iterator it = creatorVector_.begin(), end = creatorVector_.end(); it != end; ++it) {
    if (!(*it)->isType<Step>()) {
      primaryAnalysisObject_ = (*it)->create();
      analysisVector_.push_back(primaryAnalysisObject_);
    }
  }
  for (CreatorVector::const_iterator it = creatorVector_.begin(), end = creatorVector_.end(); it != end; ++it) {
    if ((*it)->isType<Step>()) {
      analysisObject_ = (*it)->create();
      analysisVector_.push_back(analysisObject_);
      pushActiveAnalysis(analysisObject_);
    }
  }

  if (!primaryAnalysisObject_)
  {
    Report::UserError() << "Analysis mode " << analysisMode_ << " is not available";
    return;
  }

  pushActiveAnalysis(primaryAnalysisObject_);

  if (!analysisObject_)
    analysisObject_ = primaryAnalysisObject_;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::printLoopInfo
// Purpose       : Prints out time loop information.
// Special Notes : Prints stats from save point start to save point finish.
//                 Special case 0,0 is entire run to this point
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/26/00
//-----------------------------------------------------------------------------
bool AnalysisManager::printLoopInfo(int start, int finish)
{
  return primaryAnalysisObject_->printLoopInfo(start, finish);
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::testDCOPOutputTime_
// Purpose       : Similar to testOutputTime, except that this is for
//                 DCOP restart files.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 10/18/07
//-----------------------------------------------------------------------------
///
/// Returns true if the DCOP Restart parameters have been set
///
/// @return true if the DCOP Restart parameters have been set
///
bool AnalysisManager::testDCOPOutputTime()
{
  return dcopRestartFlag_;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::testSaveOutputTime
// Purpose       : Similar to testOutputTime, except that this is for
//                 .SAVE files.
// Special Notes : Only outputs 1x.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 10/18/07
//-----------------------------------------------------------------------------
bool AnalysisManager::testSaveOutputTime()
{
  bool flag = true;

  if( !saveFlag_ )
  {
    flag = false;
  }
  else if (stepErrorControl_->currentTime < saveTime_)
  {
    flag = false;
  }
  else if (savedAlready_)
  {
    flag = false;
  }

  if (flag)
  {
    savedAlready_ = true;
    Xyce::dout() << "Calling SAVE outputs!" <<std::endl;
  }

  return flag;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::testRestartSaveTime_
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rob Hoekstra, SNL, Parallel ComputationalSciences.
// Creation Date : 07/31/01
//-----------------------------------------------------------------------------
bool
AnalysisManager::testRestartSaveTime()
{
  double initial_restart_interval = restartManager_.getInitialRestartInterval();
  const IO::IntervalVector &restart_intervals = restartManager_.getRestartIntervals();

  if (DEBUG_RESTART)
  {
    Xyce::dout() << "TESTING FOR RESTART SAVE" << std::endl
                 << Xyce::subsection_divider << std::endl
                 << "stepErrorControl_->currentTime: " << stepErrorControl_->currentTime << std::endl
                 << "nextSaveTime: " << nextRestartSaveTime_ << std::endl
                 << "initial_restart_interval: " << initial_restart_interval << std::endl;
    if (!restart_intervals.empty())
    {
      Xyce::dout() << "First restart interval: " << restart_intervals[0].first << std::endl;
    }
    else
    {
      Xyce::dout() << "restart_intervals is empty" << std::endl;
    }
  }

  bool flag = false;
  if (initial_restart_interval == 0.0)
  {
    flag = false;
  }
  else if (stepErrorControl_->currentTime < nextRestartSaveTime_)
  {
    flag = false;
  }
  else if (restart_intervals.empty())
  {
    while (nextRestartSaveTime_ <= stepErrorControl_->currentTime)
    {
      nextRestartSaveTime_ += initial_restart_interval;
    }
    flag = true;
  }
  else if (stepErrorControl_->currentTime < restart_intervals[0].first)
  {
    while (nextRestartSaveTime_ <= stepErrorControl_->currentTime)
    {
      nextRestartSaveTime_ += initial_restart_interval;
    }
    if (nextRestartSaveTime_ > restart_intervals[0].first)
    {
      nextRestartSaveTime_ = restart_intervals[0].first;
    }
    flag = true;
  }
  else
  {
    std::pair<double, double> currInterval, nextInterval;
    int size = restart_intervals.size();
    for (int i = 0; i < size; ++i)
    {
      if (restart_intervals[i].first <= stepErrorControl_->currentTime)
      {
        currInterval = restart_intervals[i];
        if ((i+1) < (int)restart_intervals.size())
        {
          nextInterval = restart_intervals[i+1];
        }
      }
    }
    int step = static_cast <int> ((stepErrorControl_->currentTime-currInterval.first) /
                                  currInterval.second);
    nextRestartSaveTime_ = currInterval.first + (step+1)*currInterval.second;

    if (nextInterval.first && (nextInterval.first!=currInterval.first)
        && (nextRestartSaveTime_>=nextInterval.first))
    {
      nextRestartSaveTime_ = nextInterval.first;
    }
    flag = true;
  }

  if (DEBUG_RESTART)
    Xyce::dout() << "new nextSaveTime: " << nextRestartSaveTime_ << std::endl
                 << "restart flag: " << flag << std::endl
                 << Xyce::subsection_divider << std::endl;

  return flag;
}


//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setOPAnalysisParams
// Purpose       : Handle OP statement. (.OP)
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/15/07
//-----------------------------------------------------------------------------
bool AnalysisManager::setOPAnalysisParams(
  const Util::OptionBlock &     paramsBlock)
{
  dotOpSpecified_ = true;

  return true;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setSaveOptions
// Purpose       : Sets the Save parameters.
// Special Notes : Most of these parameters are handled in the output manager,
//                 rather than here.  So, most params are a noop here, except
//                 for "TIME".
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/18/07
//-----------------------------------------------------------------------------
bool AnalysisManager::setSaveOptions(
  const Util::OptionBlock &     options)
{
  saveFlag_ = true;

  for (Util::ParamList::const_iterator it = options.begin(), end = options.end(); it != end; ++it)
    if ((*it).tag() == "TIME")
      saveTime_ = (*it).getImmutableValue<double>();

  return true;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setDCOPRestartParams
// Purpose       : Sets the DCOP restart parameters.
// Special Notes : Most of the dcop restart parameters are used and handled by
//                 the IO::OutputMgr class, so this function here doens't need
//                 to do very much.
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/18/07
//-----------------------------------------------------------------------------
bool AnalysisManager::setDCOPRestartParams(const Util::OptionBlock &option_block)
{
  dcopRestartFlag_ = true;

  for (Util::ParamList::const_iterator it = option_block.begin(), 
       end = option_block.end(); it != end; ++it)
  {
    if ((*it).tag() == "TIME")
    {
      saveTime_ = (*it).getImmutableValue<double>();
    }
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setSensOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, Sandia
// Creation Date : 6/5/13
//-----------------------------------------------------------------------------
bool AnalysisManager::setSensOptions(const Util::OptionBlock & OB)
{
  sensFlag_ = true;
  return true;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::completeOPStartStep
// Purpose       : Call to rotate next state to current state following a
//               : constrained DCOP solve when using a previous operating point
// Special Notes : Called from the Xyce-LOCA interface.
// Scope         : public
// Creator       : Dave Shirley
// Creation Date : 06/27/2006
//-----------------------------------------------------------------------------
bool AnalysisManager::completeOPStartStep( )
{
  bool bsuccess = true;

  bsuccess = dataStore_->updateStateDataArrays ();
  dataStore_->setConstantHistory();
  dataStore_->equateTmpVectors();

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::completeHomotopyStep
// Purpose       : Call to do final cleanup after a single,
//                 successful homotopy step.
// Special Notes : Called from the Xyce-LOCA interface.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/20/2006
//-----------------------------------------------------------------------------
bool
AnalysisManager::completeHomotopyStep(
  Loader::NonlinearEquationLoader &     loader,
  const std::vector<std::string> &      paramNames,
  const std::vector<double> &           paramVals,
  Linear::Vector *                      solnVecPtr )
{
  bool bsuccess = true;

  if (DEBUG_ANALYSIS)
    Xyce::dout() << "\n " << netlistFilename_
                 << " AnalysisManager::completeHomotopyStep " << std::endl;

  // Rotate the data vectors:
  bool bs1 = dataStore_->updateStateDataArrays ();
  bsuccess = bsuccess && bs1;
  dataStore_->setConstantHistory();
  dataStore_->equateTmpVectors();

  // Pass info in to the lower level solver:
  loader.homotopyStepSuccess(paramNames,paramVals);

  // Call output
  outputManagerAdapter_.outputHomotopy( paramNames, paramVals, *solnVecPtr );

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::failHomotopyStep
// Purpose       : Call to do final cleanup after a single,
//                 failed homotopy step.
// Special Notes : Called from the Xyce-LOCA interface.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/30/2006
//-----------------------------------------------------------------------------
bool AnalysisManager::failHomotopyStep(Loader::NonlinearEquationLoader &loader)
{
  if (DEBUG_ANALYSIS)
    Xyce::dout() << "\n " << netlistFilename_
                 << " AnalysisManager::failHomotopyStep " << std::endl;

// Pass info in to the lower level solver:
  loader.homotopyStepFailure();

  return true;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::initializeTransientModel
// Purpose       : ModelEvaluator Interface
// Special Notes :
// Scope         : private
// Creator       : Coffey, Schiek, Mei
// Creation Date : 05/29/09
//-----------------------------------------------------------------------------
void AnalysisManager::initializeTransientModel(
  const TimeIntg::TIAParams &   tia_params)
{
  createTimeIntegratorMethod(tia_params, TimeIntg::BackwardDifferentiation15::type);
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::evalTransientModel
// Purpose       : ModelEvaluator Interface
// Special Notes :
// Scope         : private
// Creator       : Coffey, Schiek, Mei
// Creation Date : 05/27/09
//-----------------------------------------------------------------------------
bool evalTransientModel(
  Loader::Loader &      loader,
  double                t,
  Linear::Vector *      SolVectorPtr,
  Linear::Vector *      CurrSolVectorPtr,
  Linear::Vector *      LastSolVectorPtr,
  Linear::Vector *      StaVectorPtr,
  Linear::Vector *      CurrStaVectorPtr,
  Linear::Vector *      LastStaVectorPtr,
  Linear::Vector *      StaDerivVectorPtr,
  Linear::Vector *      StoVectorPtr,
  Linear::Vector *      CurrStoVectorPtr,
  Linear::Vector *      LastStoVectorPtr,
  Linear::Vector *      stoLeadCurrQVectorPtr,
  Linear::Vector *      NextLeadFVectorPtr,
  Linear::Vector *      CurrLeadFVectorPtr,
  Linear::Vector *      LastLeadFVectorPtr,
  Linear::Vector *      LeadQVectorPtr,
  Linear::Vector *      NextJunctionVVectorPtr,
  Linear::Vector *      CurrJunctionVVectorPtr,
  Linear::Vector *      LastJunctionVVectorPtr,
  Linear::Vector *      QVectorPtr,
  Linear::Vector *      FVectorPtr,
  Linear::Vector *      BVectorPtr,
  Linear::Vector *      dFdxdVpVectorPtr,
  Linear::Vector *      dQdxdVpVectorPtr,
  Linear::Matrix *      dQdxMatrixPtr,
  Linear::Matrix *      dFdxMatrixPtr)
{
  // This is F,Q load:
  bool bsuccess = loader.loadDAEVectors(
    SolVectorPtr,
    CurrSolVectorPtr,
    LastSolVectorPtr,
    StaVectorPtr,
    CurrStaVectorPtr,
    LastStaVectorPtr,
    StaDerivVectorPtr,
    StoVectorPtr,
    CurrStoVectorPtr,
    LastStoVectorPtr,
    stoLeadCurrQVectorPtr,
    NextLeadFVectorPtr,
    CurrLeadFVectorPtr,
    LastLeadFVectorPtr,
    LeadQVectorPtr,
    NextJunctionVVectorPtr,
    CurrJunctionVVectorPtr,
    LastJunctionVVectorPtr, 
    QVectorPtr,
    FVectorPtr,
    BVectorPtr,
    dFdxdVpVectorPtr,
    dQdxdVpVectorPtr);

  // This is dQdx, dFdx load:
  bsuccess = bsuccess && loader.loadDAEMatrices(
    SolVectorPtr,
    StaVectorPtr,
    StaDerivVectorPtr,
    StoVectorPtr,
    dQdxMatrixPtr,
    dFdxMatrixPtr);

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::evalTransientModelState
// Purpose       : ModelEvaluator Interface
// Special Notes :
// Scope         : private
// Creator       : Coffey, Schiek, Mei
// Creation Date : 05/29/09
//-----------------------------------------------------------------------------
bool
evalTransientModelState(
  TimeIntg::StepErrorControl &  step_error_control,
  Loader::Loader &              loader,
  double                        t,
  Linear::Vector *                SolVectorPtr,
  Linear::Vector *                StaVectorPtr,
  Linear::Vector *                StoVectorPtr)
{
  // This is part of state vector load:
  step_error_control.currentTime = t;
  step_error_control.nextTime = t;
  loader.updateSources(); // updates source values wrt time t.
  bool bsuccess = loader.updateState(
    SolVectorPtr,
    SolVectorPtr,
    SolVectorPtr,
    StaVectorPtr,
    StaVectorPtr,
    StaVectorPtr,
    StoVectorPtr,
    StoVectorPtr,
    StoVectorPtr
                                       );
  return bsuccess;
}

// ***** Accessor methods *****
//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setBeginningIntegrationFlag
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date :
//-----------------------------------------------------------------------------
void AnalysisManager::setBeginningIntegrationFlag(bool bif)
{
  currentAnalysisStack_.back()->setBeginningIntegrationFlag(bif);
  primaryAnalysisObject_->setBeginningIntegrationFlag(bif);
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getBeginningIntegrationFlag
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date :
//-----------------------------------------------------------------------------
bool AnalysisManager::getBeginningIntegrationFlag() const
{
  return currentAnalysisStack_.back()->getBeginningIntegrationFlag();
//  return primaryAnalysisObject_->getBeginningIntegrationFlag();
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setIntegrationMethod
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date :
//-----------------------------------------------------------------------------
void AnalysisManager::setIntegrationMethod(int im)
{
  currentAnalysisStack_.back()->setIntegrationMethod(im);
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getIntegrationMethod
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date :
//-----------------------------------------------------------------------------
int AnalysisManager::getIntegrationMethod ()
{
  return currentAnalysisStack_.back()->getIntegrationMethod();
}


//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getTotalLinearSolutionTime
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date :
//-----------------------------------------------------------------------------
double AnalysisManager::getTotalLinearSolutionTime() const
{
  return primaryAnalysisObject_->getTotalLinearSolutionTime();
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getTotalResidualLoadTime
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date :
//-----------------------------------------------------------------------------
double AnalysisManager::getTotalResidualLoadTime() const
{
  return primaryAnalysisObject_->getTotalResidualLoadTime();
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getTotalJacobianLoadTime
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date :
//-----------------------------------------------------------------------------
double AnalysisManager::getTotalJacobianLoadTime() const
{
  return primaryAnalysisObject_->getTotalJacobianLoadTime();
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getDoubleDCOPEnabled ()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool AnalysisManager::getDoubleDCOPEnabled() const
{
  return primaryAnalysisObject_->getDoubleDCOPEnabled ();
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::isSimulationComplete
// Purpose       : return boolean signifying whether simulation complete or
//                 not.
//
// Special Notes :THIS VERSION IS ONLY VALID FOR TRANSIENT RUNS, where
//                 completion of the simulation means integration to final
//                 time.
// Scope         : public
// Creator       : Tom Russo, SNL, Component Information and Models
// Creation Date : 05/06/2004
//-----------------------------------------------------------------------------
bool AnalysisManager::isSimulationComplete()
{
  if (analysisMode_ == ANP_MODE_TRANSIENT)
  {
    return stepErrorControl_->isFinished();
  }
  else
  {
    Report::DevelFatal0().in("AnalysisManager::simulationComplete") 
      << "Called for non-transient run, not currently valid";
    return false;
  }
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getPauseTime
//
// Purpose       : return the time at which the simulation will pause
//
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Component Information and Models
// Creation Date : 04/29/2004
//-----------------------------------------------------------------------------
double AnalysisManager::getPauseTime() const
{
  return stepErrorControl_->pauseTime;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getStepNumber
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date : 12/16/2010
//-----------------------------------------------------------------------------
int AnalysisManager::getStepNumber() const
{
  if (!currentAnalysisStack_.empty())
    return currentAnalysisStack_.back()->getStepNumber();

  return 0;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setStepNumber
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date : 12/16/2010
//-----------------------------------------------------------------------------
void AnalysisManager::setStepNumber(int step)
{
  if (primaryAnalysisObject_)
    primaryAnalysisObject_->setStepNumber(step);
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getTranStepNumber
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date : 12/16/2010
//-----------------------------------------------------------------------------
int AnalysisManager::getTranStepNumber()
{
  int number = 0;

  if (analysisMode_ == ANP_MODE_TRANSIENT && primaryAnalysisObject_)
    number = primaryAnalysisObject_->getTranStepNumber();

  return number;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setTranStepNumber
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date : 12/16/2010
//-----------------------------------------------------------------------------
void AnalysisManager::setTranStepNumber (int step)
{
  if (primaryAnalysisObject_)
    primaryAnalysisObject_->setTranStepNumber(step);
}


//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getInitTranFlag
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date : 12/16/2010
//-----------------------------------------------------------------------------
bool AnalysisManager::getInitTranFlag() const
{
  return getStepNumber() <= 0;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getTime
// Purpose       : Gets the next time value.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/27/00
//-----------------------------------------------------------------------------
double AnalysisManager::getTime() const
{
  return stepErrorControl_->nextTime;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getFinalTime
// Purpose       : Gets the final time-step value.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/27/00
//-----------------------------------------------------------------------------
double AnalysisManager::getFinalTime() const
{
  return stepErrorControl_->finalTime;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getInitialTime
// Purpose       : Gets the initial time-step value.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/27/00
//-----------------------------------------------------------------------------
double AnalysisManager::getInitialTime() const
{
  return stepErrorControl_->initialTime;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getDCOPFlag
// Purpose       : Gets a flag indicating we are in steady state.
//                  (steady=true)
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 03/02/01
//-----------------------------------------------------------------------------
bool AnalysisManager::getDCOPFlag() const
{
  return primaryAnalysisObject_ ? primaryAnalysisObject_->getDCOPFlag() : false;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getTranOPFlag
// Purpose       : Gets a flag indicating we are in a DCOP calculation that is
//                 the initialization for a transient simulation.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 02/28/05
//-----------------------------------------------------------------------------
bool AnalysisManager::getTranOPFlag() const
{
  return ((analysisMode_ == ANP_MODE_TRANSIENT || primaryAnalysisObject_->isAnalysis(ANP_MODE_TRANSIENT))
          && (primaryAnalysisObject_->getIntegrationMethod()) == TimeIntg::NoTimeIntegration::type);
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getACOPFlag
// Purpose       : Gets a flag indicating we are in a DCOP calculation that is
//                 the initialization for a AC simulation.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 04/16/12
//-----------------------------------------------------------------------------
bool AnalysisManager::getACOPFlag() const
{
  bool return1 = (analysisMode_ == ANP_MODE_AC || primaryAnalysisObject_->isAnalysis(ANP_MODE_AC))
      && (primaryAnalysisObject_->getIntegrationMethod()) == TimeIntg::NoTimeIntegration::type;

  // if the analysis type is noise, this flag must also be set, as it is based on an 
  // AC-style calculation.
  bool return2 = (analysisMode_ == ANP_MODE_NOISE || primaryAnalysisObject_->isAnalysis(ANP_MODE_NOISE))
      && (primaryAnalysisObject_->getIntegrationMethod()) == TimeIntg::NoTimeIntegration::type;

  return (return1 || return2);
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getDCSweepFlag
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 02/28/05
//-----------------------------------------------------------------------------
bool AnalysisManager::getDCSweepFlag() const
{
  return ((analysisMode_ == ANP_MODE_DC_SWEEP || primaryAnalysisObject_->isAnalysis(ANP_MODE_DC_SWEEP)));
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getTransientFlag
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 02/28/05
//-----------------------------------------------------------------------------
bool AnalysisManager::getTransientFlag () const
{
  return (((analysisMode_ == ANP_MODE_TRANSIENT || primaryAnalysisObject_->isAnalysis(ANP_MODE_TRANSIENT))
     && (primaryAnalysisObject_->getIntegrationMethod()) != TimeIntg::NoTimeIntegration::type));
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getDoubleDCOPStep
// Purpose       : Gets the double DC Operating Point step value.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 04/25/01
//-----------------------------------------------------------------------------
int AnalysisManager::getDoubleDCOPStep() const
{
  return primaryAnalysisObject_->getDoubleDCOPStep();
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::updateDerivs
// Purpose       : Calls the time  int. method to update the corrector
//                 derivatives.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/13/00
//-----------------------------------------------------------------------------
bool AnalysisManager::updateDerivs()
{
  workingIntgMethod_->obtainCorrectorDeriv();

  return true;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::updateDivDiffs
// Purpose       : Updates the divided difference values.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/27/00
//-----------------------------------------------------------------------------
bool AnalysisManager::updateDivDiffs()
{
  dataStore_->computeDividedDifferences();

  return true;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::updateDerivsBlock
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 01/10/01
//-----------------------------------------------------------------------------
bool AnalysisManager::updateDerivsBlock(
  const std::list<IndexPair> & solGIDList,
  const std::list<IndexPair> & staGIDList)
{
  dataStore_->computeDivDiffsBlock(solGIDList, staGIDList);
  workingIntgMethod_->updateDerivsBlock (solGIDList, staGIDList);

  return true;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::registerElapsedTimer
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Dave Shirley, PSSI
// Creation Date : 02/24/06
//-----------------------------------------------------------------------------
bool AnalysisManager::registerElapsedTimer(Util::Timer * et)
{
  elapsedTimerPtr_ = et;

  return true;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getRestartDataSize
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rob Hoekstra, SNL, Computational Sciences
// Creation Date : 07/31/01
//-----------------------------------------------------------------------------
int AnalysisManager::getRestartDataSize( bool pack ) const
{
  return stepErrorControl_->getRestartDataSize( pack );
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::dumpRestartData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rob Hoekstra, SNL, Computational Sciences
// Creation Date : 07/31/01
//-----------------------------------------------------------------------------
bool AnalysisManager::dumpRestartData(
  char *                        buf,
  int                           bsize,
  int &                         pos,
  Parallel::Communicator *      comm,
  bool                          pack)
{
  return stepErrorControl_->dumpRestartData(buf, bsize, pos, comm, pack);
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::restoreRestartData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rob Hoekstra, SNL, Computational Sciences
// Creation Date : 07/31/01
//-----------------------------------------------------------------------------
bool
AnalysisManager::restoreRestartData(
  char *                        buf,
  int                           bsize,
  int &                         pos,
  Parallel::Communicator *      comm,
  bool                          pack)
{
  return stepErrorControl_->restoreRestartData(buf, bsize, pos, comm, pack, getTIAParams().initialTime);
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getSolnVarData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rob Hoekstra, SNL, Computational Sciences
// Creation Date : 07/31/01
//-----------------------------------------------------------------------------
bool AnalysisManager::getSolnVarData( const int & gid,
                                             std::vector<double> & varData ) const
{
  return dataStore_->getSolnVarData( gid, varData );
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getStateVarData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rob Hoekstra, SNL, Computational Sciences
// Creation Date : 07/31/01
//-----------------------------------------------------------------------------
bool
AnalysisManager::getStateVarData(
  const int &           gid,
  std::vector<double> & varData ) const
{
  return dataStore_->getStateVarData( gid, varData );
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getStoreVarData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date :
//-----------------------------------------------------------------------------
bool AnalysisManager::getStoreVarData( const int & gid,
                                              std::vector<double> & varData ) const
{
  return dataStore_->getStoreVarData( gid, varData );
}
//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setSolnVarData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rob Hoekstra, SNL, Computational Sciences
// Creation Date : 07/31/01
//-----------------------------------------------------------------------------
bool AnalysisManager::setSolnVarData( const int & gid,
                                             const std::vector<double> & varData )
{
  return dataStore_->setSolnVarData( gid, varData );
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setStateVarData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rob Hoekstra, SNL, Computational Sciences
// Creation Date : 07/31/01
//-----------------------------------------------------------------------------
bool AnalysisManager::setStateVarData( const int & gid,
                                              const std::vector<double> & varData )
{
  return dataStore_->setStateVarData( gid, varData );
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setStoreVarData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date :
//-----------------------------------------------------------------------------
bool AnalysisManager::setStoreVarData( const int & gid,
                                              const std::vector<double> & varData )
{
  return dataStore_->setStoreVarData( gid, varData );
}


const TimeIntg::TIAParams &
AnalysisManager::getTIAParams() const
{
  return primaryAnalysisObject_->getTIAParams();
}

TimeIntg::TIAParams &
AnalysisManager::getTIAParams()
{
  return primaryAnalysisObject_->getTIAParams();
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::registerLoader
// Purpose       : 
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Computational Sciences
// Creation Date : 04/18/02
//-----------------------------------------------------------------------------
bool AnalysisManager::setLoader(Loader::Loader &loader)
{
  loader_ = &loader;

  return true;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::registerParallelServices
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 04/19/02
//-----------------------------------------------------------------------------
bool AnalysisManager::registerParallelServices(N_PDS_Manager * pds)
{
  parallelManager_ = pds;

  return true;
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setNextSolVectorPtr
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter,SNL, Computational Sciences
// Creation Date : 04/22/2003
//-----------------------------------------------------------------------------
bool AnalysisManager::setNextSolVectorPtr (Linear::Vector * solVecPtr)
{
  return dataStore_->setNextSolVectorPtr (solVecPtr);
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setPauseTime
//
// Purpose       : Set the time at which to pause the simulation
//
// Special Notes : This is a temporary implementation that I'm using to
//                 begin the process of hiding time integrator internals from
//                 the "N_CIR_Xyce::simulateUntil" method, so that I can
//                 ultimately change those internals without the simulateUntil
//                 method
// Scope         : public
// Creator       : Tom Russo, SNL, Component Information and Models
// Creation Date : 04/29/2004
//-----------------------------------------------------------------------------
void
AnalysisManager::setPauseTime(
  double        pause_time,
  double        initial_time)
{
  stepErrorControl_->setBreakPoint(Util::BreakPoint(pause_time, Util::BreakPoint::PAUSE), initial_time);
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::getTimeIntInfo
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date :
//-----------------------------------------------------------------------------
void getTimeIntInfo(const AnalysisManager &analysis_manager, TimeIntg::TimeIntInfo & tiInfo)
{
  tiInfo.currentOrder         = analysis_manager.getWorkingIntegrationMethod().getOrder();
  tiInfo.numberOfSteps        = analysis_manager.getWorkingIntegrationMethod().getNumberOfSteps();
  tiInfo.usedOrder            = analysis_manager.getWorkingIntegrationMethod().getUsedOrder();
  tiInfo.nscsco               = analysis_manager.getWorkingIntegrationMethod().getNscsco();
  tiInfo.pdt                  = analysis_manager.getWorkingIntegrationMethod().partialTimeDeriv();

  tiInfo.nextTimeStep         = analysis_manager.getStepErrorControl().currentTimeStep;
  tiInfo.currTimeStep         = analysis_manager.getStepErrorControl().lastTimeStep;
  tiInfo.currentTime          = analysis_manager.getStepErrorControl().currentTime;
  tiInfo.nextTime             = analysis_manager.getStepErrorControl().nextTime;
  tiInfo.finalTime            = analysis_manager.getStepErrorControl().finalTime;
  tiInfo.startingTimeStep     = analysis_manager.getStepErrorControl().startingTimeStep;
  tiInfo.bpTol                = analysis_manager.getStepErrorControl().getBreakPointLess().tolerance_;

  tiInfo.dcopFlag             = analysis_manager.getAnalysisObject().getDCOPFlag();
  tiInfo.inputOPFlag          = analysis_manager.getAnalysisObject().getInputOPFlag();
  tiInfo.tranopFlag           = analysis_manager.getTranOPFlag();
  tiInfo.acopFlag             = analysis_manager.getACOPFlag();
  tiInfo.transientFlag        = analysis_manager.getTransientFlag();
  tiInfo.dcsweepFlag          = analysis_manager.getDCSweepFlag();
  tiInfo.sweepSourceResetFlag = analysis_manager.getSweepSourceResetFlag();

  tiInfo.timeStepNumber       = analysis_manager.getStepNumber();
  tiInfo.initTranFlag         = analysis_manager.getInitTranFlag();
  tiInfo.beginIntegrationFlag = analysis_manager.getBeginningIntegrationFlag();

  tiInfo.doubleDCOPStep       = analysis_manager.getDoubleDCOPStep();
  tiInfo.doubleDCOPEnabled    = analysis_manager.getDoubleDCOPEnabled();

  tiInfo.timeIntMode          = analysis_manager.getAnalysisObject().getIntegrationMethod();
}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::writeConductanceFile
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/06/2006
//-----------------------------------------------------------------------------
void writeConductanceFile(const std::vector<std::string> &device_names, Nonlinear::ConductanceExtractor &conductance_extractor, const std::string &filename)
{
  std::map<std::string,double> inputMap;

  // load inputMap from tiaParam.device_names option
  for (std::vector<std::string>::const_iterator it = device_names.begin(), end = device_names.end(); it != end; ++it) 
    inputMap[*it] = 0.0;

  // if (DEBUG_ANALYSIS && isActive(Diag::TIME_PARAMETERS))
  // {
  //   Xyce::dout() << "AnalysisManager::conductanceTest()" << std::endl;
  //   for (std::list<std::string>::const_iterator it = tia_params.device_names.begin(), end = tia_params.device_names.end(); it != end; ++it)
  //     Xyce::dout() << "current device name = \"" << *it
  //                  << "\" added to inputMap[ " << *it << " ] = " << inputMap[ *it ] << std::endl;
  // }

  int isize = inputMap.size();
  std::vector<double> outputVector(isize, 0.0);
  std::vector< std::vector<double> > jacobian(isize);
  for (int i = 0; i < isize; ++i)
  {
    jacobian[i].resize(isize, 0.0);
  }

  bool b1 = conductance_extractor.extract(inputMap, outputVector, jacobian);

  int iE1, iE2;
  int numElectrodes = isize;

  FILE *fp1;
  fp1 = fopen(filename.c_str(), "w");

  fprintf(fp1, "%s", "Conductance array: \n");
  fprintf(fp1,"%s", "              ");
  if (b1)
  {
    std::map<std::string,double>::const_iterator iterM = inputMap.begin();
    std::map<std::string,double>::const_iterator  endM = inputMap.end  ();
    for (iE2 = 0; iE2 < numElectrodes; ++iE2, ++iterM)
    {
      std::string srcname = iterM->first;
      fprintf(fp1, "\t%14s", srcname.c_str());
    }
    fprintf(fp1, "%s", "\n");

    iterM = inputMap.begin();
    for (iE1 = 0; iE1 < numElectrodes; ++iE1, ++iterM)
    {
      std::string srcname = iterM->first;
      fprintf(fp1,"%14s",srcname.c_str());
      for (iE2 = 0; iE2 < numElectrodes; ++iE2)
      {
        fprintf(fp1,"\t%14.4e",jacobian[iE1][iE2]);
      }
      fprintf(fp1,"%s", "\n");
    }
    fprintf(fp1,"%s", "\n");
  }
  else
  {
    fprintf(fp1,"%s", "\nConductance calculation failed!\n");
  }

  fclose(fp1);
}

//-----------------------------------------------------------------------------
// Namespace     : Unnamed
// Purpose       : Options parsing
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Mon Mar 16 14:24:58 2015
//-----------------------------------------------------------------------------

namespace {

// .RESULTS
struct ResultOptionsReg : public IO::PkgOptionsReg
{
  ResultOptionsReg(AnalysisManager &analysis_manager)
    : analysisManager_(analysis_manager)
  {}

  bool operator()(const Util::OptionBlock & option_block)
  {
    analysisManager_.getOutputManagerAdapter().addOutputResults(option_block);

    return true;
  }

private:
  AnalysisManager &     analysisManager_;
};

// .options OP_IO
struct DCOPRestartOptionsReg : public IO::PkgOptionsReg
{
  DCOPRestartOptionsReg( AnalysisManager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setDCOPRestartParams( options ); }

  AnalysisManager & manager_;
};


// .OP
struct OPAnalysisReg : public IO::PkgOptionsReg
{
  OPAnalysisReg( AnalysisManager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setOPAnalysisParams( options ); }

  AnalysisManager & manager_;
};

// .options SAVE
struct SaveOptionsReg : public IO::PkgOptionsReg
{
  SaveOptionsReg( AnalysisManager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setSaveOptions( options ); }

  AnalysisManager & manager_;
};

// .options SENS 
struct SensOptionsReg : public IO::PkgOptionsReg
{
  SensOptionsReg( AnalysisManager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setSensOptions( options ); }

  AnalysisManager & manager_;
};

} // namespace <unnamed>

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::registerPkgOptionsMgr
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, 1437
// Creation Date : 10/21/08
//-----------------------------------------------------------------------------
bool registerPkgOptionsMgr(AnalysisManager &analysis_manager, const std::string &netlist_filename, IO::PkgOptionsMgr &options_manager)
{
  options_manager.submitRegistration(
      "RESULT", netlist_filename, new ResultOptionsReg( analysis_manager ) );

  options_manager.submitRegistration(
      "OP", netlist_filename, new OPAnalysisReg( analysis_manager ) );

  options_manager.submitRegistration(
      "OP_IO", netlist_filename, new DCOPRestartOptionsReg( analysis_manager ) );

  options_manager.submitRegistration(
      "SAVE", netlist_filename, new SaveOptionsReg( analysis_manager ) );

  options_manager.submitRegistration(
      "SENS", netlist_filename, new SensOptionsReg( analysis_manager ) );

  return true;
}

//-----------------------------------------------------------------------------
// Function      : registerAnalysisFactory
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Mon Mar 16 14:26:41 2015
//-----------------------------------------------------------------------------
///
/// Registers the analysis factories.
///
/// In the future this is intended to become a pluggable factory allowing new analysis type to be added with no invasion
/// into the rest of the analysis code.  Whether anyone will actually create an analysis outside the system is academic
/// and creating this interface will result in defining the roles and interactions between the various analyses.
///
/// @invariant
///
/// @param netlist_filename 
/// @param options_manager 
/// @param analysis_manager 
/// @param linear_system 
/// @param nonlinear_manager 
/// @param device_manager 
/// @param builder 
/// @param topology 
///
///
void
registerAnalysisFactory(
  const std::string &           netlist_filename,
  IO::PkgOptionsMgr &           options_manager,
  AnalysisManager &             analysis_manager,
  Linear::System &              linear_system,
  Nonlinear::Manager &          nonlinear_manager,
  Device::DeviceMgr &           device_manager,
  Linear::Builder &             builder,
  Topo::Topology &              topology)
{
  registerDCSweepFactory(netlist_filename, options_manager, analysis_manager, linear_system, nonlinear_manager, topology);
  registerACFactory(netlist_filename, options_manager, analysis_manager, linear_system, nonlinear_manager, topology);
  registerTransientFactory(netlist_filename, options_manager, analysis_manager, linear_system, nonlinear_manager, topology);
  registerHBFactory(netlist_filename, options_manager, analysis_manager, linear_system, nonlinear_manager, device_manager, builder, topology);
  registerMPDEFactory(netlist_filename, options_manager, analysis_manager, linear_system, nonlinear_manager, device_manager, builder, topology);
  registerNOISEFactory(netlist_filename, options_manager, analysis_manager, linear_system, nonlinear_manager, topology);
  registerMORFactory(netlist_filename, options_manager, analysis_manager, linear_system, nonlinear_manager, topology);
  registerStepFactory(netlist_filename, options_manager, analysis_manager, linear_system, nonlinear_manager);
}

} // namespace Analysis
} // namespace Xyce
