//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_ANP_DCSweep.C,v $
// Purpose       : DC Sweep class analysis functions.
// Special Notes :
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 01/24/08
//
// Revision Information:
// ---------------------
// Revision Number: $Revision: 1.125.2.1 $
// Revision Date  : $Date: 2015/04/02 18:20:07 $
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------
#include <Xyce_config.h>

// ---------- Standard Includes ----------
#include <iostream>

// ----------   Xyce Includes   ----------
#include <N_ANP_AnalysisManager.h>
#include <N_ANP_DCSweep.h>
#include <N_ANP_HB.h>
#include <N_ANP_OutputMgrAdapter.h>
#include <N_ANP_SweepParam.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_IO_CmdParse.h>
#include <N_LOA_Loader.h>
#include <N_NLS_Manager.h>
#include <N_TIA_DataStore.h>
#include <N_TIA_NoTimeIntegration.h>
#include <N_TIA_StepErrorControl.h>
#include <N_TIA_WorkingIntegrationMethod.h>
#include <N_TOP_Topology.h>
#include <N_UTL_Diagnostic.h>
#include <N_UTL_ExtendedString.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_OptionBlock.h>
#include <N_ERH_Message.h>

#include <Teuchos_RCP.hpp>

namespace Xyce {
namespace Analysis {

DCSweep::DCSweep(
  AnalysisManager &     analysis_manager,
  Nonlinear::Manager &  nonlinear_manager,
  Topo::Topology &      topology,
  HB *                  hb_analysis)
  : AnalysisBase(analysis_manager, "DC Sweep"),
    analysisManager_(analysis_manager),
    loader_(analysis_manager.getLoader()),
    nonlinearManager_(nonlinear_manager),
    topology_(topology),
    outputManagerAdapter_(analysis_manager.getOutputManagerAdapter()),
    tiaParams_(),
    hbAnalysis_(hb_analysis),
    sensFlag_(analysis_manager.getSensFlag()),
    dcLoopInitialized_(false),
    dcLoopSize_(0)
{}

DCSweep::~DCSweep()
{}

//-----------------------------------------------------------------------------
// Function      : AnalysisManager::setTranOptions
// Purpose       :
// Special Notes : These are from '.options timeint'
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 04/18/02
//-----------------------------------------------------------------------------
bool
DCSweep::setTimeIntegratorOptions(
  const Util::OptionBlock &     option_block)
{
  for (Util::ParamList::const_iterator it = option_block.begin(), end = option_block.end(); it != end; ++it)
  {
    const Util::Param &param = (*it);

    if (param.uTag() == "DAESTATEDERIV")
      analysisManager_.setDAEStateDerivFlag(static_cast<bool> (param.getImmutableValue<int>()));
    else if (param.uTag() == "DEBUGLEVEL")
      IO::setTimeIntegratorDebugLevel(analysisManager_.getCommandLine(), param.getImmutableValue<int>());
    else if (nonlinearManager_.setReturnCodeOption(param))
      ;
    else if (tiaParams_.setTimeIntegratorOption(param))
      ;
    else if (setDCOPOption(param))
      ;
    else if (param.uTag() == "METHOD")
      ;
    else
      Report::UserError() << param.uTag() << " is not a recognized time integration option";
  }

  return true;
}


//-----------------------------------------------------------------------------
// Function      : DCSweep::setAnalysisParams
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date : 6/22/10
//-----------------------------------------------------------------------------
bool DCSweep::setAnalysisParams(const Util::OptionBlock & paramsBlock)
{
  dcSweepVector_.push_back(parseSweepParams(paramsBlock.begin(), paramsBlock.end()));

  return true;
}

//-----------------------------------------------------------------------------
// Function      : DCSweep::outputFailureStats
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 6/21/2010
//-----------------------------------------------------------------------------
bool DCSweep::outputFailureStats(std::ostream &os)
{
  if (!(dcSweepFailures_.empty()))
  {
    os << "\tFailed DC sweep steps:\t\t" << std::endl;

    for (std::vector<int>::iterator iter = dcSweepFailures_.begin(); iter != dcSweepFailures_.end(); ++iter)
    {
      os << "\t\tDC Step # " << *iter << std::endl;
    }
  }
  return true;
}

//-----------------------------------------------------------------------------
// Function      : DCSweep::run()
// Purpose       : This is the main controlling loop for DC sweep analysis.
//                 This loop calls a series of operating point calculations
//                 (ie calculations in which there is no time integration,
//                 and the time integration method is always set to "none").
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 1/28/08
//-----------------------------------------------------------------------------
bool DCSweep::doRun()
{
  return doInit() && doLoopProcess() && doFinish();
}

//-----------------------------------------------------------------------------
// Function      : DCSweep::init()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 1/28/08
//-----------------------------------------------------------------------------
bool DCSweep::doInit()
{
  bool bsuccess = true;

  if (sensFlag_)
  {
    Stats::StatTop _sensitivityStat("Sensitivity");

    nonlinearManager_.enableSensitivity(*analysisManager_.getDataStore(), *analysisManager_.getPDSManager(), topology_);
  }

  // set up the various sweep variables
  // the following should only happen once, but must be isolated to
  // prevent a step loop outside of this dc sweep from causing to
  // occur more often
  if( !dcLoopInitialized_ )
  {
    if (DEBUG_ANALYSIS && isActive(Diag::TIME_PARAMETERS))
    {
      Xyce::dout() << std::endl << std::endl
                   << Xyce::subsection_divider << std::endl
                   << "DCSweep::run()" << std::endl;
    }

    dcLoopSize_ = setupSweepLoop(loader_, dcSweepVector_.begin(), dcSweepVector_.end());

    outputManagerAdapter_.setDCAnalysisMaxSteps( dcLoopSize_ );
    outputManagerAdapter_.setDCSweepVector(dcSweepVector_);

    dcLoopInitialized_ = true;
  }

  //setup for operating pt calculation
  baseIntegrationMethod_ = TimeIntg::NoTimeIntegration::type;
  analysisManager_.createTimeIntegratorMethod(tiaParams_, baseIntegrationMethod_);

  stepNumber = 0;
  doubleDCOPFlag_ = loader_.getDoubleDCOPFlag();
  doubleDCOPStep_ = firstDCOPStep_;

  initializeSolution_();

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : DCSweep::initializeSolution()
// Purpose       : Move solution-initialization steps to separate routine so
//                 the process may be repeated cleanly when necessary.
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical and Microsystem Modeling
// Creation Date : 2/17/2010
//-----------------------------------------------------------------------------
void DCSweep::initializeSolution_()
{
  // set initial guess, if there is one to be set.
  // this setInitialGuess call is to up an initial guess in the
  // devices that have them (usually PDE devices).  This is different than
  // the "intializeProblem" call, which sets IC's.  (initial conditions are
  // different than initial guesses.
  loader_.setInitialGuess(analysisManager_.getDataStore()->nextSolutionPtr);

  // If available, set initial solution (.IC, .NODESET, etc).
  setInputOPFlag(
    outputManagerAdapter_.setupInitialConditions(
      *analysisManager_.getDataStore()->nextSolutionPtr,
      *analysisManager_.getDataStore()->flagSolutionPtr));

  // Set a constant history for operating point calculation
  analysisManager_.getDataStore()->setConstantHistory();
  analysisManager_.getDataStore()->computeDividedDifferences();
  analysisManager_.getWorkingIntegrationMethod().obtainCorrectorDeriv();
}

//-----------------------------------------------------------------------------
// Function      : DCSweep::loopProcess()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 1/28/08
//-----------------------------------------------------------------------------
bool DCSweep::doLoopProcess()
{
  static_cast<Xyce::Util::Notifier<AnalysisEvent> &>(analysisManager_).publish(AnalysisEvent(AnalysisEvent::INITIALIZE, AnalysisEvent::DC));

  int currentStep = 0;
  int finalStep = dcLoopSize_;
  while (currentStep < finalStep)
  {
    outputManagerAdapter_.setDCAnalysisStepNumber(currentStep);

    if (VERBOSE_TIME)
    {
      printStepHeader(Xyce::dout());
      analysisManager_.getStepErrorControl().outputTimeInfo(Xyce::dout());
    }

    bool reset = updateSweepParams(loader_, currentStep, dcSweepVector_.begin(), dcSweepVector_.end(), true);

    // Tell the manager if any of our sweeps are being reset in this loop iteration.
    analysisManager_.setSweepSourceResetFlag(reset);

    outputManagerAdapter_.setDCSweepVector(dcSweepVector_);

    if (currentStep != 0 && reset)
    {
      analysisManager_.getDataStore()->setZeroHistory();
      initializeSolution_();
    }

    // Perform the step:
    static_cast<Xyce::Util::Notifier<AnalysisEvent> &>(analysisManager_).publish(AnalysisEvent(AnalysisEvent::STEP_STARTED, AnalysisEvent::DC, 0.0, currentStep));

    takeStep_();

    // Set things up for the next time step, based on if this one was
    // successful.
    if (analysisManager_.getStepErrorControl().stepAttemptStatus)
    {
      static_cast<Xyce::Util::Notifier<AnalysisEvent> &>(analysisManager_).publish(AnalysisEvent(AnalysisEvent::STEP_SUCCESSFUL, AnalysisEvent::DC, 0.0, currentStep));
      processSuccessfulStep();
    }
    else // stepAttemptStatus  (ie do this if the step FAILED)
    {
      static_cast<Xyce::Util::Notifier<AnalysisEvent> &>(analysisManager_).publish(AnalysisEvent(AnalysisEvent::STEP_FAILED, AnalysisEvent::DC, 0.0, currentStep));
      processFailedStep();
    }

    // we don't really control the loop counter here.  processSuccessfulStep
    // and processFailedStep do that work.  This needs to be cleaned up. RLS
    currentStep = stepNumber;
  } // end of sweep loop

  static_cast<Xyce::Util::Notifier<AnalysisEvent> &>(analysisManager_).publish(AnalysisEvent(AnalysisEvent::FINISH, AnalysisEvent::DC));

  return true;
}

//-----------------------------------------------------------------------------
// Function      : DCSweep::processSuccessfulStep()
//
// Purpose       : Used by both function dcSweepLoop and 2-level solve
//                 function calls to process successful DC steps.
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 1/28/08
//-----------------------------------------------------------------------------
bool DCSweep::doProcessSuccessfulStep()
{
  Stats::StatTop _processSuccessfulStepStat("Successful Step");
  Stats::TimeBlock _processSuccessfulStepTimer(_processSuccessfulStepStat);

  loader_.stepSuccess(TWO_LEVEL_MODE_DC_SWEEP); // analysisManager_.getTwoLevelMode());

  // This output call is for device-specific output, such as .OP,
  // or internal plot output from PDE(TCAD) devices.
  loader_.output();

  if (sensFlag_ && !firstDoubleDCOPStep() )
  {
    nonlinearManager_.calcSensitivity(objectiveVec_, dOdpVec_, dOdpAdjVec_, scaled_dOdpVec_, scaled_dOdpAdjVec_);
  }

  // Do some statistics, as long as this isn't the first "double"
  // DCOP step. (that one doesn't count)
  if ( !firstDoubleDCOPStep() )
  {
    stepNumber += 1;
    stats_.successStepsThisParameter_ += 1;
    stats_.successfulStepsTaken_ += 1;
  }

  // update the data arrays, output:
  analysisManager_.getDataStore()->updateSolDataArrays();

  if (DEBUG_ANALYSIS & DEBUG_TIME && isActive(Diag::TIME_DUMP_SOLUTION_ARRAYS))
    analysisManager_.getDataStore()->outputSolDataArrays(Xyce::dout());

  dcSweepOutput();

  // now that output has been called, update the doubleDCOP step
  // if neccessary. (pde-only)
  doubleDCOPStep_ = lastDCOPStep_;

  return true;
}

//-----------------------------------------------------------------------------
// Function      : DCSweep::processFailedStep()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 1/28/08
//-----------------------------------------------------------------------------
bool DCSweep::doProcessFailedStep()
{
  Stats::StatTop _processFailedStat("Failed Steps");
  Stats::TimeBlock _processFailedTimer(_processFailedStat);

  loader_.stepFailure(analysisManager_.getTwoLevelMode());

  stepNumber += 1;
  dcSweepFailures_.push_back(stepNumber);
  stats_.failedStepsAttempted_  += 1;
  analysisManager_.getStepErrorControl().numberSuccessiveFailures += 1;

  return true;
}

//-----------------------------------------------------------------------------
// Function      : DCSweep::doFinish()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 1/28/08
//-----------------------------------------------------------------------------
bool DCSweep::doFinish()
{
  bool bsuccess = true;

  if (DEBUG_ANALYSIS)
    Xyce::dout() << "Calling DCSweep::doFinish() outputs!" << std::endl;

  outputManagerAdapter_.finishOutput();
  if (!(dcSweepFailures_.empty()))
  {
    bsuccess = false;
  }
  return bsuccess;
}


//-----------------------------------------------------------------------------
// Function      : DCSweep::doHandlePredictor
// Purpose       :
// Special Notes :
// Scope         : private
// Creator       : Eric Keiter, SNL
// Creation Date : 06/24/2013
//-----------------------------------------------------------------------------
bool DCSweep::doHandlePredictor()
{
  analysisManager_.getDataStore()->setErrorWtVector(tiaParams_);
  analysisManager_.getWorkingIntegrationMethod().obtainPredictor();
  analysisManager_.getWorkingIntegrationMethod().obtainPredictorDeriv();

  // In case this is the upper level of a 2-level sim, tell the
  // inner solve to do its prediction:
  loader_.startTimeStep ();

  return true;
}

//-----------------------------------------------------------------------------
// Function      : DCSweep::takeStep_
// Purpose       : Take a DC Sweep integration step.
// Special Notes :
// Scope         : private
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 01/24/08
//-----------------------------------------------------------------------------
void DCSweep::takeStep_()
{
  { // Integration step predictor
    Stats::StatTop _predictorStat("Predictor");
    Stats::TimeBlock _predictorTimer(_predictorStat);

    doHandlePredictor();
  }

  { // Load B/V source devices with time data
    Stats::StatTop _updateDeviceSourceStat("Update Device Sources");
    Stats::TimeBlock _updateDeviceSourceTimer(_updateDeviceSourceStat);

    loader_.updateSources();
  }

  { // Nonlinear solve
    Stats::StatTop _nonlinearSolveStat("Solve");
    Stats::TimeBlock _nonlinearSolveTimer(_nonlinearSolveStat);

    analysisManager_.getStepErrorControl().newtonConvergenceStatus = nonlinearManager_.solve();
  }

  { // Add change to solution
    Stats::StatTop _errorStat("Error Estimation");
    Stats::TimeBlock _errorTimer(_errorStat);

    analysisManager_.getDataStore()->stepLinearCombo();

    gatherStepStatistics(stats_, nonlinearManager_.getNonlinearSolver(), analysisManager_.getStepErrorControl().newtonConvergenceStatus);

    analysisManager_.getStepErrorControl().evaluateStepError(tiaParams_);
  }
}

//-----------------------------------------------------------------------------
// Function      : DCSweep::twoLevelStep
//
// Purpose       : Used by 2-level Newton solves to execute a single DC sweep
//                 step.
//
// Special Notes : This is mostly what happens on the inner loop of
//                 dcSweepLoop, except that DC parameters are not updated,
//                 and success/failure of the step is not determined.
//
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/10/06
//-----------------------------------------------------------------------------
bool DCSweep::twoLevelStep()
{
  loader_.updateSources();
  analysisManager_.getStepErrorControl().newtonConvergenceStatus = nonlinearManager_.solve();
  analysisManager_.getDataStore()->stepLinearCombo();
  gatherStepStatistics(stats_, nonlinearManager_.getNonlinearSolver(), analysisManager_.getStepErrorControl().newtonConvergenceStatus);
  analysisManager_.getStepErrorControl().evaluateStepError(tiaParams_);

  return analysisManager_.getStepErrorControl().stepAttemptStatus;
}

//-----------------------------------------------------------------------------
// Function      : DCSweep::dcSweepOutput
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 1/31/08
//-----------------------------------------------------------------------------
void DCSweep::dcSweepOutput()
{
  if (hbAnalysis_)
  {
    std::vector<double> timePoints, freqPoints;
    Teuchos::RCP<Linear::BlockVector> timeDomainSolnVec;
    Teuchos::RCP<Linear::BlockVector> freqDomainSolnVecReal;
    Teuchos::RCP<Linear::BlockVector> freqDomainSolnVecImaginary;
    Teuchos::RCP<Linear::BlockVector> timeDomainStoreVec;
    Teuchos::RCP<Linear::BlockVector> freqDomainStoreVecReal;
    Teuchos::RCP<Linear::BlockVector> freqDomainStoreVecImaginary;
    Teuchos::RCP<Linear::BlockVector> timeDomainLeadCurrentVec;
    Teuchos::RCP<Linear::BlockVector> freqDomainLeadCurrentVecReal;
    Teuchos::RCP<Linear::BlockVector> freqDomainLeadCurrentVecImaginary;
    Teuchos::RCP<Linear::BlockVector> timeDomainJunctionVoltageVec;
    Teuchos::RCP<Linear::BlockVector> freqDomainJunctionVoltageVecReal;
    Teuchos::RCP<Linear::BlockVector> freqDomainJunctionVoltageVecImaginary;

    // RLS Todo:  Need to update HB call here 
    // for leadcurrent and junctionvoltage vector 


    hbAnalysis_->prepareHBOutput(
      *(analysisManager_.getDataStore()->currSolutionPtr),
      timePoints,
      freqPoints,
      timeDomainSolnVec,
      freqDomainSolnVecReal,
      freqDomainSolnVecImaginary,
      timeDomainStoreVec,
      freqDomainStoreVecReal,
      freqDomainStoreVecImaginary,
      timeDomainLeadCurrentVec,
      freqDomainLeadCurrentVecReal,
      freqDomainLeadCurrentVecImaginary,
      timeDomainJunctionVoltageVec,
      freqDomainJunctionVoltageVecReal,
      freqDomainJunctionVoltageVecImaginary);

    outputManagerAdapter_.outputHB(
      timePoints,
      freqPoints,
      *timeDomainSolnVec,
      *freqDomainSolnVecReal,
      *freqDomainSolnVecImaginary,
      *timeDomainStoreVec,
      *freqDomainStoreVecReal,
      *freqDomainStoreVecImaginary,
      *timeDomainLeadCurrentVec,
      *freqDomainLeadCurrentVecReal,
      *freqDomainLeadCurrentVecImaginary,
      *timeDomainJunctionVoltageVec,
      *freqDomainJunctionVoltageVecReal,
      *freqDomainJunctionVoltageVecImaginary);
  }

  // Make sure this isn't the NLP step of a PDE DCOP.
  else if (!firstDoubleDCOPStep())
  {
    // conventional .PRINT output
    outputManagerAdapter_.dcOutput(
      stepNumber,
      *(analysisManager_.getDataStore()->currSolutionPtr),
      *(analysisManager_.getDataStore()->currStatePtr),
      *(analysisManager_.getDataStore()->currStorePtr),
      *(analysisManager_.getDataStore()->currLeadCurrentPtr),
      *(analysisManager_.getDataStore()->currLeadDeltaVPtr),
      objectiveVec_,
      dOdpVec_, dOdpAdjVec_, scaled_dOdpVec_, scaled_dOdpAdjVec_);

    // output for DCOP restart
    outputManagerAdapter_.outputDCOP(*analysisManager_.getDataStore()->currSolutionPtr);
  }

}

//-----------------------------------------------------------------------------
// Function      : DCSweep::printStepHeader()
// Purpose       : Prints out time step information.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/26/00
//-----------------------------------------------------------------------------
void DCSweep::printStepHeader(std::ostream &os)
{
  if (VERBOSE_TIME)
    dout() << std::endl << std::endl
           << "***** "<< (DEBUG_ANALYSIS ? analysisManager_.getNetlistFilename() : "")
           << "  Start of DCOP STEP                        # " << stepNumber+1
           << std::endl << std::endl;
}

//-----------------------------------------------------------------------------
// Function      : DCSweep::printLoopInfo
// Purpose       :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 12/14/2010
//-----------------------------------------------------------------------------
bool DCSweep::printLoopInfo(int start, int finish)
{
  bool bsuccess = AnalysisBase::printLoopInfo(start, finish);
  if (start == 0 && finish == 0)
  {
    outputFailureStats(Xyce::lout());
  }
  return bsuccess;
}

namespace {

//-----------------------------------------------------------------------------
// Class         : DCSweepFactory
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Thu Jan 29 12:53:02 2015
//-----------------------------------------------------------------------------
///
/// Factory for parsing DCSweep parameters from the netlist and creating DCSweep analysis.
///
class DCSweepFactory : public Factory<DCSweep>
{
public:
  //-----------------------------------------------------------------------------
  // Function      : DCSweepFactory
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:54:09 2015
  //-----------------------------------------------------------------------------
  ///
  /// Constructs the DCSweep analysis factory
  ///
  /// @invariant Stores the results of parsing.  Multiple DCSweep analysis options may be
  /// applied and each generates and additional dcsweep.
  ///
  /// @invariant The existence of the parameters specified in the constructor cannot
  /// change.
  ///
  /// @param analysis_manager 
  /// @param linear_system 
  /// @param nonlinear_manager 
  /// @param topology 
  ///
  DCSweepFactory(
    Analysis::AnalysisManager & analysis_manager,
    Linear::System &            linear_system,
    Nonlinear::Manager &        nonlinear_manager,
    Topo::Topology &            topology)
    : Factory<DCSweep>(),
      analysisManager_(analysis_manager),
      linearSystem_(linear_system),
      nonlinearManager_(nonlinear_manager),
      topology_(topology)
  {}

  virtual ~DCSweepFactory()
  {}

  //-----------------------------------------------------------------------------
  // Function      : create
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:59:00 2015
  //-----------------------------------------------------------------------------
  ///
  /// Create a new DCSweep analysis and applies the analysis and time integrator option blocks.
  ///
  /// @return new DCSweep analysis object
  ///
  ///
  DCSweep *create() const
  {
    analysisManager_.setAnalysisMode(ANP_MODE_DC_SWEEP);
    DCSweep *dc_sweep = new DCSweep(analysisManager_, nonlinearManager_, topology_);
    for (std::vector<Util::OptionBlock>::const_iterator it = dcSweepAnalysisOptionBlock_.begin(), end = dcSweepAnalysisOptionBlock_.end(); it != end; ++it)
      dc_sweep->setAnalysisParams(*it);
    dc_sweep->setTimeIntegratorOptions(timeIntegratorOptionBlock_);

    return dc_sweep;
  }

  //-----------------------------------------------------------------------------
  // Function      : setDCSweepAnalysisOptionBlock
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 13:00:14 2015
  //-----------------------------------------------------------------------------
  ///
  /// Saves the analysis parsed options block in the factory.
  ///
  /// @invariant Appends to any previously specified analysis option block.
  ///
  /// @param option_block parsed option block
  ///
  void setDCSweepAnalysisOptionBlock(const Util::OptionBlock &option_block)
  {
    for (std::vector<Util::OptionBlock>::iterator it = dcSweepAnalysisOptionBlock_.begin(), end = dcSweepAnalysisOptionBlock_.end(); it != end; ++it)
    {
      if ((*it).compareParamLists(option_block))
      {
        (*it) = option_block;
        return;
      }
    }

    // save the new one.
    dcSweepAnalysisOptionBlock_.push_back(option_block); // save a copy for later.
  }

  //-----------------------------------------------------------------------------
  // Function      : setTimeIntegratorOptionBlock
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 13:01:27 2015
  //-----------------------------------------------------------------------------
  ///
  /// Saves the time integrator parsed option block.
  ///
  /// @invariant Overwrites any previously specified time integrator option block.
  ///
  /// @param option_block parsed option block
  ///
  void setTimeIntegratorOptionBlock(const Util::OptionBlock &option_block)
  {
    timeIntegratorOptionBlock_ = option_block;
  }

public:
  AnalysisManager &                     analysisManager_;
  Linear::System &                      linearSystem_;
  Nonlinear::Manager &                  nonlinearManager_;
  Topo::Topology &                      topology_;

private:
  std::vector<Util::OptionBlock>        dcSweepAnalysisOptionBlock_;
  Util::OptionBlock                     timeIntegratorOptionBlock_;
};


// .options TIMEINT
struct TimeIntegratorOptionsReg : public IO::PkgOptionsReg
{
  TimeIntegratorOptionsReg(
    DCSweepFactory &          factory)
    : factory_(factory)
  {}

  bool operator()(const Util::OptionBlock &option_block)
  {
    factory_.setTimeIntegratorOptionBlock(option_block);

    return true;
  }

  DCSweepFactory &            factory_;
};

// .DC
struct DCSweepAnalysisReg : public IO::PkgOptionsReg
{
  DCSweepAnalysisReg(
    DCSweepFactory &          factory)
    : factory_(factory)
  {}

  bool operator()(const Util::OptionBlock &option_block)
  {
    factory_.setDCSweepAnalysisOptionBlock(option_block);

    factory_.analysisManager_.addAnalysis(&factory_);

    return true;
  }

  DCSweepFactory &            factory_;
};

} // namespace <unnamed>

bool
registerDCSweepFactory(
  const std::string &           netlist_filename,
  IO::PkgOptionsMgr &           options_manager,
  AnalysisManager &             analysis_manager,
  Linear::System &              linear_system,
  Nonlinear::Manager &          nonlinear_manager,
  Topo::Topology &              topology)
{
  DCSweepFactory *factory = new DCSweepFactory(analysis_manager, linear_system, nonlinear_manager, topology);

  analysis_manager.addAnalysisFactory(factory);

  options_manager.submitRegistration("DC", netlist_filename, new DCSweepAnalysisReg(*factory));
  options_manager.submitRegistration("TIMEINT", netlist_filename, new TimeIntegratorOptionsReg(*factory));

  return true;
}


} // namespace Analysis
} // namespace Xyce
