//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_ANP_MPDE.C,v $
// Purpose       : MPDE analysis functions.
// Special Notes :
// Creator       : Todd Coffey, 1414, Ting Mei, 1437
// Creation Date : 07/23/08
//
// Revision Information:
// ---------------------
// Revision Number: $Revision: 1.45.2.1 $
// Revision Date  : $Date: 2015/04/02 18:20:07 $
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------
#include <Xyce_config.h>

#include <N_ANP_AnalysisManager.h>
#include <N_ANP_OutputMgrAdapter.h>
#include <N_ANP_MPDE.h>
#include <N_DEV_DeviceMgr.h>
#include <N_IO_CmdParse.h>
#include <N_IO_OutputMgr.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_LOA_Loader.h>
#include <N_MPDE_Manager.h>
#include <N_TIA_DataStore.h>
#include <N_TIA_StepErrorControl.h>
#include <N_TIA_NoTimeIntegration.h>

namespace Xyce {
namespace Analysis {

//-----------------------------------------------------------------------------
// Function      : MPDE::MPDE( AnalysisManager * )
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, SNL
// Creation Date : 3/11/06
//-----------------------------------------------------------------------------
MPDE::MPDE(
  AnalysisManager &     analysis_manager,
  Linear::System &      linear_system,
  Nonlinear::Manager &  nonlinear_manager,
  Device::DeviceMgr &   device_manager,
  Linear::Builder &     builder,
  Topo::Topology &      topology)
  : AnalysisBase(analysis_manager, "MPDE"),
    StepEventListener(&analysis_manager),
    analysisManager_(analysis_manager),
    loader_(analysis_manager.getLoader()),
    linearSystem_(linear_system),
    nonlinearManager_(nonlinear_manager),
    topology_(topology),
    mpdeManager_(new N_MPDE_Manager(analysisManager_, device_manager, builder, topology, analysisManager_.getCommandLine()))
{}

void MPDE::notify(const StepEvent &event)
{
  if (event.state_ == StepEvent::STEP_STARTED)
  {
    AnalysisBase::resetForStepAnalysis();
  }
}

//-----------------------------------------------------------------------------
// Function      : MPDE::~MPDE
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
MPDE::~MPDE()
{
  delete mpdeManager_;
}

const TimeIntg::TIAParams &
MPDE::getTIAParams() const
{
  return mpdeManager_->getTIAParams();
}

TimeIntg::TIAParams &
MPDE::getTIAParams()
{
  return mpdeManager_->getTIAParams();
}

//-----------------------------------------------------------------------------
// Function      : MPDE::getDCOPFlag()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 3/24/2014
//-----------------------------------------------------------------------------
bool MPDE::getDCOPFlag() const
{
  return getIntegrationMethod() == TimeIntg::NoTimeIntegration::type;
}

//-----------------------------------------------------------------------------
// Function      : MPDE::run()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, SNL
// Creation Date : 3/11/06
//-----------------------------------------------------------------------------
bool MPDE::doRun()
{
  mpdeManager_->registerApplicationLoader(&loader_);
  mpdeManager_->run(linearSystem_, nonlinearManager_, topology_);

  return true;
}

//-----------------------------------------------------------------------------
// Function      : MPDE::init()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, SNL
// Creation Date : 3/11/06
//-----------------------------------------------------------------------------
bool MPDE::doInit()
{
//  analysisManager_.getMPDEManager()->initializeAll();

  return false;
}

//-----------------------------------------------------------------------------
// Function      : MPDE::loopProcess()
// Purpose       : Conduct the time stepping loop.
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, SNL
// Creation Date : 3/11/06
//-----------------------------------------------------------------------------
bool MPDE::doLoopProcess()
{
  return false;
}

//-----------------------------------------------------------------------------
// Function      : MPDE::processSuccessfulDCOP()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, SNL
// Creation Date : 3/11/06
//-----------------------------------------------------------------------------
bool MPDE::processSuccessfulDCOP()
{
  return false;
}

//-----------------------------------------------------------------------------
// Function      : MPDE::processSuccessfulStep()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, SNL
// Creation Date : 3/11/06
//-----------------------------------------------------------------------------
bool MPDE::doProcessSuccessfulStep()
{
  return false;
}

//-----------------------------------------------------------------------------
// Function      : MPDE::processFailedStep
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, SNL
// Creation Date : 3/11/06
//-----------------------------------------------------------------------------
bool MPDE::doProcessFailedStep()
{
  return false;
}

//-----------------------------------------------------------------------------
// Function      : MPDE::processFailedDCOP
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, SNL
// Creation Date : 3/11/06
//-----------------------------------------------------------------------------
bool MPDE::processFailedDCOP()
{
  return false;
}

//-----------------------------------------------------------------------------
// Function      : MPDE::finish
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, SNL
// Creation Date : 3/11/06
//-----------------------------------------------------------------------------
bool MPDE::doFinish()
{
  return false;
}

bool MPDE::doHandlePredictor()
{
  return true;
}

//-----------------------------------------------------------------------------
// Function      : MPDE::finalVerboseOutput
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, SNL
// Creation Date : 3/11/06
//-----------------------------------------------------------------------------
bool MPDE::finalVerboseOutput()
{
  return false;
}

namespace {

//-----------------------------------------------------------------------------
// Class         : MPDEFactory
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Thu Jan 29 12:53:02 2015
//-----------------------------------------------------------------------------
///
/// Factory for parsing MPDE parameters from the netlist and creating MPDE analysis.
///
class MPDEFactory : public Factory<MPDE>
{
public:
  //-----------------------------------------------------------------------------
  // Function      : MPDEFactory
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:54:09 2015
  //-----------------------------------------------------------------------------
  ///
  /// Constructs the MPDE analysis factory
  ///
  /// @invariant Stores the results of parsing, so if more than one of the analysis and
  /// associated lines are parsed, the second options simply overwrite the previously parsed
  /// values.
  ///
  /// @invariant The existence of the parameters specified in the constructor cannot
  /// change.
  ///
  /// @param analysis_manager 
  /// @param linear_system 
  /// @param nonlinear_manager 
  /// @param device_manager 
  /// @param builder 
  /// @param topology 
  ///
  MPDEFactory(
    Analysis::AnalysisManager & analysis_manager,
    Linear::System &            linear_system,
    Nonlinear::Manager &        nonlinear_manager,
    Device::DeviceMgr &         device_manager,
    Linear::Builder &           builder,
    Topo::Topology &            topology)
    : Factory<MPDE>(),
      analysisManager_(analysis_manager),
      linearSystem_(linear_system),
      nonlinearManager_(nonlinear_manager),
      deviceManager_(device_manager),
      builder_(builder),
      topology_(topology)
  {}

  virtual ~MPDEFactory()
  {}

  //-----------------------------------------------------------------------------
  // Function      : create
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:59:00 2015
  //-----------------------------------------------------------------------------
  ///
  /// Create a new MPDE analysis and applies the analysis and time integrator option blocks.
  ///
  /// @return new MPDE analysis object
  ///
  MPDE *create() const
  {
    analysisManager_.setAnalysisMode(ANP_MODE_MPDE);

    MPDE *mpde = new MPDE(analysisManager_, linearSystem_, nonlinearManager_, deviceManager_, builder_, topology_);

    mpde->getMPDEManager().setMPDEAnalysisParams(mpdeAnalysisOptionBlock_);
    mpde->getMPDEManager().setMPDEOptions(mpdeIntOptionBlock_);
    mpde->getMPDEManager().setTransientOptions(mpdeTimeIntegratorOptionBlock_);

    return mpde;
  }

  //-----------------------------------------------------------------------------
  // Function      : setMPDEAnalysisOptionBlock
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 13:00:14 2015
  //-----------------------------------------------------------------------------
  ///
  /// Saves the analysis parsed options block in the factory.
  ///
  /// @invariant Overwrites any previously specified analysis option block.
  ///
  /// @param option_block parsed option block
  ///
  void setMPDEAnalysisOptionBlock(const Util::OptionBlock &option_block)
  {
    mpdeAnalysisOptionBlock_ = option_block;
  }

  //-----------------------------------------------------------------------------
  // Function      : setMPDETimeIntegratorOptionBlock
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 13:01:27 2015
  //-----------------------------------------------------------------------------
  ///
  /// Saves the time integrator parsed option block.
  ///
  /// @invariant Overwrites any previously specified time integrator option block.
  ///
  /// @param option_block parsed option block
  ///
  void setMPDETimeIntegratorOptionBlock(const Util::OptionBlock &option_block)
  {
    mpdeTimeIntegratorOptionBlock_ = option_block;
  }

  //-----------------------------------------------------------------------------
  // Function      : setMPDEIntOptionBlock
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 13:01:27 2015
  //-----------------------------------------------------------------------------
  ///
  /// Saves the MPDEINT parsed option block.
  ///
  /// @invariant Overwrites any previously specified time integrator option block.
  ///
  /// @param option_block parsed option block
  ///
  void setMPDEIntOptionBlock(const Util::OptionBlock &option_block)
  {
    mpdeIntOptionBlock_ = option_block;
  }

public:
  AnalysisManager &     analysisManager_;
  Linear::System &      linearSystem_;
  Nonlinear::Manager &  nonlinearManager_;
  Device::DeviceMgr &   deviceManager_;
  Linear::Builder &     builder_;
  Topo::Topology &      topology_;

private:
  Util::OptionBlock     mpdeAnalysisOptionBlock_;
  Util::OptionBlock     mpdeIntOptionBlock_;
  Util::OptionBlock     mpdeTimeIntegratorOptionBlock_;
  Util::OptionBlock     timeIntegratorOptionBlock_;
};

// .options MPDEINT
struct MPDEIntOptionsReg : public IO::PkgOptionsReg
{
  MPDEIntOptionsReg(
    MPDEFactory &   factory )
    : factory_(factory)
  {}

  bool operator()(const Util::OptionBlock &option_block)
  {
    factory_.setMPDEIntOptionBlock(option_block);

    return true;
  }

  MPDEFactory &       factory_;
};

// .options TIMEINT-MPDE
struct MPDETimeIntegratorOptionsReg : public IO::PkgOptionsReg
{
  MPDETimeIntegratorOptionsReg(
    MPDEFactory &   factory )
    : factory_(factory)
  {}

  bool operator()(const Util::OptionBlock &option_block)
  {
    factory_.setMPDETimeIntegratorOptionBlock(option_block);

    return true;
  }

  MPDEFactory &       factory_;
};

// .MPDE
struct MPDEAnalysisReg : public IO::PkgOptionsReg
{
  MPDEAnalysisReg(
    MPDEFactory &   factory )
    : factory_(factory)
  {}

  bool operator()(const Util::OptionBlock &option_block)
  {
    factory_.setMPDEAnalysisOptionBlock(option_block);
    factory_.deviceManager_.setBlockAnalysisFlag(true);

    factory_.analysisManager_.addAnalysis(&factory_);

    return true;
  }

  MPDEFactory &               factory_;
};

} // namespace <unnamed>


bool
registerMPDEFactory(
  const std::string &           netlist_filename,
  IO::PkgOptionsMgr &           options_manager,
  AnalysisManager &             analysis_manager,
  Linear::System &              linear_system,
  Nonlinear::Manager &          nonlinear_manager,
  Device::DeviceMgr &           device_manager,
  Linear::Builder &             builder,
  Topo::Topology &              topology)
{
  MPDEFactory *factory = new MPDEFactory(analysis_manager, linear_system, nonlinear_manager, device_manager, builder, topology);

  analysis_manager.addAnalysisFactory(factory);

  options_manager.submitRegistration("MPDE", netlist_filename, new MPDEAnalysisReg(*factory));
  options_manager.submitRegistration("MPDEINT", netlist_filename, new MPDEIntOptionsReg(*factory));
  options_manager.submitRegistration("TIMEINT-MPDE", netlist_filename, new MPDETimeIntegratorOptionsReg(*factory));


  return true;
}

} // namespace Analysis
} // namespace Xyce
