//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_ANP_OutputMgrAdapter.C,v $
//
// Purpose        :
//
// Special Notes  :
//
// Creator        : Richard Schiek, SNL, Electrical and Microsystem Modeling
//
// Creation Date  : 01/24/08
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.48.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:07 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------
#include <Xyce_config.h>

#include <N_ANP_OutputMgrAdapter.h>

#include <N_ANP_NoiseData.h>

#include <N_DEV_DeviceMgr.h>
#include <N_IO_FourierMgr.h>
#include <N_IO_InitialConditions.h>
#include <N_IO_MeasureManager.h>
#include <N_IO_Objective.h>
#include <N_IO_OutputMgr.h>
#include <N_IO_OutputROM.h>
#include <N_IO_OutputMOR.h>
#include <N_IO_OutputResults.h>
#include <N_IO_OutputResponse.h>
#include <N_IO_RestartMgr.h>
#include <N_TOP_Topology.h>


namespace Xyce {
namespace Analysis {

//-----------------------------------------------------------------------------
// Function      : OutputMgrAdapter::OutputMgrAdapter( )
// Purpose       : constructor
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
OutputMgrAdapter::OutputMgrAdapter(
  Parallel::Machine                     comm,
  IO::OutputMgr &                       output_manager,
  IO::Measure::Manager &                measure_manager,
  IO::FourierMgr &                      fourier_manager,
  IO::ObjectiveManager &                objective_manager,
  IO::InitialConditionsManager &        initial_conditions_manager,
  Device::DeviceMgr &                   device_manager,
  Topo::Topology &                      topology,
  IO::RestartMgr &                      restart_manager)
  : comm_(comm),
    outputManager_(output_manager),
    measureManager_(measure_manager),
    fourierManager_(fourier_manager),
    objectiveManager_(objective_manager),
    initialConditionsManager_(initial_conditions_manager),
    deviceManager_(device_manager),
    topology_(topology),
    restartManager_(restart_manager),
    outputMOR_(0),
    outputResults_(0),
    outputResponse_(0),
    stepSweepVector_(),
    dcSweepVector_(),
    stepAnalysisStepNumber_(0),
    stepAnalysisMaxSteps_(0),
    dcAnalysisStepNumber_(0),
    dcAnalysisMaxSteps_(0)
{}

OutputMgrAdapter::~OutputMgrAdapter()
{
  delete outputMOR_;
  delete outputResults_;
}


void
OutputMgrAdapter::notify(
  const StepEvent &     event)
{
  if (event.state_ == StepEvent::STEP_STARTED)
    stepAnalysisStepNumber_ = event.count_;
  else if (event.state_ == StepEvent::INITIALIZE)
    stepAnalysisMaxSteps_ = event.count_;
}


void OutputMgrAdapter::addOutputResults(const Util::OptionBlock & option_block)
{
  if (!outputResults_)
    outputResults_ = new IO::OutputResults(outputManager_.getNetlistFilename());

  outputResults_->addResultParams(option_block);
}

void OutputMgrAdapter::addOutputResponse(const Util::OptionBlock & option_block)
{
  if (!outputResponse_)
    outputResponse_ = new IO::OutputResponse();
}

void
OutputMgrAdapter::setStepSweepVector(
  const Analysis::SweepVector &       sweep_vector)
{
  stepSweepVector_ = sweep_vector;
}

void
OutputMgrAdapter::setDCSweepVector(
  const Analysis::SweepVector &       sweep_vector)
{
  dcSweepVector_ = sweep_vector;
}

bool
OutputMgrAdapter::isRestart() const
{
  return restartManager_.isRestart();
}

void
OutputMgrAdapter::dumpRestart(Parallel::Communicator &parallel_communicator, Analysis::AnalysisManager &analysis_manager, double current_time) const
{
  IO::dumpRestartData(
    parallel_communicator,
    topology_,
    analysis_manager,
    deviceManager_,
    restartManager_.getJobName(),
    restartManager_.getPack(),
    current_time);
}


void
OutputMgrAdapter::tranOutput(
  double                time,
  Linear::Vector &        currSolutionPtr,
  Linear::Vector &        stateVecPtr,
  Linear::Vector &        storeVecPtr,
  Linear::Vector &        lead_current_vector,
  Linear::Vector &        junction_voltage_vector,
  std::vector<double> & objectiveVec_,
  std::vector<double> & dOdpVec_,
  std::vector<double> & dOdpAdjVec_,
  std::vector<double> & scaled_dOdpVec_,
  std::vector<double> & scaled_dOdpAdjVec_,
  bool                  skipPrintLineOutput)
{
  fourierManager_.updateFourierData(comm_, time, &currSolutionPtr);

  measureManager_.updateTranMeasures(comm_, time, &currSolutionPtr, &stateVecPtr, &storeVecPtr);

  outputManager_.output(
    comm_, time, deviceManager_.getParamAndReduce("TEMP"), stepAnalysisStepNumber_, stepAnalysisMaxSteps_, stepSweepVector_,
    dcAnalysisStepNumber_, dcAnalysisMaxSteps_, dcSweepVector_,
    currSolutionPtr, stateVecPtr, storeVecPtr, lead_current_vector, junction_voltage_vector, objectiveVec_,
    dOdpVec_, dOdpAdjVec_, scaled_dOdpVec_, scaled_dOdpAdjVec_,
    skipPrintLineOutput);

  if (outputResponse_) {
    outputResponse_->saveResponseVarValues(comm_, time, currSolutionPtr);
    outputResponse_->send(comm_,
                          &currSolutionPtr, 0, &stateVecPtr, &storeVecPtr,
                          &objectiveVec_, &dOdpVec_, &dOdpAdjVec_, &scaled_dOdpVec_, &scaled_dOdpAdjVec_);
  }

  // transient values for the objective function call.
  double arg1 = time;
  double arg2 = 0.0;

  objectiveManager_.outputObjective(comm_, outputManager_.getOpBuilderManager(), outputManager_, outputManager_.getCircuitTime(), arg1, arg2, currSolutionPtr, stateVecPtr, storeVecPtr);
}

void
OutputMgrAdapter::dcOutput(
  int                   dcStepNumber,
  Linear::Vector &        currSolutionPtr,
  Linear::Vector &        stateVecPtr,
  Linear::Vector &        storeVecPtr,
  Linear::Vector &        lead_current_vector,
  Linear::Vector &        junction_voltage_vector,
  std::vector<double> & objectiveVec_,
  std::vector<double> & dOdpVec_,
  std::vector<double> & dOdpAdjVec_,
  std::vector<double> & scaled_dOdpVec_,
  std::vector<double> & scaled_dOdpAdjVec_)
{
  measureManager_.updateDCMeasures(comm_, dcSweepVector_, &currSolutionPtr, &stateVecPtr, &storeVecPtr);

  outputManager_.output(
    comm_, 0.0, deviceManager_.getParamAndReduce("TEMP"), 
    stepAnalysisStepNumber_, stepAnalysisMaxSteps_, stepSweepVector_,
    dcStepNumber, dcAnalysisMaxSteps_, dcSweepVector_,
    currSolutionPtr, stateVecPtr, storeVecPtr, lead_current_vector, junction_voltage_vector, objectiveVec_,
    dOdpVec_, dOdpAdjVec_, scaled_dOdpVec_, scaled_dOdpAdjVec_);

  if (outputResponse_)
    outputResponse_->saveResponseVarValues(comm_, 0.0, currSolutionPtr);

  // dc values for the objective function call.
  double arg1 = 0.0;
  double arg2 = 0.0;

  if (dcSweepVector_.size() > 0)
    arg1 = dcSweepVector_[0].currentVal;
  if (dcSweepVector_.size() > 1)
    arg2 = dcSweepVector_[1].currentVal;

  objectiveManager_.outputObjective(comm_, outputManager_.getOpBuilderManager(), outputManager_, outputManager_.getCircuitTime(), arg1, arg2, currSolutionPtr, stateVecPtr, storeVecPtr);
}


void
OutputMgrAdapter::outputResult(
  const Linear::Vector &  currSolutionPtr,
  const Linear::Vector &  currStatePtr,
  const Linear::Vector &  currStorePtr,
  const Linear::Vector &  lead_current_vector,
  const Linear::Vector &  junction_voltage_vector)
{
  if (!outputResults_)
    outputResults_ = new IO::OutputResults(outputManager_.getNetlistFilename());

  outputResults_->output(comm_, outputManager_, outputManager_.getCircuitTime(), stepSweepVector_, stepAnalysisStepNumber_, currSolutionPtr, currStatePtr, currStorePtr, lead_current_vector, junction_voltage_vector);
}

void
OutputMgrAdapter::steppingComplete()
{
  outputManager_.steppingComplete();
  if (outputResults_)
    outputResults_->steppingComplete();
}

void
OutputMgrAdapter::finishOutput()
{
  outputManager_.finishOutput();
}

bool
OutputMgrAdapter::setupInitialConditions(
  Linear::Vector &        solnVec,
  Linear::Vector &        flagVec)
{
  return initialConditionsManager_.setupInitialConditions(comm_, getAllNodes(), solnVec, flagVec);
}

void
OutputMgrAdapter::outputDCOP(
  const Linear::Vector &  solution)
{
  initialConditionsManager_.outputDCOP(comm_, topology_.getSolutionNodeNameMap(), solution);
}


void
OutputMgrAdapter::outputMPDE(
  double                        time,
  const std::vector<double> &   fast_time_points,
  const Linear::BlockVector &   solution_vector)
{
  outputManager_.outputMPDE(comm_, time, fast_time_points, solution_vector);
}

void
OutputMgrAdapter::outputHB (
  const std::vector< double > & timePoints,
  const std::vector< double > & freqPoints,
  const Linear::BlockVector &     timeDomainSolutionVec,
  const Linear::BlockVector &     freqDomainSolutionVecReal,
  const Linear::BlockVector &     freqDomainSolutionVecImaginary,
  const Linear::BlockVector &     timeDomainStoreVec,
  const Linear::BlockVector &     freqDomainStoreVecReal,
  const Linear::BlockVector &     freqDomainStoreVecImaginary,
  const Linear::BlockVector &     timeDomainLeadCurrentVec,
  const Linear::BlockVector &     freqDomainLeadCurrentVecReal,
  const Linear::BlockVector &     freqDomainLeadCurrentVecImaginary,
  const Linear::BlockVector &     timeDomainJunctionVoltageVec,
  const Linear::BlockVector &     freqDomainJunctionVoltageVecReal,
  const Linear::BlockVector &     freqDomainJunctionVoltageVecImaginary )

{
  outputManager_.outputHB(
    comm_,
    stepAnalysisStepNumber_, stepAnalysisMaxSteps_, stepSweepVector_,
    timePoints, freqPoints,
    timeDomainSolutionVec, freqDomainSolutionVecReal, freqDomainSolutionVecImaginary, 
    timeDomainStoreVec, freqDomainStoreVecReal, freqDomainStoreVecImaginary,
    timeDomainLeadCurrentVec, freqDomainLeadCurrentVecReal, freqDomainLeadCurrentVecImaginary,
    timeDomainJunctionVoltageVec, freqDomainJunctionVoltageVecReal, freqDomainJunctionVoltageVecImaginary );
}

void
OutputMgrAdapter::outputAC (
  double                frequency,
  const Linear::Vector &  solnVecRealPtr,
  const Linear::Vector &  solnVecImaginaryPtr)
{
  measureManager_.updateACMeasures(comm_, frequency, &solnVecRealPtr, &solnVecImaginaryPtr);
  
  outputManager_.outputAC(comm_, frequency, solnVecRealPtr, solnVecImaginaryPtr);

}
void
OutputMgrAdapter::outputNoise (
  double                frequency, 
  double totalOutputNoiseDens_, double totalInputNoiseDens_, 
  const std::vector<Xyce::Analysis::NoiseData*> & noiseDataVec_)
{
  
  outputManager_.outputNoise(comm_, frequency, 
      totalOutputNoiseDens_, totalInputNoiseDens_, noiseDataVec_);
}

void
OutputMgrAdapter::outputMORTF (
  bool                                                          origSys,
  double                                                        freq,
  const Teuchos::SerialDenseMatrix<int, std::complex<double> >& H )
{
  if (!outputMOR_)
    outputMOR_ = new IO::OutputMOR(outputManager_.getNetlistFilename());

  outputMOR_->output(comm_, origSys, freq, H );
}

void
OutputMgrAdapter::resetOutputMORTF()
{
  if (!outputMOR_)
    outputMOR_ = new IO::OutputMOR(outputManager_.getNetlistFilename());

  outputMOR_->reset();
}

void
OutputMgrAdapter::outputROM(
  const Teuchos::SerialDenseMatrix<int, double> &       Ghat,
  const Teuchos::SerialDenseMatrix<int, double> &       Chat,
  const Teuchos::SerialDenseMatrix<int, double> &       Bhat,
  const Teuchos::SerialDenseMatrix<int, double> &       Lhat )
{
  IO::outputROM(comm_, outputManager_.getNetlistFilename(), Ghat, Chat, Bhat, Lhat );
}

void
OutputMgrAdapter::outputROM(
  const Linear::Matrix &                                  Ghat,
  const Linear::Matrix &                                  Chat,
  const Teuchos::SerialDenseMatrix<int, double> &       Bhat,
  const Teuchos::SerialDenseMatrix<int, double> &       Lhat )
{
  IO::outputROM(comm_, outputManager_.getNetlistFilename(),  Ghat, Chat, Bhat, Lhat );
}

double
OutputMgrAdapter::getInitialOutputInterval() const
{
  return outputManager_.getInitialOutputInterval();
}

const IO::IntervalVector &
OutputMgrAdapter::getOutputIntervals() const
{
  return outputManager_.getOutputIntervals();
}

void
OutputMgrAdapter::outputHomotopy(
  const std::vector<std::string> &      paramNames,
  const std::vector<double> &           paramVals,
  Linear::Vector &                        solnVecPtr )
{
  outputManager_.outputHomotopy (comm_, paramNames, paramVals, solnVecPtr);
}

const Xyce::NodeNameMap &
OutputMgrAdapter::getAllNodes() const
{
  return topology_.getSolutionNodeNameMap();
}

} // namespace Analysis
} // namespace Xyce
