//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_ANP_Step.C,v $
// Purpose       : .STEP Sweep class analysis functions.
// Special Notes :
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 01/24/08
//
// Revision Information:
// ---------------------
// Revision Number: $Revision: 1.73.2.1 $
// Revision Date  : $Date: 2015/04/02 18:20:07 $
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------
#include <Xyce_config.h>

#include <N_ANP_AnalysisManager.h>
#include <N_ANP_OutputMgrAdapter.h>
#include <N_ANP_SweepParam.h>
#include <N_ANP_Step.h>
#include <N_ANP_StepEvent.h>
#include <N_ERH_Message.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_IO_CmdParse.h>
#include <N_TIA_StepErrorControl.h>
#include <N_TIA_DataStore.h>
#include <N_UTL_Diagnostic.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_OptionBlock.h>
#include <N_UTL_Timer.h>

namespace Xyce {
namespace Analysis {

//-----------------------------------------------------------------------------
// Function      : Step::setAnalysisParams
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date : 6/22/10
//-----------------------------------------------------------------------------
bool Step::setAnalysisParams(const Util::OptionBlock & paramsBlock)
{
  stepSweepVector_.push_back(parseSweepParams(paramsBlock.begin(), paramsBlock.end()));
  outputManagerAdapter_.setStepSweepVector(stepSweepVector_);

  return true;
}

const TimeIntg::TIAParams &
Step::getTIAParams() const
{
  return childAnalysis_.getTIAParams();
}

TimeIntg::TIAParams &
Step::getTIAParams()
{
  return childAnalysis_.getTIAParams();
}

//-----------------------------------------------------------------------------
// Function      : Step::getDCOPFlag()
// Purpose       :
//
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 3/24/2014
//-----------------------------------------------------------------------------
bool Step::getDCOPFlag() const
{
  return childAnalysis_.getDCOPFlag();
}

//-----------------------------------------------------------------------------
// Function      : Step::run()
// Purpose       : This is the main controlling loop for Step analysis.
//
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 10/04/00
//-----------------------------------------------------------------------------
bool Step::doRun()
{
  return doInit() && doLoopProcess() && doFinish();
}

//-----------------------------------------------------------------------------
// Function      : Step::init()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/10/06
//-----------------------------------------------------------------------------
bool Step::doInit()
{
  if (DEBUG_ANALYSIS && isActive(Diag::TIME_PARAMETERS))
  {
    Xyce::dout() << std::endl << std::endl;
    Xyce::dout() << section_divider << std::endl;
    Xyce::dout() << "Step::init" << std::endl;
  }

  stepLoopSize_ = setupSweepLoop(loader_, stepSweepVector_.begin(), stepSweepVector_.end());

  Util::publish<StepEvent>(analysisManager_, StepEvent(StepEvent::INITIALIZE, stepLoopSize_));

  // analysisManager_.setStepLoopInitialized(true);

  return true;
}


//-----------------------------------------------------------------------------
// Function      : Step::loopProcess()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/10/06
//-----------------------------------------------------------------------------
bool Step::doLoopProcess()
{
  std::string msg;
  bool integration_status = true;

  for (int i = 0; i < stepLoopSize_; ++i)
  {
    // Tell the manager if any of our sweeps are being reset in this loop iteration.
    bool reset = updateSweepParams(loader_, i, stepSweepVector_.begin(), stepSweepVector_.end());

    analysisManager_.setSweepSourceResetFlag(reset);

    outputManagerAdapter_.setStepSweepVector(stepSweepVector_);

    if (DEBUG_ANALYSIS && isActive(Diag::TIME_PARAMETERS))
    {
      for (SweepVector::const_iterator it = stepSweepVector_.begin(), end = stepSweepVector_.end(); it != end; ++it)
      {
        Xyce::dout() << "Step Analysis # " << i<<"\t";
        Xyce::dout() << (*it);
      }
    }

    Util::publish<StepEvent>(analysisManager_, StepEvent(StepEvent::STEP_STARTED, i));

    // solve the loop.
    integration_status = childAnalysis_.run();
    
    StepEvent anEvent(StepEvent::STEP_COMPLETED, i);
    // this is to try to capture the final sim time when the 
    // underlying simulation (childAnalysis) is transient 
    // and then that simulation time is reported to measure manager's
    // verbose output.
    anEvent.finalSimTime_ = getTIAParams().finalTime;
    Util::publish<StepEvent>(analysisManager_, anEvent); 

    outputManagerAdapter_.outputResult(
        *(analysisManager_.getDataStore()->currSolutionPtr),
        *(analysisManager_.getDataStore()->currStatePtr),
        *(analysisManager_.getDataStore()->currStorePtr),
        *(analysisManager_.getDataStore()->currLeadCurrentPtr),
        *(analysisManager_.getDataStore()->currLeadDeltaVPtr));

  } // end of for loop, and end of step analysis.

  return integration_status;
}

//-----------------------------------------------------------------------------
// Function      : Step::processSuccessfulStep()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/10/06
//-----------------------------------------------------------------------------
bool Step::doProcessSuccessfulStep()
{
  return true;
}

//-----------------------------------------------------------------------------
// Function      : Step::processFailedStep()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/10/06
//-----------------------------------------------------------------------------
bool Step::doProcessFailedStep()
{
  return true;
}


//-----------------------------------------------------------------------------
// Function      : Step::doFinish()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/10/06
//-----------------------------------------------------------------------------
bool Step::doFinish()
{
  Util::publish<StepEvent>(analysisManager_, StepEvent(StepEvent::FINISH, stepLoopSize_));

  outputManagerAdapter_.steppingComplete();

  return true;
}

namespace {

//-----------------------------------------------------------------------------
// Class         : StepFactory
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Thu Jan 29 12:53:02 2015
//-----------------------------------------------------------------------------
///
/// Factory for parsing Step parameters from the netlist and creating Step analysis.
///
class StepFactory : public Factory<Step>
{
public:
  //-----------------------------------------------------------------------------
  // Function      : StepFactory
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:54:09 2015
  //-----------------------------------------------------------------------------
  ///
  /// Constructs the Step analysis factory
  ///
  /// @invariant Stores the results of parsing.  Multiple Step analysis options may be
  /// applied and each generates and additional step.
  ///
  /// @invariant The existence of the parameters specified in the constructor cannot
  /// change.
  ///
  /// @param analysis_manager 
  /// @param linear_system 
  /// @param nonlinear_manager 
  ///
  StepFactory(
    Analysis::AnalysisManager & analysis_manager,
    Linear::System &            linear_system,
    Nonlinear::Manager &        nonlinear_manager)
    : Factory<Step>(),
      analysisManager_(analysis_manager),
      linearSystem_(linear_system),
      nonlinearManager_(nonlinear_manager)
  {}

  virtual ~StepFactory()
  {}

  //-----------------------------------------------------------------------------
  // Function      : create
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 12:59:00 2015
  //-----------------------------------------------------------------------------
  ///
  /// Create a new Step analysis and applies the analysis and time integrator option blocks.
  ///
  /// @return new Step analysis object
  ///
  Step *create() const
  {
    Step *step = new Step(analysisManager_, analysisManager_.getAnalysisObject());
    for (std::vector<Util::OptionBlock>::const_iterator it = stepSweepAnalysisOptionBlock_.begin(), end = stepSweepAnalysisOptionBlock_.end(); it != end; ++it)
      step->setAnalysisParams(*it);

    return step;
  }

  //-----------------------------------------------------------------------------
  // Function      : setStepAnalysisOptionBlock
  // Purpose       :
  // Special Notes :
  // Scope         : public
  // Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
  // Creation Date : Thu Jan 29 13:00:14 2015
  //-----------------------------------------------------------------------------
  ///
  /// Saves the analysis parsed options block in the factory.
  ///
  /// @invariant Appends to any previously specified analysis option block.
  ///
  /// @param option_block parsed option block
  ///
  void setStepAnalysisOptionBlock(const Util::OptionBlock &option_block)
  {
    for (std::vector<Util::OptionBlock>::iterator it = stepSweepAnalysisOptionBlock_.begin(), end = stepSweepAnalysisOptionBlock_.end(); it != end; ++it)
    {
      if ((*it).compareParamLists(option_block))
      {
        (*it) = option_block;
        return;
      }
    }

    // save the new one.
    stepSweepAnalysisOptionBlock_.push_back(option_block); // save a copy for later.
  }

public:
  AnalysisManager &             analysisManager_;
  Linear::System &              linearSystem_;
  Nonlinear::Manager &          nonlinearManager_;

private:
  std::vector<Util::OptionBlock>        stepSweepAnalysisOptionBlock_;
};

// .STEP
struct StepAnalysisReg : public IO::PkgOptionsReg
{
  StepAnalysisReg(
    StepFactory &             factory)
    : factory_(factory)
  {}

  bool operator()(const Util::OptionBlock &option_block)
  {
    factory_.setStepAnalysisOptionBlock(option_block);

    factory_.analysisManager_.addAnalysis(&factory_);

    return true;
  }

  StepFactory &               factory_;
};

} // namespace <unnamed>


bool
registerStepFactory(
  const std::string &           netlist_filename,
  IO::PkgOptionsMgr &           options_manager,
  AnalysisManager &             analysis_manager,
  Linear::System &              linear_system,
  Nonlinear::Manager &          nonlinear_manager)
{
  StepFactory *factory = new StepFactory(analysis_manager, linear_system, nonlinear_manager);

  analysis_manager.addAnalysisFactory(factory);

  options_manager.submitRegistration("STEP", netlist_filename, new StepAnalysisReg(*factory));

  return true;
}

} // namespace Analysis
} // namespace Xyce
