//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Filename       : $RCSfile: N_CIR_SecondLevelSimulator.C,v $
//
// Purpose        :
//
// Special Notes  :
//
// Creator        : Eric Keiter, SNL, Parallel Computational Sciences
//
// Creation Date  : 05/27/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.5.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:07 $
//
// Current Owner  : $Author: tvrusso $
//-------------------------------------------------------------------------

#include <Xyce_config.h>

#include <N_CIR_SecondLevelSimulator.h>
#include <N_DEV_DeviceMgr.h>
#include <N_DEV_ExternData.h>
#include <N_DEV_ExternalSimulationData.h>
#include <N_LAS_System.h>
#include <N_NLS_Manager.h>
#include <N_NLS_ConductanceExtractor.h>
#include <N_UTL_FeatureTest.h>

namespace Xyce {
namespace Circuit {

Analysis::AnalysisManager *
SecondLevelSimulator::newAnalysisManager(
  const IO::CmdParse &                command_line,
  IO::RestartMgr &                    restart_manager,
  Analysis::OutputMgrAdapter &        output_manager_adapter,
  Stats::Stat                         analysis_stat)
{
  return secondLevelManager_ = new Analysis::SecondLevelManager(command_line, restart_manager, output_manager_adapter, analysis_stat);
}


// ---------------------------------------------------------------------------
// API METHODS NEEDED FOR Two-level Functions:

//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::simulateStep
// Purpose       :
// Special Notes : Used for 2-level Newton solves.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/01/2006
//---------------------------------------------------------------------------
bool SecondLevelSimulator::simulateStep(
  const Device::SolverState &           solState,
  const std::map<std::string, double> & inputMap,
  std::vector<double> &                 outputVector,
  std::vector< std::vector<double> > &  jacobian,
  TimeIntg::TwoLevelError &             tlError)
{
  bool bsuccess = false;

  if (DEBUG_CIRCUIT)
    dout() << "\nsimulateStep: " << std::endl;

  // Apply the input voltages to their appropriate sources:
  for (std::map<std::string,double>::const_iterator it = inputMap.begin(), end = inputMap.end(); it != end; ++it)
  {
    std::string name = (*it).first;
    getDeviceManager().setParam(name, (*it).second);
  }

  secondLevelManager_->setExternalSolverState(solState);

  bsuccess = secondLevelManager_->runSecondLevelStep(solState.twoLevelTimeIntgInfo, tlError);

  // calculate the conductance:
  getNonlinearManager().getConductanceExtractor().extract(inputMap, outputVector, jacobian);

  return bsuccess;
}

//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::simulateStep
// Purpose       :
// Special Notes : Used for 2-level Newton solves with Charon.
// Scope         : public
// Creator       : Roger Pawlowski, SNL
// Creation Date : 07/16/2009
//---------------------------------------------------------------------------
bool SecondLevelSimulator::simulateStep(
  const Device::ExternalSimulationData &        ext_data,
  const std::map<std::string, double> &         inputMap,
  std::vector<double> &                         outputVector,
  std::vector< std::vector<double> > &          jacobian,
  TimeIntg::TwoLevelError &                     tlError)
{
  bool bsuccess = false;

  if (DEBUG_CIRCUIT)
    dout() << "\nsimulateStep: " << std::endl;

  // Create N_DEX_SolverState object
  Device::SolverState state;

  if (ext_data.is_transient)
  {
    Xyce::lout() << "Xychron (mixed-model): Transient Xyce Step" << std::endl;

    state.currTimeStep = ext_data.current_time_step_size;
    state.lastTimeStep = ext_data.previous_time_step_size;
    state.currTime = ext_data.current_time;
    state.finalTime = ext_data.final_time;
    state.timeStepNumber = ext_data.time_step_number;
    state.startingTimeStep = ext_data.current_time_step_size;
    state.tranopFlag = false;
    state.transientFlag = true;
    state.dcopFlag = false;
    state.dcsweepFlag = false;

    // tiInfo
    state.twoLevelTimeIntgInfo.nextTimeStep = ext_data.current_time_step_size;
    state.twoLevelTimeIntgInfo.currTimeStep = ext_data.current_time_step_size;
    state.twoLevelTimeIntgInfo.nextTime = ext_data.current_time;
    state.twoLevelTimeIntgInfo.finalTime = ext_data.final_time;
    state.twoLevelTimeIntgInfo.dcopFlag = false;
    state.twoLevelTimeIntgInfo.tranopFlag = false;
    state.twoLevelTimeIntgInfo.transientFlag = true;
    state.twoLevelTimeIntgInfo.dcsweepFlag = false;
    state.twoLevelTimeIntgInfo.timeStepNumber = ext_data.time_step_number;
    state.twoLevelTimeIntgInfo.timeIntMode = 6;

    // New changes
    // pdt = 1/dt
    state.pdt = 1.0/ext_data.current_time_step_size;
    // bptol - ignore for now
    state.doubleDCOPEnabled = false;
    state.doubleDCOPStep = 1;
    // newtonIteration - doens't matter for particular devices.  For volt lim

    if (state.timeStepNumber == 0)
    {
      state.initTranFlag = true;
      state.twoLevelTimeIntgInfo.initTranFlag = true;
      state.twoLevelTimeIntgInfo.beginIntegrationFlag = true;
    }
    else
    {
      state.initTranFlag = false;
      state.twoLevelTimeIntgInfo.initTranFlag = false;
      state.twoLevelTimeIntgInfo.beginIntegrationFlag = false;
    }

    // tia pdt - set this
    state.twoLevelTimeIntgInfo.pdt = state.pdt;
    state.twoLevelTimeIntgInfo.currentOrder = 1;
  }
  else
  {
    state.tranopFlag = true;  // for a DCOP only, this should be false
    state.transientFlag = false;
    state.dcopFlag = false;
    state.dcsweepFlag = false;
    state.twoLevelTimeIntgInfo.dcopFlag = true;
    state.twoLevelTimeIntgInfo.tranopFlag = true;
    state.twoLevelTimeIntgInfo.transientFlag = false;
    state.twoLevelTimeIntgInfo.dcsweepFlag = false;
    state.twoLevelTimeIntgInfo.timeIntMode = 0;
  }

  bsuccess = simulateStep(state, inputMap, outputVector, jacobian, tlError);

  return bsuccess;
}

//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::startupSolvers
// Purpose       :
// Special Notes : Used for 2-level Newton solves.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/10/2006
//---------------------------------------------------------------------------
bool SecondLevelSimulator::startupSolvers()
{
  return secondLevelManager_->startupSecondLevelSolvers(getLinearSystem(), getNonlinearManager());
}

//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::finishSolvers 
// Purpose       :
// Special Notes : Used for 2-level Newton solves.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/10/2006
//---------------------------------------------------------------------------
bool SecondLevelSimulator::finishSolvers ()
{
  bool bsuccess = true;
  bsuccess = secondLevelManager_->finishSecondLevelSolvers();
  return bsuccess;
}

//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::homotopyStepSuccess
// Purpose       :
// Special Notes : Used for 2-level Newton solves.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/20/2006
//---------------------------------------------------------------------------
void
SecondLevelSimulator::homotopyStepSuccess(
  const std::vector<std::string> &      paramNames,
  const std::vector<double> &           paramVals)
{
  secondLevelManager_->homotopyStepSuccess(paramNames, paramVals);
}

//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::homotopyStepFailure
// Purpose       :
// Special Notes : Used for 2-level Newton solves.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/30/2006
//---------------------------------------------------------------------------
void
SecondLevelSimulator::homotopyStepFailure()
{
  secondLevelManager_->homotopyStepFailure();
}

//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::stepSuccess
// Purpose       :
// Special Notes : Used for 2-level Newton solves.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/12/2006
//---------------------------------------------------------------------------
void SecondLevelSimulator::stepSuccess(Analysis::TwoLevelMode analysis)
{
  secondLevelManager_->stepSecondLevelSuccess(analysis);
}

//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::stepFailure
// Purpose       :
// Special Notes : Used for 2-level Newton solves.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/12/2006
//---------------------------------------------------------------------------
void SecondLevelSimulator::stepFailure(Analysis::TwoLevelMode analysis)
{
  secondLevelManager_->stepSecondLevelFailure(analysis);
}


//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::getInitialQnorm
// Purpose       :
// Special Notes : Used for 2-level Newton solves.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/12/2006
//---------------------------------------------------------------------------
bool SecondLevelSimulator::getInitialQnorm (TimeIntg::TwoLevelError & tle)
{
  return secondLevelManager_->getSecondLevelInitialQnorm(tle);
}

//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::getBreakPoints
// Purpose       :
// Special Notes : Used for 2-level Newton solves.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/12/2006
//---------------------------------------------------------------------------
bool SecondLevelSimulator::getBreakPoints (std::vector<Util::BreakPoint> &breakPointTimes)
{
  return secondLevelManager_->getSecondLevelBreakPoints(breakPointTimes);
}

//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::updateStateArrays
// Purpose       :
// Special Notes : Used for 2-level Newton solves, with LOCA.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/18/2006
//---------------------------------------------------------------------------
bool SecondLevelSimulator::updateStateArrays()
{
  bool bsuccess = true;
  return bsuccess;
}

//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::setInternalParam
// Purpose       :
// Special Notes : Used for 2-level Newton solves, with LOCA.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/18/2006
//---------------------------------------------------------------------------
bool SecondLevelSimulator::setInternalParam (std::string & name, double val)
{
  getDeviceManager().setParam(name, val);
  return true;
}

//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::startTimeStep
// Purpose       :
// Special Notes : Used for 2-level Newton solves
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 03/20/2006
//---------------------------------------------------------------------------
bool SecondLevelSimulator::startTimeStep (const TimeIntg::TimeIntInfo & tiInfo)
{
  return secondLevelManager_->startSecondLevelTimeStep(getAnalysisManager().getTIAParams(), getNonlinearManager(), tiInfo);
}

//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::startTimeStep
// Purpose       :
// Special Notes : Used for 2-level Newton solves with Charon.
// Scope         : public
// Creator       : Roger Pawlowski, SNL
// Creation Date : 08/28/2009
//---------------------------------------------------------------------------
bool SecondLevelSimulator::startTimeStep (const Device::ExternalSimulationData & ext_data)
{
  TimeIntg::TimeIntInfo tiInfo;

  if (ext_data.is_transient)
  {
    tiInfo.nextTimeStep = ext_data.current_time_step_size;
    tiInfo.currTimeStep = ext_data.current_time_step_size;
    tiInfo.nextTime = ext_data.current_time;
    tiInfo.finalTime = ext_data.final_time;
    tiInfo.dcopFlag = false;
    tiInfo.tranopFlag = false;
    tiInfo.transientFlag = true;
    tiInfo.dcsweepFlag = false;
    tiInfo.timeStepNumber = ext_data.time_step_number;
    tiInfo.timeIntMode = 6;

    if (ext_data.time_step_number == 0)
    {
      tiInfo.initTranFlag = true;
      tiInfo.beginIntegrationFlag = true;
    }
    else
    {
      tiInfo.initTranFlag = false;
      tiInfo.beginIntegrationFlag = false;
    }

    // tia pdt - set this
    tiInfo.pdt = 1.0/ext_data.current_time_step_size;
    tiInfo.currentOrder = 1;
  }
  else
  {
    tiInfo.dcopFlag = true;
    tiInfo.tranopFlag = true;
    tiInfo.transientFlag = false;
    tiInfo.dcsweepFlag = false;
    tiInfo.timeIntMode = 0;
  }

  return secondLevelManager_->startSecondLevelTimeStep(getAnalysisManager().getTIAParams(), getNonlinearManager(), tiInfo);
}

//---------------------------------------------------------------------------
// Function      : SecondLevelSimulator::endTimeStep
// Purpose       :
// Special Notes : Used for 2-level Newton solves with Charon.
// Scope         : public
// Creator       : Russell Hooper, SNL
// Creation Date : 10/16/2012
//---------------------------------------------------------------------------
bool SecondLevelSimulator::endTimeStep (Device::ExternalSimulationData & ext_data)
{
  // We could opt to obtain selected time integration data here or
  // allow it to be obtained higher up, eg in charon::sc::CircuitDriver
  bool bsuccess = true;
  TimeIntg::TimeIntInfo tiInfo;
  Analysis::getTimeIntInfo(getAnalysisManager(), tiInfo);

  // Copy TimeIntg::TimeIntInfo into our ext_data container
  ext_data.currentOrder         = tiInfo.currentOrder        ;
  ext_data.numberOfSteps        = tiInfo.numberOfSteps       ;
  ext_data.usedOrder            = tiInfo.usedOrder           ;
  ext_data.nscsco               = tiInfo.nscsco              ;
  ext_data.pdt                  = tiInfo.pdt                 ;
  ext_data.nextTimeStep         = tiInfo.nextTimeStep        ;
  ext_data.currTimeStep         = tiInfo.currTimeStep        ;
  ext_data.currentTime          = tiInfo.currentTime         ;
  ext_data.nextTime             = tiInfo.nextTime            ;
  ext_data.finalTime            = tiInfo.finalTime           ;
  ext_data.startingTimeStep     = tiInfo.startingTimeStep    ;
  ext_data.bpTol                = tiInfo.bpTol               ;
  ext_data.dcopFlag             = tiInfo.dcopFlag            ;
  ext_data.acopFlag             = tiInfo.acopFlag            ;
  ext_data.inputOPFlag          = tiInfo.inputOPFlag         ;
  ext_data.tranopFlag           = tiInfo.tranopFlag          ;
  ext_data.transientFlag        = tiInfo.transientFlag       ;
  ext_data.dcsweepFlag          = tiInfo.dcsweepFlag         ;
  ext_data.timeStepNumber       = tiInfo.timeStepNumber      ;
  ext_data.initTranFlag         = tiInfo.initTranFlag        ;
  ext_data.beginIntegrationFlag = tiInfo.beginIntegrationFlag;
  ext_data.doubleDCOPStep       = tiInfo.doubleDCOPStep      ;
  ext_data.doubleDCOPEnabled    = tiInfo.doubleDCOPEnabled   ;
  ext_data.timeIntMode          = tiInfo.timeIntMode         ;
  ext_data.sweepSourceResetFlag = tiInfo.sweepSourceResetFlag;

  return bsuccess;
}

} // namespace Circuit
} // namespace Xyce
