//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright(C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Filename       : $RCSfile: N_DEV_OpBuilders.C,v $
//
// Purpose        : Output Manager
//
// Special Notes  :
//
// Creator        : David G. Baur, Raytheon
//
// Creation Date  : 11/11/2014
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.3.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:09 $
//
// Current Owner  : $Author: tvrusso $
//-------------------------------------------------------------------------

#include <Xyce_config.h>

#include <string>
#include <vector>

#include <N_DEV_fwd.h>
#include <N_UTL_fwd.h>

#include <N_DEV_Op.h>
#include <N_DEV_DeviceMgr.h>
#include <N_UTL_Algorithm.h>
#include <N_UTL_OpBuilder.h>
#include <N_UTL_Param.h>

namespace Xyce {
namespace Device {

namespace {

void parameterNameAndArgs(std::string &name, std::vector<std::string> &args, Util::ParamList::const_iterator &it) 
{
  const std::string &param_tag = (*it).tag();

  if (param_tag[0] == 'V' || param_tag[0] == 'I' || param_tag[0] == 'N')
  {
    std::ostringstream oss;
    oss << param_tag << "(";
    int arg_count = (*it).getImmutableValue<int>();
    for (int i = 0; i < arg_count; ++i)
    {
      ++it;
      if (i != 0)
        oss << ",";
      oss << (*it).tag();
      args.push_back((*it).tag());
    }
    oss << ")";
    name = oss.str();
  }
}

} // namespace <unnamed>


struct DeviceGlobalParameterOpBuilder : public Util::Op::Builder
{
  DeviceGlobalParameterOpBuilder(const DeviceMgr &device_manager)
    : deviceManager_(device_manager)
  {}

  virtual ~DeviceGlobalParameterOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<DeviceMgrGlobalParameterOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const
  {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();
    const std::string &param_string = (*it).stringValue();

    if (param_tag == "GLOBAL_PARAMETER") {
      new_op  = new DeviceMgrGlobalParameterOp(param_string, deviceManager_, param_string);
      new_op->addArg(param_string);
    }
    else
    {
      const double *result = deviceManager_.findGlobalPar(param_tag);
      if (result)
      {
        // Refactor: [DGB] Should use the address.
        new_op = new DeviceMgrGlobalParameterOp(param_tag, deviceManager_, param_tag);
      }
    }

    return new_op;
  }

private:
  const DeviceMgr &     deviceManager_;
};

struct DeviceEntityOpBuilder : public Util::Op::Builder
{
  DeviceEntityOpBuilder(const DeviceMgr &device_manager)
    : deviceManager_(device_manager)
  {}

  virtual ~DeviceEntityOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<DeviceEntityParameterOp>();
  }

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();
    const std::string &param_string = (*it).stringValue();

    const DeviceEntity *device_entity = deviceManager_.getDeviceEntity(param_tag);
    if (device_entity)
    {
      std::string param_name = Util::paramNameFromFullParamName(param_tag);
      new_op = new DeviceEntityParameterOp(param_tag, *device_entity, param_name);
    }

    return new_op;
  }

private:
  const DeviceMgr &     deviceManager_;
};


struct DeviceMgrOpBuilder : public Util::Op::Builder
{
  DeviceMgrOpBuilder(const DeviceMgr &device_manager)
    : deviceManager_(device_manager)
  {}

  virtual ~DeviceMgrOpBuilder()
  {}

  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();
    const std::string &param_string = (*it).stringValue();

    if (deviceManager_.findParam(param_tag))
    {
      new_op = new DeviceMgrParameterOp(param_tag, deviceManager_, param_tag);
    }
    return new_op;
  }

private:
  const DeviceMgr &     deviceManager_;
};


struct DeviceOptionsOpBuilder : public Util::Op::Builder
{
  DeviceOptionsOpBuilder(const DeviceOptions &device_options)
    : deviceOptions_(device_options)
  {}

  virtual ~DeviceOptionsOpBuilder()
  {}

  virtual void registerCreateFunctions(Util::Op::BuilderManager &builder_manager) const
  {
    builder_manager.addCreateFunction<DeviceOptionsOp>();
  }
  
  virtual Util::Op::Operator *makeOp(Util::ParamList::const_iterator &it) const {
    Util::Op::Operator *new_op = 0;
    const std::string &param_tag = (*it).tag();
    const std::string &param_string = (*it).stringValue();

    if (compare_nocase(param_tag.c_str(), "gmin") == 0)
    {
      new_op = new DeviceOptionsOp(param_tag, deviceOptions_, param_tag);
    }
    return new_op;
  }

private:
  const DeviceOptions &     deviceOptions_;
};

//-----------------------------------------------------------------------------
// Function      : registerOpBuilders
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Dave Shirley, PSSI
// Creation Date : 8/15/06
//-----------------------------------------------------------------------------
void registerOpBuilders(Util::Op::BuilderManager &builder_manager, Parallel::Machine comm, DeviceMgr &device_manager)
{
  builder_manager.addBuilder(new DeviceGlobalParameterOpBuilder(device_manager));
  builder_manager.addBuilder(new DeviceEntityOpBuilder(device_manager));
  builder_manager.addBuilder(new DeviceOptionsOpBuilder(device_manager.getDeviceOptions()));
//  builder_manager.addBuilder(new DeviceMgrOpBuilder(device_manager));
}

} // namespace IO
} // namespace Xyce
