//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_IO_MeasureDerivativeEvaluation.C,v $
// Purpose       : Transient analysis functions.
// Special Notes :
// Creator       : Richard Schiek, SNL, Electrical and Microsystem Modeling
// Creation Date : 03/10/2009
//
// Revision Information:
// ---------------------
// Revision Number: $Revision: 1.27.2.1 $
// Revision Date  : $Date: 2015/04/02 18:20:14 $
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

#include <N_IO_MeasureDerivativeEvaluation.h>
#include <N_ERH_ErrorMgr.h>

namespace Xyce {
namespace IO {
namespace Measure {

//-----------------------------------------------------------------------------
// Function      : DerivativeEvaluation::DerivativeEvaluation()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and Microsystems Modeling
// Creation Date : 3/10/2009
//-----------------------------------------------------------------------------
DerivativeEvaluation::DerivativeEvaluation( const Util::OptionBlock & measureBlock):
  Base(measureBlock),
  firstTimeValue_(0.0),
  lastTimeValue_(0.0),
  firstSignalValue_(0.0),
  lastSignalValue_(0.0),
  lastIndepVarValue_(0.0),
  lastDepVarValue_(0.0)
{
  // indicate that this measure type is supported and should be processed in simulation
  typeSupported_ = true;
}

void DerivativeEvaluation::prepareOutputVariables() 
{
  // this measurement should have only one dependent variable.
  // Error for now if it doesn't
  numOutVars_ = outputVars_.size();

  if ( (numOutVars_ > 1) && atGiven_ )
  {
    std::string msg = "Too many dependent variables for statistical measure, \"" + name_ + "\" Exiting.";
    N_ERH_ErrorMgr::report( N_ERH_ErrorMgr::USR_FATAL, msg);
  }

  outVarValues_.resize( numOutVars_, 0.0 );
}

//-----------------------------------------------------------------------------
// Function      : DerivativeEvaluation::reset()
// Purpose       : Called when restarting a measure function.  Resets any state
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and Microsystems Modeling
// Creation Date : 8/28/2014
//-----------------------------------------------------------------------------
void DerivativeEvaluation::reset() 
{
  resetBase();
  firstTimeValue_ = 0.0;
  firstSignalValue_ = 0.0;
}


//-----------------------------------------------------------------------------
// Function      : DerivativeEvaluation::updateTran()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and Microsystems Modeling
// Creation Date : 3/10/2009
//-----------------------------------------------------------------------------
void DerivativeEvaluation::updateTran(Parallel::Machine comm, const double circuitTime, const Linear::Vector *solnVec, const Linear::Vector *stateVec, const Linear::Vector *storeVec)
{
    // update our outVarValues_ vector
    updateOutputVars(comm, outVarValues_, circuitTime, solnVec, stateVec, storeVec, 0);

    if( !initialized_ )
    {
      // assigned last dependent and independent var to current time and outVarValue_[0] 
      // while we can't interpolate on this step, it ensurs that the initial history is
      // something realistic
      lastIndepVarValue_=circuitTime;
      lastDepVarValue_=outVarValues_[0];
      initialized_=true;
    } 

    if (atGiven_)
    { 
      // check an see if last point and this point bound the target point 
      double backDiff    = lastIndepVarValue_ - at_;
      double forwardDiff = circuitTime - at_;
    
      // if we bound the time target then either
      //  (backDiff < 0) && (forwardDiff > 0)  
      //   OR
      //  (backDiff > 0) && (forwardDiff < 0) 
      // or more simply sgn( backDiff ) = - sgn( forwardDiff ) 
      if( ((backDiff < 0.0) && (forwardDiff > 0.0)) || ((backDiff > 0.0) && (forwardDiff < 0.0)) )
      {
        // The interpolated value at time = atVal_ is calculated for debugging purposes.  It is not
        // currently used in the measure calculation. 
        double xVal = lastDepVarValue_ + (outVarValues_[0] - lastDepVarValue_)* 
	            ((at_ - lastIndepVarValue_)/(circuitTime - lastIndepVarValue_));

        // asymmetrical 3-point approximation for first derivative.  
        calculationResult_ = (outVarValues_[0] - lastDepVarValue_) / (circuitTime - lastIndepVarValue_); 
        calculationDone_ = true;
      }
    }
    else if ( ( outputValueTargetGiven_ || (numOutVars_ == 2) ) && !calculationDone_ 
	      && withinTimeWindow( circuitTime ) )
    {
      double targVal=0.0;
      bool doneIfFound=false;
      if( outputValueTargetGiven_ )
      {
        // This is the form WHEN v(a)=fixed value
        targVal = outputValueTarget_;
        doneIfFound = true;
      }
      else
      {
        // This is the form WHEN v(a)= potentially changing value
        targVal = outVarValues_[1];
        // since we can't determine if the calculation is done at this piont
        // we don't set calculationDone_ = true;
        doneIfFound = false;
      }
    
      // I think type_ is always "WHEN" here need to check 
      if ( withinRiseFallCrossWindow( outVarValues_[0], targVal ) )
      {
        // this is the simple case where Xyce output a value within tolerance 
        // to the target value 
        if( fabs(outVarValues_[0] - targVal) < minval_ )
        {
          calculationResult_ = circuitTime;
          calculationDone_ = measureLastRFC_ ? false : doneIfFound;
        }
        else
        {
          // check an see if last point and this point bound the target point 
          double backDiff    = lastDepVarValue_ - targVal;
          double forwardDiff = outVarValues_[0] - targVal;

          // if we bound the target then either
          //  (backDiff < 0) && (forwardDiff > 0)  
          //   OR
          //  (backDiff > 0) && (forwardDiff < 0) 
          // or more simply sgn( backDiff ) = - sgn( forwardDiff ) 
          if( ((backDiff < 0.0) && (forwardDiff > 0.0)) || ((backDiff > 0.0) && (forwardDiff < 0.0)) )
          {
            // The interpolated value at time = atVal_ is calculated for debugging purposes.  It is not
            // currently used in the measure calculation. 
            double xVal = lastDepVarValue_ + (outVarValues_[0] - lastDepVarValue_)* 
	                  ((at_ - lastIndepVarValue_)/(circuitTime - lastIndepVarValue_));

            // use same time interpolation algorithm as FIND-WHEN measure
            calculationTime_ = circuitTime - ( ((circuitTime - lastIndepVarValue_)/(outVarValues_[0]-lastDepVarValue_)) * (outVarValues_[0]-targVal) );

            // asymmetrical 3-point approximation for first derivative.  
            calculationResult_ = (outVarValues_[0] - lastDepVarValue_) / (circuitTime - lastIndepVarValue_);         
            calculationDone_ = measureLastRFC_ ? false : doneIfFound;
          }
        }
      }
    }
    
    lastIndepVarValue_ = circuitTime;
    lastDepVarValue_ = outVarValues_[0]; 
}


//-----------------------------------------------------------------------------
// Function      : DerivativeEvaluation::updateDC()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and Microsystems Modeling
// Creation Date : 3/10/2009
//-----------------------------------------------------------------------------
void DerivativeEvaluation::updateDC(Parallel::Machine comm, const std::vector<Analysis::SweepParam> & dcParamsVec, const Linear::Vector *solnVec, const Linear::Vector *stateVec, const Linear::Vector *storeVec)
{

}

//-----------------------------------------------------------------------------
// Function      : DerivativeEvaluation::getMeasureResult()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and Microsystems Modeling
// Creation Date : 3/10/2009
//-----------------------------------------------------------------------------
double DerivativeEvaluation::getMeasureResult()
{
  return calculationResult_;
}

//-----------------------------------------------------------------------------
// Function      : DerivativeEvaluation::printMeasureResult()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Pete Sholander, Electrical and Microsystems Modeling
// Creation Date : 2/22/2015
//-----------------------------------------------------------------------------
std::ostream& DerivativeEvaluation::printMeasureResult(std::ostream& os, bool printVerbose)
{
  basic_ios_all_saver<std::ostream::char_type> save(os);
  os << std::setprecision(precision_);
  if (!printVerbose)
  {
    os << name_ << " = " << this->getMeasureResult() << std::endl;
  }
  else
  {
    if (calculationDone_)
    {
      os << name_ << " = " << this->getMeasureResult() ;
    }
    else
    { 
      os << name_ << " = FAILED";
    }
    if (atGiven_)
    { 
      os << " for AT = " << at_ << std::endl;
    }
    else
    {    
      os << " at time = " << calculationTime_ << std::endl;
    }
  } 

  return os;
}

//-----------------------------------------------------------------------------
// Function      : DerivativeEvaluation::printMeasureWindow
// Purpose       : prints information related to measure window
// Special Notes :
// Scope         : public
// Creator       : Pete Sholander, Electrical and Microsystem Modeling
// Creation Date : 02/5/2015
//-----------------------------------------------------------------------------
std::ostream& DerivativeEvaluation::printMeasureWindow(std::ostream& os, const double endSimTime)
{
  //no op
  if (!atGiven_) 
  { 
    Base::printMeasureWindow(os,endSimTime); 
  }
  else
  {
    // no op if AT keyword was given
  }

  return os;
}

} // namespace Measure
} // namespace IO
} // namespace Xyce
