//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Filename      :  NetlistImportTool.h
//
// Purpose       : Declare the interface to read and parse a netlist for
//                 an electrical circuit.
//
// Special Notes :
//
// Creator       : Lon Waters
//
// Creation Date : 07/28/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.60.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:14 $
//
// Current Owner  : $Author: tvrusso $
//-------------------------------------------------------------------------

#ifndef NetlistImportTool_H
#define NetlistImportTool_H

#include <string>
#include <vector>

#include <N_IO_fwd.h>
#include <N_PDS_fwd.h>
#include <N_TOP_fwd.h>

#include <N_IO_CircuitBlock.h>
#include <N_IO_CircuitContext.h>
#include <N_IO_CircuitMetadata.h>
#include <N_UTL_OptionBlock.h>
#include <N_UTL_Stats.h>
#include <N_PDS_Comm.h>

namespace Xyce {
namespace IO {

class UndefinedName
{
public:
  UndefinedName()
    : name_(),
      netlistLocation_()
  {}

  UndefinedName(const std::string &name, const NetlistLocation &netlist_location)
    : name_(name),
      netlistLocation_(netlist_location)
  {}

  UndefinedName(const UndefinedName &s0)
    : name_(s0.name_),
      netlistLocation_(s0.netlistLocation_)
  {}

  UndefinedName &operator=(const UndefinedName &s0) {
    name_ = s0.name_;
    netlistLocation_ = s0.netlistLocation_;
    return *this;
  }

public:
  const std::string &getName() const {
    return name_;
  }

  void setName(const std::string &name) {
    name_ = name;
  }

  const NetlistLocation &getNetlistLocation() const {
    return netlistLocation_;
  }

  void setNetlistLocation(const NetlistLocation &netlist_location) {
    netlistLocation_ = netlist_location;
  }
  
private:
  std::string         name_;
  NetlistLocation     netlistLocation_;
};

inline bool operator<(const UndefinedName &s0, const UndefinedName &s1) {
  return s0.getName() < s1.getName();
}

typedef std::set<UndefinedName> UndefinedNameSet;

//-----------------------------------------------------------------------------
// Class         : NetlistImportTool
// Purpose       :
// Special Notes :
// Creator       : Lon Waters, SNL
// Creation Date : 07/28/00 ?
//-----------------------------------------------------------------------------
class NetlistImportTool
{
public:
  
  NetlistImportTool(Util::Op::BuilderManager &op_builder_manager);

  ~NetlistImportTool();

  // R Result
  // R- The result of constructing a circuit from a netlist.
  // I netlistFile
  // I- The file containing the netlist describing the circuit to be
  // I- constructed.
  // This function performs three basic steps. First, it reads and parses a
  // netlist from a specified file and stores the parsed fields in a structure
  // as strings. Second, it interprets the parsed data and stores the data in
  // an appropriate structure. Finally, it builds the Xyce Circuit.
  int constructCircuitFromNetlist(
    const CmdParse &                                            command_line,
    HangingResistor &                                           hanging_resistor,
    const std::string &                                         netlistFilename,
    const std::vector< std::pair< std::string, std::string> > & externalNetlistParams,
    DistributionTool &                                          distribution_tool,
    Topo::Topology &                                            topology,
    Parallel::Communicator &                                    pds_comm,
    PkgOptionsMgr &                                             options_manager,
    OutputMgr &                                                 output_manager,
    Device::DeviceMgr &                                         device_manager,
    Measure::Manager &                                          measure_manager,
    Device::DeviceNameInstanceBlockMap &                        device_names);

  bool registerDCOptions(const Util::OptionBlock & option_block);
  bool registerSTEPOptions(const Util::OptionBlock & option_block);

  const IO::AliasNodeMap &getAliasNodeMap() const {
    return aliasNodeMap_;
  }

  const UndefinedNameSet &getDeferredParameterCheck() const {
    return deferredUndefinedParameters_;
  }
  
  const std::map<std::string, Util::Param> &getMainContextFunctions() const {
    return circuitContext_.getFunctions();
  }

  const OptionBlock &getMainContextParams() const {
    return circuitContext_.getParams();
  }

  const OptionBlock &getMainContextGlobalParams() const {
    return circuitContext_.getGlobals();
  }

  bool getUseMOR() const
  {
    return useMOR_;
  }

public:

private:
  std::set<std::string>                 nodeNames_;
  std::vector<std::string>              stepParams_;
  std::vector<std::string>              dcParams_;
  UndefinedNameSet                      deferredUndefinedParameters_;

  // Maps for parsing information and streams.
  std::map<std::string, FileSSFPair>    ssfMap_;
  std::map<std::string, NetlistLocation> iflMap_;


  CircuitBlock *                        mainCircuitBlock_;
  AliasNodeMap                          aliasNodeMap_;

  std::list<CircuitContext *>           contextList_;
  CircuitContext *                      currentContextPtr_;

  CircuitMetadata                       metadata_;
  CircuitContext                        circuitContext_;
  std::map<std::string, int>            modelNames_;
  bool                                  useMOR_;
  int                                   useCount_;  // Counts copies so deletion of subcircuitList can be done properly
                                                    // in the class destructor.
};

bool registerPkgOptionsMgr(NetlistImportTool &netlist_import_tool, const std::string &netlist_filename, PkgOptionsMgr &options_manager);

int getDeviceNames(const Device::DeviceNameInstanceBlockMap &device_instance_block_map, std::vector<std::string> & names);

void checkDeviceNames(const Device::DeviceNameInstanceBlockMap &device_instance_block_map, const std::vector<std::string> & names);

// check for name collisions between nodes and devices
void checkNodeDevConflicts(const Device::DeviceNameInstanceBlockMap &device_instance_block_map, N_PDS_Comm &pdsComm);

void getLeadCurrentDevices(const Util::ParamList &variable_list, std::set<std::string> &devicesNeedingLeadCurrents);

void printLineDiagnostics(
  Parallel::Machine                             comm,
  const OutputParameterMap &                    output_parameter_map,
  const Device::DeviceMgr &                     device_manager,
  const Measure::Manager &                      measure_manager,
  const std::set<std::string> &                 node_names,
  const std::vector<std::string> &              step_params,
  const std::vector<std::string> &              dc_params,
  const Device::DeviceNameInstanceBlockMap &    device_map,
  const IO::AliasNodeMap &                      alias_node_map,
  UndefinedNameSet &                            deferred_parameter_check);

void deferredPrintLineDiagnostics(
  Parallel::Machine                             comm,
  const UndefinedNameSet &                      deferred_parameter_check,
  const Device::DeviceMgr &                     device_manager);

} // namespace IO
} // namespace Xyce

#endif // NetlistImportTool_H
