//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright(C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Filename       : $RCSfile: N_IO_InitialConditions.C,v $
//
// Purpose        : Output Manager
//
// Special Notes  :
//
// Creator        : Robert J. Hoekstra, SNL, Parallel Computational Sciences
//
// Creation Date  : 10/10/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.17.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:14 $
//
// Current Owner  : $Author: tvrusso $
//-------------------------------------------------------------------------

#include <Xyce_config.h>

#include <algorithm>
#include <iterator>

#include <N_ERH_Message.h>
#include <N_IO_InitialConditions.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_LAS_Vector.h>
#include <N_PDS_MPI.h>
#include <N_PDS_Serial.h>
#include <N_UTL_Marshal.h>
#include <N_UTL_ExtendedString.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_OptionBlock.h>
#include <N_UTL_Param.h>

namespace Xyce {
namespace IO {

namespace {

//-----------------------------------------------------------------------------
// Class         : DCOPOptionsReg
// Purpose       : functor for registering DCOP options
// Special Notes : Used by package manager submitRegistration method
// Creator       : dgbaur
// Creation Date : 06/28/2013
//-----------------------------------------------------------------------------
struct DCOPOptionsReg : public PkgOptionsReg
{
  DCOPOptionsReg( InitialConditionsManager &initial_conditions_manager )
    : initialConditionsManager_(initial_conditions_manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  {
    return initialConditionsManager_.registerDCOPOptions( options );
  }

  InitialConditionsManager &    initialConditionsManager_;
};

//-----------------------------------------------------------------------------
// Class         : ICOptionsReg
// Purpose       : functor for registering IC options
// Special Notes : Used by package manager submitRegistration method
// Creator       : dgbaur
// Creation Date : 06/28/2013
//-----------------------------------------------------------------------------
struct ICOptionsReg : public PkgOptionsReg
{
  ICOptionsReg( InitialConditionsManager &initial_conditions_manager )
    : initialConditionsManager_(initial_conditions_manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  {
    return initialConditionsManager_.registerIC ( options );
  }

  InitialConditionsManager &    initialConditionsManager_;
};



//-----------------------------------------------------------------------------
// Class         : NodeSetOptionsReg
// Purpose       : functor for registering NodeSet options
// Special Notes : Used by package manager submitRegistration method
// Creator       : dgbaur
// Creation Date : 06/28/2013
//-----------------------------------------------------------------------------
struct NodeSetOptionsReg : public PkgOptionsReg
{
  NodeSetOptionsReg( InitialConditionsManager &initial_conditions_manager )
    : initialConditionsManager_(initial_conditions_manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  {
    return initialConditionsManager_.registerNodeSet( options );
  }

  InitialConditionsManager &    initialConditionsManager_;
};



//-----------------------------------------------------------------------------
// Class         : SaveOptionsReg
// Purpose       : functor for registering Save options
// Special Notes : Used by package manager submitRegistration method
// Creator       : dgbaur
// Creation Date : 06/28/2013
//-----------------------------------------------------------------------------
struct SaveOptionsReg : public PkgOptionsReg
{
  SaveOptionsReg( InitialConditionsManager &initial_conditions_manager )
    : initialConditionsManager_(initial_conditions_manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  {
    return initialConditionsManager_.registerSave( options );
  }

  InitialConditionsManager &    initialConditionsManager_;
};

} // namespace <unnamed>

bool registerPkgOptionsMgr(InitialConditionsManager & output_manager, const std::string &netlist_filename, PkgOptionsMgr &options_manager)
{
  options_manager.submitRegistration(
    "OP_IO", netlist_filename, new DCOPOptionsReg(output_manager));

  options_manager.submitRegistration(
    "IC", netlist_filename, new ICOptionsReg(output_manager));

  options_manager.submitRegistration(
    "NODESET", netlist_filename, new NodeSetOptionsReg(output_manager));

  options_manager.submitRegistration(
    "SAVE", netlist_filename, new SaveOptionsReg(output_manager));

  return true;
}

//-----------------------------------------------------------------------------
// Function      : InitialConditionsManager::InitialConditionsManager
// Purpose       : constructor
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 10/10/00
//-----------------------------------------------------------------------------
InitialConditionsManager::InitialConditionsManager(
  const std::string &         netlist_filename)
  : netlistFilename_(netlist_filename),
    outputOnceAlreadyFlag_(false)
{}

//-----------------------------------------------------------------------------
// Function      : InitialConditionsManager::registerSave
// Purpose       : registers set of variables to set for .SAVE.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Electrical and Microsystems Modeling
// Creation Date : 10/11/07
//-----------------------------------------------------------------------------
bool InitialConditionsManager::registerSave(const Util::OptionBlock & option_block)
{
  if (DEBUG_IO)
    Xyce::dout() << " SAVE Registered::" << std::endl;

  icData_.saveFlag_ = true;

  ExtendedString sval("");

  for (Util::ParamList::const_iterator it = option_block.begin(), end = option_block.end(); it != end; ++it)
  {
    if (DEBUG_IO)
      Xyce::dout() << "(*it).tag = " << (*it).tag() << std::endl;

    if ((*it).tag() == "TYPE")
    {
      sval = (*it).stringValue();
      sval.toUpper();

      if (sval == "NODESET" || sval == ".NODESET")
      {
        icData_.saveFileType_ = ".NODESET";
      }
      else if (sval == "IC" || sval == ".IC")
      {
        icData_.saveFileType_ = ".IC";
      }
      else
      {
        Report::UserWarning0() << "Unrecognized type specified on .SAVE command.  Defaulting to .NODESET";
      }
    }
    else if ((*it).tag() == "FILE")
    {
      icData_.saveOutputFile_ = (*it).stringValue();
    }
    else if ((*it).tag() == "TIME")
    {
      // do nothing, this will be handled in the time integrator.
    }
    else if ((*it).tag() == "LEVEL")
    {
      sval = (*it).stringValue();
      sval.toUpper();

      if (sval == "ALL")
      {
        // do nothing
      }
      else if (sval == "NONE")
      {
        // none means don't output anything.  Pretend the .SAVE line isn't in the netlist.
        icData_.saveFlag_ = false;
        icData_.saveFileLevel_ = "NONE";
      }
      else if (sval == "TOP")
      {
        Report::UserWarning0() << "LEVEL=TOP in .SAVE line not supported.  Ignoring. ";
      }
      else
      {
        Report::UserWarning0() << "Unrecognized LEVEL " << sval << " specified in .SAVE command";
      }
    }
    else
    {
      Report::UserWarning0() << "Parameter " << (*it).tag() << " not recognized in .SAVE command";
    }
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : InitialConditionsManager::registerIC
// Purpose       : registers set of variables to set for .IC or .DCVOLT
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Electrical and Microsystems Modeling
// Creation Date : 09/06/07
//-----------------------------------------------------------------------------
bool InitialConditionsManager::registerIC(const Util::OptionBlock & OB)
{
  icData_.ICflag_ = true;

  // save the ICblock_.  This will be needed later.
  // We allow multiple .IC blocks in the netlist, so there needs to be an
  // STL vector of these.
  ICblockVec_.push_back(OB);

  return true;
}

//-----------------------------------------------------------------------------
// Function      : InitialConditionsManager::registerNodeSet
// Purpose       : registers set of variables to set for .NODESET
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Electrical and Microsystems Modeling
// Creation Date : 09/06/07
//-----------------------------------------------------------------------------
bool InitialConditionsManager::registerNodeSet(const Util::OptionBlock & OB)
{
 icData_.nodesetflag_ = true;

  // save the nodesetblock_.  This will be needed later.
  // We allow multiple .nodeset blocks in the netlist, so there needs to be an
  // STL vector of these.
  nodesetblockVec_.push_back(OB);

  return true;
}

//-----------------------------------------------------------------------------
// Function      : InitialConditionsManager::registerDCOPOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Dave Shirley, PSSI
// Creation Date : 05/08/06
//-----------------------------------------------------------------------------
bool InitialConditionsManager::registerDCOPOptions(const Util::OptionBlock & option_block)
{
  for (Util::ParamList::const_iterator it = option_block.begin(); it != option_block.end(); ++it)
  {
    if ((*it).tag() == "INPUT")
    {
      icData_.input_op_ = true;
      icData_.input_op_file_ = (*it).stringValue();
    }
    else if ((*it).tag() == "OUTPUT")
    {
      icData_.output_op_ = true;
      icData_.output_op_file_ = (*it).stringValue();
    }
    else if ((*it).tag() == "TIME")
    {
      // do nothing, this will be handled in the time integrator.
    }
    else
    {
      Report::UserWarning0() << "Parameter " << (*it).tag() << " not recognized in .DCOP command";
    }
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : InitialConditionsManager::getICData
// Purpose       : Provide OP data for LOCA
// Special Notes :
// Scope         : public
// Creator       : Dave Shirley, PSSI
// Creation Date : 05/04/06
//-----------------------------------------------------------------------------
InitialConditionsData::NodeNamePairMap & InitialConditionsManager::getICData(int & op_found, int & icType)
{
  op_found = icData_.op_found_;
  icType = icData_.icType_;

  return icData_.opData_;
}

//-----------------------------------------------------------------------------
// Function      : InitialConditionsManager::setupIC_or_NODESET
// Purpose       :
// Special Notes : Assumes that allNodes_ is set up.
// Scope         : private
// Creator       : Eric Keiter
// Creation Date : 09/13/07
//-----------------------------------------------------------------------------
bool
setupIC_or_NODESET(
  Parallel::Machine                             comm,
  const NodeNameMap &                           all_nodes,
  Linear::Vector &                              nextSolnVec,
  Linear::Vector &                              flagSolnVec,
  int                                           icType,
  std::vector<Util::OptionBlock> &              initBlockVec,
  InitialConditionsData::NodeNamePairMap &      op_data,
  int &                                         op_found_,
  int &                                         total_soln_)
{
  int lid = 0;
  bool success = false;

  int totalNodes = all_nodes.size();

  nextSolnVec.putScalar(0.0);
  flagSolnVec.putScalar(-1.0);

  for (NodeNameMap::const_iterator it = all_nodes.begin(), nodes_end = all_nodes.end(); it != nodes_end ; ++it)
  {
    flagSolnVec[(*it).second] = 0;
  }

  std::set<std::string> notFoundInCkt;
  std::set<std::string> matched;
  for (std::vector<Util::OptionBlock>::const_iterator it = initBlockVec.begin(), end = initBlockVec.end(); it != end; ++it)
  {
    // param loop
    for (Util::ParamList::const_iterator it_param = (*it).begin(), endPar = (*it).end(); it_param != endPar; ++it_param)
    {
      // the first tag is always "V".  At this point I variables are not supported,
      // and there is an error trap for them in the OptionBlock.C file.
      ++it_param;
      std::string node = (*it_param).tag();
      ++it_param;

      ExtendedString strValue((*it_param).tag());
      double value = 0.0;
      if (strValue.isValue())
      {
        value = strValue.Value();
      }
      else
      {
        Report::UserFatal0() << "Problems processing initial condition for node " 
                             << node << " to " << strValue << " value";
      }

      int found = false;

      NodeNameMap::const_iterator iterCI = all_nodes.find(node);
      if (iterCI != all_nodes.end())
      {
        lid = iterCI->second;
        nextSolnVec[lid] = value;
        flagSolnVec[lid] = 1;
        found = true;
      }

      if (DEBUG_IO) {
        if (found)
        {
          Xyce::dout().width(10);Xyce::dout().precision(3);Xyce::dout().setf(std::ios::scientific);
          Xyce::dout()
            << icType + " found, and set: V(" << node << "):  solution["<<lid<<"] = " << value << std::endl;
        }
      }

      Parallel::AllReduce(comm, MPI_SUM, &found, 1);

      if (found)
      {
        success = true;

        if (matched.find(node) == matched.end())
        {
          matched.insert(node);
        }
      }
      else
      {
        notFoundInCkt.insert(node);
      }
    }
  }

  nextSolnVec.importOverlap();
  flagSolnVec.importOverlap();

  op_found_ = matched.size();
  total_soln_ = totalNodes;

  std::set<std::string>::iterator matched_i = matched.begin();
  std::set<std::string>::iterator matched_end = matched.end();
  for (; matched_i != matched_end ; ++matched_i)
  {
    // only add this to the opData map if it is local to the processor.
    // The allNodes object only contains local stuff.  Otherwise there
    // isn't much point.
    NodeNameMap::const_iterator iterCI = all_nodes.find(*matched_i);
    //if (all_nodes.find(*matched_i) != all_nodes.end())
    if (iterCI != all_nodes.end())
    {
      op_data[*matched_i].first = iterCI->second;
      op_data[*matched_i].second = nextSolnVec[op_data[*matched_i].first];
    }
  }

  if (DEBUG_IO)
  {
    // Identify nodes in the circuit which were not set.
    std::set<std::string> local;
    for (NodeNameMap::const_iterator it = all_nodes.begin(), end = all_nodes.end(); it != end ; ++it)
    {
      if (matched.find((*it).first) == matched.end())
        local.insert((*it).first);
    }

    Util::Marshal mout;
    mout << local;

    std::vector<std::string> dest;
    Parallel::GatherV(comm, 0, mout.str(), dest);

    std::set<std::string> notSpecified;
    if (Parallel::rank(comm) == 0) {
      for (int p = 0; p < Parallel::size(comm); ++p) {
        Util::Marshal min(dest[p]);

        std::vector<std::string> x;
        min >> x;
        std::copy(x.begin(), x.end(), std::inserter(notSpecified, notSpecified.begin()));
      }
    }
  }

  if (Parallel::rank(comm) == 0)
  {
    if (DEBUG_IO)
    {
      if (totalNodes > matched.size())
      {
        Xyce::dout() << icType << ":  Initialized " << matched.size() << " of a possible " << totalNodes << " nodes" << std::endl;
      }
      else
      {
        Xyce::dout() << icType << ":  All " << totalNodes << " nodes initialized" << std::endl;
      }
    }

    if (notFoundInCkt.size() > 0)
    {
      lout() << icType << ":  Nodes specified on ." << icType << " line, but not present in circuit. (ignoring):" << std::endl;

      for (std::set<std::string>::iterator nm_i = notFoundInCkt.begin(), end = notFoundInCkt.end(); nm_i != end; ++nm_i)
      {
        lout() << *nm_i << std::endl;
      }
    }

    // if (DEBUG_IO)
    // {
    //   // Note: for a typical .IC line, this list of unspecified nodes will be
    //   // a very long list.  So, don't output except in debug mode.
    //   if (notSpecified.size() > 0)
    //   {
    //     dout() << icType << ":  Nodes present in circuit, but not specified on ." << icType << " line(ignoring):" << std::endl;

    //     for (std::set<std::string>::iterator nm_i = notSpecified.begin(), end = notSpecified.end(); nm_i != end; ++nm_i)
    //     {
    //       dout() << *nm_i << std::endl;
    //     }
    //   }
    // }
  }

  return success;
}

//-----------------------------------------------------------------------------
// Function      : InitialConditionsManager::outputIC_or_NODESET
// Purpose       : Outputs either an *.ic file, either in .ic or .nodeset format.
// Special Notes :
// Scope         : private
// Creator       : Eric Keiter, SNL.
// Creation Date :
//-----------------------------------------------------------------------------
void
outputIC_or_NODESET(
  Parallel::Machine             comm,
  std::ofstream &               os,
  const std::string &           saveFileType_,
  InitialConditionsData::NodeNamePairMap               all_nodes,
  const Linear::Vector &          solution)
{
  for (InitialConditionsData::NodeNamePairMap::iterator it = all_nodes.begin(), name_end = all_nodes.end(); it != name_end ; ++it)
  {
    int index =(*it).second.first;
    (*it).second.second = solution[index];
  }

  Util::Marshal mout;
  mout << all_nodes;

  std::vector<std::string> dest;
  Parallel::GatherV(comm, 0, mout.str(), dest);

  if (Parallel::rank(comm) == 0) {
    InitialConditionsData::NodeNamePairMap global;
    for (int p = 0; p < Parallel::size(comm); ++p) {
      Util::Marshal min(dest[p]);

      InitialConditionsData::NodeNamePairMap x;
      min >> x;
      for (InitialConditionsData::NodeNamePairMap::const_iterator it = all_nodes.begin(), end = all_nodes.end(); it != end ; ++it)
      {
        ExtendedString tmpName((*it).first);
        tmpName.toUpper();
        std::string::size_type pos = tmpName.rfind("BRANCH");
        if (pos == std::string::npos)
        {
          os << saveFileType_ << " V(";
          os <<(*it).first << ") = " << (*it).second.second << std::endl;
        }
      }
    }
  }
}


//-----------------------------------------------------------------------------
// Function      : InitialConditionsManager::writeDCOPRestart
// Purpose       : Output DC operating point solution vars for use in subsequent
//                 runs to speed up DCOP
// Special Notes :
// Scope         : private
// Creator       : Dave Shirley, PSSI
// Creation Date : 04/10/06
//-----------------------------------------------------------------------------
void writeDCOPRestart(Parallel::Machine comm, std::ofstream &os, const InitialConditionsData::NodeNamePairMap &all_nodes)
{
  Util::Marshal mout;
  mout << all_nodes;
  
  std::vector<std::string> dest;
  Parallel::GatherV(comm, 0, mout.str(), dest);

  if (Parallel::rank(comm) == 0) {
    InitialConditionsData::NodeNamePairMap global;
    for (int p = 0; p < Parallel::size(comm); ++p) {
      Util::Marshal min(dest[p]);

      InitialConditionsData::NodeNamePairMap x;
      min >> x;
      for (InitialConditionsData::NodeNamePairMap::const_iterator it = all_nodes.begin(), end = all_nodes.end(); it != end ; ++it)
      {
        os << (*it).first << "   " << (*it).second.second << std::endl;
      }
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : InitialConditionsManager::readDCOPRestart
// Purpose       : Attempt to input better starting estimate of DC operating
//                 point from a previous run
// Special Notes :
// Scope         : private
// Creator       : Dave Shirley, PSSI
// Creation Date : 04/11/06
//-----------------------------------------------------------------------------
bool
readDCOPRestart(
  Parallel::Machine             comm,
  std::ifstream &               is,
  const NodeNameMap &           all_nodes,
  Linear::Vector &                nextSolnVec,
  Linear::Vector &                flagSolnVec,
  InitialConditionsData::NodeNamePairMap &             op_data,
  int &                         op_found_,
  int &                         total_soln_)
{
  bool success = false;
  std::set<std::string> matched;
  std::set<std::string> read;

  flagSolnVec.putScalar(-1.0);
  nextSolnVec.putScalar(0.0);

  // Clear flags
  for (NodeNameMap::const_iterator it = all_nodes.begin(), end = all_nodes.end(); it != end ; ++it)
  {
    flagSolnVec[(*it).second] = 0;
  }

  // Populate nextSolnVec, flagSolnVec, op_data and matched for all nodes mathcing on this processor
  for (;;)
  {
    std::string node;
    double value;
    
    is >> node;
    if (node == "")
      break;
    is >> value;

    // On root proc, keep track of all node names
    if (Parallel::rank(comm) == 0)
      read.insert(node);

    NodeNameMap::const_iterator it = all_nodes.find(node);
    if (it != all_nodes.end())
    {
      int ind = (*it).second;
      nextSolnVec[ind] = value;
      flagSolnVec[ind] = 1;
      if (matched.find(node) == matched.end())
      {
        matched.insert(node);
        op_data[node].first = (*it).second;
        op_data[node].second = nextSolnVec[(*it).second];
      }
    }
  }

  // Overlap the results
  nextSolnVec.importOverlap();
  flagSolnVec.importOverlap();

  int all_node_count = all_nodes.size();

  Parallel::AllReduce(comm, MPI_SUM, &all_node_count, 1);

  Util::Marshal mout;
  mout << matched;

  std::vector<std::string> dest;
  Parallel::GatherV(comm, 0, mout.str(), dest);

  int missing_count = 0;
  if (Parallel::rank(comm) == 0) {
    std::set<std::string> x;
    for (int p = 0; p < Parallel::size(comm); ++p) {
      Util::Marshal min(dest[p]);
      min >> x;
      std::copy(x.begin(), x.end(), std::inserter(matched, matched.begin()));
    }

    std::set<std::string> missing;
    std::set_difference(read.begin(), read.end(), matched.begin(), matched.end(), std::inserter(missing, missing.begin()));

    if (read.size() > matched.size())
      Xyce::dout() << "DCOP restart:  Initialized " << matched.size() << " of a possible " << read.size() << " nodes" << std::endl;
    else
      Xyce::dout() << "DCOP restart:  All " << all_node_count << " nodes initialized" << std::endl;

    missing_count = missing.size();
    if (missing.size() > 0)
    {
      Xyce::dout() << "DCOP restart:  Nodes specified but not present in circuit:" << std::endl;
      for (std::set<std::string>::const_iterator it = missing.begin(), end = missing.end(); it != end; ++it)
        Xyce::dout() << (*it) << std::endl;
    }

    if (all_node_count - read.size() > 0)
      Xyce::dout() << "DCOP restart:  " << (all_node_count - read.size()) << " nodes present in circuit but not specified:" << std::endl;
  }

  Parallel::Broadcast(comm, &missing_count, 1, 0);

  if (missing_count == 0)
    success = true;

  return success;
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::setupInitialConditions
// Purpose       : This function is an umbrella function, under which the
//                 various types of initial conditions are potentially set up.
//                 These include, but aren't limited to .IC, .NODESET and
//                 .DCOP restart.
//
//                 In addition to setting these things up, it does some
//                 nominal checking to make sure that more than on IC hasn't
//                 been specified(ie DCOP restart and .IC can't both be set).
// Special Notes :
// Scope         : private
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 09/13/07
//-----------------------------------------------------------------------------
bool
InitialConditionsManager::setupInitialConditions(
  Parallel::Machine             comm,
  const NodeNameMap &           allNodes_,
  Linear::Vector &                solnVec,
  Linear::Vector &                flagVec)
{
    bool dcopRestart = false;
    bool dotIC = false;
    bool NODESET = false;
    icData_.icType_ = InitialConditionsData::IC_TYPE_UNDEFINED;

    if (icData_.input_op_ && icData_.ICflag_)
    {
      Report::UserFatal0()
        << "Cannot set both DCOP restart and .IC simultaneously.";
    }

    if (icData_.input_op_ && icData_.nodesetflag_)
    {
      Report::UserFatal0()
        << "Cannot set both DCOP restart and .NODESET simultaneously.";
    }

    if (icData_.ICflag_ && icData_.nodesetflag_)
    {
      Report::UserFatal0()
        << "Cannot set both .IC and .NODESET simultaneously.";
    }

    if (icData_.input_op_)
    {
      std::string inputFileName;
      std::ifstream opIn;

      if (icData_.input_op_file_ == "")
      {
        inputFileName = netlistFilename_ + ".op";
      }
      else
      {
        inputFileName = icData_.input_op_file_;
      }
      opIn.open(inputFileName.c_str(), std::ios::in);

      if (opIn.good())
      {
        Xyce::dout() << "Reading in operating point initial estimate from: "
          << inputFileName << std::endl;

        // check for dcop restart(the original)
        dcopRestart = readDCOPRestart(comm, opIn, allNodes_, solnVec, flagVec, icData_.opData_, icData_.op_found_, icData_.total_soln_);
        if (dcopRestart)
          icData_.icType_ = InitialConditionsData::IC_TYPE_DCOP_RESTART;
      }
    }
    else if (icData_.ICflag_)
    {
      // check for .IC
      if (icData_.ICflag_)
        dotIC = setupIC_or_NODESET(comm, allNodes_, solnVec, flagVec, InitialConditionsData::IC_TYPE_IC, ICblockVec_, icData_.opData_, icData_.op_found_, icData_.total_soln_);
      if (dotIC)
        icData_.icType_ = InitialConditionsData::IC_TYPE_IC;
    }
    else if (icData_.nodesetflag_)
    {
      // check for .NODESET
      if (icData_.nodesetflag_)
        NODESET = setupIC_or_NODESET(comm, allNodes_, solnVec, flagVec, InitialConditionsData::IC_TYPE_NODESET, nodesetblockVec_, icData_.opData_, icData_.op_found_, icData_.total_soln_);
      if (NODESET)
        icData_.icType_ = InitialConditionsData::IC_TYPE_NODESET;
    }

    return dcopRestart || dotIC || NODESET;
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::outputDCOP
// Purpose       :
// Special Notes :
// Scope         : private
// Creator       : Eric Keiter, SNL
// Creation Date :
//-----------------------------------------------------------------------------
void
InitialConditionsManager::outputDCOP(
  Parallel::Machine             comm,
  const NodeNameMap &           all_nodes,
  const Linear::Vector &          solution)
{
  InitialConditionsData::NodeNamePairMap             solution_node_names;

  for (NodeNameMap::const_iterator it = all_nodes.begin(), end = all_nodes.end(); it != end; ++it)
    solution_node_names[(*it).first] = std::make_pair((*it).second, 0.0);

  if (outputOnceAlreadyFlag_)
    return;

  if (icData_.output_op_)
  {
    std::ofstream opOut;

    for (InitialConditionsData::NodeNamePairMap::iterator it = solution_node_names.begin(), end = solution_node_names.end(); it != end ; ++it)
    {
      int ind = (*it).second.first;
      (*it).second.second = solution[ind];
    }

    if (Parallel::rank(comm) == 0)
    {
      std::string outputFileName;

      if (icData_.output_op_file_ == "")
        outputFileName = netlistFilename_ + ".op";
      else
        outputFileName = icData_.output_op_file_;
      opOut.open(outputFileName.c_str());
    }

    writeDCOPRestart(comm, opOut, solution_node_names);

    if (Parallel::rank(comm) == 0)
      opOut.close();
  }
  else if (icData_.saveFlag_)
  {
    std::ofstream saveOutputStream;

    if (Parallel::rank(comm) == 0)
    {
      std::string outputFileName;
      if (icData_.saveOutputFile_  == "")
      {
        outputFileName = netlistFilename_ + ".ic";
      }
      else
      {
        outputFileName = icData_.saveOutputFile_;
      }

      saveOutputStream.open(outputFileName.c_str());
    }

    // check for .IC
    outputIC_or_NODESET(comm, saveOutputStream, icData_.saveFileType_, solution_node_names, solution);
    if (Parallel::rank(comm) == 0)
      saveOutputStream.close();
  }

  outputOnceAlreadyFlag_ = true;
}

} // namespace IO
} // namespace Xyce

