//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Filename      : NetlistImportTool.C
//
// Purpose       : Implement the interface to to read and parse a netlist for
//                 an electrical circuit.
//
// Special Notes :
//
// Creator       : Lon Waters
//
// Creation Date : 07/28/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.139.2.2 $
//
// Revision Date  : $Date: 2015/04/02 18:20:14 $
//
// Current Owner  : $Author: tvrusso $
//-------------------------------------------------------------------------

#include <Xyce_config.h>

#include <iostream>
#include <map>
#include <string>
#include <vector>

#include <N_UTL_fwd.h>

#include <N_DEV_DeviceMgr.h>
#include <N_ERH_ErrorMgr.h>
#include <N_IO_CmdParse.h>
#include <N_IO_DistributionTool.h>
#include <N_IO_MORAnalysisTool.h>
#include <N_IO_NetlistImportTool.h>
#include <N_IO_OutputMgr.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_IO_MeasureManager.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_Marshal.h>

namespace Xyce {
namespace Util {

Marshal &operator<<(Marshal &mout, const IO::UndefinedName &undefined_name) 
{
  return mout << undefined_name.getName() << undefined_name.getNetlistLocation();
}

Marshal &operator>>(Marshal &min, IO::UndefinedName &undefined_name) 
{
  std::string name;
  NetlistLocation netlist_location;
  min >> name >> netlist_location;
  undefined_name.setName(name);
  undefined_name.setNetlistLocation(netlist_location);
  return min;
}

}


namespace IO {

using Xyce::DEBUG_DISTRIBUTION;

namespace {

//-----------------------------------------------------------------------------
// Class         : STEPOptionsReg
// Purpose       : functor for registering STEP options
// Special Notes : Used by package manager submitRegistration method
// Creator       : dgbaur
// Creation Date : 06/28/2013
//-----------------------------------------------------------------------------
struct STEPOptionsReg : public PkgOptionsReg
{
  STEPOptionsReg( NetlistImportTool &netlist_import_tool )
    : netlistImportTool_(netlist_import_tool)
  {}

  bool operator()( const Util::OptionBlock & options )
  {
    return netlistImportTool_.registerSTEPOptions( options );
  }

  NetlistImportTool &netlistImportTool_;
};


//-----------------------------------------------------------------------------
// Class         : DCOptionsReg
// Purpose       : functor for registering DC options
// Special Notes : Used by package manager submitRegistration method
// Creator       : dgbaur
// Creation Date : 06/28/2013
//-----------------------------------------------------------------------------
struct DCOptionsReg : public PkgOptionsReg
{
  DCOptionsReg( NetlistImportTool &netlist_import_tool )
    : netlistImportTool_(netlist_import_tool)
  {}

  bool operator()( const Util::OptionBlock & options )
  {
    return netlistImportTool_.registerDCOptions(options);
  }

  NetlistImportTool &netlistImportTool_;
};

void
errorUndefinedParameters(
  const UndefinedNameSet &      undefined_parameters)
{
  std::map<NetlistLocation, std::string> msg_map;
  int count = 0;
  for (UndefinedNameSet::const_iterator it = undefined_parameters.begin(), end = undefined_parameters.end(); it != end; ++it) {
    std::string &msg = msg_map[(*it).getNetlistLocation()];
    if (!msg.empty())
      msg += ", ";
    msg += (*it).getName();
    ++count;
  }

  for (std::map<NetlistLocation, std::string>::const_iterator it = msg_map.begin(), end = msg_map.end(); it != end; ++it) {
    Report::UserError0().at((*it).first)
      << "There " << (count == 1 ? "was " : "were ") << count << " undefined symbol"
      << (count == 1 ? "" : "s") << " in .PRINT command: " << (*it).second;
  }
}

} // namespace <unnamed>

//-----------------------------------------------------------------------------
// Function      : registerPkgOptionsMgr
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, 1437
// Creation Date : 10/21/08
//-----------------------------------------------------------------------------
bool registerPkgOptionsMgr(NetlistImportTool &netlist_import_tool, const std::string &netlist_filename, PkgOptionsMgr &options_manager)
{
  options_manager.submitRegistration(
    "DC", netlist_filename, new DCOptionsReg(netlist_import_tool));

  options_manager.submitRegistration(
    "STEP", netlist_filename, new STEPOptionsReg(netlist_import_tool));

  return true;
}


//-------------------------------------------------------------------------
// Name          : NetlistImportTool::NetlistImportTool
// Purpose       : Default constructor
// Special Notes :
// Creator       : Lon Waters
// Creation Date : 07/28/2000
//-------------------------------------------------------------------------
NetlistImportTool::NetlistImportTool(
  Util::Op::BuilderManager &    op_builder_manager)
  : mainCircuitBlock_(NULL),
    currentContextPtr_(NULL),
    circuitContext_(metadata_, contextList_, currentContextPtr_),
    useMOR_(false),
    useCount_(0),
    metadata_(op_builder_manager)
{}

//-------------------------------------------------------------------------
// Function      : NetlistImportTool::~NetlistImportTool
// Purpose       : Destructor
// Special Notes :
// Creator       : Lon Waters
// Creation Date : 07/28/2000
//-------------------------------------------------------------------------
NetlistImportTool::~NetlistImportTool()
{
  // Destroy the SpiceSeparatedFieldTool map, close the input files.
  for (std::map<std::string, FileSSFPair>::iterator it = ssfMap_.begin(), end = ssfMap_.end(); it != end; ++it)
  {
    delete (*it).second.first;
    delete (*it).second.second;
  }

  delete mainCircuitBlock_;
}

//-------------------------------------------------------------------------
// Function      : NetlistImportTool::constructCircuitFromNetlist
// Purpose       : Construct a circuit from a netlist.
// Special Notes :
// Creator       : Lon Waters
// Creation Date : 07/28/2000
//-------------------------------------------------------------------------
int
NetlistImportTool::constructCircuitFromNetlist(
  const CmdParse &                                              command_line,
  HangingResistor &                                             hanging_resistor,
  const std::string &                                           netlist_filename,
  const std::vector< std::pair< std::string, std::string> > &   externalNetlistParams,
  DistributionTool &                                            distribution_tool,
  Topo::Topology &                                              topology,
  Parallel::Communicator &                                      pds_comm,
  PkgOptionsMgr &                                               options_manager,
  OutputMgr &                                                   output_manager,
  Device::DeviceMgr &                                           device_manager,
  Measure::Manager &                                            measure_manager,
  Device::DeviceNameInstanceBlockMap &                          device_names)
{
  Parallel::Machine comm = pds_comm.comm();

  // Create a circuitBlock instance to hold netlist circuit data.
  mainCircuitBlock_ = new IO::CircuitBlock(
    netlist_filename,
    command_line,
    hanging_resistor,
    metadata_,
    modelNames_ ,
    ssfMap_,
    iflMap_,
    circuitContext_,
    useCount_,
    distribution_tool,
    topology,
    device_manager,
    device_names,
    nodeNames_,
    aliasNodeMap_,
    externalNetlistParams );

  // Parse the netlist file.
  if (DEBUG_IO)
    Xyce::dout() << "Starting netlist parsing." << std::endl;

  {
    Stats::StatTop _parseStat("Parse");
    Stats::TimeBlock _parseTimer(_parseStat);

    // Read in the circuit context and circuit options on the root processor.
    distribution_tool.parseCircuitContext(*mainCircuitBlock_);
  }

  // If we are only performing a device count, we can exit here.
  if (command_line.argExists("-count"))
  {
    return 1;
  }

  // If MOR is requested, analyze the circuit before performing full simulation
  if (mainCircuitBlock_->getMORFlag() && mainCircuitBlock_->getAnalysisName() != "MOR")
  {
    IO::MORAnalysisTool morAnalyzer(&distribution_tool, circuitContext_, mainCircuitBlock_->getOptionsTable(),
                                    command_line, ssfMap_, iflMap_);

     useMOR_ = morAnalyzer.reduceLinearSubcircuits();

     // If model-order reduction was used, there is no need to run the original circuit.
     if (useMOR_) {
       return 1;
     }
     else {
       // Remove .MOR and .OPTIONS MOR_OPTS from option table before continuing simulation
       // MOR was determined to not be viable.
       morAnalyzer.removeMOROptions();
     }
  }

  // Set the circuit context and options table.
  distribution_tool.setCircuitContext(&circuitContext_);
  distribution_tool.setCircuitOptions(mainCircuitBlock_->getOptionsTable());

  // Distribute the circuit context and circuit options to all other processors.
  distribution_tool.broadcastGlobalData(*mainCircuitBlock_, options_manager);

  {
    Stats::StatTop _distributeStat("Distribute");
    Stats::TimeBlock _distributeTimer(_distributeStat);

    // Distribute and instantiate the circuit devices.
    distribution_tool.distributeDevices(*mainCircuitBlock_);
  }

  // Check for name collisions between devices
  checkNodeDevConflicts(device_names, pds_comm);

  if (DEBUG_IO)
    Xyce::dout() << "Completed netlist parsing. ";

  // Write out preprocessed netlist, if requested.
  if (Parallel::rank(comm) == 0)
  {
    mainCircuitBlock_->writeOutNetlist();
  }

  output_manager.setTitle(mainCircuitBlock_->getTitle());

  std::set<std::string> devicesNeedingLeadCurrents;
  getLeadCurrentDevices(output_manager.getVariableList(), devicesNeedingLeadCurrents);

  // This is the last call before devices are constructed
  // So it's the last time I can isolate lead currents. However it won't be sufficient
  // as we haven't parsed expressions in devices yet.  I'll need to rethink
  // when this call is made to the device manager.  Sometime just after device instance
  // construction but before the store vector is allocated.
  device_manager.setLeadCurrentRequests(devicesNeedingLeadCurrents);

  printLineDiagnostics(comm, output_manager.getOutputParameterMap(), device_manager, measure_manager, nodeNames_, stepParams_, dcParams_, device_names, aliasNodeMap_, deferredUndefinedParameters_);

  return 1;
}

namespace {

std::set<std::string>::const_iterator findNode(const std::string &name, const std::set<std::string> &node_set, const AliasNodeMap &alias_map)
{
  std::set<std::string>::const_iterator node_it = node_set.find(name);
  if (node_it == node_set.end()) {
    AliasNodeMap::const_iterator alias_node_it = alias_map.find(name);
    if (alias_node_it != alias_map.end())
      node_it = node_set.find((*alias_node_it).second);
  }

  return node_it;
}

}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::registerDCOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Dave Shirley, PSSI
// Creation Date : 08/21/06
//-----------------------------------------------------------------------------
bool NetlistImportTool::registerDCOptions(const Util::OptionBlock & option_block)
{
  for (Util::ParamList::const_iterator it = option_block.begin(); it != option_block.end(); ++it)
  {
    if (it->tag() == "PARAM")
    {
      dcParams_.push_back((*it).stringValue());
    }
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::registerSTEPOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Dave Shirley, PSSI
// Creation Date : 08/21/06
//-----------------------------------------------------------------------------
bool NetlistImportTool::registerSTEPOptions(const Util::OptionBlock & option_block)
{
  for (Util::ParamList::const_iterator it= option_block.begin(); it != option_block.end(); ++it)
  {
    if ((*it).tag() == "PARAM")
    {
      stepParams_.push_back((*it).stringValue());
    }
  }

  return true;
}


//-----------------------------------------------------------------------------
// Function      : OutputMgr::printLineDiagnostics
// Purpose       :
// Special Notes : erkeite:  8/10/2007
//
// For commentary and some explanation of this function, see the comments
// for the deferredPrintLineDiagnostics function.
//
// Scope         : public
// Creator       : Dave Shirley, PSSI
// Creation Date : 08/18/06
//-----------------------------------------------------------------------------
void
printLineDiagnostics(
  Parallel::Machine                             comm,
  const OutputParameterMap &                    output_parameter_map,
  const Device::DeviceMgr &                     device_manager,
  const Measure::Manager &                      measure_manager,
  const std::set<std::string> &                 node_names,
  const std::vector<std::string> &              step_params,
  const std::vector<std::string> &              dc_params,
  const Device::DeviceNameInstanceBlockMap &    device_map,
  const IO::AliasNodeMap &                      alias_node_map,
  UndefinedNameSet &                            deferred_undefined_parameters)
{
  UndefinedNameSet undefined_parameters;

  for (OutputParameterMap::const_iterator it1 = output_parameter_map.begin(), end1 = output_parameter_map.end(); it1 != end1; ++it1)
  {
    const OutputParameterMap::mapped_type &parameter_vector = (*it1).second;

    for (OutputParameterMap::mapped_type::const_iterator it2 = parameter_vector.begin(), end2 = parameter_vector.end(); it2 != end2; ++it2)
    {
      const PrintParameters &print_parameters = (*it2);

      // Populate stringStat, nodeStat and instanceState with variable names from print line
      std::map<std::string, bool> stringStat;
      std::map<std::string, bool> nodeStat;
      std::map<std::string, bool> instanceStat;

      for (Util::ParamList::const_iterator it3 = print_parameters.variableList_.begin(), end3 = print_parameters.variableList_.end(); it3 != end3; ++it3)
      {
        const Util::Param &parameter = (*it3);

        std::vector<std::string> nodes;
        std::vector<std::string> instances;
        std::vector<std::string> leads;
        std::vector<std::string> strings;
        std::vector<std::string> special;
        if (Util::hasExpressionTag(parameter) )
        {
          // parameter starts with "{" but may not have been been parsed into an expression.
          // check if there is an underlying expression object with the parameter
          if (parameter.getType() == Util::EXPR)
          {
            parameter.getValue<Util::Expression>().get_names(XEXP_NODE, nodes);
            parameter.getValue<Util::Expression>().get_names(XEXP_INSTANCE, instances);
            parameter.getValue<Util::Expression>().get_names(XEXP_LEAD, leads);
            parameter.getValue<Util::Expression>().get_names(XEXP_STRING, strings);
            parameter.getValue<Util::Expression>().get_names(XEXP_SPECIAL, special);  // special returns vars like TIME
          }
          else if ( ((parameter.getType()) == Util::DBLE) || ((parameter.getType()) == Util::INT) )
          {
          }
          instances.insert(instances.end(), leads.begin(), leads.end());
          strings.insert(strings.end(), special.begin(), special.end());  // add specials to strings
        }
        else
        {
          std::string varType= parameter.tag();
          if ((varType == "I" || ((varType.size() == 2 || varType.size() == 3) && varType[0] == 'I')) && parameter.getImmutableValue<int>() > 0)
          {
            // any devices found in this I(xxx) structure need to be communicated to the device manager
            // so that the lead currents can be calculated
            if (parameter.getImmutableValue<int>() != 1)
            {
              Report::UserError() << "Only one device argument allowed in I() in .PRINT command";
            }
            else
            {
              ++it3;
              if (varType.size() == 2)
              {
                instances.push_back((*it3).tag() + "{" + varType[1] + "}");
              }
              else
              {
                instances.push_back((*it3).tag());
              }
            }
          }
          else if ((varType == "V" || ((varType.size() == 2 || varType.size() == 3) && varType[0] == 'V')) && parameter.getImmutableValue<int>() > 0)
          {
            int numIndices = parameter.getImmutableValue<int>();

            if (numIndices < 1 || numIndices > 2)
            {
              Report::UserError0() << "Only one or two node arguments allowed in V() in .PRINT command";
            }
            else
            {
              for (; numIndices > 0 ; --numIndices)
              {
                ++it3;
                nodes.push_back((*it3).tag());
              }
            }
          }
          else if (varType == "N" && parameter.getImmutableValue<int>() > 0)
          {
            // Don't attempt to check this type of variable.
            //
            // N-variables can only be fully resolved once the devices are allocated,
            // and variables which are internal to the devices have been assigned names.
            //(N-variables are the same as V-variables, except that they can include
            // internal vars).
            if (parameter.getImmutableValue<int>() != 1)
              Report::UserError() << "Only one device argument allowed in N() in .PRINT command";
            else
              ++it3;
          }
          else if (((varType == "P") || (varType == "W")) && parameter.getImmutableValue<int>() > 0)
          {
            // any devices found in this P(xxx) or W(xxx) structure need to be communicated to the device manager
            // so that the lead currents and power can be calculated
            if (parameter.getImmutableValue<int>() != 1)
            {
              Report::UserError() << "Only one device argument allowed in P() or W() in .PRINT command";
            }
            else
            {
              ++it3;
              instances.push_back((*it3).tag());
            }
          }
          else
          {
            strings.push_back(parameter.tag());
          }
        }

        for (std::vector<std::string>::const_iterator iter_s= strings.begin(); iter_s != strings.end(); ++iter_s)
        {
          const std::string &name= *iter_s;

          bool done = false;
          if (name == "TEMP" || name == "TIME" || name == "FREQUENCY" || name == "INDEX" || 
              name == "OBJFUNC" || name == "SENS" || name == "NOISE" || name == "sweep" || 
              name == "ONOISE" || name == "INOISE")
            done = true;

          if (!done)
          {
            // check if this is a step sweep value.
            for (std::vector<std::string>::const_iterator it = dc_params.begin(), end = dc_params.end(); it != end ; ++it)
            {
              if (*it == name)
              {
                done = true;
                break;
              }
            }
          }

          if (!done)
          {
            // check if this is a step sweep value.
            for (std::vector<std::string>::const_iterator it = step_params.begin(), end = step_params.end(); it != end ; ++it)
            {
              if (*it == name)
              {
                done = true;
                break;
              }
            }
          }

          double mValue;
          bool mFound = measure_manager.getMeasureValue(name, mValue);
          if (mFound)
            done = true;

          if (!done)
          {
            done = device_manager.findParam(name);

            // Note:  when this is called, the devices haven't been allocated
            // yet.  So, this particular check isn't reliable.  If the device and/or
            // parameter is not found by the device package via getParam, then
            // simply save it for a later diagnostic.
            if (!done)
            {
              deferred_undefined_parameters.insert(UndefinedName(name, print_parameters.netlistLocation_));
            }
            done = true;
          }
          stringStat[name]= done;
        }

        for (std::vector<std::string>::const_iterator iter_s = nodes.begin(); iter_s != nodes.end(); ++iter_s)
        {
          const std::string &name= *iter_s;

          // allow * to pass
          nodeStat[name] = (name == "*") || (findNode(name, node_names, alias_node_map) != node_names.end());
        }

        for (std::vector<std::string>::iterator iter_s= instances.begin(); iter_s != instances.end(); ++iter_s)
        {
          const std::string &name= *iter_s;

          if (name.substr(name.size() - 1, 1) == "}")
          {
            // allow * to pass
            instanceStat[name]= (name == "*") || (device_map.find(name.substr(0, name.size() - 3)) != device_map.end());
          }
          else
          {
            // allow * to pass
            instanceStat[name]= (name == "*") || (device_map.find(name) != device_map.end());
          }
        }
      }

      // Sum uses of names in nodeStat and instanceStat
      std::vector<int> stat;
      for (std::map<std::string, bool>::const_iterator it = nodeStat.begin(), end = nodeStat.end(); it != end; ++it)
      {
        stat.push_back((*it).second ? 1 : 0);
      }

      for (std::map<std::string, bool>::const_iterator it = instanceStat.begin(), end = instanceStat.end(); it != end; ++it)
      {
        stat.push_back((*it).second ? 1 : 0);
      }

      Parallel::AllReduce(comm, MPI_SUM, stat);

// Generate message
      std::ostringstream oss;

      int stat_index = 0;
      for (std::map<std::string, bool>::const_iterator it = nodeStat.begin(), end = nodeStat.end(); it != end; ++it)
      {
        if (stat[stat_index++] == 0)
        {
          std::ostringstream oss;
          oss << "node " << (*it).first;
          undefined_parameters.insert(UndefinedName(oss.str(), print_parameters.netlistLocation_));
        }
      }

      for (std::map<std::string, bool>::const_iterator it = instanceStat.begin(), end = instanceStat.end(); it != end; ++it)
      {
        if (stat[stat_index++] == 0)
        {
          if (((*it).first)[((*it).first).size() - 1] == '}')
          {
            std::ostringstream oss;
            oss << "I" << ((*it).first).substr(((*it).first).size() - 2, 1)
                << "("
                << ((*it).first).substr(0, ((*it).first).size() - 3)
                << ")";
            
            if ((startswith_nocase((*it).first.c_str(), "L")) || ((*it).first.find( ":L") != std::string::npos ) || ((*it).first.find( ":l") != std::string::npos ) )
              ; // deferred_undefined_parameters.insert(UndefinedName(oss.str(), print_parameters.netlistLocation_));
            else
              undefined_parameters.insert(UndefinedName(oss.str(), print_parameters.netlistLocation_));
            
          }
          else
          {
            std::ostringstream oss;
            oss << "I(" << (*it).first << ")";
            
            if ((startswith_nocase((*it).first.c_str(), "L")) || ((*it).first.find( ":L") != std::string::npos ) || ((*it).first.find( ":l") != std::string::npos ) )
              ; // deferred_undefined_parameters.insert(UndefinedName(oss.str(), print_parameters.netlistLocation_));
            else
              undefined_parameters.insert(UndefinedName(oss.str(), print_parameters.netlistLocation_));
            
          }
        }
      }

      for (std::map<std::string, bool>::const_iterator it= stringStat.begin(); it != stringStat.end(); ++it)
      {
        if (!(*it).second)
        {
          std::ostringstream oss;
          oss << (*it).first;
          undefined_parameters.insert(UndefinedName(oss.str(), print_parameters.netlistLocation_));
        }
      }
    }

  }

  if (!undefined_parameters.empty())
    errorUndefinedParameters(undefined_parameters);
    
}

//-----------------------------------------------------------------------------
// Function      : OutputMgr::deferredPrintLineDiagnostics
// Purpose       : Yet another function that checks the .print line.
//
// Special Notes : erkeite.  8/10/2007
//
//                 This function has to be called after the devices in
//                 the device package have all been allocated.  The function
//                 printLineDiagnostics is called earlier than that, so
//                 it doesn't handle device-specific parameter outputs.
//
//                 We have 3 completely different functions that do different
//                 aspects of checking the print line:
//
//                (1) printLineDiagnostics:  happens really early, as soon as
//                 topology is constructed, but before devivces are allocated.
//
//                (2) delayedPrintLineDiagnostics:  happens just a little later, to
//                 check .print line fields that are specific to device entities(ie
//                 aren't just I and V solution vars).
//
//                (3) check_output:  Happens at the beginning of the simulation,
//                 after everything is completely set up, but before any solves
//                 happen.  The advantage of this one is that it actually uses
//                 all the same function calls as the regular output call that
//                 will happen later.  So, it is probably the most thorough test
//                 of all.  However, for really big simulations, it happens relatively
//                 late in the setup.  For really large simulations, setup can take
//                 a long time, and it is often useful to get diagnostics earlier
//                 than that.
//
//                 Note from TVR 10/22/2013:  "check_output", referred to
//                 above, has been broken apart and renamed "prepareOutput".
//                 The notes from ERK above are therefore somewhat out of date.
//                 The refactor, though, is not complete, and the replacement
//                 code is not well documented.
//
// Scope         : public
// Creator       : Dave Shirley, PSSI
// Creation Date : 01/17/07
//-----------------------------------------------------------------------------
void deferredPrintLineDiagnostics(
  Parallel::Machine                             comm,
  const UndefinedNameSet &                      deferred_undefined_parameters,
  const Device::DeviceMgr &                     device_manager)
{
  std::ostringstream oss;

  Util::Marshal mout;
  mout << deferred_undefined_parameters;

  std::vector<std::string> dest;
  Parallel::AllGatherV(comm, mout.str(), dest);

  UndefinedNameSet common_parameter_set;
  for (int p = 0; p < Parallel::size(comm); ++p)
  {
    Util::Marshal min(dest[p]);
    std::vector<UndefinedName> x;
    min >> x;
    std::copy(x.begin(), x.end(), std::inserter(common_parameter_set, common_parameter_set.begin()));
  }

  UndefinedNameSet undefined_parameters;
  for (std::set<UndefinedName>::const_iterator it = common_parameter_set.begin(), end = common_parameter_set.end(); it != end; ++it)
  {
    double result = 0.0;
    if (!device_manager.getParamAndReduce((*it).getName(), result))
      undefined_parameters.insert(*it);
  }

  if (!undefined_parameters.empty())
    errorUndefinedParameters(undefined_parameters);
}

//-----------------------------------------------------------------------------
// Function      : checkNodeDevConflicts
// Purpose       : Check for name collisions between devices
// Special Notes : Non-member helper function
// Scope         : public
// Creator       : Heidi Thornquist
// Creation Date : 07/28/14
//-----------------------------------------------------------------------------
void
checkNodeDevConflicts(
  const Device::DeviceNameInstanceBlockMap &    device_instance_block_map,
  Parallel::Communicator &                      pds_comm)
{
  Parallel::Machine comm = pds_comm.comm();

  int proc_size = Parallel::size(comm);

  if (proc_size > 1)
  {
    std::vector<std::string> deviceNames;
    int byteCount = getDeviceNames(device_instance_block_map, deviceNames);

    N_ERH_ErrorMgr::safeBarrier(comm);     // All procs call (5)

    int maxByteCount;
    Parallel::AllReduce(comm, MPI_MAX, &byteCount, &maxByteCount, 1);

    std::vector<char> sendBuffer(byteCount);
    std::vector<char> recvBuffer(maxByteCount);
    int pos = 0;
    for (std::vector<std::string>::const_iterator n_i = deviceNames.begin(), n_end = deviceNames.end() ; n_i != n_end ; ++n_i)
    {
      int length = n_i->size();
      pds_comm.pack(&length, 1, &sendBuffer[0], byteCount, pos);
      pds_comm.pack(n_i->c_str(), length, &sendBuffer[0], byteCount, pos);
    }

    int proc_rank = Parallel::rank(comm);

    int sendProc = proc_rank + 1;
    int recvProc = proc_rank - 1;
    if (sendProc == proc_size) sendProc = 0;
    if (recvProc == -1) recvProc = proc_size - 1;

    while (sendProc != proc_rank)
    {
      int bsize;
      pds_comm.iRecv( &bsize, 1, recvProc );
      pds_comm.send( &byteCount, 1, sendProc );
      pds_comm.waitAll();
      pds_comm.iRecv( &recvBuffer[0], bsize, recvProc );
      pds_comm.send( &sendBuffer[0], byteCount, sendProc );
      pds_comm.waitAll();
      deviceNames.resize(0);
      pos = 0;
      while (pos < bsize)
      {
        int length = 0;

        pds_comm.unpack(&recvBuffer[0], bsize, pos, &length, 1);
        deviceNames.push_back(std::string((&recvBuffer[0])+pos, length));
        pos += length;
      }

      checkDeviceNames(device_instance_block_map, deviceNames);
      N_ERH_ErrorMgr::safeBarrier(comm);          // All procs call (5+N)

      if (++sendProc == proc_size) sendProc = 0;
      if (--recvProc == -1) recvProc = proc_size - 1;
    }
  }
}

void getLeadCurrentDevices(const Util::ParamList &variable_list, std::set<std::string> &devicesNeedingLeadCurrents)
{
  for (Util::ParamList::const_iterator iterParam = variable_list.begin() ; iterParam != variable_list.end(); ++iterParam)
  {
    std::string varType(iterParam->tag());
    //Xyce::dout() << "In getLeadCurrentDevices and varType is: \"" << varType << "\"" << std::endl;

    if (Util::hasExpressionTag(*iterParam))
    {
      std::vector<std::string> leads;

      Util::Expression exp;
      exp.set(iterParam->tag());
      exp.get_names(XEXP_LEAD, leads);

      // any lead currents found in this expression need to be communicated to the device manager.
      // Multi terminal devices have an extra designator on the name as in name{lead_name}
      // need to remove any {} in the name.
      for (std::vector<std::string>::const_iterator currLeadItr = leads.begin(); currLeadItr != leads.end(); ++currLeadItr)
      {
        size_t leadDesignator = currLeadItr->find_first_of("{");
        devicesNeedingLeadCurrents.insert( currLeadItr->substr(0, leadDesignator));
      }
    }
    else
    {
      if ( ((varType == "I" || (varType.size() == 2 && varType[0] == 'I')) || (varType == "P") || (varType == "W")) &&  (iterParam->getImmutableValue<int>() > 0))
      {
        // any devices found in this I(xxx) structure need to be communicated to the device manager
        // so that the lead currents can be calculated
        if (iterParam->getImmutableValue<int>() != 1)
        {
          Report::UserError0() << "Only one device argument allowed in I(), W() or P() in .print";
        }
        else
        {
          ++iterParam;
          if (varType.size() == 2)
          {
            devicesNeedingLeadCurrents.insert(iterParam->tag());
          }
          else
          {
            devicesNeedingLeadCurrents.insert(iterParam->tag());
          }
        }
      }
    }
  }

  // the list of devices that need lead currents.
  if (DEBUG_IO && !devicesNeedingLeadCurrents.empty())
  {
    std::set<std::string>::iterator currentDeviceNameItr = devicesNeedingLeadCurrents.begin();
    std::set<std::string>::iterator endDeviceNameItr = devicesNeedingLeadCurrents.end();
    Xyce::dout() << "Devices for which lead currents were requested: ";
    while ( currentDeviceNameItr != endDeviceNameItr)
    {
      Xyce::dout() << *currentDeviceNameItr << "  ";
      currentDeviceNameItr++;
    }
  }
}

//----------------------------------------------------------------------------
// Function       : getDeviceNames
// Purpose        : return a vector of the device names on this processor
// Special Notes  : returns byte count for packing
// Scope          : public
// Creator        : Dave Shirley, PSSI
// Creation Date  : 01/20/2006
//----------------------------------------------------------------------------
int getDeviceNames(const Device::DeviceNameInstanceBlockMap &device_instance_block_map, std::vector<std::string> & names)
{
  int byteCount = 0;

  names.clear();

  for (Device::DeviceNameInstanceBlockMap::const_iterator it = device_instance_block_map.begin(), end = device_instance_block_map.end(); it != end; ++it)
  {
    names.push_back((*it).first);
    byteCount += sizeof(int) + (*it).first.size();
  }

  return byteCount;
}

//----------------------------------------------------------------------------
// Function       : CircuitBlock::checkDeviceNames
// Purpose        : checks a vector of device names against the local device names
// Special Notes  :
// Scope          : public
// Creator        : Dave Shirley, PSSI
// Creation Date  : 01/20/2006
//----------------------------------------------------------------------------
void checkDeviceNames(const Device::DeviceNameInstanceBlockMap &device_instance_block_map, const std::vector<std::string> & names)
{
  for (std::vector<std::string>::const_iterator nName = names.begin(); nName != names.end(); ++nName)
  {
   // find duplicates across procs
    Device::DeviceNameInstanceBlockMap::const_iterator it = device_instance_block_map.find(*nName);

    if (it != device_instance_block_map.end())
    {
      Report::UserError().at((*it).second.getNetlistLocation()) << "Duplicate device " << *nName;
    }
  }
}

} // namespace IO
} // namespace Xyce
