//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Filename       : $RCSfile: N_IO_RestartMgr.C,v $
//
// Purpose        :
//
// Special Notes  :
//
// Creator        : Robert J. Hoekstra, SNL, Parallel Computational Sciences
//
// Creation Date  : 7/19/01
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.89.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:15 $
//
// Current Owner  : $Author: tvrusso $
//-------------------------------------------------------------------------

#include <Xyce_config.h>

#include <sstream>
#include <fstream>
#include <algorithm>

#include <N_ANP_AnalysisManager.h>
#include <N_DEV_DeviceMgr.h>
#include <N_ERH_ErrorMgr.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_IO_RestartMgr.h>
#include <N_IO_RestartNode.h>
#include <N_PDS_Comm.h>
#include <N_PDS_Manager.h>
#include <N_TOP_Topology.h>
#include <N_UTL_ExtendedString.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_Functors.h>
#include <N_UTL_OptionBlock.h>

namespace Xyce {
namespace IO {

namespace {

#ifdef Xyce_RESTART_NOPACK
static const bool RESTART_NOPACK = true;
#else
static const bool RESTART_NOPACK = false;
#endif

struct OptionsReg : public PkgOptionsReg
{
  OptionsReg(RestartMgr &restart_manager, const std::string &netlist_filename, int proc_size, int proc_rank)
    : restartManager_(restart_manager),
      netlistFilename_(netlist_filename),
      size_(proc_size),
      rank_(proc_rank)
  {}

  bool operator()(const Util::OptionBlock & options)
  {
    return restartManager_.registerRestartOptions(options, netlistFilename_, size_, rank_);
  }

  RestartMgr &                  restartManager_;
  const std::string &           netlistFilename_;
  const int                     size_;
  const int                     rank_;
};

} // namespace <unnamed>

//-----------------------------------------------------------------------------
// Function      : RestartMgr::registerPkgOptionsMgr
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, 1437
// Creation Date : 10/21/08
//-----------------------------------------------------------------------------
bool
registerPkgOptionsMgr(
  RestartMgr &          restart_manager,
  const std::string &   netlist_filename,
  PkgOptionsMgr &       options_manager,
  int                   proc_size,
  int                   proc_rank)
{
  options_manager.submitRegistration( "RESTART", netlist_filename, new OptionsReg(restart_manager, netlist_filename, proc_size, proc_rank));

  return true;
}

//-----------------------------------------------------------------------------
// Function      : RestartMgr::RestartMgr
// Purpose       : constructor
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 7/19/01
//-----------------------------------------------------------------------------
RestartMgr::RestartMgr()
  : restartFlag_(false),
    restartFileName_(""),
    restartJobName_(""),
    pack_(!RESTART_NOPACK),
    initialRestartInterval_(0.0)
{}

//-----------------------------------------------------------------------------
// Function      : RestartMgr::registerRestartOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 7/19/01
//-----------------------------------------------------------------------------
bool RestartMgr::registerRestartOptions(const Util::OptionBlock & option_block, const std::string &netlist_filename, int proc_size, int proc_rank)
{
  bool success = true;
  double start_time = 0.0;

  for (Util::ParamList::const_iterator it = option_block.begin(), end = option_block.end(); it != end; ++it)
  {
    ExtendedString tag(it->tag());
    tag.toUpper();

    if (tag == "TIME")
    {
      double t = it->getImmutableValue<double>();
      ++it;
      if( it == option_block.end())
      {
        Report::UserError() << ".OPTIONS Restart: Badly formed INITIAL_INTERVAL\n";
        success = false;
      }
      double iv = it->getImmutableValue<double>();
      restartIntervalPairs_.push_back(Interval(t,iv));
    }

    else if (tag == "PACK")
    {
      pack_ = it->getImmutableValue<int>();

      if (!pack_ && proc_size != 1)
      {
        Report::UserError() << ".OPTIONS Restart: Restart Data Must Be Packed for Parallel Runs, Please remove 'pack=0' from .OPTIONS RESTART line";
        return false;
      }
    }

    else if (tag == "INITIAL_INTERVAL")
    {
      initialRestartInterval_ = it->getImmutableValue<double>();
    }

    else if (tag == "FILE")
    {
      restartFileName_ = it->stringValue();
      restartFlag_ = true;
    }

    else if (tag == "JOB")
    {
      restartJobName_ = it->stringValue();
    }

    else if (tag == "START_TIME")
    {
      start_time = it->getImmutableValue<double>();
      restartFlag_ = true;
    }
  }

  if (restartFlag_ && restartFileName_.empty())
  {
    std::ostringstream oss;
    oss << restartJobName_ << start_time;
    restartFileName_ = oss.str();
  }

  if (pack_ && RESTART_NOPACK)
  {
    pack_ = false;
    Report::UserWarning() << "Restart packing disabled for this implementation of Xyce";
  }

  if (!pack_  && proc_size != 1)
  {
    Report::UserError() << ".OPTIONS Restart: Parallel Restarts must be packed!";
    success = false;
  }

  if (DEBUG_RESTART && proc_rank == 0)
  {
    Xyce::dout() << Xyce::subsection_divider << std::endl
                 << "RESTART OPTIONS SETUP: " << netlist_filename << std::endl
                 << "restartFileName: " << restartFileName_ << std::endl
                 << "restartJobName: " << restartJobName_ << std::endl
                 << "isRestart: " << restartFlag_ << std::endl
                 << "initial Interval: " << initialRestartInterval_ << std::endl
                 << "pack data: " << pack_ << std::endl;

    for (int i = 0; i < restartIntervalPairs_.size(); ++i)
      Xyce::dout() << restartIntervalPairs_[i].first << " " << restartIntervalPairs_[i].second << std::endl;

    Xyce::dout() << Xyce::subsection_divider << std::endl;
  }

  return success;
}

//-----------------------------------------------------------------------------
// Function      : dumpRestartData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 7/19/01
//-----------------------------------------------------------------------------
bool dumpRestartData(N_PDS_Comm &comm, Topo::Topology &topology, Analysis::AnalysisManager &analysis_manager, Device::DeviceMgr &device_manager, const std::string &job_name, bool pack, double time)
{
  bool success = true;

  int procID = comm.procID();
  int numProcs = comm.numProc();

  //Get Restart Node Data and setup buffers
  //---------------------------------------
  int dataSize = sizeof(int);
  std::vector<RestartNode *> nodeVec;

  topology.getRestartNodes(analysis_manager, nodeVec);

  int nodeCount = nodeVec.size();
  double nC = static_cast<double>(nodeCount);

  double gNC = 0;
  comm.sumAll( &nC, &gNC, 1);

  int globalNodeCount = static_cast<int>(gNC);

  for( int i = 0; i < nodeCount; ++i)
    dataSize += nodeVec[i]->packedByteCount();

  int maxSize = 0;
  int bsize;
  char * buf;
#ifdef Xyce_PARALLEL_MPI
  double dS = dataSize;
  double mS;
  comm.maxAll( &dS, &mS, 1);
  maxSize = static_cast<int>(mS);
  if (procID == 0)
  {
    buf = new char[maxSize];
    bsize = maxSize;
  }
  else
#endif
  {
    buf = new char[dataSize];
    bsize = dataSize;
    maxSize = dataSize;
  }

  std::ofstream outStreamSt;
  std::string outName;

  int proc = 0;
  if (procID == 0)
  {
    std::ostringstream ost;
    ost << job_name << time;
    outName.assign(ost.str());
    outStreamSt.open( outName.c_str());

    if (!outStreamSt.is_open())
    {
      Report::UserWarning0() << "Cannot Open CheckPoint File: " << outName;

      return false;
    }

    if (DEBUG_RESTART)
      Xyce::dout() << Xyce::subsection_divider << std::endl
                   << "DUMPING RESTART: " << outName << std::endl
                   << "numProcs: " << numProcs << " maxSize: " << maxSize << std::endl
                   << "proc: " << proc << " dataSize: " << dataSize << std::endl
                   << "nodeCount: " << nodeCount << std::endl
                   << "pack: " << pack << std::endl;

    int packed = pack?1:0;
    outStreamSt << numProcs << " " << maxSize << " " << packed << " " << globalNodeCount << " ";
  }


  if (pack)
  {
    //Pack Restart Nodes
    int pos = 0;
    comm.pack( &nodeCount, 1, buf, bsize, pos);
    for( int i = 0; i < nodeCount; ++i)
    {
      nodeVec[i]->pack( buf, bsize, pos, &comm);
    }

    if (procID == 0)
    {
      outStreamSt << procID << " " << dataSize << " ";
      outStreamSt.write( buf, dataSize);
    }
  }
  else
  {
    outStreamSt << " " << nodeCount << " ";
    for( int i = 0; i < nodeCount; ++i)
    {
      nodeVec[i]->dump(outStreamSt);
    }
  }

#ifdef Xyce_PARALLEL_MPI
  //PARALLEL, get and store nodes from other procs
  //----------------------------------------------
  comm.barrier();

  int size;
  for( proc = 1; proc < numProcs; ++proc)
  {
    if (procID == 0)
    {
      comm.recv( &size, 1, proc);
      comm.recv( buf, size, proc);

      outStreamSt << proc << " " << size << " ";
      outStreamSt.write( buf, size);
    }
    else if (procID == proc)
    {
      comm.send( &dataSize, 1, 0);
      comm.send( buf, dataSize, 0);
    }

    comm.barrier();
  }
#endif

  //Add Time Int stuff from proc 0
  //------------------------------
  if (procID == 0)
  {
    dataSize = analysis_manager.getRestartDataSize( pack);
    if (dataSize > maxSize)
    {
      delete [] buf;
      buf = 0;
      maxSize = dataSize;
      buf = new char[maxSize];
      bsize = maxSize;
    }

    if (DEBUG_RESTART)
      Xyce::dout() << "before tia RESTART dump: maxSize = " << maxSize << std::endl;

    int pos = 0;
    success &= analysis_manager.dumpRestartData( buf, bsize, pos, &comm, pack);

    outStreamSt << dataSize << " ";
    outStreamSt.write( buf, dataSize);
  }

  comm.barrier();

  if (DEBUG_RESTART && procID == 0)
    Xyce::dout() << Xyce::subsection_divider << std::endl;

  //Add Device Mgr stuff from proc 0
  //------------------------------
  if (procID == 0)
  {
    dataSize = device_manager.restartDataSize( pack);
    if (dataSize > maxSize)
    {
      delete [] buf;
      buf = 0;
      maxSize = dataSize;
      buf = new char[maxSize];
      bsize = maxSize;
    }

    if (DEBUG_RESTART)
      Xyce::dout() << "before dev RESTART dump: maxSize = " << maxSize << std::endl;

    int pos = 0;
    success &= device_manager.dumpRestartData( buf, bsize, pos, &comm, pack);

    outStreamSt << dataSize << " ";
    outStreamSt.write( buf, dataSize);
  }

  comm.barrier();

  if (DEBUG_RESTART && procID == 0)
    Xyce::dout() << Xyce::subsection_divider << std::endl;

  int nodeSize = nodeVec.size();
  for (int i = 0; i < nodeSize; ++i)
    delete nodeVec[i];

  delete[] buf;
  bsize = 0;

  return success;
}

//-----------------------------------------------------------------------------
// Function      : restoreRestartData
// Purpose       :
// Special Notes : version used with distributed parser
// Scope         : public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 7/19/01
//-----------------------------------------------------------------------------
bool restoreRestartData(N_PDS_Comm &comm, Topo::Topology &topology, Analysis::AnalysisManager &analysis_manager, Device::DeviceMgr &device_manager, const std::string &path)
{
  bool success = true;

  int procID = comm.procID();
  int numProcs = comm.numProc();

  char * buf = 0;
  int dataSize;
  int pos = 0;
  int bsize = 0;
  int fProcID;

  std::ifstream * inStream;

  int nodeCount;
  int oldNumProcs, maxSize;

  bool pack = RESTART_NOPACK ? false : numProcs > 1;

  //Table of restart nodes used by all procs
  std::vector<RestartNode*> nodeTable;

  //1. Proc 0 reads in data pushing NumDevices/NumProcs to every processor
  if (procID == 0)
  {
    inStream = new std::ifstream( path.c_str());

    if (!inStream->is_open())
      N_ERH_ErrorMgr::report( N_ERH_ErrorMgr::USR_FATAL_0,
                              "Cannot Open CheckPoint File: " +
                              path + " for Restart\n");

    int packed;
    int TotNumDevs;
    (*inStream) >> oldNumProcs >> maxSize >> packed >> TotNumDevs;
    pack = packed;

    char * buf1 = 0;
    int    bsize1 = 0;
    int    pos1   = 0;

    if (DEBUG_RESTART)
      Xyce::dout() << Xyce::subsection_divider << std::endl
                   << "RESTORING RESTART DATA" << std::endl
                   << "oldNumProcs: " << oldNumProcs << std::endl
                   << "maxSize: " << maxSize << std::endl
                   << "packed: " << pack << std::endl
                   << "TotNumDevs: " << TotNumDevs << std::endl;

    int DevsPerProc = TotNumDevs/numProcs + 1;

    RestartNode * nodeP;
    int CurrProc = 1;
    for( int oldProc = 0; oldProc < oldNumProcs; ++oldProc)
    {
      if (pack)
      {
        (*inStream) >> fProcID >> dataSize;
        char dummy = 'x';
        inStream->read( &dummy, 1);

        if (DEBUG_RESTART)
          Xyce::dout() << "fProcID: " << fProcID << std::endl
                       << "dataSize: " << dataSize << std::endl;

        bsize = dataSize;
        buf = new char[bsize];

        inStream->read( buf, dataSize);

        pos = 0;
        comm.unpack( buf, bsize, pos, &nodeCount, 1);
      }
      else
        (*inStream) >> nodeCount;

      if (DEBUG_RESTART)
        Xyce::dout() << "nodeCount: " << nodeCount << std::endl;

      for( int i = 0; i < nodeCount; ++i)
      {
        nodeP = new RestartNode();

        if (pack)
          nodeP->unpack( buf, bsize, pos, &comm);
        else
          nodeP->restore(*inStream);

        nodeTable.push_back( nodeP);

        if ((nodeTable.size()==DevsPerProc) ||
             ( (i==nodeCount-1) && (oldProc==oldNumProcs-1)))
        {
          //Pack up data and send to CurrProc
          int Size = 0;
          int nTSize = nodeTable.size();
          Size += sizeof(int);
          for( int j = 0; j < nTSize; ++j)
            Size += nodeTable[j]->packedByteCount();

          if (Size > bsize1)
          {
            if (bsize1) delete [] buf1;
            bsize1 = Size;
            buf1 = new char[bsize1];
          }

          pos1 = 0;
          comm.pack( &nTSize, 1, buf1, bsize1, pos1);
          for( int j = 0; j < nTSize; ++j)
            nodeTable[j]->pack( buf1, bsize1, pos1, &comm);

          if (CurrProc < numProcs)
          {
            comm.send( &Size, 1, CurrProc);
            comm.send( buf1, Size, CurrProc);
            ++CurrProc;
          }

          for_each( nodeTable.begin(), nodeTable.end(), DeletePtr<RestartNode>());
          nodeTable.clear();
        }
      }
    }

    if (bsize) delete [] buf;
    buf = buf1;
    bsize = bsize1;
  }

#ifdef Xyce_PARALLEL_MPI
  //Everybody else receive their set of devices
  if (procID != 0)
  {
    comm.recv( &bsize, 1, 0);
    buf = new char[bsize];
    comm.recv( buf, bsize, 0);
  }
  if (DEBUG_RESTART)
    Xyce::dout() << "Finished ReadIn of Restart Node Data on Proc: " << procID << std::endl;

  comm.barrier();
#endif

#ifdef Xyce_PARALLEL_MPI
  //Set Buffers to largest necessary size
  double sz = static_cast<double>(bsize);
  double msz;
  comm.maxAll( &sz, &msz, 1);
  int bsize2 = static_cast<int>(msz);
  char * buf2 = new char[bsize2];

  memcpy( buf2, buf, bsize);

  delete [] buf;
  bsize = bsize2;
  buf = new char[bsize];

  std::swap( buf, buf2);
#endif

  //Check devices in your buffer, then push around ring
  for( int i = 0; i < numProcs; ++i)
  {
    pos = 0;
    comm.unpack( buf, bsize, pos, &nodeCount, 1);
    nodeTable.resize(nodeCount);

    if (DEBUG_RESTART && procID != 0)
      Xyce::dout() << "Restart Ring Stage: " << i << " " << nodeCount << std::endl;

    for( int j = 0; j < nodeCount; ++j)
    {
      nodeTable[j] = new RestartNode();
      nodeTable[j]->unpack( buf, bsize, pos, &comm);
    }

    topology.restoreRestartNodes(analysis_manager, nodeTable);

    for_each( nodeTable.begin(), nodeTable.end(), DeletePtr<RestartNode>());
    nodeTable.clear();

#ifdef Xyce_PARALLEL_MPI
    int sendProc = procID-1;
    if (sendProc == -1) sendProc = numProcs-1;
    int recvProc = procID+1;
    if (recvProc == numProcs) recvProc = 0;

    int MaxSendSize = 1e6;

    int NumSends = 1;
    int SendSize = bsize;
    if (bsize > MaxSendSize)
    {
      NumSends = bsize/MaxSendSize;
      if (bsize != NumSends*MaxSendSize) ++NumSends;
      SendSize = MaxSendSize;
    }

    if (DEBUG_RESTART)
      Xyce::dout() << "NumSends: " << NumSends << std::endl;

    for( int j = 0; j < NumSends; ++j)
    {
      int Offset = j*SendSize;

      if (j == (NumSends-1)) SendSize = bsize - Offset;

      comm.iRecv( buf2+Offset, SendSize, sendProc);

      comm.rSend( buf+Offset, SendSize, recvProc);

      comm.waitAll();
    }

    std::swap( buf, buf2);
#endif
  }

#ifdef Xyce_PARALLEL_MPI
  bsize2 = 0;
  delete [] buf2;
  buf2 = 0;
#endif

// read in Time data
  if (procID == 0)
  {
    int proc;
    (*inStream) >> dataSize;
    char dummy = 'x';
    inStream->read( &dummy, 1);

    if (DEBUG_RESTART)
      Xyce::dout() << Xyce::subsection_divider << std::endl
                   << "RESTORING TIME INT STUFF\n" << std::endl
                   << "dataSize: " << dataSize << std::endl
                   << Xyce::subsection_divider << std::endl;

    if (dataSize > maxSize)
    {
      delete [] buf;
      maxSize = dataSize;
      buf = new char[maxSize];
      bsize = maxSize;
    }

    inStream->read( buf, dataSize);
  }

#ifdef Xyce_PARALLEL_MPI
  comm.barrier();

  comm.bcast( &dataSize, 1, 0);
  if (procID != 0)
    if (dataSize > bsize)
    {
      delete [] buf;
      buf = new char[dataSize];
      bsize = dataSize;
    }

  comm.bcast( buf, dataSize, 0);
#endif

  pos = 0;
  analysis_manager.restoreRestartData( buf, bsize, pos, &comm, pack);

// read in Dev data
  if (procID == 0)
  {
    int proc;
    (*inStream) >> dataSize;
    char dummy = 'x';
    inStream->read( &dummy, 1);

    if (DEBUG_RESTART)
      Xyce::dout() << Xyce::subsection_divider << std::endl
                   << "RESTORING DEV MGR STUFF" << std::endl
                   << "dataSize: " << dataSize << std::endl
                   << Xyce::subsection_divider << std::endl;

    if (dataSize > maxSize)
    {
      delete [] buf;
      maxSize = dataSize;
      buf = new char[maxSize];
      bsize = maxSize;
    }

    inStream->read( buf, dataSize);

    delete inStream;
  }

#ifdef Xyce_PARALLEL_MPI
  comm.barrier();

  comm.bcast( &dataSize, 1, 0);
  if (procID != 0)
    if (dataSize > bsize)
    {
      delete [] buf;
      buf = new char[dataSize];
      bsize = dataSize;
    }

  comm.bcast( buf, dataSize, 0);
#endif

  pos = 0;
  device_manager.restoreRestartData( buf, bsize, pos, &comm, pack);
  delete [] buf;

  return success;
}

} // namespace IO
} // namespace Xyce
