//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_LAS_System.h,v $
//
// Purpose        : Container class for linear system Jacobian, RHS, Soln Vec,
//                  Error Vec, and Creators N_LAS_QueryUtil and N_PDS_ParMap
//                  are registered to minimize input for matrix and vector
//                  creation
//
// Special Notes  :
//
// Creator        : Robert Hoekstra, SNL, Parallel Computational Sciences
//
// Creation Date  : 06/10/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.67.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:15 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#ifndef  Xyce_LAS_System_h
#define  Xyce_LAS_System_h

#include <list>

#include <N_ANP_fwd.h>
#include <N_LAS_fwd.h>

// to eliminate RCP warnings, putting N_PDS_Manager header here.
#include <N_PDS_Manager.h>

// ---------- Forward Declarations ----------

class Epetra_MapColoring;

namespace Xyce {
namespace Linear {

//-----------------------------------------------------------------------------
// Class         : System
// Purpose       :
// Special Notes :
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/10/00
//-----------------------------------------------------------------------------
class System
{

public:
  System();

  ~System();

  // Registration methods for necessary utilities

  // Registers the Parallel Distribution Services (PDS) manager (i.e., sets the
  // pointer.
  bool registerPDSManager(N_PDS_Manager * PDS_Manager)
  { return ( pdsMgr_ = PDS_Manager ); }

  // Registers the Query utility
  bool registerQueryUtil(QueryUtil * LAS_QUtil)
  { return ( lasQueryUtil_ = LAS_QUtil ); }

  // Registers the LAS Builder object
  bool registerBuilder(Builder * builder)
  { return ( lasBuilder_ = builder ); }

  // Registration methods for soln and state vector objects since the TIA or
  // NLS packages will probably own these

  // Registers the pointer to the nonlinear system's Jacobian matrix
  bool registerJacobianMatrix(Matrix * jacMatrixPtr)
  { return ( jacobianMatrixPtr_ = jacMatrixPtr ); }

  // Registers the pointer to the Jacobian test matrix (old DAE)
  bool registerJacTestMatrix(Matrix * jacTestMatrixPtr)
  { return ( jacTestMatrixPtr_ = jacTestMatrixPtr ); }

  // Registers the pointer to the Jacobian test matrix (new DAE)
  bool registerdFdxTestMatrix(Matrix * dFdxTestMatrixPtr)
  { return ( dFdxTestMatrixPtr_ = dFdxTestMatrixPtr ); }
  bool registerdQdxTestMatrix(Matrix * dQdxTestMatrixPtr)
  { return ( dQdxTestMatrixPtr_ = dQdxTestMatrixPtr ); }

  // Registers the pointer to the nonlinear system's residual (RHS) vector
  bool registerRHSVector (Vector * rhsVecPtr)
  { return ( rhsVectorPtr_ = rhsVecPtr ); }

  // Registers the pointer to the nonlinear system's limiting (jdxp) vector
  bool registerJDXPVector (Vector * jdxpVecPtr)
  { return ( jdxpVectorPtr_ = jdxpVecPtr ); }

  // Registers the pointer to the nonlinear system's F vector
  bool registerFVector   (Vector * fVecPtr)
  { return ( fVectorPtr_ = fVecPtr ); }

  // Registers the pointer to the voltlim dx vector
  bool registerDxVoltlimVector   (Vector * dxVLPtr)
  { return ( dxVoltlimVectorPtr_ = dxVLPtr ); }

  // Registers the pointer to the nonlinear system's limiting (jdx2) vector (old DAE)
  bool registerJDX2Vector (Vector * jdx2VecPtr)
  { return ( jdx2VectorPtr_ = jdx2VecPtr ); }

  // Registers the pointer to the nonlinear system's limiting vectors (new DAE)
  bool registerFDX2Vector (Vector * Fdx2VecPtr)
  { return ( Fdx2VectorPtr_ = Fdx2VecPtr ); }

  bool registerQDX2Vector (Vector * Qdx2VecPtr)
  { return ( Qdx2VectorPtr_ = Qdx2VecPtr ); }

  // DAE vector registrations:
  bool registerDAEQVector (Vector * qVecPtr)
  {  return (daeQVectorPtr_ = qVecPtr); }

  bool registerDAEFVector (Vector * fVecPtr)
  {  return (daeFVectorPtr_ = fVecPtr); }

  bool registerDAEBVector (Vector * bVecPtr)
  {  return (daeBVectorPtr_ = bVecPtr); }

  // dFdxdvp vector, for new-DAE voltage limiting.
  bool registerdFdxdVpVector (Vector * dFdxdVpVectorPtr)
  { return ( dFdxdVpVectorPtr_ = dFdxdVpVectorPtr); }

  // dQdxdvp vector, for new-DAE voltage limiting.
  bool registerdQdxdVpVector (Vector * dQdxdVpVectorPtr)
  { return ( dQdxdVpVectorPtr_ = dQdxdVpVectorPtr) ; }

  // DAE matrix registrations:
  bool registerDAEdQdxMatrix (Matrix * dqdxMatPtr)
  {  return (dQdxMatrixPtr_ = dqdxMatPtr); }

  bool registerDAEdFdxMatrix (Matrix * dfdxMatPtr)
  {  return (dFdxMatrixPtr_ = dfdxMatPtr); }

  // Registers the pointer to the current solution vector
  bool registerCurrSolVector(Vector ** solVecPtrPtr)
  { return ( currSolVectorPtrPtr_ = solVecPtrPtr ); }

  // Registers the pointer to the next solution vector
  bool registerNextSolVector(Vector ** solVecPtrPtr)
  { return ( nextSolVectorPtrPtr_ = solVecPtrPtr ); }

  // Registers the pointer to the flag solution vector
  bool registerFlagSolVector(Vector ** flagSolVecPtrPtr)
  { return ( flagSolVectorPtrPtr_ = flagSolVecPtrPtr ); }

  // Registers the pointer to the previous solution vector
  bool registerLastSolVector(Vector ** solVecPtrPtr)
  { return ( lastSolVectorPtrPtr_ = solVecPtrPtr ); }

  // Registers the pointer to the next solution vector's derivatives
  bool registerNextSolDerivVector(Vector ** solVecPtrPtr)
  { return ( nextSolDerivVectorPtrPtr_ = solVecPtrPtr ); }

  // Registers the pointer to the next state (auxiliary) vector's derivatives
  bool registerNextStaDerivVector(Vector ** staVecPtrPtr)
  { return ( nextStaDerivVectorPtrPtr_ = staVecPtrPtr ); }

  // Registers the pointer to the current state (auxiliary) vector
  bool registerCurrStaVector(Vector ** stateVecPtrPtr)
  { return ( currStaVectorPtrPtr_ = stateVecPtrPtr ); }

  // Registers the pointer to the next state (auxiliary) vector
  bool registerNextStaVector(Vector ** stateVecPtrPtr)
  { return ( nextStaVectorPtrPtr_ = stateVecPtrPtr ); }

  // Registers the pointer to the previous state (auxiliary) vector
  bool registerLastStaVector(Vector ** stateVecPtrPtr)
  { return ( lastStaVectorPtrPtr_ = stateVecPtrPtr ); }

  // Registers the pointer to the current store (auxiliary) vector
  bool registerCurrStoVector(Vector ** storeVecPtrPtr)
  { return ( currStoVectorPtrPtr_ = storeVecPtrPtr ); }

  // Registers the pointer to the next store (auxiliary) vector
  bool registerNextStoVector(Vector ** storeVecPtrPtr)
  { return ( nextStoVectorPtrPtr_ = storeVecPtrPtr ); }

  // Registers the pointer to the previous store (auxiliary) vector
  bool registerLastStoVector(Vector ** storeVecPtrPtr)
  { return ( lastStoVectorPtrPtr_ = storeVecPtrPtr ); }

  // Registers the pointer to the temporary state (auxiliary) vector
  bool registerTmpStaVector(Vector ** tmpStaVecPtrPtr)
  { return ( tmpStaVectorPtrPtr_ = tmpStaVecPtrPtr ); }

  // Registers the pointer to the temporary solution vector
  bool registerTmpSolVector(Vector ** tmpSolVecPtrPtr)
  { return ( tmpSolVectorPtrPtr_ = tmpSolVecPtrPtr ); }

  // Registers the pointer to the temporary state (auxiliary) vector's
  // derivatives
  bool registerTmpStaDerivVector(Vector ** tmpStaDerivVecPtrPtr)
  { return ( tmpStaDerivVectorPtrPtr_ = tmpStaDerivVecPtrPtr ); }

  // Registers the pointer to the temporary solution vector's derivatives
  bool registerTmpSolDerivVector(Vector ** tmpSolDerivVecPtrPtr)
  { return ( tmpSolDerivVectorPtrPtr_ = tmpSolDerivVecPtrPtr ); }

  bool registerTmpStaDivDiffVector(Vector ** tmpStaDivDiffVecPtrPtr)
  { return ( tmpSolDivDiffVectorPtrPtr_ = tmpStaDivDiffVecPtrPtr ); }
  bool registerTmpSolDivDiffVector(Vector ** tmpSolDivDiffVecPtrPtr)
  { return ( tmpSolDivDiffVectorPtrPtr_ = tmpSolDivDiffVecPtrPtr ); }

  // Register pointer to device mask
  bool registerDeviceMaskVector(Vector ** dmPP)
  { return (deviceMaskVectorPtrPtr_ = dmPP); }
  
  bool registerSolnColoring( Epetra_MapColoring * coloring )
  { return ( solnColoringPtr_ = coloring ); }

  bool registerICColoring( Epetra_MapColoring * coloring )
  { return ( initialConditionColoringPtr_ = coloring ); }

  // Accessor methods for linear system objects

  // Get method for the Jacobian matrix
  Matrix *  getJacobianMatrix() { return jacobianMatrixPtr_; }

  // Get method for the Jacobian test matrix (old DAE)
  Matrix *  getJacTestMatrix() { return jacTestMatrixPtr_; }
  // Get method for the Jacobian test matrix (new DAE)
  Matrix *  getdFdxTestMatrix() { return dFdxTestMatrixPtr_; }
  Matrix *  getdQdxTestMatrix() { return dQdxTestMatrixPtr_; }

  // Get method for the residual (RHS) vector
  Vector *  getRHSVector() { return rhsVectorPtr_; }

  // Get method for the limiting (jdxp) vector
  Vector *  getJDXPVector() { return jdxpVectorPtr_; }

  // Get method for the F vector
  Vector *  getFVector() { return fVectorPtr_; }

  // Get method for the voltage limiting dx vector
  Vector *  getDxVoltlimVector() { return dxVoltlimVectorPtr_; }

  // Get method for the limiting (jdx2) vector (old DAE)
  Vector *  getJDX2Vector() { return jdx2VectorPtr_; }

  // Get methods for the limiting (jdx2) vector (new DAE)
  Vector *  getFDX2Vector() { return Fdx2VectorPtr_; }
  Vector *  getQDX2Vector() { return Qdx2VectorPtr_; }

  // Get methods for DAE vectors:
  Vector * getDAEQVector () { return daeQVectorPtr_; }

  Vector * getDAEFVector () { return daeFVectorPtr_; }

  Vector * getDAEBVector () { return daeBVectorPtr_; }

  // dFdxdvp vector, for new-DAE voltage limiting.
  Vector * getdFdxdVpVector () { return  dFdxdVpVectorPtr_; }

  // dQdxdvp vector, for new-DAE voltage limiting.
  Vector * getdQdxdVpVector () { return  dQdxdVpVectorPtr_; }

  // Get methods for DAE matrices:
  Matrix * getDAEdQdxMatrix () { return dQdxMatrixPtr_; }

  Matrix * getDAEdFdxMatrix () { return dFdxMatrixPtr_; }

  // Get method for the current solution vector
  Vector *  getCurrSolVector() { return *currSolVectorPtrPtr_; }

  // Get method for the current solution vector pointer
  Vector ** getCurrSolVectorPtr() { return currSolVectorPtrPtr_; }

  // Get method for next current solution vector
  Vector *  getNextSolVector() { return *nextSolVectorPtrPtr_; }

  // Get method for next current solution vector pointer
  Vector ** getNextSolVectorPtr() { return nextSolVectorPtrPtr_; }

  // Get method for previous current solution vector 
  Vector *  getLastSolVector() { return *lastSolVectorPtrPtr_; }

  // Get method for previous current solution vector pointer
  Vector ** getLastSolVectorPtr() { return lastSolVectorPtrPtr_; }

  // Get method for flag solution vector
  Vector * getFlagSolVector() { return *flagSolVectorPtrPtr_; }

  // Get method for flag solution vector
  Vector ** getFlagSolVectorPtr() { return flagSolVectorPtrPtr_; }

  Vector *  getNextSolDerivVector()
  {
    return *nextSolDerivVectorPtrPtr_;
  }
  Vector ** getNextSolDerivVectorPtr()
  {
    return nextSolDerivVectorPtrPtr_;
  }

  Vector *  getCurrStaVector() { return *currStaVectorPtrPtr_; }
  Vector ** getCurrStaVectorPtr() { return currStaVectorPtrPtr_; }
  Vector *  getNextStaVector() { return *nextStaVectorPtrPtr_; }
  Vector ** getNextStaVectorPtr() { return nextStaVectorPtrPtr_; }
  Vector *  getLastStaVector () { return *lastStaVectorPtrPtr_; }
  Vector ** getLastStaVectorPtr() { return lastStaVectorPtrPtr_; }
  Vector *  getCurrStoVector() { return *currStoVectorPtrPtr_; }
  Vector ** getCurrStoVectorPtr() { return currStoVectorPtrPtr_; }
  Vector *  getNextStoVector() { return *nextStoVectorPtrPtr_; }
  Vector ** getNextStoVectorPtr() { return nextStoVectorPtrPtr_; }
  Vector *  getLastStoVector () { return *lastStoVectorPtrPtr_; }
  Vector ** getLastStoVectorPtr() { return lastStoVectorPtrPtr_; }

  Vector *  getNextStaDerivVector()
  {
    return *nextStaDerivVectorPtrPtr_;
  }
  Vector ** getNextStaDerivVectorPtr()
  {
    return nextStaDerivVectorPtrPtr_;
  }

  Vector *  getTmpSolVector   () { return *tmpSolVectorPtrPtr_; }
  Vector ** getTmpSolVectorPtr() { return tmpSolVectorPtrPtr_; }
  Vector *  getTmpStaVector   () { return *tmpStaVectorPtrPtr_; }
  Vector ** getTmpStaVectorPtr() { return tmpStaVectorPtrPtr_; }

  Vector ** getTmpSolDerivVectorPtr()
  {
    return tmpSolDerivVectorPtrPtr_;
  }
  Vector ** getTmpStaDerivVectorPtr()
  {
    return tmpStaDerivVectorPtrPtr_;
  }

  Vector ** getTmpSolDivDiffVectorPtr()
  {
    return tmpSolDivDiffVectorPtrPtr_;
  }
  Vector ** getTmpStaDivDiffVectorPtr()
  {
    return tmpStaDivDiffVectorPtrPtr_;
  }

  // Get method for the device mask vector
  Vector *  getDeviceMaskVector() { return *deviceMaskVectorPtrPtr_; }
  Vector **  getDeviceMaskVectorPtr() { return deviceMaskVectorPtrPtr_; }

  // Registers the Query utility
  QueryUtil * getQueryUtil()
  {
    return lasQueryUtil_;
  }

  // Problem size accessors

  // Get method for the GLOBAL solution vector size (i.e., number of unknowns)
  int getGlobalSolutionSize();

  // Get method for the GLOBAL state vector size (i.e., number of unknowns)
  int getGlobalStateSize();

  // Get method for the solution vector size (i.e., number of unknowns)
  int getSolutionSize();

  // Get method for the state (auxiliary) vector size
  int getStateSize();

  // Get method for the residual (RHS) vector size
  int getRHSSize();

  // Create residual (RHS) and Jacobian
  bool initializeSystem(); // bool block_analysis_flag);

  // Builder access
  Builder & builder() { return *lasBuilder_; }

  bool updateExternValsSolnVector(MultiVector * solnVector);
  bool updateExternValsStateVector(MultiVector * stateVector);
  bool updateExternValsStoreVector(MultiVector * storeVector);

  bool localInMatrix(const int & row) const;
  bool localInSolnVector(const int & row) const;
  bool localInStateVector(const int & row) const;

  // Debug call outputs vectors to console
  void debug() const;

private:

  N_PDS_Manager   *  pdsMgr_;

  QueryUtil *  lasQueryUtil_;
  Builder   *  lasBuilder_;

  Matrix    *  jacobianMatrixPtr_;
  Vector    *  rhsVectorPtr_;
  Vector    *  jdxpVectorPtr_;
  Vector    *  fVectorPtr_;
  Vector    *  dxVoltlimVectorPtr_;

  // old dae:
  Vector    *  jdx2VectorPtr_;
  Matrix    *  jacTestMatrixPtr_;

  // new dae:
  Vector    *  Fdx2VectorPtr_;
  Matrix    *  dFdxTestMatrixPtr_;
  Vector    *  Qdx2VectorPtr_;
  Matrix    *  dQdxTestMatrixPtr_;

  // DAE vectors:
  Vector    *  daeQVectorPtr_;
  Vector    *  daeFVectorPtr_;
  Vector    *  daeBVectorPtr_;

  Vector    *  dFdxdVpVectorPtr_;
  Vector    *  dQdxdVpVectorPtr_;

  // DAE matrices:
  Matrix    * dQdxMatrixPtr_;
  Matrix    * dFdxMatrixPtr_;

  Vector ** currSolVectorPtrPtr_;
  Vector ** nextSolVectorPtrPtr_;
  Vector ** lastSolVectorPtrPtr_;
  Vector ** flagSolVectorPtrPtr_;

  Vector ** nextSolDerivVectorPtrPtr_;
  Vector ** nextStaDerivVectorPtrPtr_;

  Vector ** currStaVectorPtrPtr_;
  Vector ** nextStaVectorPtrPtr_;
  Vector ** lastStaVectorPtrPtr_;
  Vector ** currStoVectorPtrPtr_;
  Vector ** nextStoVectorPtrPtr_;
  Vector ** lastStoVectorPtrPtr_;

  Vector **  tmpStaVectorPtrPtr_;
  Vector **  tmpSolVectorPtrPtr_;

  Vector **  tmpStaDerivVectorPtrPtr_;
  Vector **  tmpSolDerivVectorPtrPtr_;

  Vector **  tmpStaDivDiffVectorPtrPtr_;
  Vector **  tmpSolDivDiffVectorPtrPtr_;

  Vector **  deviceMaskVectorPtrPtr_;

  Epetra_MapColoring * solnColoringPtr_;

  Epetra_MapColoring * initialConditionColoringPtr_;
};

} // namespace Linear
} // namespace Xyce

#endif // Xyce_LAS_System_h

