//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_LAS_Builder.C,v $
//
// Purpose        : Builder class for parallel/serial linear objects
//
// Special Notes  :
//
// Creator        : Robert Hoekstra, SNL, Parallel Computational Sciences
//
// Creation Date  : 01/22/01
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.75.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:15 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

#include <N_UTL_fwd.h>

#include <N_ERH_ErrorMgr.h>
#include <N_LAS_Builder.h>
#include <N_LAS_LAFactory.h>
#include <N_LAS_Matrix.h>
#include <N_LAS_MultiVector.h>
#include <N_LAS_QueryUtil.h>
#include <N_LAS_Vector.h>
#include <N_PDS_Comm.h>
#include <N_PDS_GlobalAccessor.h>
#include <N_PDS_Manager.h>
#include <N_PDS_ParMap.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_Functors.h>

#include <Epetra_CrsGraph.h>
#include <Epetra_CrsMatrix.h>
#include <Epetra_Map.h>
#include <Epetra_Export.h>
#include <Epetra_MapColoring.h>
#include <EpetraExt_View_CrsGraph.h>

using std::max;
using Xyce::VERBOSE_LINEAR;

namespace Xyce {
namespace Linear {

//-----------------------------------------------------------------------------
// Function      : Builder::~
// Purpose       :
// Special Notes :
// Scope         :
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
Builder::~Builder()
{}

//-----------------------------------------------------------------------------
// Function      : Builder::createMultiVector
// Purpose       : returns Soln/RHS sized multiVector
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/10/00
//-----------------------------------------------------------------------------
MultiVector * Builder::createMultiVector( const int numVectors, const double value ) const
{
  return LAFactory::newMultiVector( value,
                                    *(pdsMgr_->getParallelMap(Parallel::SOLUTION)),
                                    numVectors,
                                    *(pdsMgr_->getParallelMap(Parallel::SOLUTION_OVERLAP_GND)) );
}

//-----------------------------------------------------------------------------
// Function      : Builder::createStateMultiVector
// Purpose       : returns State sized multiVector
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/10/00
//-----------------------------------------------------------------------------
MultiVector * Builder::createStateMultiVector( const int numVectors, const double value ) const
{
  return LAFactory::newMultiVector( value,
                                    *(pdsMgr_->getParallelMap(Parallel::STATE)),
                                    numVectors,
                                    *(pdsMgr_->getParallelMap(Parallel::STATE_OVERLAP)) );
}

//-----------------------------------------------------------------------------
// Function      : Builder::createStoreMultiVector
// Purpose       : returns Store sized multiVector
// Special Notes :
// Scope         : Public
// Creator       : Eric Keiter
// Creation Date :
//-----------------------------------------------------------------------------
MultiVector * Builder::createStoreMultiVector( const int numVectors, const double value ) const
{
  return LAFactory::newMultiVector( value,
                                    *(pdsMgr_->getParallelMap(Parallel::STORE)),
                                    numVectors,
                                    *(pdsMgr_->getParallelMap(Parallel::STORE_OVERLAP)) );
}

//-----------------------------------------------------------------------------
// Function      : Builder::createVector
// Purpose       : returns Soln/RHS sized vector
// Special Notes : Takes an initial value argument.
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Computational Sciences
// Creation Date : 03/02/01
//-----------------------------------------------------------------------------
Vector * Builder::createVector( const double value ) const
{
    return LAFactory::newVector( value,
                                 *(pdsMgr_->getParallelMap(Parallel::SOLUTION)),
                                 *(pdsMgr_->getParallelMap(Parallel::SOLUTION_OVERLAP_GND)) );
}

//-----------------------------------------------------------------------------
// Function      : Builder::createStateVector
// Purpose       : returns State sized vector
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/10/00
//-----------------------------------------------------------------------------
Vector * Builder::createStateVector( const double value ) const
{
    return LAFactory::newVector( value,
                                 *(pdsMgr_->getParallelMap(Parallel::STATE)),
                                 *(pdsMgr_->getParallelMap(Parallel::STATE_OVERLAP)) );
}

//-----------------------------------------------------------------------------
// Function      : Builder::createStoreVector
// Purpose       : returns Store sized vector
// Special Notes :
// Scope         : Public
// Creator       : Eric Keiter, SNL
// Creation Date :
//-----------------------------------------------------------------------------
Vector * Builder::createStoreVector( const double value ) const
{
    return LAFactory::newVector( value,
                                 *(pdsMgr_->getParallelMap(Parallel::STORE)),
                                 *(pdsMgr_->getParallelMap(Parallel::STORE_OVERLAP)) );
}


//-----------------------------------------------------------------------------
// Function      : Builder::createStoreVector
// Purpose       : returns Store sized vector
// Special Notes :
// Scope         : Public
// Creator       : Eric Keiter, SNL
// Creation Date :
//-----------------------------------------------------------------------------
Vector * Builder::createLeadCurrentVector( const double value ) const
{
    return LAFactory::newVector( value,
                                 *(pdsMgr_->getParallelMap(Parallel::LEADCURRENT)),
                                 *(pdsMgr_->getParallelMap(Parallel::LEADCURRENT_OVERLAP)) );
}

//-----------------------------------------------------------------------------
// Function      : Builder::createMatrix
// Purpose       : returns Matrix initialized based on QueryUtil and ParMap
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/10/00
//-----------------------------------------------------------------------------
Matrix * Builder::createMatrix(const double initialValue) const
{

  Matrix * mat = 0;

  Epetra_CrsGraph * overlapGraph = pdsMgr_->getMatrixGraph( Parallel::JACOBIAN_OVERLAP_GND );
  Epetra_CrsGraph * baseGraph = pdsMgr_->getMatrixGraph( Parallel::JACOBIAN );

  mat = new Matrix( overlapGraph, baseGraph );

  return mat;
}

//-----------------------------------------------------------------------------
// Function      : Builder::createSolnColoring
// Purpose       : Color Map representing variable types in solution vector
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 03/08/04
//-----------------------------------------------------------------------------
Epetra_MapColoring * Builder::createSolnColoring() const
{
  const std::vector<char> & charColors = lasQueryUtil_->rowList_VarType();

  int size = charColors.size();
  std::vector<int> colors( size );
  for( int i = 0; i < size; ++i )
  {
    switch( charColors[i] )
    {
      case 'V': colors[i] = 0;
                break;
      case 'I': colors[i] = 1;
                break;
      default : colors[i] = 2;
                break;
    }
  }

  return new Epetra_MapColoring( *(pdsMgr_->getParallelMap(Parallel::SOLUTION)->petraBlockMap()),
                                 &colors[0], 0 );
}


//-----------------------------------------------------------------------------
// Function      : Builder::createInitialConditionColoring
// Purpose       :
// Special Notes : The .IC and .NODESET capabilities will use the variables
//                 which are colored 0.  This will be all voltage nodes not
//                 connected to independent sources.
// Scope         : Public
// Creator       : Eric R. Keiter,  SNL
// Creation Date : 10/15/07
//-----------------------------------------------------------------------------
Epetra_MapColoring * Builder::createInitialConditionColoring() const
{
  const std::vector<char> & charColors = lasQueryUtil_->rowList_VarType();
  const std::vector<int> & vsrcGIDColors = lasQueryUtil_->vsrcGIDVec();

  int size = charColors.size();
  int vsrcSize = vsrcGIDColors.size();
  std::vector<int> colors( size );
  for( int i = 0; i < size; ++i )
  {
    switch( charColors[i] )
    {
      case 'V': colors[i] = 0;
                break;
      case 'I': colors[i] = 1;
                break;
      default : colors[i] = 2;
                break;
    }
  }

  N_PDS_ParMap * solnMap = pdsMgr_->getParallelMap(Parallel::SOLUTION);
  for( int i=0; i < vsrcSize; ++i )
  {
    int vsrcID = vsrcGIDColors[i];
    // Convert the ID from local to global if it is valid and the build is parallel.
    if (vsrcID >= 0)
    {
#ifdef Xyce_PARALLEL_MPI
      vsrcID = solnMap->globalToLocalIndex( vsrcGIDColors[i] );
#endif
      if (vsrcID < size && vsrcID >= 0)
        colors[vsrcID] = 1;
    }
  }

  return new Epetra_MapColoring( *(solnMap->petraBlockMap()), &colors[0], 0 );
}

//-----------------------------------------------------------------------------
// Function      : Builder::generateParMaps
// Purpose       : Creates parallel maps for SOLN and STATE
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 1/24/01
//-----------------------------------------------------------------------------
bool Builder::generateParMaps()
{
  int numLocalRows = lasQueryUtil_->numLocalRows();
  std::vector<int> arrayGIDs(numLocalRows);
  arrayGIDs = lasQueryUtil_->rowList_GID();

  int numLocalStateVars = lasQueryUtil_->numLocalStateVars();
  std::vector<int> arrayStateGIDs(numLocalStateVars);
  arrayStateGIDs = lasQueryUtil_->rowList_StateGID();

  int numLocalStoreVars = lasQueryUtil_->numLocalStoreVars();
  std::vector<int> arrayStoreGIDs(numLocalStoreVars);
  arrayStoreGIDs = lasQueryUtil_->rowList_StoreGID();
  
  int numLocalLeadCurrentVars = lasQueryUtil_->numLocalLeadCurrentVars();
  std::vector<int> arrayLeadCurrentGIDs(numLocalLeadCurrentVars);
  arrayLeadCurrentGIDs = lasQueryUtil_->rowList_LeadCurrentGID();

  int numGlobalRows = lasQueryUtil_->numGlobalRows();
  N_PDS_ParMap * solnMap = pdsMgr_->createParallelMap(numGlobalRows, numLocalRows, arrayGIDs);

  int numGlobalStateVars = lasQueryUtil_->numGlobalStateVars();
  N_PDS_ParMap * stateMap = pdsMgr_->createParallelMap(numGlobalStateVars, numLocalStateVars, arrayStateGIDs);

  int numGlobalStoreVars = lasQueryUtil_->numGlobalStoreVars();
  N_PDS_ParMap * storeMap = pdsMgr_->createParallelMap(numGlobalStoreVars, numLocalStoreVars, arrayStoreGIDs);

  int numGlobalLeadCurrentVars = lasQueryUtil_->numGlobalLeadCurrentVars();
  N_PDS_ParMap * leadCurrentMap = pdsMgr_->createParallelMap(numGlobalLeadCurrentVars, numLocalLeadCurrentVars, arrayLeadCurrentGIDs);
    
  pdsMgr_->addParallelMap(Parallel::SOLUTION, solnMap);
  pdsMgr_->addParallelMap(Parallel::STATE, stateMap);
  pdsMgr_->addParallelMap(Parallel::STORE, storeMap);
  pdsMgr_->addParallelMap(Parallel::LEADCURRENT, leadCurrentMap);

  int procCnt = pdsMgr_->getPDSComm()->numProc();
  int numExternRows = lasQueryUtil_->numExternRows();
  int totalRows = numLocalRows + numExternRows;

  int size = lasQueryUtil_->rowList_ExternGID().size();
  for(int i = 0; i < size; ++i)
    arrayGIDs.push_back(lasQueryUtil_->rowList_ExternGID()[i].first);

  int numGlobalExternRows = lasQueryUtil_->numGlobalExternRows();
  int totalGlobalRows = numGlobalRows + numGlobalExternRows;
  N_PDS_ParMap * overlapSolnMap = pdsMgr_->createParallelMap(
	totalGlobalRows,
        totalRows,
        arrayGIDs );

  arrayGIDs.push_back(-1);
  totalGlobalRows += procCnt;
  totalRows++;
  N_PDS_ParMap * overlapGndSolnMap = pdsMgr_->createParallelMap(
	totalGlobalRows,
        totalRows,
        arrayGIDs,
        -1 );

  int numExternStateVars = lasQueryUtil_->numExternStateVars();
  int totalStateVars = numLocalStateVars + numExternStateVars;
  arrayStateGIDs.reserve(totalStateVars);

  size = lasQueryUtil_->rowList_ExternStateGID().size();
  for( int i = 0; i < size; ++i )
    arrayStateGIDs.push_back(lasQueryUtil_->rowList_ExternStateGID()[i].first);

  int numGlobalExternStateVars = lasQueryUtil_->numGlobalExternStateVars();
  int totalGlobalStateVars = numGlobalStateVars + numGlobalExternStateVars;
  N_PDS_ParMap * overlapStateMap = pdsMgr_->createParallelMap(
        totalGlobalStateVars,
        totalStateVars,
        arrayStateGIDs );

  int numExternStoreVars = lasQueryUtil_->numExternStoreVars();
  int totalStoreVars = numLocalStoreVars + numExternStoreVars;
  arrayStoreGIDs.reserve(totalStoreVars);

  size = lasQueryUtil_->rowList_ExternStoreGID().size();
  for( int i = 0; i < size; ++i )
    arrayStoreGIDs.push_back(lasQueryUtil_->rowList_ExternStoreGID()[i].first);

  int numGlobalExternStoreVars = lasQueryUtil_->numGlobalExternStoreVars();
  int totalGlobalStoreVars = numGlobalStoreVars + numGlobalExternStoreVars;
  N_PDS_ParMap * overlapStoreMap = pdsMgr_->createParallelMap(
        totalGlobalStoreVars,
        totalStoreVars,
        arrayStoreGIDs );

  int numExternLeadCurrentVars = lasQueryUtil_->numExternLeadCurrentVars();
  int totalLeadCurrentVars = numLocalLeadCurrentVars + numExternLeadCurrentVars;
  arrayLeadCurrentGIDs.reserve(totalLeadCurrentVars);

  size = lasQueryUtil_->rowList_ExternLeadCurrentGID().size();
  for( int i = 0; i < size; ++i )
    arrayLeadCurrentGIDs.push_back(lasQueryUtil_->rowList_ExternLeadCurrentGID()[i].first);

  int numGlobalExternLeadCurrentVars = lasQueryUtil_->numGlobalExternLeadCurrentVars();
  int totalGlobalLeadCurrentVars = numGlobalLeadCurrentVars + numGlobalExternLeadCurrentVars;
  N_PDS_ParMap * overlapLeadCurrentMap = pdsMgr_->createParallelMap(
        totalGlobalLeadCurrentVars,
        totalLeadCurrentVars,
        arrayLeadCurrentGIDs );

  pdsMgr_->addParallelMap( Parallel::SOLUTION_OVERLAP, overlapSolnMap );
  pdsMgr_->addParallelMap( Parallel::SOLUTION_OVERLAP_GND, overlapGndSolnMap );
  pdsMgr_->addParallelMap( Parallel::STATE_OVERLAP, overlapStateMap );
  pdsMgr_->addParallelMap( Parallel::STORE_OVERLAP, overlapStoreMap );
  pdsMgr_->addParallelMap( Parallel::LEADCURRENT_OVERLAP, overlapLeadCurrentMap );

  N_PDS_GlobalAccessor * solnGA = pdsMgr_->addGlobalAccessor( Parallel::SOLUTION );
  solnGA->registerExternGIDVector( lasQueryUtil_->rowList_ExternGID() );
  solnGA->generateMigrationPlan();

  N_PDS_GlobalAccessor * stateGA = pdsMgr_->addGlobalAccessor( Parallel::STATE );
  stateGA->registerExternGIDVector( lasQueryUtil_->rowList_ExternStateGID() );
  stateGA->generateMigrationPlan();

  N_PDS_GlobalAccessor * storeGA = pdsMgr_->addGlobalAccessor( Parallel::STORE );
  storeGA->registerExternGIDVector( lasQueryUtil_->rowList_ExternStoreGID() );
  storeGA->generateMigrationPlan();

  return true;
}

//-----------------------------------------------------------------------------
// Function      : Builder::generateGraphs
// Purpose       : Generation of Matrix Graphs, stored with ParMgr
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 8/23/02
//-----------------------------------------------------------------------------
bool Builder::generateGraphs()
{
  const std::vector< std::vector<int> > & rcData = lasQueryUtil_->rowList_ColList();
  int numLocalRows_Overlap = rcData.size();

  int maxNZs = 0;
  std::vector<int> arrayNZs( numLocalRows_Overlap );
  for( int i = 0; i < numLocalRows_Overlap; ++i )
  {
    arrayNZs[i] = rcData[i].size();
    maxNZs = max( maxNZs, arrayNZs[i] );
  }

  Epetra_BlockMap & overlapGndMap = *(pdsMgr_->getParallelMap( Parallel::SOLUTION_OVERLAP_GND )->petraBlockMap());
  Epetra_BlockMap & overlapMap = *(pdsMgr_->getParallelMap( Parallel::SOLUTION_OVERLAP )->petraBlockMap());
  Epetra_BlockMap & localMap = *(pdsMgr_->getParallelMap( Parallel::SOLUTION )->petraBlockMap());

  Epetra_CrsGraph * overlapGndGraph = new Epetra_CrsGraph( Copy, overlapGndMap, &arrayNZs[0] );
  Epetra_CrsGraph * overlapGraph = new Epetra_CrsGraph( Copy, overlapMap, &arrayNZs[0] );

  std::vector<int> indices( maxNZs );
  for( int i = 0; i < numLocalRows_Overlap; ++i )
  {
    copy( rcData[i].begin(), rcData[i].end(), indices.begin() );
    if( arrayNZs[i] )
      overlapGndGraph->InsertGlobalIndices( overlapGndMap.GID(i), arrayNZs[i], &indices[0] );

    if( overlapGndMap.GID(i) != -1 && arrayNZs[i] )
    {
      if( indices[0] == -1 )
        overlapGraph->InsertGlobalIndices( overlapMap.GID(i), arrayNZs[i]-1, &indices[1] );
      else
        overlapGraph->InsertGlobalIndices( overlapMap.GID(i), arrayNZs[i], &indices[0] );
    }
  }
  overlapGndGraph->FillComplete();
  overlapGraph->FillComplete();
  pdsMgr_->addMatrixGraph( Parallel::JACOBIAN_OVERLAP_GND, overlapGndGraph );
  pdsMgr_->addMatrixGraph( Parallel::JACOBIAN_OVERLAP, overlapGraph );

#ifndef Xyce_PARALLEL_MPI
  //determine what the column map should be for the local view of the graph
  //store a new one if it is different than the overlap map
  const Epetra_BlockMap & oColMap = overlapGraph->ColMap();
  int oColMapSize = oColMap.NumMyElements();
  std::vector<int> nColMapElements;
  for( int i = 0; i < oColMapSize; ++i )
    if( overlapMap.MyGID( oColMap.GID(i) ) ) nColMapElements.push_back( oColMap.GID(i) );

  N_PDS_ParMap * localColMap = 0;

  double lSize = static_cast<double>(nColMapElements.size());
  double tSize;
  pdsMgr_->getPDSComm()->sumAll( &lSize, &tSize, 1 );
  if( tSize < overlapMap.NumGlobalElements() )
  {
    int totalSize = static_cast<int>(tSize);
    localColMap = pdsMgr_->createParallelMap(
        totalSize,
        nColMapElements.size(),
        nColMapElements );
    pdsMgr_->addParallelMap( Parallel::SOLUTION_COLUMN, localColMap );
  }

  Epetra_BlockMap * nColMap = &overlapMap;
  if( localColMap ) nColMap = localColMap->petraBlockMap();

  EpetraExt::CrsGraph_View * viewTransform = new EpetraExt::CrsGraph_View( &localMap, nColMap );
  Epetra_CrsGraph * localGraph = &((*viewTransform)( *overlapGraph ));
  pdsMgr_->addMatrixGraph( Parallel::JACOBIAN, localGraph, viewTransform );
#else
  if( pdsMgr_->getPDSComm()->isSerial() )
  {
    //determine what the column map should be for the local view of the graph
    //store a new one if it is different than the overlap map
    const Epetra_BlockMap & oColMap = overlapGraph->ColMap();
    int oColMapSize = oColMap.NumMyElements();
    std::vector<int> nColMapElements;
    for( int i = 0; i < oColMapSize; ++i )
      if( overlapMap.MyGID( oColMap.GID(i) ) ) nColMapElements.push_back( oColMap.GID(i) );

    N_PDS_ParMap * localColMap = 0;

    double lSize = static_cast<double>(nColMapElements.size());
    double tSize;
    pdsMgr_->getPDSComm()->sumAll( &lSize, &tSize, 1 );
    if( tSize < overlapMap.NumGlobalElements() )
    {
      int totalSize = static_cast<int>(tSize);
      localColMap = pdsMgr_->createParallelMap(
          totalSize,
          nColMapElements.size(),
          nColMapElements );
      pdsMgr_->addParallelMap( Parallel::SOLUTION_COLUMN, localColMap );
    }

    Epetra_BlockMap * nColMap = &overlapMap;
    if( localColMap ) nColMap = localColMap->petraBlockMap();

    EpetraExt::CrsGraph_View * viewTransform = new EpetraExt::CrsGraph_View( &localMap, nColMap );
    Epetra_CrsGraph * localGraph = &((*viewTransform)( *overlapGraph ));
    pdsMgr_->addMatrixGraph( Parallel::JACOBIAN, localGraph, viewTransform );
  }
  else
  {
    Epetra_Export exporter( overlapMap, localMap );
    Epetra_CrsGraph * localGraph = new Epetra_CrsGraph( Copy, localMap, 0 );
    localGraph->Export( *overlapGraph, exporter, Add );
    if (VERBOSE_LINEAR)
      Xyce::lout() << "Exported To Local Graph!\n"  << std::endl;

    localGraph->FillComplete();

    if (VERBOSE_LINEAR)
      Xyce::lout() << "Local Graph Transformed!\n"  << std::endl;

    pdsMgr_->addMatrixGraph( Parallel::JACOBIAN, localGraph );
  }
#endif


  return true;
}

//-----------------------------------------------------------------------------
// Function      : Builder::getSolutionMap
// Purpose       :
// Special Notes : This is specifically for blockAnalysis types (like MPDE & HB)
// so we can get a valid map from the builder.
// Scope         : Public
// Creator       : Todd Coffey, 1414
// Creation Date : 09/05/08
//-----------------------------------------------------------------------------
RCP<const Epetra_Map> Builder::getSolutionMap() const
{
  return(Teuchos::rcp(pdsMgr_->getParallelMap(Parallel::SOLUTION)->petraMap(),false));
}

//-----------------------------------------------------------------------------
// Function      : Builder::getSolutionMaps
// Purpose       :
// Special Notes : This is specifically for ModelEvaluator Interface
// Scope         : Public
// Creator       : Todd Coffey, 1414
// Creation Date : 06/02/09
//-----------------------------------------------------------------------------
void Builder::getSolutionMaps(
    RCP<N_PDS_ParMap>& soln_map,
    RCP<N_PDS_ParMap>& soln_ognd_map
    ) const
{
  soln_map = Teuchos::rcp(pdsMgr_->getParallelMap(Parallel::SOLUTION),false);
  soln_ognd_map = Teuchos::rcp(pdsMgr_->getParallelMap(Parallel::SOLUTION_OVERLAP_GND),false);
}

//-----------------------------------------------------------------------------
// Function      : Builder::getStateMap
// Purpose       :
// Special Notes : This is specifically for ModelEvaluator Interface
// Scope         : Public
// Creator       : Todd Coffey, 1414
// Creation Date : 06/02/09
//-----------------------------------------------------------------------------
void Builder::getStateMap(
    RCP<N_PDS_ParMap> & state_map
    ) const
{
  state_map = Teuchos::rcp(pdsMgr_->getParallelMap(Parallel::STATE),false);
}

//-----------------------------------------------------------------------------
// Function      : Builder::getStoreMap
// Purpose       :
// Special Notes : This is specifically for ModelEvaluator Interface
// Scope         : Public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
void Builder::getStoreMap(
    RCP<N_PDS_ParMap> & store_map
    ) const
{
  store_map = Teuchos::rcp(pdsMgr_->getParallelMap(Parallel::STORE),false);
}

//-----------------------------------------------------------------------------
// Function      : Builder::getdFdxGraphs
// Purpose       :
// Special Notes : This is specifically for ModelEvaluator Interface
// Scope         : Public
// Creator       : Todd Coffey, 1414
// Creation Date : 06/02/09
//-----------------------------------------------------------------------------
void Builder::getdFdxGraphs(
    RCP<Epetra_CrsGraph>& dFdx_graph,
    RCP<Epetra_CrsGraph>& dFdx_ognd_graph
    ) const
{
  dFdx_graph = Teuchos::rcp(pdsMgr_->getMatrixGraph( Parallel::DAE_DFDX_JAC ));
  dFdx_ognd_graph = Teuchos::rcp(pdsMgr_->getMatrixGraph( Parallel::DAE_DFDX_JAC_OVERLAP_GND ));
}

//-----------------------------------------------------------------------------
// Function      : Builder::getdQdxGraphs
// Purpose       :
// Special Notes : This is specifically for ModelEvaluator Interface
// Scope         : Public
// Creator       : Todd Coffey, 1414
// Creation Date : 06/02/09
//-----------------------------------------------------------------------------
void Builder::getdQdxGraphs(
    RCP<Epetra_CrsGraph>& dQdx_graph,
    RCP<Epetra_CrsGraph>& dQdx_ognd_graph
    ) const
{
  dQdx_graph = Teuchos::rcp(pdsMgr_->getMatrixGraph( Parallel::DAE_DQDX_JAC ));
  dQdx_ognd_graph = Teuchos::rcp(pdsMgr_->getMatrixGraph( Parallel::DAE_DQDX_JAC_OVERLAP_GND ));
}

} // namespace Linear
} // namespace Xyce
