//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_LAS_HBBlockJacobiEpetraOperator.C,v $
//
// Purpose        :
//
// Creator        : Heidi Thornquist, 1437
//
// Creation Date  : 09/04/08
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.9.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:15 $
//
// Current Owner  : $Author: tvrusso $
//-------------------------------------------------------------------------

// ---------- Standard Includes ----------

// ----------   Xyce Includes   ----------

#include <Xyce_config.h>

#include <N_ERH_ErrorMgr.h>
#include <N_LAS_HBBlockJacobiEpetraOperator.h>
#include <N_LAS_HBBuilder.h>
#include <N_LAS_MultiVector.h>
#include <N_LAS_Vector.h>
#include <N_PDS_ParMap.h>

// ----------   Trilinos Includes   ----------

#include <Epetra_LinearProblem.h>
#include <Epetra_MultiVector.h>
#include <Epetra_Vector.h>
#include <Epetra_Comm.h>
#include <Epetra_Map.h>
#include <Amesos_BaseSolver.h>

using Teuchos::RCP;

namespace Xyce {
namespace Linear {

//-----------------------------------------------------------------------------
// Function      : matrixFreeEpetraOperator
// Purpose       : non-member constructor
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 9/4/08
//-----------------------------------------------------------------------------
RCP<HBBlockJacobiEpetraOperator> blockJacobiOperator(
    const std::vector<Teuchos::RCP<Epetra_LinearProblem> >& epetraProblems,
    const std::vector<Teuchos::RCP<Amesos_BaseSolver> >& amesosSolvers,
    const Teuchos::RCP<HBBuilder>& hbBuilder
    )
{
  RCP<HBBlockJacobiEpetraOperator> epetraOperator =
    rcp(new HBBlockJacobiEpetraOperator);
  epetraOperator->initialize(epetraProblems,
      amesosSolvers,
      hbBuilder
      );
  return epetraOperator;
}


//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::HBBlockJacobiEpetraOperator
// Purpose       : Constructor
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
HBBlockJacobiEpetraOperator::HBBlockJacobiEpetraOperator()
{
  isInitialized_ = false;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::HBBlockJacobiEpetraOperator
// Purpose       : Destructor
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
HBBlockJacobiEpetraOperator::~HBBlockJacobiEpetraOperator()
{
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::initialize
// Purpose       : Initialization
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
void HBBlockJacobiEpetraOperator::initialize(
      const std::vector<Teuchos::RCP<Epetra_LinearProblem> >& epetraProblems,
      const std::vector<Teuchos::RCP<Amesos_BaseSolver> >& amesosSolvers,
      const Teuchos::RCP<HBBuilder>& hbBuilder
    )
{
  epetraProblems_ = epetraProblems;
  amesosSolvers_ = amesosSolvers;
  hbBuilder_ = hbBuilder;
  N_ = epetraProblems_.size();
  M_ = (N_-1)/2;
  isInitialized_ = true;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::SetUseTranspose
// Purpose       : Define if transpose Apply and ApplyInverse is to be used.
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
int HBBlockJacobiEpetraOperator::SetUseTranspose(bool UseTranspose)
{
  // This is not supported for the HB load layers.
  return -1;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::ApplyInverse
// Purpose       : Apply matrix free preconditioner with Epetra_MultiVectors
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
int HBBlockJacobiEpetraOperator::ApplyInverse(
  const Epetra_MultiVector& X,
  Epetra_MultiVector& Y
  ) const
{
  // Convert these to MultiVectors and call the other Apply

  // Cast away the const until the Apply which will enforce it.
  // This is necessary because there is no const view function in MultiVector
//  Epetra_MultiVector* Xptr = const_cast<Epetra_MultiVector*>(&X);
//  MultiVector las_X(Xptr);  // This is the wrong thing to do, when it goes out of scope, it deletes the Epetra_MultiVector Ptr.
//  MultiVector las_Y(&Y);

  // COPY the multi-vector data into new objects on the stack.
  Epetra_MultiVector* Xcopy = new Epetra_MultiVector(X); // This gets deleted by the MultiVector below
  Epetra_MultiVector* Ycopy = new Epetra_MultiVector(Y); // This gets deleted by the MultiVector below
  MultiVector las_X(Xcopy, true); // this co-ops the Epetra_MultiVector and uses (and owns) its memory
  MultiVector las_Y(Ycopy, true); // this co-ops the Epetra_MultiVector and uses (and owns) its memory
  int status = ApplyInverse(las_X,las_Y);
  // COPY the Ycopy data back into Y
  Y = las_Y.epetraObj();
  return(status);
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::ApplyInverse
// Purpose       : Apply matrix free preconditioner with MultiVectors
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
int HBBlockJacobiEpetraOperator::ApplyInverse(
  const MultiVector& X,
  MultiVector& Y
  ) const
{
  if (!isInitialized_)
  {
    std::string msg = "HBBlockJacobiEpetraOperator::ApplyInverse:  I'm not initialized!";
    N_ERH_ErrorMgr::report(N_ERH_ErrorMgr::DEV_FATAL_0, msg);
  }
  bool status = true;

  // Determine number of time-domain variables.
  int n = X.localLength() / (2*N_);

  // Apply the j-th block to all the vectors in the multivector.
  for (int nB=0; nB<N_; ++nB) {

    Epetra_MultiVector * nB_RHS = epetraProblems_[nB]->GetRHS();
    double * nB_Soln = epetraProblems_[nB]->GetLHS()->Values();

    // Apply preconditioner to each vector
    for (int i=0 ; i<X.numVectors() ; ++i) {

      const Vector x(X.epetraVector(i), true);
      Vector y(Y.epetraVector(i), true);

      // Load the real and imaginary part of the RHS.
      // NOTE:  We have to convert from complex pair storage to real-equivalent form.
      for (int j=0; j<n; ++j) {
	nB_RHS->ReplaceMyValue(j, 0, x[j*(2*N_)+2*nB]);        // real
	nB_RHS->ReplaceMyValue(n+j, 0, x[j*(2*N_)+2*nB+1]);    // imaginary
      }

      // Solve the linear system
      amesosSolvers_[nB]->Solve();

      // Copy the solution back into y
      for (int j=0; j<n; ++j) {
	y[j*(2*N_)+2*nB]   = nB_Soln[j];
	y[j*(2*N_)+2*nB+1] = nB_Soln[n+j];
      }
 
    }
  }

  if (status)
  {
    return 0;
  }
  else
  {
    return -1;
  }
}
//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::Apply
// Purpose       : Apply inverse of matrix free preconditioner with Epetra_MultiVectors
// Special Notes : Not supported!
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
int HBBlockJacobiEpetraOperator::Apply(
  const Epetra_MultiVector& X,
  Epetra_MultiVector& Y
  ) const
{
  std::string msg = "HBBlockJacobiEpetraOperator::Apply is not supported!";
  N_ERH_ErrorMgr::report(N_ERH_ErrorMgr::DEV_FATAL_0, msg);
  return -1;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::Apply
// Purpose       : Apply inverse of matrix free preconditioner with MultiVectors
// Special Notes : Not supported!
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
int HBBlockJacobiEpetraOperator::Apply(
  const MultiVector& X,
  MultiVector& Y
  ) const
{
  std::string msg = "HBBlockJacobiEpetraOperator::Apply is not supported!";
  N_ERH_ErrorMgr::report(N_ERH_ErrorMgr::DEV_FATAL_0, msg);
  return -1;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::NormInf
// Purpose       : Norm Inf of matrix
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
double HBBlockJacobiEpetraOperator::NormInf() const
{
  std::string msg = "HBBlockJacobiEpetraOperator::NormInf is not supported!";
  N_ERH_ErrorMgr::report(N_ERH_ErrorMgr::DEV_FATAL_0, msg);
  return -1.0;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::Label
// Purpose       : Label for operator
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
const char * HBBlockJacobiEpetraOperator::Label() const
{
  return "Matrix Free Harmonic Balance Block Jacobi Preconditioner";
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::UseTranspose
// Purpose       : Query for useTranspose setting
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
bool HBBlockJacobiEpetraOperator::UseTranspose() const
{
  // Use Transpose is not supported, so always return false.
  return false;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::HasNormInf
// Purpose       : Query for normInf support
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
bool HBBlockJacobiEpetraOperator::HasNormInf() const
{
  // Norm Inf is not supported, so always return false.
  return false;
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::Comm
// Purpose       : Return Epetra_Comm object
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
const Epetra_Comm & HBBlockJacobiEpetraOperator::Comm() const
{
  if (!isInitialized_)
  {
    std::string msg = "HBBlockJacobiEpetraOperator::Comm:  I'm not initialized!";
    N_ERH_ErrorMgr::report(N_ERH_ErrorMgr::DEV_FATAL_0, msg);
  }
  return((epetraProblems_[0])->GetMatrix()->Comm());
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::OperatorDomainMap
// Purpose       : Return Epetra_Map corresponding to domain of operator
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
const Epetra_Map & HBBlockJacobiEpetraOperator::OperatorDomainMap() const
{
  if (!isInitialized_)
  {
    std::string msg = "HBBlockJacobiEpetraOperator::OperatorDomainMap:  I'm not initialized!";
    N_ERH_ErrorMgr::report(N_ERH_ErrorMgr::DEV_FATAL_0, msg);
  }
  return(*(hbBuilder_->getSolutionMap()));
}

//-----------------------------------------------------------------------------
// Function      : HBBlockJacobiEpetraOperator::OperatorRangeMap
// Purpose       : Return Epetra_Map corresponding to range of operator
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, 1437
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
const Epetra_Map & HBBlockJacobiEpetraOperator::OperatorRangeMap() const
{
  if (!isInitialized_)
  {
    std::string msg = "HBBlockJacobiEpetraOperator::OperatorRangeMap:  I'm not initialized!";
    N_ERH_ErrorMgr::report(N_ERH_ErrorMgr::DEV_FATAL_0, msg);
  }
  return(*(hbBuilder_->getSolutionMap()));
}

} // namespace Linear
} // namespace Xyce
