//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Filename       : $RCSfile: N_LAS_MultiVector.C,v $
//
// Purpose        : Implementation file for the Abstract interface to the
//                  multi-vector types (RDP, RSP, CDP or CSP).
//
// Special Notes  :
//
// Creator        : Scott A. Hutchinson, SNL, Parallel Computational Sciences
//
// Creation Date  : 05/20/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.126.2.2 $
//
// Revision Date  : $Date: 2015/04/02 18:20:15 $
//
// Current Owner  : $Author: tvrusso $
//-------------------------------------------------------------------------

#include <Xyce_config.h>

// ---------- Standard Includes ----------
#include <cstdio>

// ----------   Xyce Includes   ----------
#include <N_UTL_fwd.h>

#include <N_ERH_ErrorMgr.h>
#include <N_LAS_MultiVector.h>
#include <N_LAS_Vector.h>
#include <N_PDS_Comm.h>
#include <N_PDS_ParMap.h>
#include <N_UTL_FeatureTest.h>

// ---------  Other Includes  -----------

#include <Epetra_MultiVector.h>
#include <Epetra_Vector.h>
#include <Epetra_Import.h>
#include <Epetra_Export.h>
#include <Epetra_Map.h>
#include <Epetra_Comm.h>

#ifdef Xyce_PARALLEL_MPI
#include <Epetra_MpiComm.h>
#endif

#include <EpetraExt_View_MultiVector.h>
#include <Teuchos_BLAS.hpp>

namespace Xyce {
namespace Linear {

//-----------------------------------------------------------------------------
// Function      : MultiVector::MultiVector
// Purpose       : Constructor
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/20/00
//-----------------------------------------------------------------------------
MultiVector::MultiVector(N_PDS_ParMap & map, int numVectors)
:  parallelMap_(rcp(&map,false)),
   overlapMap_(rcp(&map,false)),
   importer_(0),
   exporter_(0),
   viewTransform_(0),
   isOwned_(true)
{
  if (map.numGlobalEntities() < 0)
    N_ERH_ErrorMgr::report(N_ERH_ErrorMgr::DEV_FATAL,
    "MultiVector::MultiVector - vector length too short. Vectors must be > 0 in length.");

  // Create a new Petra MultiVector and set the pointer.
  aMultiVector_ = new Epetra_MultiVector( *map.petraBlockMap(), numVectors );

  oMultiVector_ = aMultiVector_;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::MultiVector
// Purpose       : Constructor
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 06/06/02
//-----------------------------------------------------------------------------
MultiVector::MultiVector(
  N_PDS_ParMap &        map,
  N_PDS_ParMap &        ol_map,
  int                   numVectors )
  : parallelMap_(rcp(&map,false)),
    overlapMap_(rcp(&ol_map,false)),
    importer_(0),
    exporter_(0),
    viewTransform_(0),
    isOwned_(true)
{
  if (map.numGlobalEntities() < 0)
    N_ERH_ErrorMgr::report(N_ERH_ErrorMgr::DEV_FATAL,
                           "MultiVector::MultiVector - vector length too short. Vectors must be > 0 in length.");

  // Create a new Petra MultiVector and set the pointer.
  oMultiVector_ = new Epetra_MultiVector(*ol_map.petraBlockMap(), numVectors);

  viewTransform_ = new EpetraExt::MultiVector_View(
    *overlapMap_->petraBlockMap(), *parallelMap_->petraBlockMap());
  aMultiVector_ = &((*viewTransform_)(*oMultiVector_));
  if (map.pdsComm().numProc() > 1)
  {
    exporter_ = new Epetra_Export( *overlapMap_->petraBlockMap(),
                                   *parallelMap_->petraBlockMap() );
  }

  importer_ = new Epetra_Import( *overlapMap_->petraBlockMap(), *parallelMap_->petraBlockMap() );
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::MultiVector
// Purpose       : Copy Constructor
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/20/00
//-----------------------------------------------------------------------------
MultiVector::MultiVector( const MultiVector & right )
: parallelMap_( right.parallelMap_ ),
  overlapMap_( right.overlapMap_ ),
  oMultiVector_( new Epetra_MultiVector( *(right.oMultiVector_) ) ),
  importer_(0),
  exporter_(0),
  viewTransform_(0),
  isOwned_(true)
{
  if (right.aMultiVector_ == right.oMultiVector_)
    aMultiVector_ = oMultiVector_;
  else
  {
    viewTransform_ = new EpetraExt::MultiVector_View( *overlapMap_->petraBlockMap(),
                                                      *parallelMap_->petraBlockMap() );
    aMultiVector_ = &((*viewTransform_)( *oMultiVector_ ));
  }

  // Generate new exporter instead of using copy constructor, there is an issue with Epetra_MpiDistributor
  if( right.exporter_ ) 
  {
    exporter_ = new Epetra_Export( *overlapMap_->petraBlockMap(),
                                   *parallelMap_->petraBlockMap() );
  }
  if( right.importer_ ) importer_ = new Epetra_Import( *right.importer_ );
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::MultiVector
// Purpose       : Constructor
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 04/09/03
//-----------------------------------------------------------------------------
MultiVector::MultiVector( Epetra_MultiVector * overlapMV, Epetra_Map& parMap, bool isOwned )
: oMultiVector_( overlapMV ),
  importer_(0),
  exporter_(0),
  viewTransform_(0),
  isOwned_(isOwned)
{
  // Make sure there is anything to communicate before creating a transform, importer, or exporter
  if (overlapMV->MyLength() == parMap.NumMyElements())
    aMultiVector_ = oMultiVector_;
  else
  {
    viewTransform_ = new EpetraExt::MultiVector_View( overlapMV->Map(), parMap );
    aMultiVector_ = &((*viewTransform_)(*oMultiVector_));
    if( parMap.Comm().NumProc() > 1 )
      exporter_ = new Epetra_Export( overlapMV->Map(), parMap );

    importer_ = new Epetra_Import( overlapMV->Map(), parMap );
  }
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::MultiVector
// Purpose       : Constructor
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 04/09/03
//-----------------------------------------------------------------------------
MultiVector::MultiVector( Epetra_MultiVector * origMV, bool isOwned )
: aMultiVector_( origMV ),
  oMultiVector_( origMV ),
  importer_(0),
  exporter_(0),
  viewTransform_(0),
  isOwned_(isOwned)
{
#ifdef Xyce_PARALLEL_MPI
  Epetra_Comm& ecomm = const_cast<Epetra_Comm &>( origMV->Comm() );
  Epetra_MpiComm * empicomm = dynamic_cast<Epetra_MpiComm *>( &ecomm );

  if (empicomm)
  {
    pdsComm_= Teuchos::rcp( Xyce::Parallel::createPDSComm( 0, 0, empicomm->Comm() ) );
  }
  else
  { 
    // This is a parallel build executed using one processor.
    pdsComm_ = Teuchos::rcp( Xyce::Parallel::createPDSComm( 0, 0, MPI_COMM_WORLD ) );
  }

#else
  pdsComm_ = Teuchos::rcp( Xyce::Parallel::createPDSComm(0, 0, MPI_COMM_NULL) );
#endif
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::operator=
// Purpose       : assignment
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/20/00
//-----------------------------------------------------------------------------
MultiVector & MultiVector::operator=( const MultiVector & right )
{
  if (this != &right)
  {
    if( (oMultiVector_->Map().NumGlobalElements() == right.oMultiVector_->Map().NumGlobalElements())
        && (oMultiVector_->Map().NumMyElements() == right.oMultiVector_->Map().NumMyElements()) )
    {
      *oMultiVector_ = *right.oMultiVector_;
    }

    if( (aMultiVector_->Map().NumGlobalElements() == right.aMultiVector_->Map().NumGlobalElements())
        && (aMultiVector_->Map().NumMyElements() == right.aMultiVector_->Map().NumMyElements()) )
    {
      *aMultiVector_ = *right.aMultiVector_;
    }
    else
    {
      if (VERBOSE_LINEAR)
        N_ERH_ErrorMgr::report( N_ERH_ErrorMgr::DEV_FATAL_0, "MultiVector being assigned with different Mapping\n");
    }
  }

  return *this;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::~MultiVector
// Purpose       : Default destructor
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/20/00
//-----------------------------------------------------------------------------
MultiVector::~MultiVector()
{
  delete importer_;
  delete exporter_;
  delete viewTransform_; //destroys of aMultiVector_ as well
  if (isOwned_)
  {
    delete oMultiVector_;
  }
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::globalLength
// Purpose       :
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/22/00
//-----------------------------------------------------------------------------
int MultiVector::globalLength() const
{
  return aMultiVector_->GlobalLength();
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::localLength
// Purpose       :
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/22/00
//-----------------------------------------------------------------------------
int MultiVector::localLength() const
{
  return aMultiVector_->MyLength();
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::numVectors
// Purpose       :
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/22/00
//-----------------------------------------------------------------------------
int MultiVector::numVectors() const
{
  return aMultiVector_->NumVectors();
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::dotProduct
// Purpose       : Returns the dot product of "this" vector and another.
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/22/00
//-----------------------------------------------------------------------------
double MultiVector::dotProduct( const MultiVector & y ) const
{
  double result = 0.0;
  int PetraError = aMultiVector_->Dot(*(y.aMultiVector_), &result);

  if (DEBUG_LINEAR)
    processError( "MultiVector::dotProduct - ", PetraError );

  return result;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::scale
// Purpose       : Scales a MultiVector by a constant value.
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/23/00
//-----------------------------------------------------------------------------
void MultiVector::scale(const double a)
{
  int PetraError = aMultiVector_->Scale(a);

  if (DEBUG_LINEAR)
    processError( "MultiVector::scale - ", PetraError);
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::scale
// Purpose       : Scales a MultiVector by a constant value, but
//                 puts it into this.
//                 this = a*x
// Special Notes :
// Scope         : Public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/16/00
//-----------------------------------------------------------------------------
void MultiVector::scale(const double a, const MultiVector &x)
{
  int PetraError = aMultiVector_->Scale(a, *(x.aMultiVector_));

  if (DEBUG_LINEAR)
    processError( "MultiVector::scale - ", PetraError);
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::multiply
// Purpose       : Performs element-wise multiplication of two vectors
//                 this = this @ x
//                 where @ represents element-wise multiplication
// Special Notes :
// Scope         : Public
// Creator       : Roger P. Pawlowski, SNL, Parallel Computational Sciences
// Creation Date : 3/24/03
//-----------------------------------------------------------------------------
void MultiVector::multiply(const MultiVector &x)
{
  int PetraError = aMultiVector_->Multiply(1.0, *aMultiVector_,
					   *(x.aMultiVector_), 0.0);

  if (DEBUG_LINEAR)
    processError( "MultiVector::scale - ", PetraError);
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::daxpy
// Purpose       : Linear combination of two MultiVectors:
//                 this = y + a*x
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/23/00
//-----------------------------------------------------------------------------
void MultiVector::daxpy(const MultiVector &y, const double a,
                              const MultiVector &x)
{
  int PetraError = aMultiVector_->Update(1.0, *(y.aMultiVector_), a,
                                         *(x.aMultiVector_), 0.0);

  if (DEBUG_LINEAR)
    processError( "MultiVector::daxpy - ", PetraError);
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::linearCombo
// Purpose       : Linear combination of two MultiVectors:
//                 this = a*x + b*y
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Computational Sciences
// Creation Date : 01/08/01
//-----------------------------------------------------------------------------
void MultiVector::linearCombo(const double a, const MultiVector &x,
                                    const double b, const MultiVector &y)
{
  int PetraError = aMultiVector_->Update(a, *(x.aMultiVector_), b,
                                         *(y.aMultiVector_), 0.0);

  if (DEBUG_LINEAR)
    processError( "MultiVector::linearCombo - ", PetraError);
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::linearCombo
// Purpose       : Linear combination of three MultiVectors:
//                 this = a*x + b*y + c*z
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Computational Sciences
// Creation Date : 01/08/01
//-----------------------------------------------------------------------------
void MultiVector::linearCombo(const double a, const MultiVector &x,
                                    const double b, const MultiVector &y,
                                    const double c, const MultiVector &z)
{
  update( a, x, b, y, c, z, 0.0);
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::update
// Purpose       :
// Special Notes : ERK. From the epetra documentation:
//
//                 this = s*this + a*A
//
// Scope         : Public
// Creator       : Rob Hoekstra, SNL, Computational Sciences
// Creation Date : 02/04/02
//-----------------------------------------------------------------------------
void MultiVector::update( double a, const MultiVector & A,
                                double s )
{
  aMultiVector_->Update( a, *(A.aMultiVector_), s );
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::update
// Purpose       :
// Special Notes : ERK.  From the epetra documentation:
//
//                 this = s*this + a*A + b*B
//
// Scope         : Public
// Creator       : Rob Hoekstra, SNL, Computational Sciences
// Creation Date : 02/04/02
//-----------------------------------------------------------------------------
void MultiVector::update( double a, const MultiVector & A,
                                double b, const MultiVector & B,
                                double s )
{
  aMultiVector_->Update( a, *(A.aMultiVector_),
                         b, *(B.aMultiVector_),
                         s );
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::update
// Purpose       :
// Special Notes :
// Scope         : Public
// Creator       : Rob Hoekstra, SNL, Computational Sciences
// Creation Date : 02/04/02
//-----------------------------------------------------------------------------
void MultiVector::update( double a, const MultiVector & A,
                                double b, const MultiVector & B,
                                double c, const MultiVector & C,
                                double s )
{
  aMultiVector_->Update( a, *(A.aMultiVector_),
                         b, *(B.aMultiVector_),
                         s );
  aMultiVector_->Update( c, *(C.aMultiVector_),
                         1.0 );
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::addVec
// Purpose       : Add multiple of a MultiVector:  this = this + a*y
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/23/00
//-----------------------------------------------------------------------------
void MultiVector::addVec(const double a, const MultiVector &y)
{
  int PetraError = aMultiVector_->Update(a, *(y.aMultiVector_), 1.0);

  if (DEBUG_LINEAR)
    processError( "MultiVector::addVec - ", PetraError);
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::gemm
// Purpose       : This routine performs a variety of MM operations.
//                 Variations are due to the fact that A, B and C can be local
//                 replicated or global distributed multi_vectors and that we
//                 may or may not operate with the transpose of A and B.
// Special Notes : See Petra documentation for possible cases.
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/23/00
//-----------------------------------------------------------------------------
void MultiVector::GEMM(const bool transA, const bool transB,
                             const double alpha, const double beta,
                             const MultiVector &A,
                             const MultiVector &B)
{

  char tA, tB;

  transA ? (tA='N') : (tA='T');
  transB ? (tB='N') : (tB='T');

  int PetraError = aMultiVector_->Multiply(tA, tB, alpha, *(A.aMultiVector_), *(B.aMultiVector_), beta);

  if (DEBUG_LINEAR)
  {
    static const char *methodMsg = "MultiVector::GEMM - ";

    Xyce::dout() << "sizes of this:  numvectors = ";
    Xyce::dout() << aMultiVector_->NumVectors ();
    Xyce::dout() << "  mylength = ";
    Xyce::dout() << aMultiVector_->MyLength ();
    Xyce::dout() << std::endl;


    Xyce::dout() << "sizes of A:  numvectors = ";
    Xyce::dout() << A.aMultiVector_->NumVectors ();

    Xyce::dout() << "  mylength = ";
    Xyce::dout() << A.aMultiVector_->MyLength ();
    Xyce::dout() << std::endl;

    Xyce::dout() << "sizes of B:  numvectors = ";
    Xyce::dout() << B.aMultiVector_->NumVectors();

    Xyce::dout() << "  mylength = ";
    Xyce::dout() << B.aMultiVector_->MyLength ();
    Xyce::dout() << std::endl;

    Xyce::dout() << "PetraError = " << PetraError << std::endl;
    processError(methodMsg, PetraError);

  }
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::lpNorm
// Purpose       : Returns lp norms of each vector in MultiVector
// Special Notes : Only p=1 and p=2 implemented now since this is all Petra
//                 supports.
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/23/00
//-----------------------------------------------------------------------------
int MultiVector::lpNorm(const int p, double * result) const
{
  int PetraError = -1;
  static const char *methodMsg = "MultiVector::lpNorm - ";

  if (p == 1)
    PetraError = aMultiVector_->Norm1(result);
  else if (p == 2)
    PetraError = aMultiVector_->Norm2(result);
  else
    Xyce::Report::DevelFatal0().in(methodMsg) << "Requested norm is not supported";

  if (DEBUG_LINEAR)
    processError(methodMsg, PetraError);

  return PetraError;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::infNorm
// Purpose       : Returns infinity norm of each vector in MultiVector
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Computational Sciences
// Creation Date : 01/16/01
//-----------------------------------------------------------------------------
int MultiVector::infNorm(double * result) const
{
  static const char *methodMsg = "MultiVector::infNorm - ";
  int PetraError = aMultiVector_->NormInf(result);

  if (DEBUG_LINEAR) 
    processError(methodMsg, PetraError);

  return PetraError;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::infNormIndex
// Purpose       : Returns index of the maximum absolute entry in MultiVector
// Special Notes :
// Scope         : Public
// Creator       : Heidi K. Thornquist, SNL, Electrical Systems Modeling
// Creation Date : 11/21/11
//-----------------------------------------------------------------------------
int MultiVector::infNormIndex(int * index) const
{
  Teuchos::BLAS<int,double> blas;

  int numProcs = aMultiVector_->Comm().NumProc();
  int numVectors = aMultiVector_->NumVectors();
  int myLength = aMultiVector_->MyLength();
  std::vector<int> indexTemp( numVectors, 0 ), indexTempAll( numVectors*numProcs, 0 );
  std::vector<double> doubleTemp( numVectors, 0.0 ), doubleTempAll( numVectors*numProcs, 0.0 );
  double ** pointers = aMultiVector_->Pointers();

  for (int i=0; i < numVectors; i++)
  {
    // Remember that IAMAX returns 1-based indexing, so subtract 1 to get the actual index.
    int jj = blas.IAMAX(myLength, pointers[i], 1) - 1;
    if (jj>-1)
    {
      indexTemp[i] = aMultiVector_->Map().GID(jj);
      doubleTemp[i] = std::abs(pointers[i][jj]);
    }
  }

  // Use the Epetra communicator to gather all the local maximum values and indices
  int result = aMultiVector_->Comm().GatherAll(&indexTemp[0], &indexTempAll[0], numVectors);
  result += aMultiVector_->Comm().GatherAll(&doubleTemp[0], &doubleTempAll[0], numVectors);

  // Compute the global infNorm and index
  for (int i=0; i < numVectors; i++)
  {
    // Remember that IAMAX returns 1-based indexing, so subtract 1 to get the actual index.
    int ii = blas.IAMAX( numProcs, &doubleTempAll[i], numVectors ) - 1;
    index[i] = indexTempAll[ii*numVectors + i];
  }

  return result;
}


//-----------------------------------------------------------------------------
// Function      : MultiVector::wRMSNorm
// Purpose       : Returns weighted root-mean-square of each vector in
//                 MultiVector
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 10/12/00
//-----------------------------------------------------------------------------
int MultiVector::wRMSNorm(const MultiVector & weights, double * result) const
{
  int PetraError = aMultiVector_->NormWeighted( *(weights.aMultiVector_), result );

  if (DEBUG_LINEAR)
    processError( "MultiVector::wRMSNorm - ", PetraError);

  return PetraError;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::wMaxNorm
// Purpose       : Returns the weighted
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Computational Sciences
// Creation Date : 03/19/01
//-----------------------------------------------------------------------------
int MultiVector::wMaxNorm(MultiVector & weights,
                                MultiVector & tmpVector,
                                double * result) const
{
  int length  = aMultiVector_->MyLength();
  int numVecs = numVectors();

  for (int i = 0;  i < numVecs; ++i)
    for (int j = 0; j < length; ++j)
      tmpVector[i][j] = fabs((*(this))[i][j]) / weights[i][j];

  int PetraError = tmpVector.aMultiVector_->MaxValue(result);

  return PetraError;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::minValue
// Purpose       : Return the minimum value for the multivector
// Special Notes :
// Scope         : Public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/16/00
//-----------------------------------------------------------------------------
int MultiVector::minValue(double * result) const
{
  int PetraError = aMultiVector_->MinValue(result);

  return PetraError;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::maxValue
// Purpose       : Return the maximum value for the multivector
// Special Notes :
// Scope         : Public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/16/00
//-----------------------------------------------------------------------------
int MultiVector::maxValue(double * result) const
{
  int PetraError = aMultiVector_->MaxValue(result);

  return PetraError;
}


//-----------------------------------------------------------------------------
// Function      : MultiVector::reduce
// Purpose       : Takes a replicated local multi_vector and performs a global
//                 reduction on each entry.
// Special Notes : No_op with new Petra.
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/23/00
//-----------------------------------------------------------------------------
void MultiVector::reduce()
{
  N_ERH_ErrorMgr::report( N_ERH_ErrorMgr::DEV_FATAL,
    "MultiVector::reduce Not available in current PETRA!\n" );
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::random
// Purpose       : Generates random numbers drawn from a uniform distribution
//                 on the interval (-1,1) using a multiplicative congruential
//                 generator with modulus 2^31 - 1.
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/23/00
//-----------------------------------------------------------------------------
void MultiVector::random()
{
  int PetraError = aMultiVector_->Random();

  if (DEBUG_LINEAR)
    processError( "MultiVector::random - ", PetraError);
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::putScalar
// Purpose       : Fills MultiVector with the value "scalar".
// Special Notes :
// Scope         : Public
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/23/00
//-----------------------------------------------------------------------------
void MultiVector::putScalar(const double scalar)
{
  int PetraError = oMultiVector_->PutScalar(scalar);

  if (DEBUG_LINEAR)
    processError( "MultiVector::putScalar - ", PetraError);
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::addScalar
// Purpose       : Adds to MultiVector with the value "scalar".
// Special Notes :
// Scope         : Public
// Creator       : Robert J. Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 01/18/01
//-----------------------------------------------------------------------------
void MultiVector::addScalar(const double scalar)
{
  int length  = aMultiVector_->MyLength();
  int numVecs = numVectors();

  for (int i = 0; i < numVecs; ++i)
    for (int j = 0; j < length; ++j)
      (*aMultiVector_)[i][j] += scalar;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::absValue
// Purpose       : Abs value of elements of MultiVector
// Special Notes :
// Scope         : Public
// Creator       : Robert J. Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 01/18/01
//-----------------------------------------------------------------------------
void MultiVector::absValue(const MultiVector & A)
{
  int PetraError = oMultiVector_->Abs(*(A.oMultiVector_));

  if (DEBUG_LINEAR)
    processError( "MultiVector::absValue - ", PetraError);
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::reciprocal
// Purpose       : Reciprocal elements of MultiVector
// Special Notes :
// Scope         : Public
// Creator       : Robert J. Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 02/07/01
//-----------------------------------------------------------------------------
void MultiVector::reciprocal(const MultiVector & A)
{
  int PetraError = oMultiVector_->Reciprocal(*(A.oMultiVector_));

  if (DEBUG_LINEAR)
    processError( "MultiVector::reciprocal - ", PetraError);
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::getVectorView
// Purpose       : Const view of individual vector in MultiVector
// Special Notes :
// Scope         : Public
// Creator       : Todd Coffey, 1414
// Creation Date : 9/11/08
//-----------------------------------------------------------------------------
RCP<const Vector> MultiVector::getVectorView(int index) const
{
  RCP<const Vector> vec = rcp(new Vector((*oMultiVector_)(index),false), true);
  return vec;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::getNonConstVectorView
// Purpose       : NonConst view of individual vector in MultiVector
// Special Notes :
// Scope         : Public
// Creator       : Todd Coffey, 1414
// Creation Date : 9/11/08
//-----------------------------------------------------------------------------
RCP<Vector> MultiVector::getNonConstVectorView(int index)
{
  RCP<Vector> vec = rcp(new Vector((*oMultiVector_)(index),false), true);
  return vec;
}

// //-----------------------------------------------------------------------------
// // Function      : MultiVector::operator []
// // Purpose       : "[]" operator for MultiVector.
// // Special Notes :
// // Scope         : Public
// // Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// // Creation Date : 05/23/00
// //-----------------------------------------------------------------------------
// double *& MultiVector::operator[](int index)
// {
//   return (*oMultiVector_)[index];
// }

// //-----------------------------------------------------------------------------
// // Function      : MultiVector::operator []
// // Purpose       : "[]" operator for MultiVector.
// // Special Notes : This version returns a "const double".
// // Scope         : Public
// // Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// // Creation Date : 05/23/00
// //-----------------------------------------------------------------------------
// const double *& MultiVector::operator[](int index) const
// {
//   return const_cast <const double *&> ((*oMultiVector_)[index]);
// }

//-----------------------------------------------------------------------------
// Function      : MultiVector::fillComplete
// Purpose       :
// Special Notes :
// Scope         : Public
// Creator       : Robert J Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 08/29/03
//-----------------------------------------------------------------------------
void MultiVector::fillComplete()
{
  if ( exporter_ )
  {
    aMultiVector_->Export( *oMultiVector_, *exporter_, Add );
  }
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::vectorImport
// Purpose       : Import using Petra_Import object
// Special Notes :
// Scope         : Public
// Creator       : Robert J Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 02/09/01
//-----------------------------------------------------------------------------
bool MultiVector::vectorImport(MultiVector * vec,
                                     Epetra_Import * importer)
{
  aMultiVector_->Import(*(vec->aMultiVector_), *importer, Insert);
  return true;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::vectorExport
// Purpose       : Export using Petra_Import object
// Special Notes : Reverse of import function
// Scope         : Public
// Creator       : Robert J Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 02/09/01
//-----------------------------------------------------------------------------
bool MultiVector::vectorExport(MultiVector * vec,
                                     Epetra_Import * importer)
{
  aMultiVector_->Export(*(vec->aMultiVector_), *importer, Insert);
  return true;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::importOverlap
// Purpose       :
// Special Notes :
// Scope         : Public
// Creator       : Robert J Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 06/06/02
//-----------------------------------------------------------------------------
bool MultiVector::importOverlap()
{
  bool flag = false;

  if( importer_ )
    flag = oMultiVector_->Import( *aMultiVector_, *importer_, Insert );

  return flag;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::exportContribution
// Purpose       :
// Special Notes :
// Scope         : Public
// Creator       : Robert J Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 06/06/02
//-----------------------------------------------------------------------------
bool MultiVector::exportContribution( bool do_sum )
{
  if( exporter_ )
    if( do_sum )
      return aMultiVector_->Export( *oMultiVector_, *exporter_, Add );
    else
      return aMultiVector_->Export( *oMultiVector_, *exporter_, Zero );
  else
    return false;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::writeToFile
// Purpose       : Dumps out the multivector entries to a file.
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 06/19/00
//-----------------------------------------------------------------------------
void MultiVector::writeToFile(const char *filename, bool useLIDs )
{
  int numProcs = aMultiVector_->Comm().NumProc();
  int localRank = aMultiVector_->Comm().MyPID();
  int masterRank = 0;

  for( int p = 0; p < numProcs; ++p )
  {
    //A barrier inside the loop so each processor waits its turn.
    aMultiVector_->Comm().Barrier();

    if(p == localRank)
    {
      FILE *file = NULL;

      if(masterRank == localRank)
      {
        //This is the master processor, open a new file.
        file = fopen(filename,"w");

        //Write the RDP_MultiVector dimension n into the file.
        fprintf(file,"%d\n",globalLength());
      }
      else
      {
        //This is not the master proc, open file for appending
        file = fopen(filename,"a");
      }

      //Now loop over the local portion of the RDP_MultiVector.
      int length  = localLength();
      int numVecs = numVectors();

      for (int i = 0; i < numVecs; ++i)
        for (int j = 0; j < length; ++j)
        {
          int loc = aMultiVector_->Map().GID(j);
          if( useLIDs ) loc = j;
          fprintf(file,"%d %d %20.13e\n",i,loc,(*aMultiVector_)[i][j]);
        }
      fclose(file);
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::readFromFile
// Purpose       : Reads in the multivector entries from a file.
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 08/29/03
//-----------------------------------------------------------------------------
void MultiVector::readFromFile( const char * filename )
{
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::processError
// Purpose       : Concrete implementation which processes Petra (in this case)
//                 error codes taken from the Petra member function returns.
// Special Notes : Petra specific.  NOTE ALSO - this function is currently
//                 within the "Xyce_DEBUG_LINEAR" ifdef and so any calls to
//                 this should also be so bracketed.
// Scope         : Private
// Creator       : Scott A. Hutchinson, SNL, Parallel Computational Sciences
// Creation Date : 05/22/00
//-----------------------------------------------------------------------------
void MultiVector::processError(const char *methodMsg, int error) const
{

  const std::string PetraOK();
  const std::string PetraError();

  // Process the error
  switch (error)
  {
  case 0:
    Xyce::dout() << methodMsg << ": Function returned without warnings or errors." << std::endl;
    break;

  default:
    Xyce::Report::DevelFatal0().in(methodMsg) << "Function returned with an error.";
  }
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::getElementByGlobalIndex
// Purpose       : Get element from vector using global index.
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/6/00
//-----------------------------------------------------------------------------
const double & MultiVector::getElementByGlobalIndex(
  const int & global_index, const int & vec_index) const
{
  if( aMultiVector_ != oMultiVector_ )
    return (*oMultiVector_)[vec_index][overlapMap_->globalToLocalIndex(global_index)];
  else if( parallelMap_ == Teuchos::null )
    return (*aMultiVector_)[vec_index][ aMultiVector_->Map().LID(global_index) ];
  else
  {
    int i = parallelMap_->globalToLocalIndex(global_index);

    if (i != -1)
      return ((*aMultiVector_)[vec_index])[i];
    else {
      std::map<int,double>::const_iterator it = externVectorMap_.find(global_index);
      if (it != externVectorMap_.end())
        return (*it).second;
      else
      {
        char message[128];
        sprintf(message, "getElementByGlobalIndex: failed to find MultiVector "
                "global index. global_index = %d", global_index);
        std::string msg(message);

        N_ERH_ErrorMgr::report(N_ERH_ErrorMgr::DEV_FATAL, msg);
        return (*externVectorMap_.find(-1)).second;
      }
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::setElementByGlobalIndex
// Purpose       : Set element from vector using global index.
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/6/00
//-----------------------------------------------------------------------------
bool MultiVector::setElementByGlobalIndex(const int & global_index,
                                                const double & val,
                                                const int & vec_index)
{
  if( aMultiVector_ != oMultiVector_ )
    (*oMultiVector_)[vec_index][overlapMap_->globalToLocalIndex(global_index)] = val;
  else if( parallelMap_ == Teuchos::null )
    (*oMultiVector_)[vec_index][ oMultiVector_->Map().LID(global_index) ] = val;
  else
  {
    if (global_index != -1)
    {
      int i = parallelMap_->globalToLocalIndex(global_index);

      if (i != -1)
      {
        ( (*aMultiVector_)[vec_index] )[i] = val;
        return true;
      }
      else
      {
        Xyce::Report::DevelFatal().in("setElementByGlobalIndex") << "Failed to find MultiVector global index: " << global_index;
        return false;
      }
    }
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::sumElementByGlobalIndex
// Purpose       : Set element from vector using global index.
// Special Notes :
// Scope         : Public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/7/00
//-----------------------------------------------------------------------------
bool MultiVector::sumElementByGlobalIndex(const int & global_index,
                                                const double & val,
                                                const int & vec_index)
{
  if( aMultiVector_ != oMultiVector_ )
    (*oMultiVector_)[vec_index][overlapMap_->globalToLocalIndex(global_index)] += val;
  else if( parallelMap_ == Teuchos::null )
    (*oMultiVector_)[vec_index][ oMultiVector_->Map().LID(global_index) ] += val;
  else
  {
    if (global_index != -1 )
    {
      int i = parallelMap_->globalToLocalIndex(global_index);

      if (i != -1)
      {
        ( (*aMultiVector_)[vec_index] )[i] += val;
        return true;
      }
      else
      {
        N_ERH_ErrorMgr::report(N_ERH_ErrorMgr::DEV_FATAL,
                             std::string(" sumElementByGlobalIndex: failed to find "
                                    "MultiVector global index "));
        return false;
      }
    }
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : epetraVector
// Purpose       :
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 02/19/03
//-----------------------------------------------------------------------------
Epetra_Vector * MultiVector::epetraVector( int index ) const
{
  return new Epetra_Vector( View, *aMultiVector_, index );
}

//-----------------------------------------------------------------------------
// Function      : printPetraObject
// Purpose       :
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 07/14/00
//-----------------------------------------------------------------------------
void MultiVector::printPetraObject(std::ostream &os) const
{
  if (aMultiVector_ != oMultiVector_)
  {
    os << *aMultiVector_;
  }
  os << *oMultiVector_;

  if (VERBOSE_LINEAR)
  {
    std::map<int, double>::const_iterator it_idM = externVectorMap_.begin();
    std::map<int, double>::const_iterator end_idM = externVectorMap_.end();
    if (it_idM != end_idM) os << "<Extern Vector Map>" << std::endl;
    for (; it_idM != end_idM; ++it_idM)
    {
      os << "  " << it_idM->first << "\t" << it_idM->second << std::endl;
    }
    os << std::endl;
  }
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::packedByteCount
// Purpose       :
// Special Notes :
// Scope         : Public
// Creator       : Robert J. Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 7/31/01
//-----------------------------------------------------------------------------
int MultiVector::packedByteCount() const
{
  int count = 0;
  count += sizeof(int); //# vectors
  count += sizeof(int); //local size
  count += numVectors() * aMultiVector_->MyLength() * sizeof(double); //local vectors

  count += sizeof(int); //extern map size
  count += externVectorMap_.size() * (sizeof(int)+sizeof(double));

  return count;
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::pack
// Purpose       :
// Special Notes :
// Scope         : Public
// Creator       : Robert J. Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 7/31/01
//-----------------------------------------------------------------------------
void MultiVector::pack( char * buf, int bsize, int & pos,
			N_PDS_Comm * comm ) const
{
  int nVs = numVectors();
  comm->pack( &nVs, 1, buf, bsize, pos );
  int locLth = aMultiVector_->MyLength();
  comm->pack( &locLth, 1, buf, bsize, pos );

  double * A;
  int stride;
  aMultiVector_->ExtractView( &A, &stride );

  comm->pack( A, locLth * nVs, buf, bsize, pos );

  int size = externVectorMap_.size();
  comm->pack( &size, 1, buf, bsize, pos );

  int loc;
  double val;
  std::map<int,double>::const_iterator iterXM = externVectorMap_.begin();
  std::map<int,double>::const_iterator endXM = externVectorMap_.end ();
  for(; iterXM != endXM; ++iterXM )
  {
    loc = iterXM->first;
    val = iterXM->second;
    comm->pack( &loc, 1, buf, bsize, pos );
    comm->pack( &val, 1, buf, bsize, pos );
  }
}

//-----------------------------------------------------------------------------
// Function      : MultiVector::unpack
// Purpose       :
// Special Notes :
// Scope         : Public
// Creator       : Robert J. Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 7/31/01
//-----------------------------------------------------------------------------
void MultiVector::unpack( char * buf, int bsize, int & pos,
			N_PDS_Comm * comm )
{
  int numV, locL;
  comm->unpack( buf, bsize, pos, &numV, 1 );
  comm->unpack( buf, bsize, pos, &locL, 1 );

  if( (numV != numVectors()) || (locL != aMultiVector_->MyLength()) )
    N_ERH_ErrorMgr::report( N_ERH_ErrorMgr::DEV_FATAL,
	"UNPACKING VECTOR FAILED!\n" );

  std::vector<double> tmpVec( numV * locL );
  comm->unpack( buf, bsize, pos, &(tmpVec[0]), locL * numV );

  int loc = 0;
  for( int i = 0; i < numV; ++i )
    for( int j = 0; j < locL; ++j )
      (*aMultiVector_)[i][j] = tmpVec[loc++];

  int size;
  double val;
  comm->unpack( buf, bsize, pos, &size, 1 );
  externVectorMap_.clear();
  for( int i = 0; i < size; ++i )
  {
    comm->unpack( buf, bsize, pos, &loc, 1 );
    comm->unpack( buf, bsize, pos, &val, 1 );
    externVectorMap_[loc] = val;
  }
}

} // namespace Linear
} // namespace Xyce
