//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_LAS_System.C,v $
//
// Purpose        : Container class for linear system
//                  Jacobian, RHS, Soln Vec, Error Vec, and Creators
//                  N_LAS_QueryUtil and N_PDS_ParMap are registered to
//                  minimize input for matrix and vector creation
//
// Special Notes  :
//
// Creator        : Robert Hoekstra, SNL, Parallel Computational Sciences
//
// Creation Date  : 06/10/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.87.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:15 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

#include <iostream>

#include <N_ANP_AnalysisManager.h>
#include <N_LAS_Builder.h>
#include <N_LAS_Matrix.h>
#include <N_LAS_MultiVector.h>
#include <N_LAS_QueryUtil.h>
#include <N_LAS_System.h>
#include <N_LAS_Vector.h>
#include <N_PDS_GlobalAccessor.h>
#include <N_PDS_ParMap.h>
#include <N_UTL_FeatureTest.h>

#include <Epetra_MapColoring.h>

namespace Xyce {
namespace Linear {

//-----------------------------------------------------------------------------
// Function      : System::System
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : David G. Baur  Raytheon  Sandia National Laboratories 1355
// Creation Date : Mon Oct  6 07:33:03 2014
//-----------------------------------------------------------------------------
System::System()
  : pdsMgr_(0),
    lasQueryUtil_(0),
    lasBuilder_(0),
    jacobianMatrixPtr_(0),
    rhsVectorPtr_(0),
    jdxpVectorPtr_(0),
    fVectorPtr_(0),
    daeQVectorPtr_(0),
    daeFVectorPtr_(0),
    daeBVectorPtr_(0),
    dFdxdVpVectorPtr_(0),
    dQdxdVpVectorPtr_(0),
    dQdxMatrixPtr_(0),
    dFdxMatrixPtr_(0),
    currSolVectorPtrPtr_(0),
    nextSolVectorPtrPtr_(0),
    lastSolVectorPtrPtr_(0),
    flagSolVectorPtrPtr_(0),
    nextSolDerivVectorPtrPtr_(0),
    nextStaDerivVectorPtrPtr_(0),
    currStaVectorPtrPtr_(0),
    nextStaVectorPtrPtr_(0),
    lastStaVectorPtrPtr_(0),
    currStoVectorPtrPtr_(0),
    nextStoVectorPtrPtr_(0),
    lastStoVectorPtrPtr_(0),
    tmpStaVectorPtrPtr_(0),
    tmpSolVectorPtrPtr_(0),
    tmpStaDerivVectorPtrPtr_(0),
    tmpSolDerivVectorPtrPtr_(0),
    tmpStaDivDiffVectorPtrPtr_(0),
    tmpSolDivDiffVectorPtrPtr_(0) ,
    deviceMaskVectorPtrPtr_(0) ,
    solnColoringPtr_(0),
    initialConditionColoringPtr_(0),
    dxVoltlimVectorPtr_(0),
    jdx2VectorPtr_(0),
    jacTestMatrixPtr_(0),
    Fdx2VectorPtr_(0),
    dFdxTestMatrixPtr_(0),
    Qdx2VectorPtr_(0),
    dQdxTestMatrixPtr_(0)
  {}

//-----------------------------------------------------------------------------
// Function      : System::~System
// Purpose       : destructor
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/10/00
//-----------------------------------------------------------------------------
System::~System()
{
  delete jacobianMatrixPtr_;
  delete jdxpVectorPtr_;
  delete rhsVectorPtr_;
  delete fVectorPtr_;
  delete solnColoringPtr_;
  delete initialConditionColoringPtr_;
  delete dxVoltlimVectorPtr_;
  // old DAE:
  delete jdx2VectorPtr_;
  delete jacTestMatrixPtr_;
  // new DAE:
  delete Fdx2VectorPtr_;
  delete dFdxTestMatrixPtr_;
  delete Qdx2VectorPtr_;
  delete dQdxTestMatrixPtr_;
  delete dFdxdVpVectorPtr_;
  delete dQdxdVpVectorPtr_;
}

//-----------------------------------------------------------------------------
// Function      : System::getGlobalSolutionSize
// Purpose       : returns size of global solution vector
// Special Notes :
// Scope         : Public
// Creator       : Eric R. Keiter, SNL
// Creation Date : 2/14/06
//-----------------------------------------------------------------------------
int System::getGlobalSolutionSize()
{
  Vector * tmpVec = lasBuilder_->createVector();
  int size = tmpVec->globalLength();
  delete tmpVec;
  return size;
}

//-----------------------------------------------------------------------------
// Function      : System::getGlobalStateSize
// Purpose       : returns size of global solution vector
// Special Notes :
// Scope         : Public
// Creator       : Eric R. Keiter, SNL
// Creation Date : 5/5/10
//-----------------------------------------------------------------------------
int System::getGlobalStateSize()
{
  Vector * tmpVec = lasBuilder_->createStateVector();
  int size = tmpVec->globalLength();
  delete tmpVec;
  return size;
}

//-----------------------------------------------------------------------------
// Function      : System::getSolutionSize
// Purpose       : returns size of local solution vector
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/10/00
//-----------------------------------------------------------------------------
int System::getSolutionSize()
{
  Vector * tmpVec = lasBuilder_->createVector();
  int size = tmpVec->localLength();
  delete tmpVec;
  return size;
}

//-----------------------------------------------------------------------------
// Function      : System::getStateSize
// Purpose       : returns size of local state vector
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/10/00
//-----------------------------------------------------------------------------
int System::getStateSize()
{
  Vector * tmpVec = lasBuilder_->createStateVector();
  int size = tmpVec->localLength();
  delete tmpVec;
  return size;
}

//-----------------------------------------------------------------------------
// Function      : System::getRHSSize
// Purpose       : returns size of local RHS vector
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/10/00
//-----------------------------------------------------------------------------
int System::getRHSSize()
{
  Vector * tmpVec = lasBuilder_->createVector();
  int size = tmpVec->localLength();
  delete tmpVec;
  return size;
}

//-----------------------------------------------------------------------------
// Function      : System::initializeSystem
// Purpose       :
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/10/00
//-----------------------------------------------------------------------------
bool System::initializeSystem() // bool block_analysis_flag)
{
  bool bSuccess = true;

  // if (!block_analysis_flag) {
  //   registerSolnColoring( lasBuilder_->createSolnColoring() );
  //   registerICColoring( lasBuilder_->createInitialConditionColoring() );
  // }

  bSuccess = bSuccess && registerRHSVector( lasBuilder_->createVector() );
  bSuccess = bSuccess && registerJDXPVector( lasBuilder_->createVector() );
  bSuccess = bSuccess && registerFVector( lasBuilder_->createVector() );
  if (DEBUG_VOLTLIM)
  {
    bSuccess = bSuccess && registerDxVoltlimVector( lasBuilder_->createVector() );

    // old DAE:
    bSuccess = bSuccess && registerJDX2Vector( lasBuilder_->createVector() );
    bSuccess = bSuccess && registerJacTestMatrix( lasBuilder_->createMatrix() );

    // new DAE:
    bSuccess = bSuccess && registerFDX2Vector( lasBuilder_->createVector() );
    bSuccess = bSuccess && registerdFdxTestMatrix( lasBuilder_->createMatrix() );
    bSuccess = bSuccess && registerQDX2Vector( lasBuilder_->createVector() );
    bSuccess = bSuccess && registerdQdxTestMatrix( lasBuilder_->createMatrix() );
  }

  // these are needed for new-DAE:
  registerdFdxdVpVector ( lasBuilder_->createVector() );
  registerdQdxdVpVector ( lasBuilder_->createVector() );

  return ( bSuccess && registerJacobianMatrix( lasBuilder_->createMatrix() ) );
}

//-----------------------------------------------------------------------------
// Function      : System::updateExternValsSolnVector
// Purpose       : updates off proc values of soln vectors
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/18/00
//-----------------------------------------------------------------------------
bool System::updateExternValsSolnVector(MultiVector * solnVector)
{
#ifdef Xyce_PARALLEL_MPI
  N_PDS_GlobalAccessor * Accessor = pdsMgr_->getGlobalAccessor( Parallel::SOLUTION );

  Accessor->migrateMultiVector(solnVector);
#endif

  return true;
}

//-----------------------------------------------------------------------------
// Function      : System::updateExternValsStateVector
// Purpose       : updates off proc values of state vectors
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/18/00
//-----------------------------------------------------------------------------
bool System::updateExternValsStateVector(MultiVector * stateVector)
{
#ifdef Xyce_PARALLEL_MPI
  N_PDS_GlobalAccessor * Accessor = pdsMgr_->getGlobalAccessor( Parallel::STATE );

  Accessor->migrateMultiVector(stateVector);
#endif

  return true;
}

//-----------------------------------------------------------------------------
// Function      : System::updateExternValsStoreVector
// Purpose       : updates off proc values of store vectors
// Special Notes :
// Scope         : Public
// Creator       : Eric Keiter
// Creation Date :
//-----------------------------------------------------------------------------
bool System::updateExternValsStoreVector(MultiVector * storeVector)
{
#ifdef Xyce_PARALLEL_MPI
  N_PDS_GlobalAccessor * Accessor = pdsMgr_->getGlobalAccessor( Parallel::STORE );

  Accessor->migrateMultiVector(storeVector);
#endif

  return true;
}

//-----------------------------------------------------------------------------
// Function      : System::localInMatrix
// Purpose       : tests if row is local for jacobian
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/18/00
//-----------------------------------------------------------------------------
bool System::localInMatrix(const int & row) const
{
  return localInSolnVector( row );
}

//-----------------------------------------------------------------------------
// Function      : System::localInSolnVector
// Purpose       : tests if row is local for Soln Vector
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/18/00
//-----------------------------------------------------------------------------
bool System::localInSolnVector(const int & row) const
{
  return count(lasQueryUtil_->rowList_GID().begin(),
               lasQueryUtil_->rowList_GID().end(),
               row );
}

//-----------------------------------------------------------------------------
// Function      : System::localInStateVector
// Purpose       : tests if row is local for State Vector
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 6/18/00
//-----------------------------------------------------------------------------
bool System::localInStateVector(const int & row) const
{
  return count(lasQueryUtil_->rowList_StateGID().begin(),
               lasQueryUtil_->rowList_StateGID().end(),
               row );
}

//-----------------------------------------------------------------------------
// Function      : System::debug
// Purpose       : debug output
// Special Notes :
// Scope         : Public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 11/8/00
//-----------------------------------------------------------------------------
void System::debug() const
{
  Xyce::dout() << "Linear System Debug Output" << std::endl;
  Xyce::dout() << "--------------------------" << std::endl;

  Xyce::dout() << "RHS Vector:" << std::endl;
  rhsVectorPtr_->printPetraObject(Xyce::dout());

  Xyce::dout() << "Current Solution Vector:" << std::endl;
  (*currSolVectorPtrPtr_)->printPetraObject(Xyce::dout());

  Xyce::dout() << "Next Solution Vector:" << std::endl;
  (*nextSolVectorPtrPtr_)->printPetraObject(Xyce::dout());

  Xyce::dout() << "Next Solution Derivative Vector:" << std::endl;
  (*nextSolDerivVectorPtrPtr_)->printPetraObject(Xyce::dout());

  Xyce::dout() << "Current State Vector:" << std::endl;
  (*currStaVectorPtrPtr_)->printPetraObject(Xyce::dout());

  Xyce::dout() << "Next State Vector:" << std::endl;
  (*nextStaVectorPtrPtr_)->printPetraObject(Xyce::dout());

  Xyce::dout() << "Temp Solution Vector:" << std::endl;
  (*tmpSolVectorPtrPtr_)->printPetraObject(Xyce::dout());

  Xyce::dout() << "Temp Solution Derivative Vector:" << std::endl;
  (*tmpSolDerivVectorPtrPtr_)->printPetraObject(Xyce::dout());

  Xyce::dout() << "Temp State Vector:" << std::endl;
  (*tmpStaVectorPtrPtr_)->printPetraObject(Xyce::dout());

  Xyce::dout() << "Temp State Derivative Vector:" << std::endl;
  (*tmpStaDerivVectorPtrPtr_)->printPetraObject(Xyce::dout());

}

} // namespace Linear
} // namespace Xyce
