//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_LOA_CktLoader.C,v $
//
// Purpose        : This file contains class definitions for the loader
//                  services package.
//
// Special Notes  :
//
// Creator        : Eric R. Keiter, SNL, Parallel Computational Sciences
//
// Creation Date  : 05/26/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.90.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:16 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

// ----------   Xyce Includes   ----------
#include <N_UTL_fwd.h>

#include <N_DEV_Algorithm.h>
#include <N_DEV_DeviceMgr.h>
#include <N_DEV_ExternDevice.h>
#include <N_DEV_SolverState.h>
#include <N_LOA_CktLoader.h>
#include <N_PDS_MPI.h>
#include <N_PDS_Serial.h>

namespace Xyce {
namespace Loader {

//-----------------------------------------------------------------------------
// Function      : CktLoader::CktLoader
// Purpose       : constructor
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 5/27/00
//-----------------------------------------------------------------------------
CktLoader::CktLoader(
  Device::DeviceMgr &         device_manager)
  : deviceManager_(device_manager)
{}

//-----------------------------------------------------------------------------
// Function      : CktLoader::~CktLoader
// Purpose       : destructor
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 5/27/00
//-----------------------------------------------------------------------------
CktLoader::~CktLoader()
{

}

bool
CktLoader::setParam(
  std::string &   name,
  double                val,
  bool overrideOriginal) 
{
  return deviceManager_.setParam(name,val, overrideOriginal);
}

bool
CktLoader::analyticSensitivitiesAvailable(
  std::string &   name)
{
  return deviceManager_.analyticSensitivitiesAvailable(name);
}


void
CktLoader::getAnalyticSensitivities(
  std::string &   name, 
  std::vector<double> & dfdpVec, 
  std::vector<double> & dqdpVec,
  std::vector<double> & dbdpVec,
  std::vector<int> &    FindicesVec,
  std::vector<int> &    QindicesVec,
  std::vector<int> &    BindicesVec) const
{
  return deviceManager_.getAnalyticSensitivities(name, dfdpVec, dqdpVec, dbdpVec, 
                                                  FindicesVec, QindicesVec, BindicesVec);
}

double
CktLoader::getParamAndReduce(
  const std::string &   name) const
{
  return deviceManager_.getParamAndReduce(name);
}

bool
CktLoader::getParamAndReduce(
  const std::string &   name,
  double &              val) const
{
  return deviceManager_.getParamAndReduce(name,val);
}

bool CktLoader::initializeProblem(
                                        Linear::Vector * nextSolVectorPtr,
                                        Linear::Vector * currSolVectorPtr,
                                        Linear::Vector * lastSolVectorPtr,
                                        Linear::Vector * nextStaVectorPtr,
                                        Linear::Vector * currStaVectorPtr,
                                        Linear::Vector * lastStaVectorPtr,
                                        Linear::Vector * StateDerivVectorPtr,
                                        Linear::Vector * nextStoVectorPtr,
                                        Linear::Vector * currStoVectorPtr,
                                        Linear::Vector * lastStoVectorPtr,
                                        Linear::Vector * QVectorPtr,
                                        Linear::Vector * FVectorPtr,
                                        Linear::Vector * BVectorPtr,
                                        Linear::Vector * dFdxdVpVectorPtr,
                                        Linear::Vector * dQdxdVpVectorPtr) const
{
  return CktLoader::deviceManager_.setICs
      (nextSolVectorPtr,
       currSolVectorPtr,
       lastSolVectorPtr,
       nextStaVectorPtr,
       currStaVectorPtr,
       lastStaVectorPtr,
       StateDerivVectorPtr,
       nextStoVectorPtr,
       currStoVectorPtr,
       lastStoVectorPtr,
       QVectorPtr,
       FVectorPtr,
       BVectorPtr,
       dFdxdVpVectorPtr,
       dQdxdVpVectorPtr);
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::loadDeviceErrorWeightMask
// Purpose       : Load mask for use in calculating weighted norms.
// Special Notes : Loads the mask into the LAS_System.  Need only be called
//                 once.
// Scope         : public
// Creator       : Tom Russo, SNL, Electrical and Microsystems Modeling
// Creation Date : 1/19/07
//-----------------------------------------------------------------------------
bool
CktLoader::loadDeviceErrorWeightMask(
  Linear::Vector *      deviceMask) const
{
  return deviceManager_.loadErrorWeightMask(deviceMask);
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::loadDAEMatrices
// Purpose       : This function
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 03/03/04
//-----------------------------------------------------------------------------
bool CktLoader::loadDAEMatrices(
  Linear::Vector *        tmpSolVectorPtr,
  Linear::Vector *        tmpStateVectorPtr,
  Linear::Vector *        tmpStateDerivVectorPtr,
  Linear::Vector *        tmpStoreVectorPtr,
  Linear::Matrix *        tmpdQdxMatrixPtr,
  Linear::Matrix *        tmpdFdxMatrixPtr)
{
  return CktLoader::deviceManager_.loadDAEMatrices(
    tmpSolVectorPtr,
    tmpStateVectorPtr,
    tmpStateDerivVectorPtr,
    tmpStoreVectorPtr,
    tmpdQdxMatrixPtr,
    tmpdFdxMatrixPtr);

}

//-----------------------------------------------------------------------------
// Function      : CktLoader::loadDAEVectors
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 03/03/04
//-----------------------------------------------------------------------------
bool CktLoader::loadDAEVectors   (
                                        Linear::Vector * nextSolVectorPtr,
                                        Linear::Vector * currSolVectorPtr,
                                        Linear::Vector * lastSolVectorPtr,
                                        Linear::Vector * nextStaVectorPtr,
                                        Linear::Vector * currStaVectorPtr,
                                        Linear::Vector * lastStaVectorPtr,
                                        Linear::Vector * StateDerivVectorPtr,
                                        Linear::Vector * nextStoVectorPtr,
                                        Linear::Vector * currStoVectorPtr,
                                        Linear::Vector * lastStoVectorPtr,
                                        Linear::Vector * stoLeadCurrQVectorPtr,
                                        Linear::Vector * nextLeadFVectorPtr,
                                        Linear::Vector * currLeadFVectorPtr,
                                        Linear::Vector * lastLeadFVectorPtr,
                                        Linear::Vector * nextLeadQVectorPtr,
                                        Linear::Vector * nextJunctionVVectorPtr,
                                        Linear::Vector * currJunctionVVectorPtr,
                                        Linear::Vector * lastJunctionVVectorPtr,
                                        Linear::Vector * QVectorPtr,
                                        Linear::Vector * FVectorPtr,
                                        Linear::Vector * BVectorPtr,
                                        Linear::Vector * dFdxdVpVectorPtr,
                                        Linear::Vector * dQdxdVpVectorPtr)
{
  return CktLoader::deviceManager_.loadDAEVectors
      (nextSolVectorPtr,
       currSolVectorPtr,
       lastSolVectorPtr,
       nextStaVectorPtr,
       currStaVectorPtr,
       lastStaVectorPtr,
       StateDerivVectorPtr,
       nextStoVectorPtr,
       currStoVectorPtr,
       lastStoVectorPtr,
       stoLeadCurrQVectorPtr,
       nextLeadFVectorPtr,
       currLeadFVectorPtr,
       lastLeadFVectorPtr,
       nextLeadQVectorPtr,
       nextJunctionVVectorPtr,
       currJunctionVVectorPtr,
       lastJunctionVVectorPtr,
       QVectorPtr,
       FVectorPtr,
       BVectorPtr,
       dFdxdVpVectorPtr,
       dQdxdVpVectorPtr);
}


//-----------------------------------------------------------------------------
// Function      : CktLoader::updateState
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Computational Sciences
// Creation Date : 11/17/05
//-----------------------------------------------------------------------------
bool
CktLoader::updateState(
  Linear::Vector * nextSolVectorPtr,
  Linear::Vector * currSolVectorPtr,
  Linear::Vector * lastSolVectorPtr,
  Linear::Vector * nextStaVectorPtr,
  Linear::Vector * currStaVectorPtr,
  Linear::Vector * lastStaVectorPtr,
  Linear::Vector * nextStoVectorPtr,
  Linear::Vector * currStoVectorPtr,
  Linear::Vector * lastStoVectorPtr)
{
  return deviceManager_.updateState(
    nextSolVectorPtr,
    currSolVectorPtr,
    lastSolVectorPtr,
    nextStaVectorPtr,
    currStaVectorPtr,
    lastStaVectorPtr,
    nextStoVectorPtr,
    currStoVectorPtr,
    lastStoVectorPtr);
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::loadBVectorsforAC
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Ting Mei, SNL
// Creation Date :
//-----------------------------------------------------------------------------
bool CktLoader::loadBVectorsforAC
  (Linear::Vector * bVecRealPtr,
   Linear::Vector * bVecImagPtr)
{
  return deviceManager_.loadBVectorsforAC (bVecRealPtr, bVecImagPtr);
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::getNumNoiseSources
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
int CktLoader::getNumNoiseSources ()
{
  return deviceManager_.getNumNoiseSources ();
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::getNumNoiseDevices
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
int CktLoader::getNumNoiseDevices ()
{
  return deviceManager_.getNumNoiseDevices ();
}


//-----------------------------------------------------------------------------
// Function      : CktLoader::setupNoiseSources
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
void CktLoader::setupNoiseSources 
  (std::vector<Xyce::Analysis::NoiseData*> & noiseDataVec)
{
  return deviceManager_.setupNoiseSources (noiseDataVec);
}



//-----------------------------------------------------------------------------
// Function      : CktLoader::getNoiseSources
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 12/8/2014
//-----------------------------------------------------------------------------
void CktLoader::getNoiseSources 
  (std::vector<Xyce::Analysis::NoiseData*> & noiseDataVec)
{
  return deviceManager_.getNoiseSources (noiseDataVec);
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::getBMatrixStampforMOR
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist and Ting Mei, SNL
// Creation Date :
//-----------------------------------------------------------------------------
bool CktLoader::getBMatrixEntriesforMOR(std::vector<int>& bMatEntriesVec,
                                                     std::vector<int>& bMatPosEntriesVec)
{
  return deviceManager_.getBMatrixEntriesforMOR( bMatEntriesVec, bMatPosEntriesVec );
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::getInductorsEntriesforMOR
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist and Ting Mei, SNL
// Creation Date :
//-----------------------------------------------------------------------------
/*bool CktLoader::getInductorsEntriesforMOR(std::vector<int>& inductorEntriesVec)
{
  return deviceManager_.getInductorsEntriesforMOR( inductorEntriesVec );
}
*/
//-----------------------------------------------------------------------------
// Function      : CktLoader::setInitialGuess
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 5/02/03
//-----------------------------------------------------------------------------
bool CktLoader::setInitialGuess(Linear::Vector * solVectorPtr)
{
  return CktLoader::deviceManager_.setInitialGuess (solVectorPtr);
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::updateSources
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/13/00
//-----------------------------------------------------------------------------
bool CktLoader::updateSources()
{
  return CktLoader::deviceManager_.updateSources();
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::output
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 4/13/01
//-----------------------------------------------------------------------------
bool CktLoader::output() const
{
  return CktLoader::deviceManager_.output();
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::finishOutput
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/19/04
//-----------------------------------------------------------------------------
bool CktLoader::finishOutput() const
{
  return CktLoader::deviceManager_.finishOutput();
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::xgetLinearSystemFlag
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/17/00
//-----------------------------------------------------------------------------
bool CktLoader::getLinearSystemFlag() const
{
  return CktLoader::deviceManager_.getLinearSystemFlag();
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::getDoubleDCOPFlag
// Purpose       : This is an accessor to allow the time integrator to determine
//                 if the current problem includes a PDE device.  If it does,
//                 then it makes sense to use two-pass DCOP calulation.  Hence
//                 the "DoubleDCOPFlag" in the name.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 04/25/01
//-----------------------------------------------------------------------------
bool CktLoader::getDoubleDCOPFlag () const
{
  return CktLoader::deviceManager_.getPDESystemFlag();
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::getLimiterFlag
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 03/11/04
//-----------------------------------------------------------------------------
bool CktLoader::getLimiterFlag ()
{
  return CktLoader::deviceManager_.getVoltageLimiterFlag ();
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::getBreakPoints
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/11/01
//-----------------------------------------------------------------------------
bool CktLoader::getBreakPoints ( std::vector<Util::BreakPoint> & breakPointTimes ) const
{
  return CktLoader::deviceManager_.getBreakPoints(breakPointTimes);
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::getMaxTimeStepSize ()
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 7/31/01
//-----------------------------------------------------------------------------
double CktLoader::getMaxTimeStepSize ()
{
  return deviceManager_.getMaxTimeStepSize ();
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::enablePDEContinuation
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/22/02
//-----------------------------------------------------------------------------
int CktLoader::enablePDEContinuation()
{
  return deviceManager_.enablePDEContinuation();
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::disablePDEContinuation
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/22/02
//-----------------------------------------------------------------------------
bool CktLoader::disablePDEContinuation ()
{
  return deviceManager_.disablePDEContinuation();
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::getNumInterfaceNodes
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 12/03/02
//-----------------------------------------------------------------------------
void CktLoader::getNumInterfaceNodes (std::vector<int> & numINodes)
{
  deviceManager_.getNumInterfaceNodes (numINodes);
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::loadCouplingRHS
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 12/03/02
//-----------------------------------------------------------------------------
bool CktLoader::loadCouplingRHS(int iSubProblem, int iCouple, Linear::Vector * dfdvPtr)
{
  return deviceManager_.loadCouplingRHS (iSubProblem, iCouple, dfdvPtr);
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::calcCouplingTerms
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 12/03/02
//-----------------------------------------------------------------------------
bool CktLoader::calcCouplingTerms (int iSubProblem, int iCouple, const Linear::Vector * dxdvPtr)
{
  return deviceManager_.calcCouplingTerms (iSubProblem, iCouple, dxdvPtr);
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::getHomotopyBlockSize
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Roger Pawlowski, SNL, Parallel Computational Sciences
// Creation Date : 01/26/2005
//-----------------------------------------------------------------------------
int CktLoader::getHomotopyBlockSize() const
{
  return deviceManager_.getHomotopyBlockSize();
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::allDevsConverged
// Purpose       : Check whether any device has taken an action that renders
//                  normal convergence checks invalid (i.e. that the current
//                  step must be assumed unconverged).
// Special Notes :
// Scope         : public
// Creator       : Tom Russo, SNL, Component Information and Models
// Creation Date : 03/22/05
//-----------------------------------------------------------------------------
bool CktLoader::allDevicesConverged(
  Parallel::Machine     comm)
{
  return deviceManager_.allDevicesConverged(comm);
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::innerDevsConverged
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 3/21/06
//-----------------------------------------------------------------------------
bool
CktLoader::innerDevicesConverged(
  Parallel::Machine     comm)
{
  return Device::devicesConverged(comm, deviceManager_.getExternDevices());
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::stepSuccess
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date : 03/12/06
//-----------------------------------------------------------------------------
void CktLoader::stepSuccess(Xyce::Analysis::TwoLevelMode analysis)
{
  const Device::InstanceVector &extern_devices = deviceManager_.getExternDevices();

  for (Device::InstanceVector::const_iterator it = extern_devices.begin(); it != extern_devices.end(); ++it)
  {
    Device::ExternDevice::Instance &extern_device = static_cast<Device::ExternDevice::Instance &>(*(*it));

    extern_device.stepSuccess(analysis);
  }
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::stepFailure
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date : 03/12/06
//-----------------------------------------------------------------------------
void CktLoader::stepFailure(Xyce::Analysis::TwoLevelMode analysis)
{
  const Device::InstanceVector &extern_devices = deviceManager_.getExternDevices();

  for (Device::InstanceVector::const_iterator it = extern_devices.begin(); it != extern_devices.end(); ++it)
  {
    Device::ExternDevice::Instance &extern_device = static_cast<Device::ExternDevice::Instance &>(*(*it));

    extern_device.stepFailure(analysis);
  }
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::acceptStep
// Purpose       : Communicate to devices that the current step has been
//                 accepted
// Special Notes : Most devices need not know.  The Transmission line does.
// Scope         : public
// Creator       : Tom Russo, SNL
// Creation Date : 01/23/07
//-----------------------------------------------------------------------------
void CktLoader::acceptStep ()
{
  deviceManager_.acceptStep();
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::getInitialQnorm
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date : 03/18/07
//-----------------------------------------------------------------------------
bool CktLoader::getInitialQnorm (std::vector<TimeIntg::TwoLevelError> & tleVec )
{
  return deviceManager_.getInitialQnorm(tleVec);
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::getInnerLoopErrorSums
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date : 03/15/06
//-----------------------------------------------------------------------------
bool CktLoader::getInnerLoopErrorSums (std::vector<TimeIntg::TwoLevelError> & tleVec)
{
  return deviceManager_.getInnerLoopErrorSums (tleVec);
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::updateStateArrays
// Purpose       : Tells the inner loop solve to update state arrays.
// Special Notes : Needed to support voltlim with loca, with 2-level.
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date : 03/15/06
//-----------------------------------------------------------------------------
bool CktLoader::updateStateArrays ()
{
  return deviceManager_.updateStateArrays ();
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::startTimeStep ()
// Purpose       : Tells the inner loop solve to do its prediction.
// Special Notes : Needed to support 2-level.
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date : 03/15/06
//-----------------------------------------------------------------------------
bool CktLoader::startTimeStep ()
{
  return deviceManager_.startTimeStep ();
}

//-----------------------------------------------------------------------------
// Function      : CktLoader::setExternalSolverState
// Purpose       :
// Special Notes : Needed to support 2-level.
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date : 03/21/06
//-----------------------------------------------------------------------------
void CktLoader::setExternalSolverState(const Device::SolverState & ss)
{
  return deviceManager_.setExternalSolverState (ss);
}

} // namespace Loader
} // namespace Xyce
