//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_LOA_HBLoader.C,v $
// Purpose       :
// Special Notes :
// Creator       :
// Creation Date :
//
// Revision Information:
// ---------------------
// Revision Number: $Revision: 1.48.2.1 $
// Revision Date  : $Date: 2015/04/02 18:20:16 $
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#include <Xyce_config.h>


// ---------- Standard Includes ----------
#include <iostream>

// ----------   Xyce Includes   ----------

#include <N_LOA_HBLoader.h>
#include <N_MPDE_Discretization.h>
#include <N_ERH_ErrorMgr.h>

#include <N_LAS_Builder.h>
#include <N_LAS_HBBuilder.h>
#include <N_LAS_Matrix.h>
#include <N_LAS_BlockVector.h>
#include <N_LAS_BlockMatrix.h>
#include <N_LAS_BlockSystemHelpers.h>

#include <N_UTL_fwd.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_Math.h>

#include <N_PDS_ParMap.h>
#include <N_DEV_DeviceMgr.h>

#include <Epetra_MultiVector.h>
#include <Epetra_BlockMap.h>

using Teuchos::rcp;
using Teuchos::RCP;
using Teuchos::rcp_dynamic_cast;

namespace Xyce {
namespace Loader {


  // Default constructor
HBLoader::HBLoader(
  const Teuchos::RCP<const N_MPDE_Discretization> discPtr,
  Device::DeviceMgr &                 device_manager,
  Linear::Builder &                   builder)
  : CktLoader(device_manager),
    deviceManager_(device_manager),
    builder_(builder),
    fastTimeDiscPtr_(discPtr),
    periodicTimesOffset_(0),
    period_(0),
    matrixFreeFlag_(false)
{
  // Now initialize all the time domain working vectors.
  appVecPtr_ = rcp(builder_.createVector());
  appNextStaVecPtr_ = rcp(builder_.createStateVector());
  appCurrStaVecPtr_ = rcp(builder_.createStateVector());
  appLastStaVecPtr_ = rcp(builder_.createStateVector());

  appdQdxPtr_ = rcp(builder_.createMatrix());
  appdFdxPtr_ = rcp(builder_.createMatrix());

  appNextStoVecPtr_ = rcp(builder_.createStoreVector());
  appCurrStoVecPtr_ = rcp(builder_.createStoreVector());
  appLastStoVecPtr_ = rcp(builder_.createStoreVector());
  appStoLeadCurrQVecPtr_ = rcp(builder_.createStoreVector()); 
  
  appNextLeadFVecPtr_ = rcp(builder.createLeadCurrentVector());
  appCurrLeadFVecPtr_ = rcp(builder.createLeadCurrentVector());
  appLastLeadFVecPtr_ = rcp(builder.createLeadCurrentVector());
  appLeadQVecPtr_     = rcp(builder.createLeadCurrentVector());
  appNextJunctionVVecPtr_ = rcp(builder.createLeadCurrentVector());
  appCurrJunctionVVecPtr_ = rcp(builder.createLeadCurrentVector());
  appLastJunctionVVecPtr_ = rcp(builder.createLeadCurrentVector());
}

//-----------------------------------------------------------------------------
// Function      : HBLoader::registerHBBuilder
// Purpose       : Registration method for the HB builder
// Special Notes :
// Scope         : public
// Creator       : Heidi Thornquist, Sandia Labs
// Creation Date : 11/08/13
//-----------------------------------------------------------------------------
void HBLoader::registerHBBuilder( Teuchos::RCP<Linear::HBBuilder> hbBuilderPtr )
{
  hbBuilderPtr_ = hbBuilderPtr; 

  // Now initialize all the frequency domain working vectors.
  bXtPtr_ = hbBuilderPtr_->createTimeDomainBlockVector();
  bVtPtr_ = hbBuilderPtr_->createTimeDomainBlockVector();
  bmdQdxPtr_ = rcp_dynamic_cast<Linear::BlockMatrix>(rcp(hbBuilderPtr_->createMatrix()));
  bmdFdxPtr_ = rcp_dynamic_cast<Linear::BlockMatrix>(rcp(hbBuilderPtr_->createMatrix()));

  // Vectors related to lead currents
  bStoreVecFreqPtr_ = hbBuilderPtr_->createExpandedRealFormTransposeStoreBlockVector();
  bStoreLeadCurrQVecFreqPtr_ = hbBuilderPtr_->createExpandedRealFormTransposeStoreBlockVector();

  bLeadCurrentVecFreqPtr_ = hbBuilderPtr_->createExpandedRealFormTransposeLeadCurrentBlockVector();
  bLeadCurrentQVecFreqPtr_ = hbBuilderPtr_->createExpandedRealFormTransposeLeadCurrentBlockVector();
}

//-----------------------------------------------------------------------------
// Function      : HBLoader::setFreq
// Purpose       : Assign times for fast time scale
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
void HBLoader::setHBFreqs( const std::vector<double> & freqs)
{
   freqs_ = freqs;
}



//-----------------------------------------------------------------------------
// Function      : HBLoader::setFastTimes
// Purpose       : Assign times for fast time scale
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
void HBLoader::setFastTimes( const std::vector<double> & times )
{
  times_ = times;
}

//-----------------------------------------------------------------------------
// Function      : HBLoader::loadDAEMatrices
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool HBLoader::loadDAEMatrices( Linear::Vector * X,
                                     Linear::Vector * S,
                                     Linear::Vector * dSdt,
                                     Linear::Vector * Store,
                                     Linear::Matrix * dQdx,
                                     Linear::Matrix * dFdx)
{
  if ( matrixFreeFlag_ )
  {
    if (DEBUG_HB)
    {
      Xyce::dout() << std::endl
                   << Xyce::section_divider << std::endl
                   << "  HBLoader::loadDAEMatrices:  matrixFree case" << std::endl;
    }

    dQdx->put(0.0);
    dFdx->put(0.0);
    // Do nothing in the Matrix Free Case.
    return(true);
  }
  else
  {
    if (DEBUG_HB)
    {
      Xyce::dout() << std::endl
                   << Xyce::section_divider << std::endl
                   << "  HBLoader::loadDAEMatrices: Time dependent with matrix case" << std::endl;
    }

    return loadTimeDepDAEMatrices(X,S,dSdt, Store, dQdx,dFdx);
  }

}

//-----------------------------------------------------------------------------
// Function      : HBLoader::loadTimeDepDAEMatrices
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Todd Coffey, 1414
// Creation Date : 09/03/08
//-----------------------------------------------------------------------------
bool HBLoader::loadTimeDepDAEMatrices( Linear::Vector * X,
                                       Linear::Vector * S,
                                       Linear::Vector * dSdt,
                                       Linear::Vector * Store,
                                       Linear::Matrix * dQdx,
                                       Linear::Matrix * dFdx)
{
  if (DEBUG_HB)
  {
    Xyce::dout() << std::endl
                 << Xyce::section_divider << std::endl
                 << "  HBLoader::loadTimeDepDAEMatrices" << std::endl;
  }

  //Zero out matrices
  dQdx->put(0.0);
  dFdx->put(0.0);

  Linear::Vector appdSdt( *appNextStaVecPtr_ );

  Linear::BlockMatrix & bdQdx = *dynamic_cast<Linear::BlockMatrix*>(dQdx);
  Linear::BlockMatrix & bdFdx = *dynamic_cast<Linear::BlockMatrix*>(dFdx);
  Linear::BlockVector & bX = *dynamic_cast<Linear::BlockVector*>(X);
#ifdef Xyce_FLEXIBLE_DAE_LOADS
  Linear::BlockVector & bS = *dynamic_cast<Linear::BlockVector*>(S);
  Linear::BlockVector & bdSdt = *dynamic_cast<Linear::BlockVector*>(dSdt);
  Linear::BlockVector & bStore = *dynamic_cast<Linear::BlockVector*>(Store);
#endif // Xyce_FLEXIBLE_DAE_LOADS

  int BlockCount = bX.blockCount();

  for( int i = 0; i < BlockCount; ++i )
  {
    if (DEBUG_HB)
    {
      Xyce::dout() << "Processing diagonal matrix block " << i << " of " << BlockCount-1 << std::endl;
    }

#ifdef Xyce_FLEXIBLE_DAE_LOADS
    //Set Time for fast time scale somewhere
//    state_.fastTime = times_[i];
    deviceManager_.setFastTime( times_[i] );

    //Update the sources
    appLoaderPtr_->updateSources();

    *appVecPtr_ = bX.block(i);
    *appNextStaVecPtr_ = bS.block(i);
    appdSdt = bdSdt.block(i);
    *appNextStoVecPtr_ = bStore.block(i);

    appLoaderPtr_->loadDAEMatrices( &*appVecPtr_, &*appNextStaVecPtr_, &appdSdt, &*appNextStoVecPtr_, &*appdQdxPtr_, &*appdFdxPtr_);

    bdQdx.block(i,i).add( *appdQdxPtr_ );
    bdFdx.block(i,i).add( *appdFdxPtr_ );
#else
    //For now, the matrices are loaded during the loadDAEVectors method
    //Just copied here
    bdQdx.block(i,i).add( bmdQdxPtr_->block(i,i) );
    bdFdx.block(i,i).add( bmdFdxPtr_->block(i,i) );

#endif // Xyce_FLEXIBLE_DAE_LOADS
  }

  // Fast Time scale discretization terms:
  // These are d(dQ/dt1)/dx terms, but go into bdFdx.
  // For this procedure, need to re-use the app matrix, appdQdx.
  Linear::Matrix & dQdxdt = *appdQdxPtr_;

  int Start = fastTimeDiscPtr_->Start();
  int Width = fastTimeDiscPtr_->Width();

  const std::vector<double> & Coeffs = fastTimeDiscPtr_->Coeffs();

  for( int i = 0; i < BlockCount; ++i )
  {
    if (DEBUG_HB)
    {
      Xyce::dout() << "Processing off diagonal matrix blocks on row " << i << " of " << BlockCount-1 << std::endl;
    }

    int Loc;
    int indexT1 = i + Start + periodicTimesOffset_;
    int indexT2 = indexT1 + Width - 1;
    double invh2 = 1.0 / (periodicTimes_[indexT2] - periodicTimes_[indexT1]);

    for( int j = 0; j < Width; ++j )
    {
      Loc = i + (j+Start);

      if( Loc < 0 )
      {
        Loc += BlockCount;
      }
      else if( Loc > (BlockCount-1) )
      {
        Loc -= BlockCount;
      }

      dQdxdt.put(0.0);
      dQdxdt.add( bdQdx.block(Loc,Loc) );
      dQdxdt.scale( Coeffs[j]*invh2 );
      bdFdx.block(i,Loc).add( dQdxdt );
    }
  }

  dQdx->fillComplete();
  dFdx->fillComplete();

  if (DEBUG_HB)
  {
    Xyce::dout() << "HB bX:" << std::endl;
    bX.printPetraObject(dout());
    Xyce::dout() << "HB bdQdx:" << std::endl;
    bdQdx.printPetraObject(dout());
    Xyce::dout() << "HB bdFdx:" << std::endl;
    bdFdx.printPetraObject(dout());
#ifdef Xyce_FLEXIBLE_DAE_LOADS
    Xyce::dout() << "HB bS:" << std::endl;
    bS.printPetraObject(dout());
    Xyce::dout() << "HB dSdt:" << std::endl;
    bdSdt.printPetraObject(dout());
    Xyce::dout() << "HB bStore:" << std::endl;
    bStore.printPetraObject(dout());
#endif // Xyce_FLEXIBLE_DAE_LOADS

    Xyce::dout() << Xyce::section_divider << std::endl;
  }
  
  return true;
}

//-----------------------------------------------------------------------------
// Function      : HBLoader::applyDAEMatrices
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool HBLoader::applyDAEMatrices( Linear::Vector * Xf,
                                 Linear::Vector * S,
                                 Linear::Vector * dSdt,
                                 Linear::Vector * Store,
                                 const Linear::Vector & Vf,
                                 Linear::Vector * dQdxV,
                                 Linear::Vector * dFdxV )
{
  if ( !matrixFreeFlag_ )
  {
    std::string msg="HBLoader::applyDAEMatrices.  This function should only be called in the matrix free case.";
    N_ERH_ErrorMgr::report ( N_ERH_ErrorMgr::DEV_FATAL, msg);
  }
  if (DEBUG_HB) 
  {
    Xyce::dout() << std::endl
                 << Xyce::section_divider << std::endl
                 << "  HBLoader::applyDAEMatrices" << std::endl;
  }

  //Zero out matvec vectors
  dQdxV->putScalar(0.0); // This is dQdx * V on output (only used in HB-env simulation)
  dFdxV->putScalar(0.0); // This is dFdx * V on output (This is main output for HB)

  bXtPtr_->putScalar(0.0);
  bVtPtr_->putScalar(0.0);

  Linear::BlockVector & bXf = *dynamic_cast<Linear::BlockVector*>(Xf);
  // We have to do something special with Vf because AztecOO (or Belos)
  // probably used the Epetra_LinearProblem's Epetra_Maps to create the input
  // vector here.  In this case, Vf is just an Linear::Vector and not an
  // Linear::BlockVector.
  const Linear::BlockVector bVf(Vf, bXf.blockSize());

  permutedIFT(bXf, &*bXtPtr_);
  permutedIFT(bVf, &*bVtPtr_);

  Linear::BlockVector & bX = *bXtPtr_;

  Linear::BlockVector * bdQdxV = dynamic_cast<Linear::BlockVector*>(dQdxV);
  Linear::BlockVector * bdFdxV = dynamic_cast<Linear::BlockVector*>(dFdxV);

  Teuchos::RCP<Linear::BlockVector>  bdQdxVt = hbBuilderPtr_->createTimeDomainBlockVector();
  Teuchos::RCP<Linear::BlockVector>  bdFdxVt = hbBuilderPtr_->createTimeDomainBlockVector();

  int BlockCount = bX.blockCount();
  for( int i = 0; i < BlockCount; ++i )
  {
    if (DEBUG_HB)
    {
      Xyce::dout() << "Processing diagonal matrix block " << i << " of " << BlockCount-1 << std::endl;
    }

    // Get the already stored time-domain Jacobian matrices
    appdQdxPtr_ = vecAppdQdxPtr_[i];
    appdFdxPtr_ = vecAppdFdxPtr_[i];

    if (DEBUG_HB)
    {
      Xyce::dout() << "bVtPtr block i = " << i << " : " << std::endl;
      bVtPtr_->block(i).printPetraObject(dout());

      Xyce::dout() << "appdQdxPtr_ = " << i << " : " <<  std::endl;
      appdQdxPtr_->printPetraObject(dout());

      Xyce::dout() << "appdFdxPtr_ = " << i << " : " <<  std::endl;
      appdFdxPtr_->printPetraObject(dout());
    }
    

    appdQdxPtr_->matvec(false, bVtPtr_->block(i), bdQdxVt->block(i));
    appdFdxPtr_->matvec(false, bVtPtr_->block(i), bdFdxVt->block(i));

    if (DEBUG_HB)
    {
      Xyce::dout() << "bdQdxVt block i = " << i << " : " << std::endl;
      bdQdxVt->block(i).printPetraObject(dout());

      Xyce::dout() << "bdFdxVt block i = " << i << " : " << std::endl;
      bdFdxVt->block(i).printPetraObject(dout());
    }
  }

  permutedFFT(*bdQdxVt, bdQdxV);
  permutedFFT(*bdFdxVt, bdFdxV);

  int blockCount = bXf.blockCount();
  int blockSize = bXf.block(0).globalLength();
//  double omega = 2.0 * M_PI/ period_;

  int size_ = freqs_.size();
  int posFreq = (size_-1)/2;
  double omega = 2.0 * M_PI * freqs_[posFreq];
  for( int i = 0; i < blockCount; ++i )
  {
    Linear::Vector QVec(bXf.block(i));
    Linear::Vector freqVec = bdQdxV->block(i);

    // Only one processor owns each block of the frequency-domain vector
    if (freqVec.localLength() > 0)
    {
     
      omega = 2.0 * M_PI * freqs_[posFreq];

      QVec[0] = -freqVec[1]*omega;
      QVec[1] = freqVec[0]*omega;

      for (int j=1; j < (blockSize/2+1)/2; ++j)
      {
        omega = 2.0 * M_PI * freqs_[posFreq+j];
        QVec[2*j] = -freqVec[2*j+1]*omega;
        QVec[2*(blockSize/2-j)] = -freqVec[2*j+1]*omega;

        QVec[2*j+1] = freqVec[2*j]*omega;
        QVec[2*(blockSize/2-j)+1] = -freqVec[2*j]*omega;
      }
    }

    bdFdxV->block(i).update(1.0, QVec , 1.0);
  }

  if (DEBUG_HB)
  {
    Xyce::dout() << "HB bX:" << std::endl;
    bX.printPetraObject(dout());
    Xyce::dout() << "HB bdQdxV:" << std::endl;
    bdQdxV->printPetraObject(dout());
    Xyce::dout() << "HB bdFdxV:" << std::endl;
    bdFdxV->printPetraObject(dout());

    Xyce::dout() << Xyce::section_divider << std::endl;
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : HBLoader::updateState
// Purpose       :
// Special Notes : ERK.  This function needs to be a no-op.  The reason
//                 is that the state information needs to be the same
//                 at the time of updateState, loadDAEVectors and
//                 loadDAEMatrices.  Thus, they have to all happen inside
//                 of the same "fast time" loop.  So, this functionality
//                 has been moved up into loadDAEVectors.
//
//                 Note: for similar reasons, loadDAEMatrices is called from
//                 within that function as well.
//
// Scope         : public
// Creator       : Todd Coffey, 1414
// Creation Date : 01/17/07
//-----------------------------------------------------------------------------
bool HBLoader::updateState
 (Linear::Vector * nextSolVectorPtr,
  Linear::Vector * currSolVectorPtr,
  Linear::Vector * lastSolVectorPtr,
  Linear::Vector * nextStaVectorPtr,
  Linear::Vector * currStaVectorPtr,
  Linear::Vector * lastStaVectorPtr,
  Linear::Vector * nextStoVectorPtr,
  Linear::Vector * currStoVectorPtr,
  Linear::Vector * lastStoVectorPtr
  )
{
  bool bsuccess = true;

  // For HB case, this needs to be a no-op.

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : HBLoader::loadDAEVectors
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       :
// Creation Date :
//-----------------------------------------------------------------------------
bool HBLoader::loadDAEVectors( Linear::Vector * Xf,
                               Linear::Vector * currX,
                               Linear::Vector * lastX,
                               Linear::Vector * S,
                               Linear::Vector * currS,
                               Linear::Vector * lastS,
                               Linear::Vector * dSdt,
                               Linear::Vector * Store,
                               Linear::Vector * currStore,
                               Linear::Vector * lastStore,
                               Linear::Vector * storeLeadCurrQ,
                               Linear::Vector * nextLeadFVectorPtr,
                               Linear::Vector * currLeadFVectorPtr,
                               Linear::Vector * lastLeadFVectorPtr,
                               Linear::Vector * nextLeadQVectorPtr,
                               Linear::Vector * nextJunctionVVectorPtr,
                               Linear::Vector * currJunctionVVectorPtr,
                               Linear::Vector * lastJunctionVVectorPtr,
                               Linear::Vector * Q,
                               Linear::Vector * F,
                               Linear::Vector * B,
                               Linear::Vector * dFdxdVp,
                               Linear::Vector * dQdxdVp )
{
  if (DEBUG_HB)
  {
    Xyce::dout() << std::endl
                 << Xyce::section_divider << std::endl
                 << "  HBLoader::loadDAEVectors" << std::endl;
  }

  //Zero out vectors
  appVecPtr_->putScalar(0.0);
  appNextStaVecPtr_->putScalar(0.0);
  appCurrStaVecPtr_->putScalar(0.0);
  appLastStaVecPtr_->putScalar(0.0);
  Linear::Vector appdSdt( *appNextStaVecPtr_ );

  appNextStoVecPtr_->putScalar(0.0);
  appCurrStoVecPtr_->putScalar(0.0);
  appLastStoVecPtr_->putScalar(0.0);
  appStoLeadCurrQVecPtr_->putScalar(0.0);

  Linear::Vector appQ( *appVecPtr_ );
  Linear::Vector appF( *appVecPtr_ );
  Linear::Vector appB( *appVecPtr_ );

  Linear::Vector appdFdxdVp( *appVecPtr_ );
  Linear::Vector appdQdxdVp( *appVecPtr_ );

  // This is a temporary load storage vector.
  Linear::Vector dQdt2( *appVecPtr_ );

  bXtPtr_->putScalar(0.0);

  Linear::BlockVector & bXf = *dynamic_cast<Linear::BlockVector*>(Xf);

  permutedIFT(bXf, &*bXtPtr_);

  // 12/8/06 tscoffe:   Note:  "b" at beginning of variable name means Linear::BlockVector
  Linear::BlockVector & bX = *bXtPtr_;
  Linear::BlockVector & bS = *dynamic_cast<Linear::BlockVector*>(S);
  Linear::BlockVector & bcurrS = *dynamic_cast<Linear::BlockVector*>(currS);
  Linear::BlockVector & blastS = *dynamic_cast<Linear::BlockVector*>(lastS);
  Linear::BlockVector & bdSdt = *dynamic_cast<Linear::BlockVector*>(dSdt);
  Linear::BlockVector & bStore = *dynamic_cast<Linear::BlockVector*>(Store);
  Linear::BlockVector & bcurrStore = *dynamic_cast<Linear::BlockVector*>(currStore);
  Linear::BlockVector & blastStore = *dynamic_cast<Linear::BlockVector*>(lastStore);
  
  Linear::BlockVector & bNextLeadF = *dynamic_cast<Linear::BlockVector*>(nextLeadFVectorPtr);
  Linear::BlockVector & bCurrLeadF = *dynamic_cast<Linear::BlockVector*>(currLeadFVectorPtr);
  Linear::BlockVector & bLastLeadF = *dynamic_cast<Linear::BlockVector*>(lastLeadFVectorPtr);
  Linear::BlockVector & bLeadQ = *dynamic_cast<Linear::BlockVector*>(nextLeadQVectorPtr);
  Linear::BlockVector & bNextJunctionV = *dynamic_cast<Linear::BlockVector*>(nextJunctionVVectorPtr);
  Linear::BlockVector & bCurrJunctionV = *dynamic_cast<Linear::BlockVector*>(currJunctionVVectorPtr);
  Linear::BlockVector & bLastJunctionV = *dynamic_cast<Linear::BlockVector*>(lastJunctionVVectorPtr);
  
  Linear::BlockVector * bQ = dynamic_cast<Linear::BlockVector*>(Q);
  Linear::BlockVector * bF = dynamic_cast<Linear::BlockVector*>(F);
  Linear::BlockVector * bB = dynamic_cast<Linear::BlockVector*>(B);

  Linear::BlockVector * bdFdxdVp = dynamic_cast<Linear::BlockVector*>(dFdxdVp);
  Linear::BlockVector * bdQdxdVp = dynamic_cast<Linear::BlockVector*>(dQdxdVp);


//  Linear::Vector * storeLeadCurrQ,

  Linear::BlockVector & bstoreLeadCurrQ = *dynamic_cast<Linear::BlockVector*>(storeLeadCurrQ);

  Linear::BlockVector  bQt(*bXtPtr_);
  Linear::BlockVector  bFt(*bXtPtr_);
  Linear::BlockVector  bBt(*bXtPtr_);

  Linear::BlockVector  bdQdxdVpt(*bXtPtr_);
  Linear::BlockVector  bdFdxdVpt(*bXtPtr_);


//  Linear::BlockVector  bstoreLeadCurrQt(*bXtPt  );

#ifndef Xyce_FLEXIBLE_DAE_LOADS
  bmdQdxPtr_->put(0.0);
  bmdFdxPtr_->put(0.0);
#endif

  int BlockCount = bX.blockCount();
  int BlockSize =  bX.blockSize();

#if 0
  Xyce::dout() << "bX.blockCount = "<< BlockCount <<std::endl;
  Xyce::dout() << "bX.blockSize = "<< BlockSize <<std::endl;
#endif

  // We are storing the time domain Jacobians, initialize the memory here 
  if ((int)vecAppdQdxPtr_.size()!=BlockCount)
  {
    vecAppdQdxPtr_.resize( BlockCount );
    vecAppdFdxPtr_.resize( BlockCount );
    for ( int i=0; i < BlockCount; ++i )
    {
      vecAppdQdxPtr_[i] = rcp(builder_.createMatrix());
      vecAppdFdxPtr_[i] = rcp(builder_.createMatrix());
    }
  }

  // Now perform implicit application of frequency domain Jacobian. 
  for( int i = 0; i < BlockCount; ++i )
  {
    if (DEBUG_HB)
    {
      Xyce::dout() << "Processing vectors for block " << i << " of " << BlockCount-1 << std::endl;
    }

    //Set Time for fast time scale somewhere
//    state_.fastTime = times_[i];
    deviceManager_.setFastTime( times_[i] );

    if (DEBUG_HB)
    {
      Xyce::dout() << "Calling updateSources on the appLoader" << std::endl;
    }

    //Update the sources
    appLoaderPtr_->updateSources();  // this is here to handle "fast" sources.

    *appVecPtr_ = bX.block(i);
    *appNextStaVecPtr_ = bS.block(i);
    *appCurrStaVecPtr_ = bcurrS.block(i);
    *appLastStaVecPtr_ = blastS.block(i);
    appdSdt = bdSdt.block(i);
    *appNextStoVecPtr_ = bStore.block(i);
    *appCurrStoVecPtr_ = bcurrStore.block(i);
    *appLastStoVecPtr_ = blastStore.block(i);
    *appStoLeadCurrQVecPtr_ = bstoreLeadCurrQ.block(i);
    *appNextLeadFVecPtr_  = bNextLeadF.block(i);
    *appCurrLeadFVecPtr_  = bCurrLeadF.block(i);
    *appLastLeadFVecPtr_  = bLastLeadF.block(i);
    *appLeadQVecPtr_      =  bLeadQ.block(i);
    *appNextJunctionVVecPtr_  =  bNextJunctionV.block(i);
    *appCurrJunctionVVecPtr_  =  bCurrJunctionV.block(i);
    *appLastJunctionVVecPtr_  =  bLastJunctionV.block(i);
    
//blastStore.block(i); // Need to update to correct block!

    if (DEBUG_HB)
    {
      Xyce::dout() << "Updating State for block " << i << " of " << BlockCount-1 << std::endl;
    }

    // Note: This updateState call is here (instead of in the
    // HBLoader::updateState function) because it has to be called
    // for the same fast time point.
    appLoaderPtr_->updateState
      ( &*appVecPtr_,
        &*appVecPtr_,  // note, this is a placeholder! ERK
        &*appVecPtr_,  // note, this is a placeholder! ERK
        &*appNextStaVecPtr_, &*appCurrStaVecPtr_ , &*appLastStaVecPtr_,
        &*appNextStoVecPtr_, &*appCurrStoVecPtr_ , &*appLastStoVecPtr_
        );

    bS.block(i) = *appNextStaVecPtr_;
    bcurrS.block(i) = *appCurrStaVecPtr_;
    blastS.block(i) = *appLastStaVecPtr_;
    bStore.block(i) = *appNextStoVecPtr_;
    bcurrStore.block(i) = *appCurrStoVecPtr_;
    blastStore.block(i) = *appLastStoVecPtr_;

    if (DEBUG_HB)
    {
      Xyce::dout() << "Calling loadDAEVectors on the appLoader" << std::endl;
    }

    // This has to be done because the app loader does NOT zero these vectors out.
    appQ.putScalar(0.0);
    appF.putScalar(0.0);
    appB.putScalar(0.0);
    appdFdxdVp.putScalar(0.0);
    appdQdxdVp.putScalar(0.0);

    appLoaderPtr_->loadDAEVectors
      ( &*appVecPtr_,
        &*appVecPtr_,  // note, this is a placeholder! ERK
        &*appVecPtr_,  // note, this is a placeholder! ERK
        &*appNextStaVecPtr_, &*appCurrStaVecPtr_, &*appLastStaVecPtr_, &appdSdt,
        &*appNextStoVecPtr_, &*appCurrStoVecPtr_, &*appLastStoVecPtr_, &*appStoLeadCurrQVecPtr_,
        &*appNextLeadFVecPtr_, &*appCurrLeadFVecPtr_, &*appLastLeadFVecPtr_, &*appLeadQVecPtr_, 
        &*appNextJunctionVVecPtr_, &*appCurrJunctionVVecPtr_,&*appLastJunctionVVecPtr_,
        &appQ, &appF, &appB,
        &appdFdxdVp, &appdQdxdVp );

    bQt.block(i) = appQ;
    bFt.block(i) = appF;
    bBt.block(i) = appB;

    bdQdxdVpt.block(i) = appdQdxdVp;
    bdFdxdVpt.block(i) = appdFdxdVp;

    bstoreLeadCurrQ.block(i) = *appStoLeadCurrQVecPtr_;

    bStore.block(i) = *appNextStoVecPtr_;  //lead current get loaded during loadDAEVectors

    bNextLeadF.block(i) = *appNextLeadFVecPtr_;  // lead currents loaded into lead current vector.
    bLeadQ.block(i) = *appLeadQVecPtr_;

//    Xyce::dout() << "HB Loader Store Vector TD" << std::endl;
//    bStore.printPetraObject(std::cout);

//    Xyce::dout() << "HB Loader Store  Q Vector TD" << std::endl;
//    bstoreLeadCurrQ.printPetraObject(std::cout);

    // Store the time domain Jacobian for future use.
    vecAppdQdxPtr_[i]->put(0.0);
    vecAppdFdxPtr_[i]->put(0.0);

    // Load dQdx and dFdx into the storage location for this time point.
    appLoaderPtr_->loadDAEMatrices( &*appVecPtr_, &*appNextStaVecPtr_, &appdSdt, &*appNextStoVecPtr_, &*vecAppdQdxPtr_[i],  &*vecAppdFdxPtr_[i]);
  }

  permutedFFT(bBt, bB);
 
  permutedFFT(bQt, bQ); 

  permutedFFT(bFt, bF);

  permutedFFT(bdQdxdVpt, bdQdxdVp);

  permutedFFT(bdFdxdVpt, bdFdxdVp);

  bStoreVecFreqPtr_->putScalar(0.0);
  bStoreLeadCurrQVecFreqPtr_->putScalar(0.0);

  permutedFFT(bStore, &*bStoreVecFreqPtr_);
  permutedFFT(bstoreLeadCurrQ, &*bStoreLeadCurrQVecFreqPtr_);  

  bLeadCurrentVecFreqPtr_->putScalar(0.0);
  bLeadCurrentQVecFreqPtr_->putScalar(0.0);
  permutedFFT(bNextLeadF, &*bLeadCurrentVecFreqPtr_);
  permutedFFT(bLeadQ, &*bLeadCurrentQVecFreqPtr_);

//  Xyce::dout() << "HB Store Vector FD" << std::endl;
//  bStoreVecFreqPtr_->printPetraObject(std::cout);
//  bStoreLeadCurrQVecFreqPtr_->printPetraObject(std::cout);
//  bLeadCurrentVecFreqPtr_->printPetraObject(Xyce::dout());   

  int blockCount = bXf.blockCount();
  int blockSize = bXf.block(0).globalLength();
//  double omega = 2.0 * M_PI/ period_;
  
  int size_ = freqs_.size();
  int posFreq = (size_-1)/2;
  double omega = 2.0 * M_PI * freqs_[posFreq];

  for( int i = 0; i < blockCount; ++i )
  {
    // Create work vectors from the current frequency block vector
    // NOTE:  This needs to be done for each block to make sure that the
    //        map is the same as the bF block.
    Linear::Vector QVec(bQ->block(i));
    Linear::Vector freqVec = bQ->block(i);

    Linear::Vector dQdxdVpVec(bdQdxdVp->block(i));
    Linear::Vector freqVec1 = bdQdxdVp->block(i);

    // Only one processor owns each block of the frequency-domain vector
    if (freqVec.localLength() > 0)
    {
      omega = 2.0 * M_PI * freqs_[posFreq]; 

      QVec[0] = -freqVec[1]*omega;
      QVec[1] = freqVec[0]*omega;

      dQdxdVpVec[0] = -freqVec1[1]*omega;
      dQdxdVpVec[1] = freqVec1[0]*omega;

      for (int j=1; j < (blockSize/2+1)/2; ++j)
      {
        
        omega = 2.0 * M_PI * freqs_[posFreq + j];

        QVec[2*j] = -freqVec[2*j+1]*omega;
        QVec[2*(blockSize/2-j)] = -freqVec[2*j+1]*omega;

        QVec[2*j+1] = freqVec[2*j]*omega;
        QVec[2*(blockSize/2-j)+1] = -freqVec[2*j]*omega;

        dQdxdVpVec[2*j] = -freqVec1[2*j+1]*omega;
        dQdxdVpVec[2*(blockSize/2-j)] = -freqVec1[2*j+1]*omega;

        dQdxdVpVec[2*j+1] = freqVec1[2*j]*omega;
        dQdxdVpVec[2*(blockSize/2-j)+1] = -freqVec1[2*j]*omega;

      }
    }

    bF->block(i).update(1.0, QVec , 1.0);

    bdFdxdVp->block(i).update(1.0, dQdxdVpVec, 1.0);
  }

  blockCount = bStoreVecFreqPtr_->blockCount();
  blockSize =  bStoreVecFreqPtr_->blockSize();

  for( int i = 0; i < blockCount; ++i )
  {
    // Create work vectors from the current frequency block vector
    // NOTE:  This needs to be done for each block to make sure that the
    //        map is the same as the bF block.
    Linear::Vector stoLeadCurrdQdtVec(bStoreLeadCurrQVecFreqPtr_->block(i));
    Linear::Vector stoLeadCurrQVec = bStoreLeadCurrQVecFreqPtr_->block(i);

    // Only one processor owns each block of the frequency-domain vector
    if (stoLeadCurrQVec.localLength() > 0)
    {

      omega = 2.0 * M_PI * freqs_[posFreq];
      stoLeadCurrdQdtVec[0] = stoLeadCurrQVec[1]*omega;
      stoLeadCurrdQdtVec[1] = stoLeadCurrQVec[0]*omega;

      for (int j=1; j < (blockSize/2+1)/2; ++j)
      {
        omega = 2.0 * M_PI * freqs_[posFreq+ j];

        stoLeadCurrdQdtVec[2*j] = -stoLeadCurrQVec[2*j+1]*omega;
        stoLeadCurrdQdtVec[2*(blockSize/2-j)] = -stoLeadCurrQVec[2*j+1]*omega;

        stoLeadCurrdQdtVec[2*j+1] = stoLeadCurrQVec[2*j]*omega;
        stoLeadCurrdQdtVec[2*(blockSize/2-j)+1] = -stoLeadCurrQVec[2*j]*omega;
      }
    }

    bStoreVecFreqPtr_->block(i).update(1.0, stoLeadCurrdQdtVec, 1.0);
//    Xyce::dout() << "HB Store Vector  dqdt + f(x) FD" << std::endl;
//    bStoreVecFreqPtr_->printPetraObject(std:: cout);
  }

  // 
  // take care of the lead current vector 
  blockCount = bLeadCurrentVecFreqPtr_->blockCount();
  blockSize =  bLeadCurrentVecFreqPtr_->blockSize();

  for( int i = 0; i < blockCount; ++i )
  {
    // Create work vectors from the current frequency block vector
    // NOTE:  This needs to be done for each block to make sure that the
    //        map is the same as the bF block.
    Linear::Vector leadCurrdQdtVec(bLeadCurrentQVecFreqPtr_->block(i));
    Linear::Vector leadCurrQVec = bLeadCurrentQVecFreqPtr_->block(i);

    // Only one processor owns each block of the frequency-domain vector
    if (leadCurrQVec.localLength() > 0)
    {

      omega = 2.0 * M_PI * freqs_[posFreq];
      leadCurrdQdtVec[0] = leadCurrQVec[1]*omega;
      leadCurrdQdtVec[1] = leadCurrQVec[0]*omega;

      for (int j=1; j < (blockSize/2+1)/2; ++j)
      {
        omega = 2.0 * M_PI * freqs_[posFreq+ j];

        leadCurrdQdtVec[2*j] = -leadCurrQVec[2*j+1]*omega;
        leadCurrdQdtVec[2*(blockSize/2-j)] = -leadCurrQVec[2*j+1]*omega;

        leadCurrdQdtVec[2*j+1] = leadCurrQVec[2*j]*omega;
        leadCurrdQdtVec[2*(blockSize/2-j)+1] = -leadCurrQVec[2*j]*omega;
      }
    }

    bLeadCurrentVecFreqPtr_->block(i).update(1.0, leadCurrdQdtVec, 1.0);
  }


//  permutedIFT(*bStoreVecFreqPtr_, &bStore);
  if (DEBUG_HB)
  {
    Xyce::dout() << "HB X Vector" << std::endl;
    bX.printPetraObject(std::cout);
    //  Xyce::dout() << "HB S Vector" << std::endl;
    //  bS.printPetraObject(Xyce::dout());
    //  Xyce::dout() << "HB dSdt Vector" << std::endl;
    //  bdSdt.printPetraObject(Xyce::dout());
    Xyce::dout() << "HB Store Vector" << std::endl;
    bStore.printPetraObject(std::cout);
    Xyce::dout() << "HB Q Vector" << std::endl;
    bQ->printPetraObject(std::cout);
    Xyce::dout() << "HB F Vector" << std::endl;
    bF->printPetraObject(std::cout);
    Xyce::dout() << "HB bdFdxdVp Vector" << std::endl;
    bdFdxdVp->printPetraObject(std::cout);
    Xyce::dout() << "HB bdQdxdVp Vector" << std::endl;
    bdQdxdVp->printPetraObject(std::cout);

#ifndef Xyce_FLEXIBLE_DAE_LOADS
    Xyce::dout() << "HB bmdQdx_" << std::endl;
    bmdQdxPtr_->printPetraObject(std::cout);
    Xyce::dout() << "HB bmdFdx_" << std::endl;
    bmdFdxPtr_->printPetraObject(std::cout);
#endif // Xyce_FLEXIBLE_DAE_LOADS
    Xyce::dout() << Xyce::section_divider << std::endl;
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : HBLoader::loadDeviceErrorWeightMask
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 11/1/2014
//-----------------------------------------------------------------------------
bool
HBLoader::loadDeviceErrorWeightMask(
  Linear::Vector *      deviceMask) const
{
#if 0
  Linear::BlockVector & bDevMask = *dynamic_cast<Linear::BlockVector*>(deviceMask);

#if 0
  Xyce::dout() << "HBLoader::loadDeviceErrorWeightMask.  Original (nonblock) deviceMask.size = "
    << appVecPtr_->globalLength() <<std::endl;
#endif

  appVecPtr_->putScalar(1.0);
  bool returnValue = deviceManager_.loadErrorWeightMask(&*appVecPtr_);

  int blockCount = bDevMask.blockCount();
  int blockSize =  bDevMask.blockSize();

#if 0
  Xyce::dout() << "bDevMask.blockCount = "<< blockCount <<std::endl;
  Xyce::dout() << "bDevMask.blockSize = "<< blockSize <<std::endl;
  appVecPtr_->printPetraObject(Xyce::dout());
#endif

  //Teuchos::RCP<N_PDS_ParMap> baseMap = Teuchos::rcp_const_cast<N_PDS_ParMap>( hbBuilderPtr_->getBaseStoreMap() );

  for( int i = 0; i < blockCount; ++i )
  {
    // See if this variable is owned by the local processor.
    // If so, this processor owns the entire j-th block of the vector
    //int lid = baseMap->globalToLocalIndex( i );

    Linear::Vector& localVecRef =  bDevMask.block(i);

    for (int j=0;j<blockSize;++j)
    {
      localVecRef[j] = (*appVecPtr_)[i];
    }
  }

#if 0
  bDevMask.printPetraObject(Xyce::dout());
#endif

  return returnValue;
#else
  return true;
#endif
}

//-----------------------------------------------------------------------------
// Function      : HBLoader::permutedFFT
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Ting Mei
// Creation Date : 09/05/08
//---------------------------------------------------------------------------
void HBLoader::permutedFFT(const Linear::BlockVector & xt, Linear::BlockVector * xf)
{
  // Call the function to compute the permuted FFT from the block system helper functions.
  computePermutedDFT( *dftInterface_, xt, xf );
}

//-----------------------------------------------------------------------------
// Function      : HBLoader::permutedIFT
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Ting Mei
// Creation Date : 09/05/08
//---------------------------------------------------------------------------
void HBLoader::permutedIFT(const Linear::BlockVector & xf, Linear::BlockVector * xt)
{
  // Call the function to compute the permuted IFT from the block system helper functions.
  computePermutedIFT( *dftInterface_, xf, xt );
}

} // namespace Loader
} // namespace Xyce
