//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-------------------------------------------------------------------------
// Filename       : $RCSfile: N_NLS_Manager.C,v $
//
// Purpose        : Body for the nonlinear solver manager class implementation
//                  which will allow for the creation and selection of NLS
//                  algorithms.
//
// Special Notes  : GOF Strategy Pattern
//
// Creator        : Scott A. Hutchinson, SNL, Parallel Computational Sciences
//
// Creation Date  : 04/28/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.218.2.1 $
//
// Revision Date  : $Date $
//
// Current Owner  : $Author: tvrusso $
//-------------------------------------------------------------------------

#include <Xyce_config.h>

// ----------  Xyce Includes   ----------

#include <N_ERH_ErrorMgr.h>
#include <N_IO_CmdParse.h>
#include <N_IO_OutputMgr.h>
#include <N_IO_PkgOptionsMgr.h>
#include <N_NLS_ConductanceExtractor.h>
#include <N_NLS_DampedNewton.h>
#include <N_NLS_Manager.h>
#include <N_NLS_NOX_Interface.h>
#include <N_NLS_NonLinInfo.h>
#include <N_NLS_Sensitivity.h>
#include <N_NLS_TwoLevelNewton.h>
#include <N_TIA_DataStore.h>
#include <N_UTL_Expression.h>
#include <N_UTL_FeatureTest.h>

// ----------  Trilinos Includes   ----------

namespace Xyce {
namespace Nonlinear {

namespace
{

struct Manager_OptionsReg : public IO::PkgOptionsReg
{
  Manager_OptionsReg( Manager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setOptions( options ); }

  Manager & manager_;
};

struct Manager_TranOptionsReg : public IO::PkgOptionsReg
{
  Manager_TranOptionsReg( Manager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setTranOptions( options ); }

  Manager & manager_;
};

struct Manager_HBOptionsReg : public IO::PkgOptionsReg
{
  Manager_HBOptionsReg( Manager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setHBOptions( options ); }

  Manager & manager_;
};


struct Manager_LSOptionsReg : public IO::PkgOptionsReg
{
  Manager_LSOptionsReg( Manager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setLinSolOptions( options ); }

  Manager & manager_;
};

struct Manager_LocaOptionsReg : public IO::PkgOptionsReg
{
  Manager_LocaOptionsReg( Manager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setLocaOptions( options ); }

  Manager & manager_;
};

struct Manager_SensOptionsReg : public IO::PkgOptionsReg
{
  Manager_SensOptionsReg( Manager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setSensOptions( options ); }

  Manager & manager_;
};

struct Manager_SensitivityOptionsReg : public IO::PkgOptionsReg
{
  Manager_SensitivityOptionsReg( Manager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setSensitivityOptions( options ); }

  Manager & manager_;
};

struct Manager_TwoLvlOptionsReg : public IO::PkgOptionsReg
{
  Manager_TwoLvlOptionsReg( Manager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setTwoLevelOptions( options ); }

  Manager & manager_;
};

struct Manager_TwoLvlTranOptionsReg : public IO::PkgOptionsReg
{
  Manager_TwoLvlTranOptionsReg( Manager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setTwoLevelTranOptions( options ); }

  Manager & manager_;
};

struct Manager_DCOPRestartOptionsReg : public IO::PkgOptionsReg
{
  Manager_DCOPRestartOptionsReg( Manager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setDCOPRestartOptions (options); }

  Manager & manager_;
};


struct Manager_ICOptionsReg : public IO::PkgOptionsReg
{
  Manager_ICOptionsReg( Manager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setICOptions (options); }

  Manager & manager_;
};


struct Manager_NodeSetOptionsReg : public IO::PkgOptionsReg
{
  Manager_NodeSetOptionsReg( Manager & manager )
    : manager_(manager)
  {}

  bool operator()( const Util::OptionBlock & options )
  { return manager_.setNodeSetOptions (options); }

  Manager & manager_;
};

} // namespace <unnamed>

//-----------------------------------------------------------------------------
// Function      : registerPkgOptionsMgr
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, 1437
// Creation Date : 10/21/08
//-----------------------------------------------------------------------------
bool registerPkgOptionsMgr(Manager &manager, const std::string netlist_filename, IO::PkgOptionsMgr &options_manager)
{
  options_manager.submitRegistration(
      "NONLIN", netlist_filename, new Manager_OptionsReg( manager ) );

  options_manager.submitRegistration(
      "NONLIN-TRAN", netlist_filename, new Manager_TranOptionsReg( manager ) );

  options_manager.submitRegistration(
      "NONLIN-HB", netlist_filename, new Manager_HBOptionsReg( manager ) );

  options_manager.submitRegistration(
      "LINSOL", netlist_filename, new Manager_LSOptionsReg( manager ) );

  options_manager.submitRegistration(
      "LOCA", netlist_filename, new Manager_LocaOptionsReg( manager ) );

  options_manager.submitRegistration(
      "SENS", netlist_filename, new Manager_SensOptionsReg( manager ) );

  options_manager.submitRegistration(
      "SENSITIVITY", netlist_filename, new Manager_SensitivityOptionsReg( manager ) );

  options_manager.submitRegistration(
      "NONLIN-TWOLEVEL", netlist_filename, new Manager_TwoLvlOptionsReg( manager ) );

  options_manager.submitRegistration(
      "NONLIN-TWOLEVEL-TRAN", netlist_filename, new Manager_TwoLvlTranOptionsReg( manager ) );

  options_manager.submitRegistration(
      "DCOP", netlist_filename, new Manager_DCOPRestartOptionsReg( manager ) );

  options_manager.submitRegistration(
      "IC", netlist_filename, new Manager_ICOptionsReg( manager ) );

  options_manager.submitRegistration(
      "NODESET", netlist_filename, new Manager_NodeSetOptionsReg( manager ) );
  return true;
}

//-----------------------------------------------------------------------------
// Function      : Manager::Manager
// Purpose       : constructor
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 5/01/00
//-----------------------------------------------------------------------------
Manager::Manager(
  const IO::CmdParse & command_line)
  : commandLine_(command_line),
    nonlinearSolver_(0),
    conductanceExtractorPtr_(0),
    nlsSensitivityPtr_(0),
    lasPrecPtr_(0),
    matrixFreeFlag_(false),
    noxFlag_(false),
    noxFlagInner_(false),
    noxFlagTransient_(false),
    optionBlockMap_(),
    initializeAllFlag_(false),
    retCodes_(),
    exprPtr_(0)
{}


//-----------------------------------------------------------------------------
// Function      : Manager::~Manager
// Purpose       : destructor
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 5/01/00
//-----------------------------------------------------------------------------

Manager::~Manager()
{
  delete nonlinearSolver_;
  delete nlsSensitivityPtr_;
  delete conductanceExtractorPtr_;
  delete exprPtr_;
}

//-----------------------------------------------------------------------------
// Function      : Manager::setOptions
// Purpose       : Sets the nonlinear solver options.
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 9/29/00
//-----------------------------------------------------------------------------
bool Manager::setOptions(const Util::OptionBlock & option_block)
{
  optionBlockMap_["dcop"] = option_block;
  return true;
}

//-----------------------------------------------------------------------------
// Function      : Manager::setTranOptions
// Purpose       : Sets the nonlinear solver options.
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 9/05/01
//-----------------------------------------------------------------------------
bool Manager::setTranOptions(const Util::OptionBlock & option_block)
{
  optionBlockMap_["transient"] = option_block;
  return true;
}

//-----------------------------------------------------------------------------
// Function      : Manager::setHBOptions
// Purpose       : Sets the nonlinear solver options.
// Special Notes :
// Scope         : public
// Creator       : Ting Mei, SNL
// Creation Date : 02/03/2009
//-----------------------------------------------------------------------------
bool Manager::setHBOptions(const Util::OptionBlock & option_block)
{
  optionBlockMap_["hb"] = option_block;
  return true;
}

//-----------------------------------------------------------------------------
// Function      : Manager::getHBOptions
// Purpose       : Sets the nonlinear solver options.
// Special Notes :
// Scope         : public
// Creator       : Ting Mei, SNL
// Creation Date : 02/03/2009
//-----------------------------------------------------------------------------
Util::OptionBlock &
Manager::getHBOptions()
{
  return optionBlockMap_["hb"];
}

//-----------------------------------------------------------------------------
// Function      : Manager::setLinSolOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Robert Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 11/9/00
//-----------------------------------------------------------------------------
bool Manager::setLinSolOptions(const Util::OptionBlock & option_block)
{
  optionBlockMap_["petra"] = option_block;
  return true;
}

//-----------------------------------------------------------------------------
// Function      : Manager::setLocaOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 05/02/03
//-----------------------------------------------------------------------------
bool Manager::setLocaOptions(const Util::OptionBlock & option_block)
{
  optionBlockMap_["loca"] = option_block;
  return true;
}

//-----------------------------------------------------------------------------
// Function      : Manager::setTwoLevelLocaOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 05/02/03
//-----------------------------------------------------------------------------
bool Manager::setTwoLevelLocaOptions(const Util::OptionBlock & option_block)
{
  optionBlockMap_["twolevelloca"] = option_block;
  return true;
}

//-----------------------------------------------------------------------------
// Function      : Manager::setTwoLevelOptions
// Purpose       : This option setter will create a new class structure.
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/20/02
//-----------------------------------------------------------------------------
bool Manager::setTwoLevelOptions(const Util::OptionBlock & option_block)
{
  optionBlockMap_["twolevel"] = option_block;

  return true;
}

//-----------------------------------------------------------------------------
// Function      : Manager::setTwoLevelTranOptions
// Purpose       : This option setter will create a new class structure.
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/20/02
//-----------------------------------------------------------------------------
bool Manager::setTwoLevelTranOptions(const Util::OptionBlock & option_block)
{
  optionBlockMap_["twoleveltran"] = option_block;

  return true;
}

//-----------------------------------------------------------------------------
// Function      : setReturnCodeOption
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 04/18/02
//-----------------------------------------------------------------------------
bool
Manager::setReturnCodeOption(
  const Util::Param &           param)
{
  if (param.uTag() == "NLNEARCONV" )
  {
    // nlNearConvFlag = static_cast<bool> (param.getImmutableValue<int>());
    Nonlinear::ReturnCodes codes = getReturnCodes();
    codes.nearConvergence = param.getImmutableValue<int>() ? 3 : -3;
    setReturnCodes(codes);
  }
  else if (param.uTag() == "NLSMALLUPDATE" )
  {
    // nlSmallUpdateFlag = static_cast<bool> (param.getImmutableValue<int>());
    Nonlinear::ReturnCodes codes = getReturnCodes();
    codes.smallUpdate = param.getImmutableValue<int>() ? 4 : -4;
    setReturnCodes(codes);
  }
  else
  {
    return false;
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : Manager::using Nox_
// Purpose       : This function determines if we are using NOX or not.
// Special Notes :
// Scope         : private
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/30/02
//-----------------------------------------------------------------------------
void Manager::usingNox()
{
  noxFlag_ = true;
  noxFlagInner_ = true;
  noxFlagTransient_ = true;

  const Util::OptionBlock &option_blockdcop = optionBlockMap_["dcop"];
  const Util::OptionBlock &option_blocktran = optionBlockMap_["transient"];

  // scan for changes to the dcop values of noxFlag_
  for (Util::ParamList::const_iterator it = option_blockdcop.begin(); it != option_blockdcop.end(); ++ it)
  {
    if ((*it).uTag() == "NOX")
    {
      noxFlag_ = (*it).getImmutableValue<int>();
      noxFlagInner_ = noxFlag_;
    }
  }

  // now check for a nox flag on the .options twolevel line, if it exists.
  OptionBlockMap::const_iterator obmIter = optionBlockMap_.find("twolevel") ;
  if ( obmIter != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_blocktwoLevel = obmIter->second;
    for (Util::ParamList::const_iterator it = option_blocktwoLevel.begin(); it != option_blocktwoLevel.end(); ++ it)
    {
      if ((*it).uTag() == "NOX")
      {
        noxFlagInner_ = (*it).getImmutableValue<int>();
      }
    }
  }


  // scan for changes to the transient value for noxFlagTransient_
  for (Util::ParamList::const_iterator it = option_blocktran.begin(); it != option_blocktran.end(); ++ it)
  {
    if ((*it).uTag() == "NOX")
    {
      noxFlagTransient_ = (*it).getImmutableValue<int>();
    }
  }

  obmIter = optionBlockMap_.find("hb") ;
  if ( obmIter != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_blockhb = obmIter->second;
    for (Util::ParamList::const_iterator it = option_blockhb.begin(); it != option_blockhb.end(); ++ it)
    {
      if ((*it).uTag() == "NOX")
      {
        noxFlag_ = (*it).getImmutableValue<int>();
        noxFlagInner_ = noxFlag_;
      }
    }
  }

  // now check if the command line has specified nox.  The command line
  // overrides the netlist.
  if( commandLine_.getArgumentValue( "-nox" ) == "off" )
  {
    noxFlag_ = false;
    noxFlagInner_ = false;
    noxFlagTransient_ = false;
  }

  if (DEBUG_NONLINEAR)
    Xyce::dout() << "noxFlag is: " << (noxFlag_ ? "true" : "false") << std::endl
                 << "noxFlagTransient is: " <<  (noxFlagTransient_ ? "true" : "false") << std::endl;
}

//-----------------------------------------------------------------------------
// Function      : Manager::allocateSolver
// Purpose       : This function determines which solver to allocate, and
//                 allocates it.
//
//                 Right now the possibilities are:
//
//                    DampedNewton
//                    NOXInterface
//                    TwoLevelNewton
//
// Special Notes :
// Scope         : private
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/28/02
//-----------------------------------------------------------------------------
bool
Manager::allocateSolver(
  Analysis::AnalysisManager &           analysis_manager,
  Loader::NonlinearEquationLoader &     nonlinear_equation_loader,
  Linear::System &                      linear_system,
  TimeIntg::DataStore &                 data_store,
  Parallel::Manager &                   parallel_manager,
  IO::InitialConditionsManager &        initial_conditions_manager,
  IO::OutputMgr &                       output_manager)
{
  bool bsuccess = true;
  bool bs1 = true;

  // determine if we are using NOX or not.
  usingNox();

  // if ".options nonlin two-level" appeared in the netlist, then
  // allocate the two-level solver.  Otherwise allocate one of the single
  // level solvers.

  if (optionBlockMap_.find("twolevel") != optionBlockMap_.end()
      || optionBlockMap_.find("twoleveltran") != optionBlockMap_.end())
  {
    twoLevelNewtonFlag_ = true;
    nonlinearSolver_ = new TwoLevelNewton(noxFlag_, noxFlagInner_, commandLine_);
  }
  else
  {
    twoLevelNewtonFlag_ = false;
    if (noxFlag_)
    {
      delete nonlinearSolver_;
      nonlinearSolver_ = new N_NLS_NOX::Interface(commandLine_);
    }
    else
    {
      delete nonlinearSolver_;
      nonlinearSolver_ = new DampedNewton(commandLine_);
    }
  }

  // now register everything, now that the solver class is set up.
  bs1 = nonlinearSolver_->registerLinearSystem(&linear_system);    bsuccess = bsuccess && bs1;
  bs1 = nonlinearSolver_->registerAnalysisManager(&analysis_manager); bsuccess = bsuccess && bs1;
  bs1 = nonlinearSolver_->registerLoader(&nonlinear_equation_loader);          bsuccess = bsuccess && bs1;
  bs1 = nonlinearSolver_->registerTIADataStore(&data_store);        bsuccess = bsuccess && bs1;

  if (lasPrecPtr_)
    bs1 = nonlinearSolver_->registerPrecondFactory(lasPrecPtr_); bsuccess = bsuccess && bs1;

  bs1 = nonlinearSolver_->registerOutputMgr(&output_manager); bsuccess = bsuccess && bs1;
  bs1 = nonlinearSolver_->registerInitialConditionsManager(&initial_conditions_manager); bsuccess = bsuccess && bs1;
  bs1 = nonlinearSolver_->registerParallelMgr(&parallel_manager); bsuccess = bsuccess && bs1;

  OptionBlockMap::const_iterator it = optionBlockMap_.find("dcop") ;
  if ( it != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_block = (*it).second;
    bs1 = nonlinearSolver_->setOptions (option_block);
    bsuccess = bsuccess && bs1;
  }

  it = optionBlockMap_.find("transient") ;
  if ( it != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_block = (*it).second;
    bs1 = nonlinearSolver_->setTranOptions (option_block);
    bsuccess = bsuccess && bs1;
  }

  it = optionBlockMap_.find("hb") ;
  if ( it != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_block = (*it).second;
    bs1 = nonlinearSolver_->setHBOptions (option_block);
    bsuccess = bsuccess && bs1;
  }

  it = optionBlockMap_.find("loca") ;
  if ( it != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_block = (*it).second;
    bs1 = nonlinearSolver_->setLocaOptions (option_block);
    bsuccess = bsuccess && bs1;
  }

  it = optionBlockMap_.find("twolevel") ;
  if ( it != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_block = (*it).second;
    bs1 = nonlinearSolver_->setTwoLevelOptions (option_block);
    bsuccess = bsuccess && bs1;
  }

  it = optionBlockMap_.find("twoleveltran") ;
  if ( it != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_block = (*it).second;
    bs1 = nonlinearSolver_->setTwoLevelTranOptions (option_block);
    bsuccess = bsuccess && bs1;
  }

  it = optionBlockMap_.find("twolevelloca") ;
  if ( it != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_block = (*it).second;
    bs1 = nonlinearSolver_->setTwoLevelLocaOptions (option_block);
    bsuccess = bsuccess && bs1;
  }

  it = optionBlockMap_.find("petra") ;
  if ( it != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_block = (*it).second;
    bs1 = nonlinearSolver_->setPetraOptions(option_block);
    bsuccess = bsuccess && bs1;
  }

  it = optionBlockMap_.find("dcop_restart") ;
  if ( it != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_block = (*it).second;
    bs1 = nonlinearSolver_->setDCOPRestartOptions(option_block);
    bsuccess = bsuccess && bs1;
  }

  it = optionBlockMap_.find("ic") ;
  if ( it != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_block = (*it).second;
    bs1 = nonlinearSolver_->setICOptions(option_block);
    bsuccess = bsuccess && bs1;
  }

  it = optionBlockMap_.find("nodeset") ;
  if ( it != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_block = (*it).second;
    bs1 = nonlinearSolver_->setNodeSetOptions(option_block);
    bsuccess = bsuccess && bs1;
  }

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : Manager::initializeAll
// Purpose       : This can only be called after the linear system
//                 (Linear::System) has been registered, and after all the
//                 options have been set.
//
// Special Notes : This function obtains the solution, temporary solution and
//                 rhs vectors from the LAS system class.
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/12/00
//-----------------------------------------------------------------------------
bool
Manager::initializeAll(
  Analysis::AnalysisManager &           analysis_manager,
  Loader::NonlinearEquationLoader &     nonlinear_equation_loader,
  Linear::System &                      linear_system,
  TimeIntg::DataStore &                 data_store,
  Parallel::Manager &                   parallel_manager,
  IO::InitialConditionsManager &        initial_conditions_manager,
  IO::OutputMgr &                       output_manager,
  Topo::Topology &                      topology)
{
  bool bsuccess = true;
  bool bs1 = true;
  bool bs2 = true;

  bs1 = allocateSolver(analysis_manager, nonlinear_equation_loader, linear_system, data_store, parallel_manager, initial_conditions_manager, output_manager);
  bsuccess = bsuccess && bs1;

  nonlinearSolver_->setMatrixFreeFlag(matrixFreeFlag_);
  bs2 = nonlinearSolver_->initializeAll();  bsuccess = bsuccess && bs2;

  nonlinearSolver_->setReturnCodes(retCodes_);

  initializeAllFlag_ = true;

  if (!conductanceExtractorPtr_)
  {
    conductanceExtractorPtr_ = new ConductanceExtractor(*nonlinearSolver_, topology);
  }

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : Manager::allocateTranSolver
// Purpose       : Allocate a different solver for transient problems if
//                 the user has requested a different one.
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and Microsystems Modeling
// Creation Date : 12/22/08
//-----------------------------------------------------------------------------
void
Manager::allocateTranSolver(
  Analysis::AnalysisManager &           analysis_manager,
  Loader::NonlinearEquationLoader &     nonlinear_equation_loader,
  Linear::System &                      linear_system,
  TimeIntg::DataStore &                 data_store,
  Parallel::Manager &                   parallel_manager,
  IO::OutputMgr &                       output_manager,
  Topo::Topology &                      topology)
{
  bool bsuccess = true;
  bool bs1 = true;

  // only do a reallocation if the the solver type is changing from dcop to transient
  if ((noxFlag_ && !noxFlagTransient_) || (!noxFlag_&& noxFlagTransient_))
  {
    if (noxFlagTransient_)
    {
      delete nonlinearSolver_;
      nonlinearSolver_ = new N_NLS_NOX::Interface(commandLine_);
    }
    else
    {
      delete nonlinearSolver_;
      nonlinearSolver_ = new DampedNewton(commandLine_);
    }

    // now register everything, now that the solver class is set up.
    nonlinearSolver_->registerLinearSystem(&linear_system);
    nonlinearSolver_->registerAnalysisManager(&analysis_manager);
    nonlinearSolver_->registerLoader(&nonlinear_equation_loader);
    nonlinearSolver_->registerTIADataStore(&data_store);
    nonlinearSolver_->registerOutputMgr(&output_manager);
    nonlinearSolver_->registerParallelMgr(&parallel_manager);

    nonlinearSolver_->setMatrixFreeFlag(matrixFreeFlag_);
    nonlinearSolver_->initializeAll();
    nonlinearSolver_->setReturnCodes(retCodes_);

    OptionBlockMap::const_iterator it = optionBlockMap_.find("transient") ;
    if (it != optionBlockMap_.end())
    {
      const Util::OptionBlock &option_block = (*it).second;
      bs1 = nonlinearSolver_->setTranOptions (option_block);
      bsuccess = bsuccess && bs1;
    }

    initializeAllFlag_ = true;

    delete conductanceExtractorPtr_;
    conductanceExtractorPtr_ = new ConductanceExtractor (*nonlinearSolver_, topology);
  }
}

//-----------------------------------------------------------------------------
// Function      : Manager::getNonLinInfo
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, Parallel Computational Sciences
// Creation Date : 12/05/02
//-----------------------------------------------------------------------------
NonLinInfo Manager::getNonLinInfo() const
{
  NonLinInfo nlInfo;

  nlInfo.newtonIter    = nonlinearSolver_->getNumIterations();
  nlInfo.twoLevelNewtonCouplingMode  = nonlinearSolver_->getCouplingMode ();
  nlInfo.locaFlag      = nonlinearSolver_->getLocaFlag ();

  if (nlInfo.locaFlag)
  {
    nlInfo.continuationStep       = nonlinearSolver_->getContinuationStep ();
    nlInfo.firstContinuationParam = nonlinearSolver_->isFirstContinuationParam ();
    nlInfo.firstSolveComplete     = nonlinearSolver_->isFirstSolveComplete ();
  }

  return nlInfo;
}

//-----------------------------------------------------------------------------
// Function      : solve
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Tamara Kolda, SNL, CSMR
// Creation Date : 1/30/02
//-----------------------------------------------------------------------------
int Manager::solve()
{
  int status = 0;

  status = nonlinearSolver_->solve();
  if (status >= 0)
  {
    if (!exprPtr_)
      exprPtr_ = new Util::Expression(std::string("0"));

    exprPtr_->set_accepted_time();
  }

  return status;
}

//-----------------------------------------------------------------------------
// Function      : setAnalysisMode
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Tamara Kolda, SNL, CSMR
// Creation Date : 1/30/02
//-----------------------------------------------------------------------------
void Manager::setAnalysisMode(AnalysisMode mode)
{
  nonlinearSolver_->setAnalysisMode(mode);
}

//-----------------------------------------------------------------------------
// Function      : resetAll
// Purpose       : like setAnalysisMode, but will also result in NOX
//                 resetting a few extra things.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, 9233
// Creation Date : 1/30/02
//-----------------------------------------------------------------------------
void Manager::resetAll(AnalysisMode mode)
{
  nonlinearSolver_->resetAll (mode);
}

//-----------------------------------------------------------------------------
// Function      : Manager::setSensOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 11/18/02
//-----------------------------------------------------------------------------
bool Manager::setSensOptions (const Util::OptionBlock & option_block)
{
  optionBlockMap_["sens"] = option_block;
  return true;
}

//-----------------------------------------------------------------------------
// Function      : Manager::setSensitivityOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 11/18/02
//-----------------------------------------------------------------------------
bool Manager::setSensitivityOptions (const Util::OptionBlock & option_block)
{
  optionBlockMap_["sensitivity"] = option_block;
  return true;
}

//-----------------------------------------------------------------------------
// Function      : Manager::enableSensitivity
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 04/30/03
//-----------------------------------------------------------------------------
bool
Manager::enableSensitivity(
  TimeIntg::DataStore & data_store,  
  Parallel::Manager &   parallel_manager,
  Topo::Topology &      topology)
{
  bool bsuccess = true;

  if (!nlsSensitivityPtr_)
  {
    Stats::StatTop _sensitivityStat("Setup");

    bool b1 = setupSensitivity(data_store, parallel_manager, topology);
    bsuccess = bsuccess && b1;
  }
  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : Manager::icSensitivity
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL
// Creation Date :
//-----------------------------------------------------------------------------
bool Manager::icSensitivity (
  std::vector<double> & objectiveVec,
  std::vector<double> & dOdpVec,
  std::vector<double> & dOdpAdjVec,
  std::vector<double> & scaled_dOdpVec,
  std::vector<double> & scaled_dOdpAdjVec)
{
  if (!nlsSensitivityPtr_)
  {
    Report::DevelFatal0().in("Manager::isSensitivity") <<  "Manager::enableSensitivity must be called first";
    return false;
  }

  return nlsSensitivityPtr_->icSensitivity(objectiveVec, dOdpVec, dOdpAdjVec, scaled_dOdpVec, scaled_dOdpAdjVec);
}

//-----------------------------------------------------------------------------
// Function      : Manager::calcSensitivity
// Purpose       : This is the controller function for performing a direct
//                 sensitivity calculation.  It is generally called from
//                 the time integration package, as only that package really
//                 knows when to do it... (I may change this later.)
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/31/02
//-----------------------------------------------------------------------------
bool Manager::calcSensitivity
  (
   std::vector<double> & objectiveVec,
   std::vector<double> & dOdpVec,
   std::vector<double> & dOdpAdjVec,
   std::vector<double> & scaled_dOdpVec,
   std::vector<double> & scaled_dOdpAdjVec)
{
  bool bsuccess = true;

  if (!nlsSensitivityPtr_)
  {
    Report::DevelFatal0().in("Manager::calcSensitivity") <<  "Manager::enableSensitivity must be called first";
    return false;
  }

  bsuccess = nlsSensitivityPtr_->solve(objectiveVec, dOdpVec, dOdpAdjVec, scaled_dOdpVec, scaled_dOdpAdjVec);

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : Manager::setupSensitivity
// Purpose       :
// Special Notes :
// Scope         : private
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 10/31/02
//-----------------------------------------------------------------------------
bool
Manager::setupSensitivity(
  TimeIntg::DataStore & data_store,  
  Parallel::Manager &   parallel_manager,
  Topo::Topology &      topology)
{
  if (nlsSensitivityPtr_)
  {
    Report::DevelFatal0().in("Manager::setupSensitivity") <<  "Manager::enableSensitivity may only be called once";
    return false;
  }

  bool bsuccess = true;
  bool bs1 = true;

  nlsSensitivityPtr_ = new Sensitivity(*nonlinearSolver_, topology, commandLine_);
  bs1 = nlsSensitivityPtr_->registerParallelMgr(&parallel_manager); bsuccess = bsuccess && bs1;
  bs1 = nlsSensitivityPtr_->registerTIADataStore(&data_store); bsuccess = bsuccess && bs1;

  OptionBlockMap::const_iterator it = optionBlockMap_.find("sens") ;
  if ( it != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_block = it->second;
    bs1 = nlsSensitivityPtr_->setOptions(option_block);
    bsuccess = bsuccess && bs1;
  }

  it = optionBlockMap_.find("sensitivity") ;
  if ( it != optionBlockMap_.end() )
  {
    const Util::OptionBlock &option_block = it->second;
    bs1 = nlsSensitivityPtr_->setSensitivityOptions(option_block);
    bsuccess = bsuccess && bs1;
  }

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : Manager::setReturnCodes
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 7/02/03
//-----------------------------------------------------------------------------
void Manager::setReturnCodes(const ReturnCodes & retCodeTmp)
{
  retCodes_ = retCodeTmp;
  if (initializeAllFlag_)
  {
    nonlinearSolver_->setReturnCodes(retCodes_);
  }
}

//-----------------------------------------------------------------------------
// Function      : Manager::getReturnCodes
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rich Schiek, Electrical and MEMS Modeling
// Creation Date : 9/28/2009
//-----------------------------------------------------------------------------
const ReturnCodes &Manager::getReturnCodes() const
{
  // the ReturnCodes structure is very simple, so just
  // return a copy of it.
  return retCodes_;
}

//-----------------------------------------------------------------------------
// Function      : Manager::setDCOPRestartOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 09/17/07
//-----------------------------------------------------------------------------
bool Manager::setDCOPRestartOptions (const Util::OptionBlock& option_block )
{
  optionBlockMap_["dcop_restart"] = option_block;
  return true;
}

//-----------------------------------------------------------------------------
// Function      : Manager::setICOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 09/17/07
//-----------------------------------------------------------------------------
bool Manager::setICOptions (const Util::OptionBlock& option_block )
{
  optionBlockMap_["ic"] = option_block;
  return true;
}

//-----------------------------------------------------------------------------
// Function      : Manager::setNodeSetOptions
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 09/25/07
//-----------------------------------------------------------------------------
bool Manager::setNodeSetOptions (const Util::OptionBlock& option_block )
{
  optionBlockMap_["nodeset"] = option_block;
  return true;
}

} // namespace Nonlinear
} // namespace Xyce
