//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_TIA_DataStore.h,v $
//
// Purpose       : This file handles the class that defines the data arrays
//                 needed for the time integration algorithms.
//
// Special Notes :
//
// Creator       : Buddy Watts
//
// Creation Date : 6/1/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.110.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:18 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#ifndef Xyce_N_TIA_DataStore_h
#define Xyce_N_TIA_DataStore_h

// ---------- Standard Includes ----------
#include <list>
#include <vector>

// ----------   Xyce Includes   ----------
#include <N_LAS_fwd.h>
#include <N_TIA_fwd.h>

#include <N_TIA_TwoLevelError.h>
#include <N_UTL_IndexPair.h>

namespace Xyce {
namespace TimeIntg {

//-----------------------------------------------------------------------------
// Class         : DataStore
// Purpose       : This is the class for defining data arrays needed in the
//                 time integration algorithms.
// Special Notes :
// Creator       : Buddy Watts, SNL
// Creation Date : 6/01/00
//-----------------------------------------------------------------------------
class DataStore
{
  public:
  DataStore(int solution_size, int state_size, int max_order, const Linear::System &linear_system, const Linear::Builder &linear_system_builder);
    ~DataStore();

    void allocateSensitivityArrays(Linear::Builder &builder, int numParams);
    void deleteSensitivityArrays();

  private:
    DataStore(const DataStore& rhs);
    DataStore &operator=(const DataStore& rhs);

  // DataStore Functions
  public:
    void initializeDataArrays();
    void enableOrderOneStart();

    void updateSolDataArrays();
    bool updateStateDataArrays();

    void outputSolDataArrays(std::ostream &os);
    void setConstantHistory();
    void setConstantSensitivityHistory();
    void setZeroHistory();
    void setErrorWtVector(const TIAParams &tia_params);
    double WRMS_errorNorm();

    double partialErrorNormSum();
    double partialQErrorNormSum();

    double partialSum_m1(int currentOrder);
    double partialSum_m2(int currentOrder);

    double globalLength ();
    void computeDividedDifferences();

    void computeDivDiffsBlock(const std::list<IndexPair> & solGIDList,
                              const std::list<IndexPair> & staGIDList);

    void printOutPointers ();
    bool equateTmpVectors ();
    bool usePreviousSolAsPredictor ();

    void outputPredictedSolution(std::ostream &os);
    void outputPredictedDerivative(std::ostream &os);

    void stepLinearCombo ();

    double partialSum_p1(int currentOrder, int maxOrder);
    double partialSum_q1();

    double delta_x_errorNorm_m1();
    double delta_x_errorNorm_m2();
    double delta_x_errorNorm_p1();
    double delta_x_errorNorm_q1();

    bool getSolnVarData( const int & gid, std::vector<double> & varData );
    bool getStateVarData( const int & gid, std::vector<double> & varData );
    bool setSolnVarData( const int & gid, const std::vector<double> & varData );
    bool setStateVarData( const int & gid, const std::vector<double> & varData );
    bool getStoreVarData( const int & gid, std::vector<double> & varData );
    bool setStoreVarData( const int & gid, const std::vector<double> & varData );

    bool setNextSolVectorPtr (Linear::Vector * solVecPtr);
    bool unsetNextSolVectorPtr ();

  bool resetAll(double absolute_error_tolerance, double relative_error_tolerance);

  bool resetFastTimeData ();

private:
  bool                  nextSolPtrSwitched_;

  public:
  const Linear::System &        linearSystem_;
  
  // TIA Arrays (pointers) for  Integration Solution Process:
    unsigned int maxOrder;
    unsigned int solutionSize;
    unsigned int stateSize;

    // temporary vectors:
    Linear::Vector * tmpSolVectorPtr;
    Linear::Vector * tmpStaVectorPtr;
    Linear::Vector * tmpStaDerivPtr;
    Linear::Vector * tmpLeadCurrentVectorPtr;
    Linear::Vector * tmpLeadDeltaVPtr;
    Linear::Vector * tmpStaDivDiffPtr;

    Linear::Vector * tmpStoVectorPtr;

    // Predictors
    Linear::Vector * xn0Ptr;

    // Solutions:
    Linear::Vector * currSolutionPtr;
    Linear::Vector * lastSolutionPtr;
    Linear::Vector * oldeSolutionPtr;
    Linear::Vector * nextSolutionPtr;
    Linear::Vector * flagSolutionPtr;

    Linear::Vector * savedNextSolutionPtr;

    // States:
    Linear::Vector * currStatePtr;
    Linear::Vector * lastStatePtr;
    Linear::Vector * oldeStatePtr;
    Linear::Vector * nextStatePtr;

    // Storage:
    Linear::Vector * currStorePtr;
    Linear::Vector * lastStorePtr;
    Linear::Vector * oldeStorePtr;
    Linear::Vector * nextStorePtr;
    
    // for lead current calculations.  F component is
    // held in the store vector, Q component is here
    // Note: These will be removed when all lead current calc's 
    // are moved to their own vector.
    Linear::Vector * currStoreLeadCurrQPtr;
    Linear::Vector * lastStoreLeadCurrQPtr;
    Linear::Vector * oldeStoreLeadCurrQPtr;
    Linear::Vector * nextStoreLeadCurrQPtr;
    
    // Lead current and power vectors 
    Linear::Vector * currLeadCurrentPtr;
    Linear::Vector * lastLeadCurrentPtr;
    Linear::Vector * oldLeadCurrentPtr;
    Linear::Vector * nextLeadCurrentPtr;
    
    Linear::Vector * currLeadDeltaVPtr;
    Linear::Vector * lastLeadDeltaVPtr;
    Linear::Vector * oldLeadDeltaVPtr;
    Linear::Vector * nextLeadDeltaVPtr;
    
    // for lead current calculations.  F component is
    // held in the store vector, Q component is here
    Linear::Vector * currLeadCurrentQPtr;
    Linear::Vector * lastLeadCurrentQPtr;
    Linear::Vector * oldLeadCurrentQPtr;
    Linear::Vector * nextLeadCurrentQPtr;    

    // sensitivity vectors:
    std::vector<Linear::Vector*> sensRHSPtrVector;

    std::vector<Linear::Vector*> nextDfdpPtrVector;

    std::vector<Linear::Vector*> currDqdpPtrVector;
    std::vector<Linear::Vector*> lastDqdpPtrVector;
    std::vector<Linear::Vector*> oldeDqdpPtrVector;
    std::vector<Linear::Vector*> nextDqdpPtrVector;

    std::vector<Linear::Vector*> nextDbdpPtrVector;

    std::vector<Linear::Vector*> currDXdpPtrVector;
    std::vector<Linear::Vector*> lastDXdpPtrVector;
    std::vector<Linear::Vector*> oldeDXdpPtrVector;
    std::vector<Linear::Vector*> nextDXdpPtrVector;

    // matvecs:
    std::vector<Linear::Vector*> currDQdxDXdpPtrVector;
    std::vector<Linear::Vector*> lastDQdxDXdpPtrVector;
    std::vector<Linear::Vector*> oldeDQdxDXdpPtrVector;
    std::vector<Linear::Vector*> nextDQdxDXdpPtrVector;

    std::vector<Linear::Vector*> currDQdxDXdpDerivPtrVector;
    std::vector<Linear::Vector*> lastDQdxDXdpDerivPtrVector;
    std::vector<Linear::Vector*> oldeDQdxDXdpDerivPtrVector;
    std::vector<Linear::Vector*> nextDQdxDXdpDerivPtrVector;

    // Derivatives of Solutions:
    Linear::Vector * currSolutionDerivPtr;
    Linear::Vector * lastSolutionDerivPtr;
    Linear::Vector * oldeSolutionDerivPtr;
    Linear::Vector * nextSolutionDerivPtr;

    // Derivatives of States:
    Linear::Vector * currStateDerivPtr;
    Linear::Vector * lastStateDerivPtr;
    Linear::Vector * oldeStateDerivPtr;
    Linear::Vector * nextStateDerivPtr;

    // Derivatives of Store for lead curent calculations
    // Note: These will be removed when all lead current calc's 
    // are moved to their own vector.
    Linear::Vector * currStoreLeadCurrQDerivPtr;
    Linear::Vector * lastStoreLeadCurrQDerivPtr;
    Linear::Vector * oldeStoreLeadCurrQDerivPtr;
    Linear::Vector * nextStoreLeadCurrQDerivPtr;
    
    // Derivatives of Store for lead curent calculations
    Linear::Vector * currLeadCurrentQDerivPtr;
    Linear::Vector * lastLeadCurrentQDerivPtr;
    Linear::Vector * oldLeadCurrentQDerivPtr;
    Linear::Vector * nextLeadCurrentQDerivPtr;


    // Derivatives of dq/dp for sensitivity calculations
    std::vector<Linear::Vector*> currDqdpDerivPtrVector;
    std::vector<Linear::Vector*> lastDqdpDerivPtrVector;
    std::vector<Linear::Vector*> oldeDqdpDerivPtrVector;
    std::vector<Linear::Vector*> nextDqdpDerivPtrVector;

    // Sensitivity residual matvec term:
    Linear::Vector * resMatVecPtr;

    // Scaled Divided Differences
    Linear::Vector * currSolutionDivDiffPtr;
    Linear::Vector * lastSolutionDivDiffPtr;
    Linear::Vector * oldeSolutionDivDiffPtr;
    Linear::Vector * nextSolutionDivDiffPtr;

    // Scaled Divided Differences
    Linear::Vector * currStateDivDiffPtr;
    Linear::Vector * lastStateDivDiffPtr;
    Linear::Vector * oldeStateDivDiffPtr;
    Linear::Vector * nextStateDivDiffPtr;

    // Error Vectors
    Linear::Vector * errWtVecPtr;
    Linear::Vector * absErrTolPtr;
    Linear::Vector * relErrTolPtr;

    // Jacobian and RHS
    Linear::Matrix * JMatrixPtr;
    Linear::Vector * RHSVectorPtr;
    Linear::Vector * JdxpVectorPtr; // only used in the device manager for debug purposes these days.

    // NonLinear Solution Vectors
    Linear::Vector * newtonCorrectionPtr;

    // Mask for error norms (to allow some equations not to take part in
    // weighted norms)
    Linear::Vector * deviceErrorWeightMask_;

  // limiter flag:
    bool limiterFlag;

    // 2-level information:
    std::vector<TwoLevelError> innerErrorInfoVec;

    // new-DAE data (originally from the new-DAE derrived class)
    // Error Vectors
    Linear::Vector * qErrWtVecPtr;

    // DAE formulation vectors
    Linear::Vector * daeQVectorPtr;
    Linear::Vector * daeFVectorPtr;
    Linear::Vector * daeBVectorPtr;

    // voltage limiting vectors
    Linear::Vector * dFdxdVpVectorPtr;
    Linear::Vector * dQdxdVpVectorPtr;

    // DAE formulation matrices
    Linear::Matrix * dQdxMatrixPtr;
    Linear::Matrix * dFdxMatrixPtr;

    // temporary stop-gap, needed for sensitivity analysis.
    //Linear::Matrix * saved_dQdxMatrixPtr;

    // HB temporary Matvec storage vectors
    Linear::Vector * dQdxVecVectorPtr;
    Linear::Vector * dFdxVecVectorPtr;

    // History arrays
    std::vector<Linear::Vector*> xHistory;
    std::vector<Linear::Vector*> qHistory;
    std::vector<Linear::Vector*> sHistory;    // state history
    std::vector<Linear::Vector*> stoHistory;  // store history
    std::vector<Linear::Vector*> stoLeadCurrQHistory;  // store history for lead current Q component.
    std::vector<Linear::Vector*> leadCurrentHistory;  // history for lead current Q component.
    std::vector<Linear::Vector*> leadCurrentQHistory;  // history for lead current Q component.
    std::vector<Linear::Vector*> leadDeltaVHistory;  // history for junction voltage
    
    // sensitivity histories
    std::vector< std::vector<Linear::Vector*> > dfdpHistory;
    std::vector< std::vector<Linear::Vector*> > dqdpHistory;
    std::vector< std::vector<Linear::Vector*> > dbdpHistory;
    std::vector< std::vector<Linear::Vector*> > dXdpHistory;

    std::vector< std::vector<Linear::Vector*> > dQdxdXdpHistory;

    // Predictors
    Linear::Vector * qn0Ptr;
    Linear::Vector * qpn0Ptr;

    Linear::Vector * sn0Ptr;
    Linear::Vector * spn0Ptr;

    Linear::Vector * ston0Ptr;
    Linear::Vector * stopn0Ptr;

    Linear::Vector * stoQn0Ptr;
    Linear::Vector * stoQpn0Ptr;

    Linear::Vector * leadCurrentn0Ptr;
    Linear::Vector * leadCurrentpn0Ptr;

    Linear::Vector * leadCurrentQn0Ptr;
    Linear::Vector * leadCurrentQpn0Ptr;
    Linear::Vector * leadDeltaVn0Ptr;
    Linear::Vector * leadDeltaVpn0Ptr;


    // Nonlinear solution vector:
    Linear::Vector * qNewtonCorrectionPtr;
    Linear::Vector * sNewtonCorrectionPtr;
    Linear::Vector * stoNewtonCorrectionPtr;
    Linear::Vector * stoLeadCurrQNewtonCorrectionPtr;
    Linear::Vector * leadCurrentNewtonCorrectionPtr;
    Linear::Vector * leadCurrentQNewtonCorrectionPtr;
    Linear::Vector * leadDeltaVNewtonCorrectionPtr;

    // Step-size selection temporary vectors
    Linear::Vector * delta_x;
    Linear::Vector * delta_q;

    // Temporary vectors for WaMPDE interpolation
    Linear::Vector * tmpXn0APtr;
    Linear::Vector * tmpXn0BPtr;

    // These are for MPDE fast time scale points
    std::vector<double> timeSteps;
    std::vector<bool> timeStepsBreakpointFlag;
    std::vector<Linear::Vector*> fastTimeSolutionVec;
    std::vector<Linear::Vector*> fastTimeStateVec;
    std::vector<Linear::Vector*> fastTimeQVec;
    std::vector<Linear::Vector*> fastTimeStoreVec;

    std::vector<double> objectiveVec_;
    std::vector<double> dOdpVec_;
    std::vector<double> dOdpAdjVec_;
    std::vector<double> scaled_dOdpVec_;
    std::vector<double> scaled_dOdpAdjVec_;
    std::vector<double> paramOrigVals_; 

private:
    std::vector<int> indexVVars;
    std::vector<int> indexMaskedVars;  
};

} // namespace TimeIntg
} // namespace Xyce

#endif // Xyce_N_TIA_DataStore_h

