//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_TIA_Gear12.h,v $
//
// Purpose       : This file defines the classes for the backward
//                 differentiation, order 1-2 method.
//
// Special Notes :
//
// Creator       : Ting Mei
//
// Creation Date : 10/31/07
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.25.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:18 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#ifndef Xyce_N_TIA_GEAR12_H
#define Xyce_N_TIA_GEAR12_H

// ---------- Standard Includes ----------

// ----------   Xyce Includes   ----------
#include <N_TIA_DataStore.h>
#include <N_TIA_StepErrorControl.h>
#include <N_TIA_TimeIntegrationMethods.h>
#include <N_PDS_ParMap.h>
#include <N_UTL_Math.h>
#include <N_UTL_IndexPair.h>

namespace Xyce {
namespace TimeIntg {

//-----------------------------------------------------------------------------
// Class         : Gear12
// Purpose       : Gear12 Formula Integration Class
//                (derived from TimeIntegrationMethod)
// Special Notes :
// Creator       : Ting Mei, SNL
// Creation Date : 10/31/07
//-----------------------------------------------------------------------------
class Gear12 : public TimeIntegrationMethod
{
public:
  static const int type = 8;
  static const char *name;

  static TimeIntegrationMethod *factory(const TIAParams &tia_params, StepErrorControl &step_error_control, DataStore &data_store);

  Gear12(
    const TIAParams &   tia_params,
    StepErrorControl &  step_error_control,
    DataStore &         data_store);

  ~Gear12()
  {}

  const char *getName() const /* override */ {
    return "Gear 12";
  }

  // Predict solution at next time point.
  void obtainPredictor() /* override */;

  // Evaluate the predictor derivative formula.
  void obtainPredictorDeriv()  /* override */
  {}

  // Evaluate the corrector derivative formula.
  void obtainCorrectorDeriv() /* override */
  {}

  // Compute an estimate of the error in the integration step.
  void updateDerivsBlock(const std::list< IndexPair > & solGIDList, const std::list< IndexPair > & staGIDList) /* override */
  {}

  // Compute an estimate of the error in the integration step.
  double computeErrorEstimate() const /* override */
  {
    return sec.ck_*ds.WRMS_errorNorm();
  }

  // Interpolate solution approximation at prescribed time point.
  bool interpolateSolution(double timepoint, Linear::Vector * tmpSolVectorPtr, std::vector<Linear::Vector*> & historyVec) /* override */;

  // Print output using interpolation when order is high
  bool printOutputSolution(
    Analysis::OutputMgrAdapter &        outputManagerAdapter,
    const TIAParams &                   tia_params,
    const double                        time,
    Linear::Vector *                      solnVecPtr,
    const bool                          doNotInterpolate,
    const std::vector<double> &         outputInterpolationTimes,
    bool                                skipPrintLineOutput) /* override */;

  // Print transient output from MPDE simulation
  bool printMPDEOutputSolution(
    Analysis::OutputMgrAdapter &        outputManagerAdapter,
    const double                        time,
    Linear::Vector *                      solnVecPtr,
    const std::vector<double> &         fastTimes );  /* override */

  // Print transient output from WaMPDE simulation
  bool printWaMPDEOutputSolution(
    Analysis::OutputMgrAdapter &        outputManagerAdapter,
    const double                        time,
    Linear::Vector *                      solnVecPtr,
    const std::vector<double> &         fastTimes,
    const int                           phiGID) /* override */;

  // .SAVE output using interpolation when order is high
  bool saveOutputSolution(
    Analysis::OutputMgrAdapter &        outputManagerAdapter,
    const TIAParams &                   tia_params,
    Linear::Vector *                      solnVecPtr,
    const double                        saveTime,
    const bool                          doNotInterpolate) /* override */; 

  // Computes the step adjustment.
  double computeExpoStepAdjust(double stepadjust) /* override */;

  // Gets the time-integration order.
  int getOrder() const /* override */
  {
    return sec.currentOrder_;
  }

  int getUsedOrder() const /* override */
  {
    return sec.usedOrder_;
  }
  
  int getNumberOfSteps() const /* override */
  {
    return sec.numberOfSteps_;
  }

  int getNscsco() const /* override */
  {
    return sec.nscsco_;
  }

  int getMaxOrder() const /* override */
  {
    return sec.maxOrder_;
  }

  void getInitialQnorm(TwoLevelError & tle) const /* override */;

  void getTwoLevelError(TwoLevelError & tle) const /* override */;

  void updateStateDeriv() /* override */;

  // calculates dQ/dt component of store vector and adds it to store vector
  void updateLeadCurrent() /* override */;

  // calculates dQ/dt component of lead current Q vector and adds it to the lead current vector
  void updateLeadCurrentVec() /* override */;

  // Returns the current partial time derivative for either the solution or
  // state vector.
  double partialTimeDeriv() const /* override */;

  // Gets the leading coefficient for the specified time-integration method.
  double getLeadingCoeff() const { return leadingCoeff;  /* override */}

  // sets the leading coefficient for the specified time-integration method.
  void setLeadingCoeff(double & LC) { leadingCoeff = LC;  /* override */}

  // Evaluate corrector residual for nonlinear solver
  void obtainResidual() /* override */;

  // obtain residuals for transient sensitivity analysis
  void obtainSensitivityResiduals() /* override */;

  // obtain final derivatives w.r.t. time for transient sensitivity analysis
  void loadFinalSensitivityDerivatives() /* override */
  {}

  // Evaluate corrector Jacobian for nonlinear solver
  void obtainJacobian() /* override */;

  // Update history array after a successful step
  void updateHistory() /* override */;

  // Restore history array after a failed step
  void restoreHistory() /* override */;

  // Update method coefficients
  void updateCoeffs() /* override */;

  // Initialize method with initial solution & step-size
  void initialize(const TIAParams &tia_params) /* override */;

  // setup 2-level data.
  void setTwoLevelTimeInfo(const TimeIntInfo & tiInfo) /* override */;

  // Reject a step(this turns back history and chooses new order & step-size)
  void rejectStep(const TIAParams &tia_params) /* override */;

  // Reject a step, but only restore the history, as the new step will be
  // imposed by an outside program.
  void rejectStepForHabanero() /* override */;

  // Complete a step(this updates history and chooses new order & step-size)
  void completeStep(const TIAParams &tia_params) /* override */;

private:
  // Interpolate MPDE solution approximation at prescribed time point.
  bool interpolateMPDESolution(std::vector<double>& timepoint, Linear::Vector * tmpSolVectorPtr) /* override */;

  // Initialize arrays related to sensitivities
  void initializeSensitivities();

  // Predict solution at next time point.
  void obtainSensitivityPredictors();

  // Update sensitivity history arrays after a successful step
  void updateSensitivityHistory();

  // Check whether to reduce order independent of local error test
  void checkReduceOrder();

  double                timept_;        ///< Keep track of last interpolation point in printMPDEOutputSolution
  DataStore &           ds;             ///< Reference to the TIA data-store object.
  StepErrorControl &    sec;            ///< Reference to step-error control object
  double                leadingCoeff;   ///< Time-integration method leading coefficient value.
};

} // namespace TimeIntg
} // namespace Xyce

#endif     //Xyce_N_TIA_GEAR12_H

