//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_TIA_NoTimeIntegration.h,v $
//
// Purpose       : This file defines the classes for the "no time integration"
//                 method.
//
// Special Notes :
//
// Creator       : Eric Keiter
//
// Creation Date : 7/21/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.47.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:18 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#ifndef Xyce_N_TIA_NO_TIME_INTEGRATION_H
#define Xyce_N_TIA_NO_TIME_INTEGRATION_H

// ---------- Standard Includes ----------
#include <N_UTL_Math.h>

// ----------   Xyce Includes   ----------
#include <N_TIA_fwd.h>
#include <N_TIA_DataStore.h>
#include <N_TIA_TimeIntegrationMethods.h>

#define MATRIX_FAILSAFE 1
#define NUM_LIMIT  1.0e-20

namespace Xyce {
namespace TimeIntg {

//-----------------------------------------------------------------------------
// Class         : NoTimeIntegration
// Purpose       : Class objects for use during DC analysis (applies only when
//                 all time derivatives are set to 0) (derived from
//                 TimeIntegrationMethod)
// Special Notes :
// Creator       : Buddy Watts, SNL
// Creation Date : 6/01/00
//-----------------------------------------------------------------------------
class NoTimeIntegration : public TimeIntegrationMethod
{
public:
  static const int type = 0;
  static const char *name;

  static TimeIntegrationMethod *factory(const TIAParams &tia_params, StepErrorControl &step_error_control, DataStore &data_store);

  NoTimeIntegration(
    const TIAParams & tiaP,
    StepErrorControl & secTmp,
    DataStore & dsTmp);

  ~NoTimeIntegration();

  const char *getName() const /* override */ {
    return "None";
  }

  // Predict solution at next time point (No Integration) /* override */.
  void obtainPredictor() /* override */
  {
    ds.usePreviousSolAsPredictor ();
  }

  // Evaluate the predictor derivative formula (No Integration) /* override */.
  void obtainPredictorDeriv() /* override */
  {}

  // Evaluate the corrector derivative formula (No Integration) /* override */.
  void obtainCorrectorDeriv() /* override */;

  // Compute an estimate of the error in the integration step (No Integration) /* override */.
  void updateDerivsBlock(const std::list< IndexPair > & solGIDList,
                         const std::list< IndexPair > & staGIDList) /* override */
  {}

  // Compute an estimate of the error in the integration step (No Integration) /* override */.
  double computeErrorEstimate() const /* override */ { return 0.0; }

  // Interpolate solution approximation at prescribed time point (No
  // Integration) /* override */.
  bool interpolateSolution(double timepoint, Linear::Vector * tmpSolVectorPtr, std::vector<Linear::Vector*> & historyVec ) /* override */
  {
    return false;
  }

  bool printOutputSolution(
    Analysis::OutputMgrAdapter &        outputManagerAdapter,
    const TIAParams &                   tia_params,
    const double                        time,
    Linear::Vector *                      solnVecPtr,
    const bool                          doNotInterpolate,
    const std::vector<double> &         outputInterpolationTimes,
    bool                                skipPrintLineOutput) /* override */;

  bool saveOutputSolution(
    Analysis::OutputMgrAdapter &        outputManagerAdapter,
    const TIAParams &                   tia_params,
    Linear::Vector *                      solnVecPtr,
    const double                        saveTime,
    const bool                          doNotInterpolate) /* override */;

  // Computes the step adjustment (No Integration) /* override */.
  double computeExpoStepAdjust(double stepadjust) /* override */ { return 0.0; }

  // Gets the time-integration order (No Integration) /* override */.
  int getOrder() const /* override */
  {
    return 0;
  }
  
  int getUsedOrder() const /* override */
  {
    return 0;
  }

  int getNumberOfSteps() const /* override */
  {
    return 0;
  }

  int getNscsco() const /* override */
  {
    return 0;
  }

  int getMaxOrder() const /* override */
  {
    return -2;
  }

  void getInitialQnorm(TwoLevelError & tle) const /* override */;
  void getTwoLevelError(TwoLevelError & tle) const /* override */;

  double partialTimeDeriv() const /* override */
  {
#ifdef MATRIX_FAILSAFE
    return NUM_LIMIT;
#else
    return 0.0;
#endif
  }

  // Gets the leading coefficient for the specified time-integration method.
  double getLeadingCoeff() const /* override */ { return leadingCoeff; }

  // sets the leading coefficient for the specified time-integration method.
  void setLeadingCoeff(double & LC) /* override */ { leadingCoeff = LC; }

  // 03/08/04 erkeite:  New functions necessary new-DAE:
  // Evaluate residual for nonlinear solver
  void obtainResidual() /* override */;

  // obtain residuals for transient sensitivity analysis
  void obtainSensitivityResiduals() /* override */;

  // obtain final derivatives w.r.t. time for transient sensitivity analysis
  void loadFinalSensitivityDerivatives() /* override */ {return;}

  // Evaluate Jacobian for nonlinear solver
  void obtainJacobian() /* override */;

  // Apply Jacobian for nonlinear solver
  void applyJacobian(const Linear::Vector& input, Linear::Vector& result) /* override */;

  void initialize(const TIAParams &tia_params) /* override */
  {}

  void setTwoLevelTimeInfo(const TimeIntInfo & tiInfo) /* override */
  {}
  
  void rejectStep(const TIAParams &tia_params) /* override */;
  void completeStep(const TIAParams &tia_params) /* override */;

private:
  double                alphas;         ///< $\alpha_s$ fixed-leading coefficient of this BDF method
  DataStore &           ds;             ///< Reference to the TIA data-store object.
  StepErrorControl &    sec;            ///< Reference to step-error control object
  double                leadingCoeff;   ///< Time-integration method leading coefficient value.
};

} // namespace TimeIntg
} // namespace Xyce

#endif // Xyce_N_TIA_NO_TIME_INTEGRATION_H
