//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_TIA_DataStore.C,v $
//
// Purpose       : This file creates & initializes the data arrays needed for
//                 the time integration algorithms.
//
// Special Notes :
//
// Creator       : Buddy Watts
//
// Creation Date : 6/1/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.181.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:18 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

// ---------- Standard Includes ----------
#include <iostream>

#include <N_TIA_DataStore.h>

#include <N_LAS_Builder.h>
#include <N_LAS_Matrix.h>
#include <N_LAS_MultiVector.h>
#include <N_LAS_System.h>
#include <N_LAS_Vector.h>
#include <N_TIA_TIAParams.h>
#include <N_UTL_fwd.h>
#include <N_UTL_DeleteList.h>
#include <N_UTL_Diagnostic.h>
#include <N_UTL_FeatureTest.h>
#include <N_UTL_MachDepParams.h>

namespace Xyce {
namespace TimeIntg {

//-----------------------------------------------------------------------------
// Function      : DataStore::DataStore
// Purpose       : constructor
// Special Notes :
// Scope         : public
// Creator       : Buddy Watts, SNL
// Creation Date : 6/01/00
//-----------------------------------------------------------------------------
DataStore::DataStore(
  int                           solution_size,
  int                           state_size,
  int                           max_order,
  const Linear::System &        linear_system,
  const Linear::Builder &       linear_system_builder)
  : linearSystem_(linear_system),
    nextSolPtrSwitched_(false),
    limiterFlag(false),
    maxOrder(max_order),
    solutionSize(solution_size),
    stateSize(state_size),
    tmpSolVectorPtr(0),
    tmpStaVectorPtr(0),
    tmpStaDerivPtr(0),
    tmpStaDivDiffPtr(0),
    tmpStoVectorPtr(0),
    tmpLeadCurrentVectorPtr(0),
    tmpLeadDeltaVPtr(0),
    xn0Ptr (0),
    currSolutionPtr(0),
    lastSolutionPtr(0),
    oldeSolutionPtr(0),
    nextSolutionPtr(0),
    flagSolutionPtr(0),
    savedNextSolutionPtr(0),
    currStatePtr(0),
    lastStatePtr(0),
    oldeStatePtr(0),
    nextStatePtr(0),
    currStorePtr(0),
    lastStorePtr(0),
    oldeStorePtr(0),
    nextStorePtr(0),
    currStoreLeadCurrQPtr(0),
    lastStoreLeadCurrQPtr(0),
    oldeStoreLeadCurrQPtr(0),
    nextStoreLeadCurrQPtr(0),
    currStoreLeadCurrQDerivPtr(0),
    lastStoreLeadCurrQDerivPtr(0),
    oldeStoreLeadCurrQDerivPtr(0),
    nextStoreLeadCurrQDerivPtr(0),
    currLeadCurrentPtr(0),
    lastLeadCurrentPtr(0),
    oldLeadCurrentPtr(0),
    nextLeadCurrentPtr(0),
    currLeadDeltaVPtr(0),
    lastLeadDeltaVPtr(0),
    oldLeadDeltaVPtr(0),
    nextLeadDeltaVPtr(0),
    currLeadCurrentQPtr(0),
    lastLeadCurrentQPtr(0),
    oldLeadCurrentQPtr(0),
    nextLeadCurrentQPtr(0),
    currLeadCurrentQDerivPtr(0),
    lastLeadCurrentQDerivPtr(0),
    oldLeadCurrentQDerivPtr(0),
    nextLeadCurrentQDerivPtr(0),
    currSolutionDerivPtr(0),
    lastSolutionDerivPtr(0),
    oldeSolutionDerivPtr(0),
    nextSolutionDerivPtr(0),
    currStateDerivPtr(0),
    lastStateDerivPtr(0),
    oldeStateDerivPtr(0),
    nextStateDerivPtr(0),
    resMatVecPtr(0),
    currSolutionDivDiffPtr(0),
    lastSolutionDivDiffPtr(0),
    oldeSolutionDivDiffPtr(0),
    nextSolutionDivDiffPtr(0),
    currStateDivDiffPtr(0),
    lastStateDivDiffPtr(0),
    oldeStateDivDiffPtr(0),
    nextStateDivDiffPtr(0),
    errWtVecPtr(0),
    absErrTolPtr(0),
    relErrTolPtr(0),
    JMatrixPtr(0),
    RHSVectorPtr(0),
    JdxpVectorPtr(0),
    newtonCorrectionPtr(0),
    deviceErrorWeightMask_(0),
    qErrWtVecPtr(0),
    daeQVectorPtr(0),
    daeFVectorPtr(0),
    daeBVectorPtr(0),
    dQdxMatrixPtr(0),
    dFdxMatrixPtr(0),
    //saved_dQdxMatrixPtr(0),
    dQdxVecVectorPtr(0),
    dFdxVecVectorPtr(0),
    dFdxdVpVectorPtr(0),
    dQdxdVpVectorPtr(0),
    qn0Ptr(0),
    qpn0Ptr(0),
    sn0Ptr(0),
    spn0Ptr(0),
    ston0Ptr(0),
    stopn0Ptr(0),
    leadCurrentn0Ptr(0),
    leadCurrentpn0Ptr(0),
    leadCurrentQn0Ptr(0),
    leadCurrentQpn0Ptr(0),
    leadDeltaVn0Ptr(0),
    leadDeltaVpn0Ptr(0),
    qNewtonCorrectionPtr(0),
    sNewtonCorrectionPtr(0),
    stoNewtonCorrectionPtr(0),
    stoLeadCurrQNewtonCorrectionPtr(0),
    leadCurrentNewtonCorrectionPtr(0),
    leadCurrentQNewtonCorrectionPtr(0),
    leadDeltaVNewtonCorrectionPtr(0),
    delta_x(0),
    delta_q(0),
    tmpXn0APtr(0),
    tmpXn0BPtr(0)
{
  // temporary vectors:
  tmpSolVectorPtr = linear_system_builder.createVector();
  tmpStaVectorPtr = linear_system_builder.createStateVector();
  tmpStaDerivPtr = linear_system_builder.createStateVector();
  tmpStaDivDiffPtr = linear_system_builder.createStateVector();
  tmpStoVectorPtr = linear_system_builder.createStoreVector();
  tmpLeadCurrentVectorPtr = linear_system_builder.createLeadCurrentVector();
  tmpLeadDeltaVPtr = linear_system_builder.createLeadCurrentVector();

  xn0Ptr = linear_system_builder.createVector();

  // solution vectors:
  currSolutionPtr = linear_system_builder.createVector();
  lastSolutionPtr = linear_system_builder.createVector();
  nextSolutionPtr = linear_system_builder.createVector();
  flagSolutionPtr = linear_system_builder.createVector();

  // state vectors:
  currStatePtr    = linear_system_builder.createStateVector();
  lastStatePtr    = linear_system_builder.createStateVector();
  nextStatePtr    = linear_system_builder.createStateVector();

  // store vectors:
  currStorePtr    = linear_system_builder.createStoreVector();
  lastStorePtr    = linear_system_builder.createStoreVector();
  nextStorePtr    = linear_system_builder.createStoreVector();
  currStoreLeadCurrQPtr = linear_system_builder.createStoreVector();
  lastStoreLeadCurrQPtr = linear_system_builder.createStoreVector();
  nextStoreLeadCurrQPtr = linear_system_builder.createStoreVector();

  // lead current and delta V vectors (for power calc).
  currLeadCurrentPtr= linear_system_builder.createLeadCurrentVector();
  lastLeadCurrentPtr= linear_system_builder.createLeadCurrentVector();
  //oldLeadCurrentPtr= linear_system_builder.createLeadCurrentVector();
  nextLeadCurrentPtr= linear_system_builder.createLeadCurrentVector();
  currLeadDeltaVPtr= linear_system_builder.createLeadCurrentVector();
  lastLeadDeltaVPtr= linear_system_builder.createLeadCurrentVector();
  //oldLeadDeltaVPtr= linear_system_builder.createLeadCurrentVector();
  nextLeadDeltaVPtr= linear_system_builder.createLeadCurrentVector();
  currLeadCurrentQPtr= linear_system_builder.createLeadCurrentVector();
  lastLeadCurrentQPtr= linear_system_builder.createLeadCurrentVector();
  //oldLeadCurrentQPtr= linear_system_builder.createLeadCurrentVector();
  nextLeadCurrentQPtr= linear_system_builder.createLeadCurrentVector();
  currLeadCurrentQDerivPtr= linear_system_builder.createLeadCurrentVector();
  lastLeadCurrentQDerivPtr= linear_system_builder.createLeadCurrentVector();
  //oldLeadCurrentQDerivPtr= linear_system_builder.createLeadCurrentVector();
  nextLeadCurrentQDerivPtr= linear_system_builder.createLeadCurrentVector();

  // solution derivative vectors:
  currSolutionDerivPtr = linear_system_builder.createVector();
  lastSolutionDerivPtr = linear_system_builder.createVector();
  nextSolutionDerivPtr = linear_system_builder.createVector();

  // state derivative vectors:
  currStateDerivPtr = linear_system_builder.createStateVector();
  lastStateDerivPtr = linear_system_builder.createStateVector();
  nextStateDerivPtr = linear_system_builder.createStateVector();

  // store derivative vec for lead currents.
  currStoreLeadCurrQDerivPtr = linear_system_builder.createStoreVector();
  lastStoreLeadCurrQDerivPtr = linear_system_builder.createStoreVector();
  nextStoreLeadCurrQDerivPtr = linear_system_builder.createStoreVector();

  // solution scaled divided differences:
  currSolutionDivDiffPtr = linear_system_builder.createVector();
  lastSolutionDivDiffPtr = linear_system_builder.createVector();
  nextSolutionDivDiffPtr = linear_system_builder.createVector();

  // state scaled divided differences:
  currStateDivDiffPtr = linear_system_builder.createStateVector();
  lastStateDivDiffPtr = linear_system_builder.createStateVector();
  nextStateDivDiffPtr = linear_system_builder.createStateVector();

  // error vectors:
  errWtVecPtr      = linear_system_builder.createVector();
  absErrTolPtr     = linear_system_builder.createVector();
  relErrTolPtr     = linear_system_builder.createVector();

  errWtVecPtr->putScalar(1.0);

  // device mask pointer
  deviceErrorWeightMask_ = linear_system_builder.createVector();
  deviceErrorWeightMask_->putScalar(1.0);

  // nonlinear solution vectors:
  newtonCorrectionPtr = linear_system_builder.createVector();

  // new-DAE stuff:
  // Error Vectors
  qErrWtVecPtr      = linear_system_builder.createVector();

  // DAE formulation vectors
  daeQVectorPtr      = linear_system_builder.createVector();
  daeFVectorPtr      = linear_system_builder.createVector();
  daeBVectorPtr      = linear_system_builder.createVector();

  // DAE formulation matrices
  dQdxMatrixPtr = linear_system_builder.createMatrix();
  dFdxMatrixPtr = linear_system_builder.createMatrix();
  //saved_dQdxMatrixPtr = linear_system_builder.createMatrix();

  dQdxVecVectorPtr = linear_system_builder.createVector();
  dFdxVecVectorPtr = linear_system_builder.createVector();

  // History arrays
  for (int i = 0; i < max_order + 1; ++i)
  {
    xHistory.push_back(linear_system_builder.createVector());
    qHistory.push_back(linear_system_builder.createVector());
    sHistory.push_back(linear_system_builder.createStateVector());
    stoHistory.push_back(linear_system_builder.createStoreVector());
    stoLeadCurrQHistory.push_back(linear_system_builder.createStoreVector());
    leadCurrentHistory.push_back(linear_system_builder.createLeadCurrentVector());
    leadCurrentQHistory.push_back(linear_system_builder.createLeadCurrentVector());
    leadDeltaVHistory.push_back(linear_system_builder.createLeadCurrentVector());
  }

  // Predictors
  qn0Ptr = linear_system_builder.createVector();
  qpn0Ptr = linear_system_builder.createVector();
  sn0Ptr = linear_system_builder.createStateVector();
  spn0Ptr = linear_system_builder.createStateVector();
  ston0Ptr = linear_system_builder.createStoreVector();
  stopn0Ptr = linear_system_builder.createStoreVector();
  stoQn0Ptr = linear_system_builder.createStoreVector();
  stoQpn0Ptr = linear_system_builder.createStoreVector();
  leadCurrentn0Ptr = linear_system_builder.createLeadCurrentVector();
  leadCurrentpn0Ptr = linear_system_builder.createLeadCurrentVector();
  leadCurrentQn0Ptr = linear_system_builder.createLeadCurrentVector();
  leadCurrentQpn0Ptr = linear_system_builder.createLeadCurrentVector();
  leadDeltaVn0Ptr = linear_system_builder.createLeadCurrentVector();
  leadDeltaVpn0Ptr = linear_system_builder.createLeadCurrentVector();
 
  // Nonlinear solution vector:
  qNewtonCorrectionPtr = linear_system_builder.createVector();
  sNewtonCorrectionPtr = linear_system_builder.createStateVector();
  stoNewtonCorrectionPtr = linear_system_builder.createStoreVector();
  stoLeadCurrQNewtonCorrectionPtr = linear_system_builder.createStoreVector();
  leadCurrentNewtonCorrectionPtr = linear_system_builder.createLeadCurrentVector();
  leadCurrentQNewtonCorrectionPtr = linear_system_builder.createLeadCurrentVector();
  leadDeltaVNewtonCorrectionPtr = linear_system_builder.createLeadCurrentVector();
 
  // Step-size selection temporary vectors
  delta_x = linear_system_builder.createVector();
  delta_q = linear_system_builder.createVector();

  // Temporary vector for MPDE & WaMPDE interpolation
  tmpXn0APtr = linear_system_builder.createVector();
  tmpXn0BPtr = linear_system_builder.createVector();
}

//-----------------------------------------------------------------------------
// Function      : DataStore::DataStore
// Purpose       : destructor
// Special Notes :
// Scope         : public
// Creator       : Buddy Watts, SNL
// Creation Date : 6/01/00
//-----------------------------------------------------------------------------
DataStore::~DataStore()
{
  delete tmpSolVectorPtr;
  delete tmpStaVectorPtr;
  delete tmpStaDerivPtr;
  delete tmpStaDivDiffPtr;
  delete tmpStoVectorPtr;
  delete tmpLeadCurrentVectorPtr;
  delete tmpLeadDeltaVPtr;
  delete xn0Ptr;

  delete currSolutionPtr;
  delete lastSolutionPtr;

  delete nextSolutionPtr;
  delete flagSolutionPtr;

  delete currStatePtr;
  delete lastStatePtr;

  delete nextStatePtr;

  delete currStorePtr;
  delete lastStorePtr;
  delete nextStorePtr;

  // Note: These will go away when the lead currents are moved to their own vector
  delete currStoreLeadCurrQPtr;
  delete lastStoreLeadCurrQPtr;
  delete nextStoreLeadCurrQPtr;
  delete currStoreLeadCurrQDerivPtr;
  delete lastStoreLeadCurrQDerivPtr;
  delete nextStoreLeadCurrQDerivPtr;

  // Lead current and power vectors
  delete currLeadCurrentPtr;
  delete lastLeadCurrentPtr;
  //delete oldLeadCurrentPtr;
  delete nextLeadCurrentPtr;
  delete currLeadDeltaVPtr;
  delete lastLeadDeltaVPtr;
  //delete oldLeadDeltaVPtr;
  delete nextLeadDeltaVPtr;

  // for lead current calculations.  F component is
  // held in the store vector, Q component is here
  delete currLeadCurrentQPtr;
  delete lastLeadCurrentQPtr;
  //delete oldLeadCurrentQPtr;
  delete nextLeadCurrentQPtr;
  delete currLeadCurrentQDerivPtr;
  delete lastLeadCurrentQDerivPtr;
  //delete oldLeadCurrentQDerivPtr;
  delete nextLeadCurrentQDerivPtr;

  delete currSolutionDerivPtr;
  delete lastSolutionDerivPtr;

  delete nextSolutionDerivPtr;

  delete currStateDerivPtr;
  delete lastStateDerivPtr;

  delete nextStateDerivPtr;

  delete currSolutionDivDiffPtr;
  delete lastSolutionDivDiffPtr;

  delete nextSolutionDivDiffPtr;

  delete currStateDivDiffPtr;
  delete lastStateDivDiffPtr;

  delete nextStateDivDiffPtr;

  delete errWtVecPtr;
  delete absErrTolPtr;
  delete relErrTolPtr;

  delete deviceErrorWeightMask_;

  delete newtonCorrectionPtr;

  //new-DAE:
  // Error Vectors
  delete qErrWtVecPtr;

  // DAE formulation vectors
  delete daeQVectorPtr;
  delete daeFVectorPtr;
  delete daeBVectorPtr;

  // DAE formulation matrices
  delete dQdxMatrixPtr;
  delete dFdxMatrixPtr;
  //delete saved_dQdxMatrixPtr;

  // HB temporary vectors
  delete dQdxVecVectorPtr;
  delete dFdxVecVectorPtr;

  Xyce::deleteList(xHistory.begin(), xHistory.end());
  Xyce::deleteList(qHistory.begin(), qHistory.end());
  Xyce::deleteList(sHistory.begin(), sHistory.end());
  Xyce::deleteList(stoHistory.begin(), stoHistory.end());
  Xyce::deleteList(stoLeadCurrQHistory.begin(), stoLeadCurrQHistory.end());
  Xyce::deleteList(leadCurrentHistory.begin(), leadCurrentHistory.end());
  Xyce::deleteList(leadCurrentQHistory.begin(), leadCurrentQHistory.end());
  Xyce::deleteList(leadDeltaVHistory.begin(), leadDeltaVHistory.end());

  delete qn0Ptr;
  delete qpn0Ptr;
  delete sn0Ptr;
  delete spn0Ptr;
  delete ston0Ptr;
  delete stopn0Ptr;
  delete stoQn0Ptr;
  delete stoQpn0Ptr;
  delete leadCurrentn0Ptr;
  delete leadCurrentpn0Ptr;
  delete leadCurrentQn0Ptr;
  delete leadCurrentQpn0Ptr;
  delete leadDeltaVn0Ptr;
  delete leadDeltaVpn0Ptr;


  // Nonlinear solution vector:
  delete qNewtonCorrectionPtr;
  delete sNewtonCorrectionPtr;
  delete stoNewtonCorrectionPtr;
  delete stoLeadCurrQNewtonCorrectionPtr;
  delete leadCurrentNewtonCorrectionPtr;
  delete leadCurrentQNewtonCorrectionPtr;
  delete leadDeltaVNewtonCorrectionPtr;
 
  // Step-size selection temporary vectors
  delete delta_x;
  delete delta_q;

  // Temporary vector for WaMPDE interpolation
  delete tmpXn0APtr;
  delete tmpXn0BPtr;

  // Delete data in the fast time storage for HB and MPDE
  Xyce::deleteList(fastTimeSolutionVec.begin(), fastTimeSolutionVec.end());
  Xyce::deleteList(fastTimeStateVec.begin(), fastTimeStateVec.end());
  Xyce::deleteList(fastTimeQVec.begin(), fastTimeQVec.end());
  Xyce::deleteList(fastTimeStoreVec.begin(), fastTimeStoreVec.end());

  // delete the sensitivity related stuff, if necessary
  deleteSensitivityArrays();
}

//-----------------------------------------------------------------------------
// Function      : DataStore::deleteSensitivityArrays
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 8/6/2014
//-----------------------------------------------------------------------------
void DataStore::deleteSensitivityArrays()
{
  int numParams = currDqdpPtrVector.size();

  delete resMatVecPtr;

  for (int ip=0;ip<numParams;++ip)
  {
    delete sensRHSPtrVector[ip];
    delete nextDfdpPtrVector[ip];

    delete currDqdpPtrVector[ip];
    delete lastDqdpPtrVector[ip];
    delete nextDqdpPtrVector[ip];

    delete nextDbdpPtrVector[ip];

    delete currDXdpPtrVector[ip];
    delete lastDXdpPtrVector[ip];
    delete nextDXdpPtrVector[ip];

    delete currDQdxDXdpPtrVector[ip];
    delete lastDQdxDXdpPtrVector[ip];
    delete nextDQdxDXdpPtrVector[ip];

    delete currDQdxDXdpDerivPtrVector[ip];
    delete lastDQdxDXdpDerivPtrVector[ip];
    delete nextDQdxDXdpDerivPtrVector[ip];

    delete currDqdpDerivPtrVector[ip];
    delete lastDqdpDerivPtrVector[ip];
    delete nextDqdpDerivPtrVector[ip];

    // history
    for (int i = 0; i < dfdpHistory[ip].size(); ++i)
    {
      delete dfdpHistory[ip][i];
      delete dqdpHistory[ip][i];
      delete dbdpHistory[ip][i];
      delete dXdpHistory[ip][i];
      delete dQdxdXdpHistory[ip][i];
    }
    dfdpHistory[ip].clear();
    dqdpHistory[ip].clear();
    dbdpHistory[ip].clear();
    dXdpHistory[ip].clear();
    dQdxdXdpHistory[ip].clear();
  }

  sensRHSPtrVector.clear();

  nextDfdpPtrVector.clear();

  currDqdpPtrVector.clear();
  lastDqdpPtrVector.clear();
  nextDqdpPtrVector.clear();

  nextDbdpPtrVector.clear();

  currDXdpPtrVector.clear();
  lastDXdpPtrVector.clear();
  nextDXdpPtrVector.clear();

  currDQdxDXdpPtrVector.clear();
  lastDQdxDXdpPtrVector.clear();
  nextDQdxDXdpPtrVector.clear();

  currDQdxDXdpDerivPtrVector.clear();
  lastDQdxDXdpDerivPtrVector.clear();
  nextDQdxDXdpDerivPtrVector.clear();

  currDqdpDerivPtrVector.clear();
  lastDqdpDerivPtrVector.clear();
  nextDqdpDerivPtrVector.clear();

  // history
  dfdpHistory.clear();
  dqdpHistory.clear();
  dbdpHistory.clear();
  dXdpHistory.clear();
  dQdxdXdpHistory.clear();
}

//-----------------------------------------------------------------------------
// Function      : DataStore::allocateSensitivityArrays
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 8/6/2014
//-----------------------------------------------------------------------------
void DataStore::allocateSensitivityArrays(Linear::Builder &builder, int numParams)
{
  resMatVecPtr = builder.createVector();

  sensRHSPtrVector.resize(numParams);

  nextDfdpPtrVector.resize(numParams);

  currDqdpPtrVector.resize(numParams);
  lastDqdpPtrVector.resize(numParams);
  oldeDqdpPtrVector.resize(numParams);
  nextDqdpPtrVector.resize(numParams);

  nextDbdpPtrVector.resize(numParams);

  currDXdpPtrVector.resize(numParams);
  lastDXdpPtrVector.resize(numParams);
  oldeDXdpPtrVector.resize(numParams);
  nextDXdpPtrVector.resize(numParams);

  currDQdxDXdpPtrVector.resize(numParams);
  lastDQdxDXdpPtrVector.resize(numParams);
  oldeDQdxDXdpPtrVector.resize(numParams);
  nextDQdxDXdpPtrVector.resize(numParams);

  currDQdxDXdpDerivPtrVector.resize(numParams);
  lastDQdxDXdpDerivPtrVector.resize(numParams);
  oldeDQdxDXdpDerivPtrVector.resize(numParams);
  nextDQdxDXdpDerivPtrVector.resize(numParams);

  currDqdpDerivPtrVector.resize(numParams);
  lastDqdpDerivPtrVector.resize(numParams);
  oldeDqdpDerivPtrVector.resize(numParams);
  nextDqdpDerivPtrVector.resize(numParams);

  // history
  dfdpHistory.resize(numParams);
  dqdpHistory.resize(numParams);
  dbdpHistory.resize(numParams);
  dXdpHistory.resize(numParams);
  dQdxdXdpHistory.resize(numParams);

  for (int ip=0;ip<numParams;++ip)
  {
    sensRHSPtrVector[ip] = builder.createVector();

    nextDfdpPtrVector[ip] = builder.createVector();

    currDqdpPtrVector[ip] = builder.createVector();
    lastDqdpPtrVector[ip] = builder.createVector();
    nextDqdpPtrVector[ip] = builder.createVector();

    nextDbdpPtrVector[ip] = builder.createVector();

    currDXdpPtrVector[ip] = builder.createVector();
    lastDXdpPtrVector[ip] = builder.createVector();
    nextDXdpPtrVector[ip] = builder.createVector();

    currDQdxDXdpPtrVector[ip] = builder.createVector();
    lastDQdxDXdpPtrVector[ip] = builder.createVector();
    nextDQdxDXdpPtrVector[ip] = builder.createVector();

    currDQdxDXdpDerivPtrVector[ip] = builder.createVector();
    lastDQdxDXdpDerivPtrVector[ip] = builder.createVector();
    nextDQdxDXdpDerivPtrVector[ip] = builder.createVector();

    currDqdpDerivPtrVector[ip] = builder.createVector();
    lastDqdpDerivPtrVector[ip] = builder.createVector();
    nextDqdpDerivPtrVector[ip] = builder.createVector();

    // history
    for (int i = 0; i < maxOrder + 1; ++i)
    {
      dfdpHistory[ip].push_back(builder.createVector());
      dqdpHistory[ip].push_back(builder.createVector());
      dbdpHistory[ip].push_back(builder.createVector());
      dXdpHistory[ip].push_back(builder.createVector());
      dQdxdXdpHistory[ip].push_back(builder.createVector());
    }
  }
}

// ----------------------------------------------------------------------------
// -----------------------  DataStore Class Functions -------------------------
// ----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Function      : DataStore::printOutPointers
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 6/23/00
//-----------------------------------------------------------------------------
void DataStore::printOutPointers()
{
  Xyce::dout() << "olde ptr = " << oldeSolutionPtr << std::endl
               << "last ptr = " << lastSolutionPtr << std::endl
               << "curr ptr = " << currSolutionPtr << std::endl
               << "next ptr = " << nextSolutionPtr << std::endl;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::setConstantHistory
// Purpose       : This function is called  after the operating point
//                 calculation has been called.  Once the operating point
//                 solution has been obtained, the code should regard that
//                 solution as having been the existing constant solution since
//                 the dawn of time.
// Special Notes : The most recent solution, etc., are in the "next" vectors.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 6/23/00
//-----------------------------------------------------------------------------
void DataStore::setConstantHistory()
{
  if (DEBUG_TIME && isActive(Diag::TIME_HISTORY))
    Xyce::dout() << "\nDataStore::setConstantHistory" << std::endl;

  // Solutions:
  *(lastSolutionPtr) = *(nextSolutionPtr);
  *(currSolutionPtr) = *(nextSolutionPtr);

  // Derivative of Solutions:
  *(lastSolutionDerivPtr) = *(nextSolutionDerivPtr);
  *(currSolutionDerivPtr) = *(nextSolutionDerivPtr);

  // Scaled Divided Differences of solutions:
  *(lastSolutionDivDiffPtr) = *(nextSolutionDivDiffPtr);
  *(currSolutionDivDiffPtr) = *(nextSolutionDivDiffPtr);

  // States:
  *(lastStatePtr) = *(nextStatePtr);
  *(currStatePtr) = *(nextStatePtr);

  // Stores:
  *(lastStorePtr) = *(nextStorePtr);
  *(currStorePtr) = *(nextStorePtr);

  *(lastStoreLeadCurrQPtr) = *(nextStoreLeadCurrQPtr);
  *(currStoreLeadCurrQPtr) = *(nextStoreLeadCurrQPtr);

  // Derivative of States:
  *(lastStateDerivPtr) = *(nextStateDerivPtr);
  *(currStateDerivPtr) = *(nextStateDerivPtr);

  // lead current derivative info in store
  *(lastStoreLeadCurrQDerivPtr) = *(nextStoreLeadCurrQDerivPtr);
  *(currStoreLeadCurrQDerivPtr) = *(nextStoreLeadCurrQDerivPtr);

  // Scaled Divided Differences of states:
  *(lastStateDivDiffPtr) = *(nextStateDivDiffPtr);
  *(currStateDivDiffPtr) = *(nextStateDivDiffPtr);
  
  // lead current and junction voltage vectors 
  *(lastLeadCurrentPtr) = *(nextLeadCurrentPtr);
  *(currLeadCurrentPtr) = *(nextLeadCurrentPtr);
  *(lastLeadCurrentQPtr) = *(nextLeadCurrentQPtr);
  *(currLeadCurrentQPtr) = *(nextLeadCurrentQPtr);
  *(lastLeadDeltaVPtr) = *(nextLeadDeltaVPtr);
  *(currLeadDeltaVPtr) = *(nextLeadDeltaVPtr);
  *(lastLeadCurrentQDerivPtr) = *(nextLeadCurrentQDerivPtr);
  *(currLeadCurrentQDerivPtr) = *(nextLeadCurrentQDerivPtr);


  setConstantSensitivityHistory();
}

//-----------------------------------------------------------------------------
// Function      : DataStore::setConstantSensitivityHistory
// Purpose       :
// Special Notes : The most recent solution, etc., are in the "next" vectors.
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 8/28/2014
//-----------------------------------------------------------------------------
void DataStore::setConstantSensitivityHistory()
{
  // Sensitivities:
  for (int ip=0;ip<nextDfdpPtrVector.size();++ip)
  {

    *(lastDqdpPtrVector[ip]) = *(nextDqdpPtrVector[ip]);
    *(currDqdpPtrVector[ip]) = *(nextDqdpPtrVector[ip]);

    *(lastDqdpDerivPtrVector[ip]) = *(nextDqdpDerivPtrVector[ip]);
    *(currDqdpDerivPtrVector[ip]) = *(nextDqdpDerivPtrVector[ip]);

    *(lastDXdpPtrVector[ip]) = *(nextDXdpPtrVector[ip]);
    *(currDXdpPtrVector[ip]) = *(nextDXdpPtrVector[ip]);

    *(lastDQdxDXdpPtrVector[ip]) = *(nextDQdxDXdpPtrVector[ip]);
    *(currDQdxDXdpPtrVector[ip]) = *(nextDQdxDXdpPtrVector[ip]);

    *(lastDQdxDXdpDerivPtrVector[ip]) = *(nextDQdxDXdpDerivPtrVector[ip]);
    *(currDQdxDXdpDerivPtrVector[ip]) = *(nextDQdxDXdpDerivPtrVector[ip]);
  }
}

//-----------------------------------------------------------------------------
// Function      : DataStore::resetAll
//
// Purpose       : This function resets everything so that a transient loop
//                 can be started from the beginning.
//
// Special Notes : This function was needed for HB.
//
// Scope         : public
// Creator       : T. Mei, SNL
// Creation Date : 02/26/09
//-----------------------------------------------------------------------------
bool
DataStore::resetAll(
  double        absolute_error_tolerance,
  double        relative_error_tolerance)
{
  absErrTolPtr->putScalar(absolute_error_tolerance); // tiaParams_.absErrorTol);
  relErrTolPtr->putScalar(relative_error_tolerance); // tiaParams_.relErrorTol);

  return true;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::resetFastTimeData
//
// Purpose       : This function deletes the information from all the vectors
//                 that store fast time data for HB and MPDE
//
// Special Notes : This function was needed for HB.
//
// Scope         : public
// Creator       : Heidi Thornquist, SNL
// Creation Date : 06/05/13
//-----------------------------------------------------------------------------
bool DataStore::resetFastTimeData()
{
  // Clear the time step vectors
  timeSteps.clear();
  timeStepsBreakpointFlag.clear();

  // Delete any stored up any solution or state info
  Xyce::deleteList(fastTimeSolutionVec.begin(), fastTimeSolutionVec.end());
  Xyce::deleteList(fastTimeStateVec.begin(), fastTimeStateVec.end());
  Xyce::deleteList(fastTimeQVec.begin(), fastTimeQVec.end());
  Xyce::deleteList(fastTimeStoreVec.begin(), fastTimeStoreVec.end());

  fastTimeSolutionVec.clear();
  fastTimeStateVec.clear();
  fastTimeQVec.clear();
  fastTimeStoreVec.clear();

  return true;
}



//-----------------------------------------------------------------------------
// Function      : DataStore::updateSolDataArrays
// Purpose       : Update the necessary integration data arrays for
//                 preparation of the next integration step. This is done after
//                 a successful step has been taken.
// Special Notes :
// Scope         : public
// Creator       : Buddy Watts, SNL
// Creation Date : 6/01/00
//-----------------------------------------------------------------------------
void DataStore::updateSolDataArrays()
{
  if (DEBUG_TIME && isActive(Diag::TIME_STEP))
  {
    Xyce::dout() << "\nDataStore::updateSolDataArrays " << std::endl;
  }

  // if the next solution has been switched out (probably because of NOX),
  // then reset it
  if (nextSolPtrSwitched_)
    unsetNextSolVectorPtr();

  // Solutions:
  oldeSolutionPtr = lastSolutionPtr;
  lastSolutionPtr = currSolutionPtr;
  currSolutionPtr = nextSolutionPtr;
  nextSolutionPtr = oldeSolutionPtr;

  // Derivative of Solutions:
  oldeSolutionDerivPtr = lastSolutionDerivPtr;
  lastSolutionDerivPtr = currSolutionDerivPtr;
  currSolutionDerivPtr = nextSolutionDerivPtr;
  nextSolutionDerivPtr = oldeSolutionDerivPtr;

  // Scaled Divided Differences of solutions:
  oldeSolutionDivDiffPtr = lastSolutionDivDiffPtr;
  lastSolutionDivDiffPtr = currSolutionDivDiffPtr;
  currSolutionDivDiffPtr = nextSolutionDivDiffPtr;
  nextSolutionDivDiffPtr = oldeSolutionDivDiffPtr;

  // States:
  oldeStatePtr = lastStatePtr;
  lastStatePtr = currStatePtr;
  currStatePtr = nextStatePtr;
  nextStatePtr = oldeStatePtr;

  // Stores:
  oldeStorePtr = lastStorePtr;
  lastStorePtr = currStorePtr;
  currStorePtr = nextStorePtr;
  nextStorePtr = oldeStorePtr;
  oldeStoreLeadCurrQPtr = lastStoreLeadCurrQPtr;
  lastStoreLeadCurrQPtr = currStoreLeadCurrQPtr;
  currStoreLeadCurrQPtr = nextStoreLeadCurrQPtr;
  nextStoreLeadCurrQPtr = oldeStoreLeadCurrQPtr;

  // Derivative of States:
  oldeStateDerivPtr = lastStateDerivPtr;
  lastStateDerivPtr = currStateDerivPtr;
  currStateDerivPtr = nextStateDerivPtr;
  nextStateDerivPtr = oldeStateDerivPtr;

  // lead curent component of store
  oldeStoreLeadCurrQDerivPtr = lastStoreLeadCurrQDerivPtr;
  lastStoreLeadCurrQDerivPtr = currStoreLeadCurrQDerivPtr;
  currStoreLeadCurrQDerivPtr = nextStoreLeadCurrQDerivPtr;
  nextStoreLeadCurrQDerivPtr = oldeStoreLeadCurrQDerivPtr;

  // Scaled Divided Differences of states:
  oldeStateDivDiffPtr = lastStateDivDiffPtr;
  lastStateDivDiffPtr = currStateDivDiffPtr;
  currStateDivDiffPtr = nextStateDivDiffPtr;
  nextStateDivDiffPtr = oldeStateDivDiffPtr;
  
  // lead current and junction voltage vectors
  oldLeadCurrentPtr = lastLeadCurrentPtr;
  lastLeadCurrentPtr = currLeadCurrentPtr;
  currLeadCurrentPtr = nextLeadCurrentPtr;
  nextLeadCurrentPtr = oldLeadCurrentPtr;

  oldLeadCurrentQPtr = lastLeadCurrentQPtr;
  lastLeadCurrentQPtr = currLeadCurrentQPtr;
  currLeadCurrentQPtr = nextLeadCurrentQPtr;
  nextLeadCurrentQPtr = oldLeadCurrentQPtr;
 
  oldLeadCurrentQDerivPtr = lastLeadCurrentQDerivPtr;
  lastLeadCurrentQDerivPtr = currLeadCurrentQDerivPtr;
  currLeadCurrentQDerivPtr = nextLeadCurrentQDerivPtr;
  nextLeadCurrentQDerivPtr = oldLeadCurrentQDerivPtr;
 
  oldLeadDeltaVPtr = lastLeadDeltaVPtr;
  lastLeadDeltaVPtr = currLeadDeltaVPtr;
  currLeadDeltaVPtr = nextLeadDeltaVPtr;
  nextLeadDeltaVPtr = oldLeadDeltaVPtr;

  int iparamSizeQ = currDqdpPtrVector.size();
  for (int ip=0;ip<iparamSizeQ;++ip)
  {
    oldeDqdpPtrVector[ip] = lastDqdpPtrVector[ip];
    lastDqdpPtrVector[ip] = currDqdpPtrVector[ip];
    currDqdpPtrVector[ip] = nextDqdpPtrVector[ip];
    nextDqdpPtrVector[ip] = oldeDqdpPtrVector[ip];
  }

  int iparamSizeQderiv = currDqdpDerivPtrVector.size();
  for (int ip=0;ip<iparamSizeQderiv;++ip)
  {
    oldeDqdpDerivPtrVector[ip] = lastDqdpDerivPtrVector[ip];
    lastDqdpDerivPtrVector[ip] = currDqdpDerivPtrVector[ip];
    currDqdpDerivPtrVector[ip] = nextDqdpDerivPtrVector[ip];
    nextDqdpDerivPtrVector[ip] = oldeDqdpDerivPtrVector[ip];
  }

  int iparamSizeX = currDXdpPtrVector.size();
  for (int ip=0;ip<iparamSizeX;++ip)
  {
    oldeDXdpPtrVector[ip] = lastDXdpPtrVector[ip];
    lastDXdpPtrVector[ip] = currDXdpPtrVector[ip];
    currDXdpPtrVector[ip] = nextDXdpPtrVector[ip];
    nextDXdpPtrVector[ip] = oldeDXdpPtrVector[ip];
  }

  int iparamSizeQX = currDQdxDXdpPtrVector.size();
  for (int ip=0;ip<iparamSizeQX;++ip)
  {
    oldeDQdxDXdpPtrVector[ip] = lastDQdxDXdpPtrVector[ip];
    lastDQdxDXdpPtrVector[ip] = currDQdxDXdpPtrVector[ip];
    currDQdxDXdpPtrVector[ip] = nextDQdxDXdpPtrVector[ip];
    nextDQdxDXdpPtrVector[ip] = oldeDQdxDXdpPtrVector[ip];
  }


  int iparamSizeQXd = currDQdxDXdpDerivPtrVector.size();
  for (int ip=0;ip<iparamSizeQXd;++ip)
  {
    oldeDQdxDXdpDerivPtrVector[ip] = lastDQdxDXdpDerivPtrVector[ip];
    lastDQdxDXdpDerivPtrVector[ip] = currDQdxDXdpDerivPtrVector[ip];
    currDQdxDXdpDerivPtrVector[ip] = nextDQdxDXdpDerivPtrVector[ip];
    nextDQdxDXdpDerivPtrVector[ip] = oldeDQdxDXdpDerivPtrVector[ip];
  }

  // copy contents of "curr" into "next".  This is to insure
  // that at a minimum, the initial guess for the Newton solve
  // will at least be the results of the previous Newton solve.
  *(nextSolutionPtr) = *(currSolutionPtr);
  *(nextStatePtr)    = *(currStatePtr);
  *(nextStorePtr)    = *(currStorePtr);
  *(nextLeadCurrentPtr) = *(currLeadCurrentPtr);
  *(nextLeadDeltaVPtr) = *(currLeadDeltaVPtr);
}

//-----------------------------------------------------------------------------
// Function      : DataStore::updateStateDataArrays
//
//
// Purpose       : Same as updateSolDataArrays, but this function only
//                 advances the state vector, and leaves the
//                 solution alone.
//
// Special Notes : The main usage of this function is LOCA.
//                 After each continuation step, LOCA needs to call
//                 this function.  LOCA keeps track of solution vectors
//                 on its own, which is why updateSolDataArrays
//                 is inappropriate for LOCA.
//
//                 This is necessary for voltage limiting to be
//                 consistent with LOCA.
//
// Scope         : public
// Creator       : Eric R. Keiter, SNL, 9233.
// Creation Date : 3/06/05
//-----------------------------------------------------------------------------
bool DataStore::updateStateDataArrays()
{
  if (DEBUG_TIME && isActive(Diag::TIME_STEP))
    Xyce::dout() << "\nDataStore::updateStateDataArrays " << std::endl;

  // States:
  oldeStatePtr = lastStatePtr;
  lastStatePtr = currStatePtr;
  currStatePtr = nextStatePtr;
  nextStatePtr = oldeStatePtr;

  // Stores:
  oldeStorePtr = lastStorePtr;
  lastStorePtr = currStorePtr;
  currStorePtr = nextStorePtr;
  nextStorePtr = oldeStorePtr;

  // q component of lead current
  oldeStoreLeadCurrQPtr = lastStoreLeadCurrQPtr;
  lastStoreLeadCurrQPtr = currStoreLeadCurrQPtr;
  currStoreLeadCurrQPtr = nextStoreLeadCurrQPtr;
  nextStoreLeadCurrQPtr = oldeStoreLeadCurrQPtr;

  // Derivative of States:
  oldeStateDerivPtr = lastStateDerivPtr;
  lastStateDerivPtr = currStateDerivPtr;
  currStateDerivPtr = nextStateDerivPtr;
  nextStateDerivPtr = oldeStateDerivPtr;

  // Scaled Divided Differences of states:
  oldeStateDivDiffPtr = lastStateDivDiffPtr;
  lastStateDivDiffPtr = currStateDivDiffPtr;
  currStateDivDiffPtr = nextStateDivDiffPtr;
  nextStateDivDiffPtr = oldeStateDivDiffPtr;

  // lead current and junction voltage vectors
  oldLeadCurrentPtr = lastLeadCurrentPtr;
  lastLeadCurrentPtr = currLeadCurrentPtr;
  currLeadCurrentPtr = nextLeadCurrentPtr;
  nextLeadCurrentPtr = oldLeadCurrentPtr;
 
  oldLeadCurrentQPtr = lastLeadCurrentQPtr;
  lastLeadCurrentQPtr = currLeadCurrentQPtr;
  currLeadCurrentQPtr = nextLeadCurrentQPtr;
  nextLeadCurrentQPtr = oldLeadCurrentQPtr;
  
  oldLeadDeltaVPtr = lastLeadDeltaVPtr;
  lastLeadDeltaVPtr = currLeadDeltaVPtr;
  currLeadDeltaVPtr = nextLeadDeltaVPtr;
  nextLeadDeltaVPtr = oldLeadDeltaVPtr;
  

  // Now, make the "next" stuff the same as the "curr" stuff.
  // This is done because at the end of the tranop, but before
  // the transient phase starts, the function setConstantHistory
  // will be called.  When it is called, the most recent values
  // for state variables need to be in the "next" vectors.
  //
  // As long as LOCA solves are never used for transient, and only
  // for DC and tranop solves, this is OK.  This is, of course,
  // a bit of a kludge.
  *(nextStatePtr) = *(currStatePtr);
  *(nextStorePtr) = *(currStorePtr);
  *(nextLeadCurrentPtr) = *(currLeadCurrentPtr);
  *(nextLeadDeltaVPtr) = *(currLeadDeltaVPtr);

  return true;
}

#ifndef OLD_PLOT
//-----------------------------------------------------------------------------
// Function      : DataStore::outputSolDataArrays
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/29/00
//-----------------------------------------------------------------------------
void DataStore::outputSolDataArrays(std::ostream &os)
{
  char tmp[256];
  os << std::endl
     << Xyce::section_divider << std::endl
     << std::endl
     << "  Solution Vectors:\n          Current               Last               Olde               Error" << std::endl;

  for (unsigned int k = 0; k < solutionSize; ++k)
  {
    os << (*(currSolutionPtr))[k] << (*(lastSolutionPtr))[k] << (*(oldeSolutionPtr))[k] << (*(newtonCorrectionPtr))[k] << std::endl;
  }

  os << Xyce::section_divider << std::endl;
}

#else

//-----------------------------------------------------------------------------
// Function      : DataStore::outputSolDataArrays
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/29/00
//-----------------------------------------------------------------------------
void DataStore::outputSolDataArrays(std::ostream &os)
{
  for (unsigned int k = 0; k < solutionSize; ++k)
  {
    os << "\t" <<  (*(currSolutionPtr))[0][k] << std::endl;
  }
  os << std::endl;
}

#endif

//-----------------------------------------------------------------------------
// Function      : DataStore::enableOrderOneStart
// Purpose       : Initialize arrays for "re-starting" integration at order 1.
// Special Notes :
// Scope         : public
// Creator       : Buddy Watts, SNL
// Creation Date : 6/01/00
//-----------------------------------------------------------------------------
void DataStore::enableOrderOneStart()
{
  // solution vectors:
  *lastSolutionPtr = *currSolutionPtr;
  *oldeSolutionPtr = *currSolutionPtr;

  nextSolutionDerivPtr->putScalar(0.0);
  currSolutionDerivPtr->putScalar(0.0);
  currSolutionDivDiffPtr->putScalar(0.0);

  // state vectors:
  *lastStatePtr = *currStatePtr;
  *oldeStatePtr = *currStatePtr;

  nextStateDerivPtr->putScalar(0.0);
  currStateDerivPtr->putScalar(0.0);
  currStateDivDiffPtr->putScalar(0.0);

  // store vectors:
  *lastStorePtr = *currStorePtr;
  *oldeStorePtr = *currStorePtr;
  *lastStoreLeadCurrQPtr = *currStoreLeadCurrQPtr;
  *oldeStoreLeadCurrQPtr = *currStoreLeadCurrQPtr;
  nextStoreLeadCurrQDerivPtr->putScalar(0.0);
  currStoreLeadCurrQDerivPtr->putScalar(0.0);

  // lead current and junction voltage vectors 
  *lastLeadCurrentPtr = *currLeadCurrentPtr;
  *oldLeadCurrentPtr = *currLeadCurrentPtr;
  
  *lastLeadCurrentQPtr = *currLeadCurrentQPtr;
  *oldLeadCurrentQPtr = *currLeadCurrentQPtr;
  
  *lastLeadDeltaVPtr = *currLeadDeltaVPtr;
  *oldLeadDeltaVPtr = *currLeadDeltaVPtr;
  
  nextLeadCurrentQDerivPtr->putScalar(0.0);
  currLeadCurrentQDerivPtr->putScalar(0.0);

  // sensitivities

  int iparamSizeQ = currDqdpPtrVector.size();
  for (int ip=0;ip<iparamSizeQ;++ip)
  {
    *(oldeDqdpPtrVector[ip]) = *(currDqdpPtrVector[ip]);
    *(lastDqdpPtrVector[ip]) = *(currDqdpPtrVector[ip]);
    currDqdpPtrVector[ip]->putScalar(0.0);
    nextDqdpPtrVector[ip]->putScalar(0.0);
  }

  int iparamSizeQd = currDqdpDerivPtrVector.size();
  for (int ip=0;ip<iparamSizeQ;++ip)
  {
    *(oldeDqdpDerivPtrVector[ip]) = *(currDqdpDerivPtrVector[ip]);
    *(lastDqdpDerivPtrVector[ip]) = *(currDqdpDerivPtrVector[ip]);
    currDqdpDerivPtrVector[ip]->putScalar(0.0);
    nextDqdpDerivPtrVector[ip]->putScalar(0.0);
  }

  int iparamSizeX = currDXdpPtrVector.size();
  for (int ip=0;ip<iparamSizeX;++ip)
  {
    *(oldeDXdpPtrVector[ip]) = *(currDXdpPtrVector[ip]);
    *(lastDXdpPtrVector[ip]) = *(currDXdpPtrVector[ip]);
    currDXdpPtrVector[ip]->putScalar(0.0);
    nextDXdpPtrVector[ip]->putScalar(0.0);
  }

  int iparamSizeQX = currDQdxDXdpPtrVector.size();
  for (int ip=0;ip<iparamSizeQX;++ip)
  {
    *(oldeDQdxDXdpPtrVector[ip]) = *(currDQdxDXdpPtrVector[ip]);
    *(lastDQdxDXdpPtrVector[ip]) = *(currDQdxDXdpPtrVector[ip]);
    currDQdxDXdpPtrVector[ip]->putScalar(0.0);
    nextDQdxDXdpPtrVector[ip]->putScalar(0.0);
  }


  int iparamSizeQXd = currDQdxDXdpDerivPtrVector.size();
  for (int ip=0;ip<iparamSizeQX;++ip)
  {
    *(oldeDQdxDXdpDerivPtrVector[ip]) = *(currDQdxDXdpDerivPtrVector[ip]);
    *(lastDQdxDXdpDerivPtrVector[ip]) = *(currDQdxDXdpDerivPtrVector[ip]);
    currDQdxDXdpDerivPtrVector[ip]->putScalar(0.0);
    nextDQdxDXdpDerivPtrVector[ip]->putScalar(0.0);
  }

}

//-----------------------------------------------------------------------------
// Function      : DataStore::outputPredictedSolution
// Purpose       : Set ErrorEstimate array values to zero.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/27/00
//-----------------------------------------------------------------------------
void DataStore::outputPredictedSolution(std::ostream &os)
{
  os << Xyce::subsection_divider << std::endl
     << "  Predicted Solution:" << std::endl;

  for (unsigned int k = 0; k< solutionSize; ++k)
    os << (*(nextSolutionPtr))[k] << std::endl;

  os << Xyce::subsection_divider << std::endl;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::outputPredictedDerivative
// Purpose       : Set ErrorEstimate array values to zero.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/27/00
//-----------------------------------------------------------------------------
void DataStore::outputPredictedDerivative(std::ostream &os)
{
  os << Xyce::subsection_divider << std::endl
     << "  Predicted Derivative:" << std::endl;

  for (unsigned int k = 0; k < solutionSize; ++k)
    os << (*(nextSolutionDerivPtr))[k] << std::endl;

  os << Xyce::subsection_divider << std::endl;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::partialErrorNormSum
// Purpose       : Needed by 2-level solves.
//
// Special Notes : A weighted RMS norm is this:
//
//                  norm = sqrt ( 1/n * sum (x/w)^2 )
//
//                  What this function returns is: sum (x/w)^2
//
//                  It will later be summed with other patial sums
//                  to get the complete WRMS value.
//
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 03/15/06
//-----------------------------------------------------------------------------
double DataStore::partialErrorNormSum ()
{
  double errorNorm = 0.0;
  newtonCorrectionPtr->wRMSNorm(*errWtVecPtr, &errorNorm);
  double sum = errorNorm*errorNorm;
  double length = newtonCorrectionPtr->globalLength();

  sum *= length;

  return sum;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::globalLength
// Purpose       : Needed by 2-level solves.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 03/15/06
//-----------------------------------------------------------------------------
double DataStore::globalLength ()
{
  return newtonCorrectionPtr->globalLength();
}

//-----------------------------------------------------------------------------
// Function      : DataStore::computeDividedDifferences
// Purpose       : Compute the scaled (by current stepsize) divided difference
//                 approximation to the derivative.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/27/00
//-----------------------------------------------------------------------------
void DataStore::computeDividedDifferences()
{
  // First do the solution divided difference:
  nextSolutionDivDiffPtr->linearCombo(1.0, *(nextSolutionPtr), -1.0, *(currSolutionPtr));

  // Now do the state divided difference:
  nextStateDivDiffPtr->linearCombo(1.0, *(nextStatePtr), -1.0, *(currStatePtr));

  return;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::computeDivDiffsBlock
// Purpose       : Compute the scaled (by current stepsize) divided difference
//                 approximation to the derivative.  This is the same as
//                 the  function "computeDividedDifferences", except that
//                 the  operation is only performed on a sub-block of the
//                 vectors.
//
// Special Notes : Not done yet...
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 01/10/01
//-----------------------------------------------------------------------------
void DataStore::computeDivDiffsBlock (
         const std::list<IndexPair> & solGIDList,
         const std::list<IndexPair> & staGIDList)
{}

//-----------------------------------------------------------------------------
// Function      : DataStore::equateTmpVectors
// Purpose       : This function equates the 6 temporary vectors  with
//                 their "next" vector equivalents.  This function
//                 is neccessary for the nonlinear solver damping loop.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 01/24/02
//-----------------------------------------------------------------------------
bool DataStore::equateTmpVectors()
{
  // next solution vector:
  *tmpSolVectorPtr = *nextSolutionPtr;

  // next state vector:
  *tmpStaVectorPtr = *nextStatePtr;

  // next store vector:
  *tmpStoVectorPtr = *nextStorePtr;

  // next state divided difference vector:
  *tmpStaDivDiffPtr = *nextStateDivDiffPtr;

  // next state derivative vector:
  *tmpStaDerivPtr  = *nextStateDerivPtr;
  
  // next lead currrent and junction voltage vectors
  *tmpLeadCurrentVectorPtr = *nextLeadCurrentPtr;
  *tmpLeadDeltaVPtr = *nextLeadDeltaVPtr;

  return true;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::usePreviousSolAsPredictor
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 6/28/01
//-----------------------------------------------------------------------------
bool DataStore::usePreviousSolAsPredictor ()
{
  bool bsuccess = true;

  *nextSolutionPtr = *currSolutionPtr;
  *nextStatePtr    = *currStatePtr;
  *nextStorePtr    = *currStorePtr;

  return bsuccess;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::setNextSolVectorPtr
// Purpose       :
// Special Notes : Only needed for NOX, and other solvers that prefer to
//                 own the solution vector.
//
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 4/22/03
//-----------------------------------------------------------------------------
bool DataStore::setNextSolVectorPtr (Linear::Vector * solVecPtr)
{
  // only save the old pointer if it hasn't been switched yet.
  if (!nextSolPtrSwitched_)
  {
    savedNextSolutionPtr = nextSolutionPtr;
    nextSolPtrSwitched_ = true;
  }
  nextSolutionPtr = solVecPtr;

  return true;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::unsetNextSolVectorPtr
// Purpose       :
// Special Notes : This also copies over the solution, in addition to the
//                 pointer.
//
//                 This is only called when it is time to rotate the
//                 pointers for the next time step.  If we have been
//                 running with NOX, or with some other solver that prefers
//                 to own the solution vector, this is neccessary, and the
//                 switch flag should be set.  Otherwise, not.
//
//                 Basically, if we've been running with NOX, then the next
//                 solution vector ptr has probably been switched out at least
//                 once.  We need to maintain the history, so we make a
//                 copy of this switched solution vector, and the
//                 restore the old pointer.
//
//                 This is kludgy, but will have to do for now.
//
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 4/22/03
//-----------------------------------------------------------------------------
bool DataStore::unsetNextSolVectorPtr ()
{
  if (nextSolPtrSwitched_)
  {
    *savedNextSolutionPtr = *nextSolutionPtr;
    nextSolutionPtr = savedNextSolutionPtr;
    nextSolPtrSwitched_ = false;
  }

  return true;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::setZeroHistory
// Purpose       : Sets everything to zero.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL
// Creation Date : 1/24/07
//-----------------------------------------------------------------------------
void DataStore::setZeroHistory()
{
  if (DEBUG_TIME && isActive(Diag::TIME_HISTORY))
    Xyce::dout() << "\nDataStore::setZeroHistory" << std::endl;

  // Solutions:
  nextSolutionPtr->putScalar(0.0);
  nextSolutionDerivPtr->putScalar(0.0);
  nextSolutionDivDiffPtr->putScalar(0.0);

  // States:
  nextStatePtr->putScalar(0.0);
  nextStateDerivPtr->putScalar(0.0);
  nextStateDivDiffPtr->putScalar(0.0);
  nextStorePtr->putScalar(0.0);
  nextStoreLeadCurrQPtr->putScalar(0.0);
  nextLeadCurrentPtr->putScalar(0.0);
  nextLeadCurrentQPtr->putScalar(0.0);
  nextLeadDeltaVPtr->putScalar(0.0);

  for (int ip=0;ip<nextDfdpPtrVector.size();++ip)
  {
    nextDfdpPtrVector[ip]->putScalar(0.0);
    nextDqdpPtrVector[ip]->putScalar(0.0);
    nextDqdpDerivPtrVector[ip]->putScalar(0.0);
    nextDbdpPtrVector[ip]->putScalar(0.0);
    nextDXdpPtrVector[ip]->putScalar(0.0);
    nextDQdxDXdpPtrVector[ip]->putScalar(0.0);
    nextDQdxDXdpDerivPtrVector[ip]->putScalar(0.0);
  }

  qErrWtVecPtr->putScalar(0.0);

  // DAE formulation vectors
  daeQVectorPtr->putScalar(0.0);
  daeFVectorPtr->putScalar(0.0);
  daeBVectorPtr->putScalar(0.0);

  // Predictors
  xn0Ptr->putScalar(0.0);
  qn0Ptr->putScalar(0.0);
  qpn0Ptr->putScalar(0.0);
  sn0Ptr->putScalar(0.0);
  spn0Ptr->putScalar(0.0);
  stoQn0Ptr->putScalar(0.0);
  stoQpn0Ptr->putScalar(0.0);
  leadCurrentQn0Ptr->putScalar(0.0);
  leadCurrentQpn0Ptr->putScalar(0.0);
  leadDeltaVn0Ptr->putScalar(0.0);
  leadDeltaVpn0Ptr->putScalar(0.0);

  // Nonlinear solution vector:
  qNewtonCorrectionPtr->putScalar(0.0);
  sNewtonCorrectionPtr->putScalar(0.0);
  stoNewtonCorrectionPtr->putScalar(0.0);
  stoLeadCurrQNewtonCorrectionPtr->putScalar(0.0);
  leadCurrentNewtonCorrectionPtr->putScalar(0.0);
  leadCurrentQNewtonCorrectionPtr->putScalar(0.0);
  leadDeltaVNewtonCorrectionPtr->putScalar(0.0);

  // Step-size selection temporary vectors
  delta_x->putScalar(0.0);
  delta_q->putScalar(0.0);

  // Temporary vector for WaMPDE interpolation
  tmpXn0APtr->putScalar(0.0);
  tmpXn0BPtr->putScalar(0.0);

  // This just sets the "oldDAE" history vectors to zero.
  setConstantHistory ();

  // new-DAE history:
  int sizeOfHistory = xHistory.size();
  for (int i = 0; i < sizeOfHistory; ++i)
  {
    xHistory[i]->putScalar(0.0);
    qHistory[i]->putScalar(0.0);
    sHistory[i]->putScalar(0.0);
    stoHistory[i]->putScalar(0.0);
    stoLeadCurrQHistory[i]->putScalar(0.0);
    leadCurrentHistory[i]->putScalar(0.0);
    leadCurrentQHistory[i]->putScalar(0.0);
    leadDeltaVHistory[i]->putScalar(0.0);


    for (int ip=0;ip<dfdpHistory.size();++ip)
    {
      std::vector<Linear::Vector *> dfdp = dfdpHistory[ip];
      std::vector<Linear::Vector *> dqdp = dqdpHistory[ip];
      std::vector<Linear::Vector *> dbdp = dbdpHistory[ip];
      std::vector<Linear::Vector *> dXdp = dXdpHistory[ip];
      std::vector<Linear::Vector *> dQdxdXdp = dQdxdXdpHistory[ip];

      dfdp[i]->putScalar(0.0);
      dqdp[i]->putScalar(0.0);
      dbdp[i]->putScalar(0.0);
      dXdp[i]->putScalar(0.0);
      dQdxdXdp[i]->putScalar(0.0);
    }
  }
}

//-----------------------------------------------------------------------------
// Function      : DataStore::setErrorWtVector
// Purpose       : Set the Error Weight Vector (defined in terms of the
//                 solution approximation and error tolerances).
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL,Parallel Computational Sciences
// Creation Date : 5/25/05
//-----------------------------------------------------------------------------
void
DataStore::setErrorWtVector(
  const TIAParams &             tia_params)
{
  // presort the variables into types
  if (indexVVars.empty() && indexMaskedVars.empty())
  {
    // now fill the index arrays
    for (int k = 0; k < solutionSize; ++k)
    {
      if ((*deviceErrorWeightMask_)[k] == 0.0)
        indexMaskedVars.push_back(k);
      else
        indexVVars.push_back(k);
    }
  }

  if (DEBUG_TIME && isActive(Diag::TIME_ERROR))
  {
    Xyce::dout() << Xyce::section_divider << std::endl
                 << "DataStore::setErrorWtVector" << std::endl << std::endl
                 << "   errorWtVector    currSolution     relErrorTol    absErrorTol" << std::endl
                 << "   --------------  --------------  --------------  --------------" << std::endl;
  }

  if (tia_params.newLte)
  {
    double currMaxValue = 0.0;
    currSolutionPtr->infNorm(&currMaxValue);

    errWtVecPtr->putScalar(currMaxValue);

    if (DEBUG_TIME && isActive(Diag::TIME_ERROR))
    {
      std::vector<int> index(1, -1);
      currSolutionPtr->infNormIndex( &index[0] );
      Xyce::dout() << "currMaxValue = " << currMaxValue << ", currMaxValueIndex = " << index[0] << std::endl;
    }
  }
  else
  {
    errWtVecPtr->absValue(*currSolutionPtr);

    if (DEBUG_TIME && isActive(Diag::TIME_ERROR))
    {
      double currMaxValue = 0.0;
      currSolutionPtr->infNorm(&currMaxValue);
      std::vector<int> index(1, -1);
      currSolutionPtr->infNormIndex( &index[0] );
      Xyce::dout() << "currMaxValueoldLte = " << currMaxValue << ", currMaxValueIndex = " << index[0]  << std::endl;
    }
  }

  qErrWtVecPtr->absValue(*daeQVectorPtr);

  // Voltage variables
  for (std::vector<int>::const_iterator it = indexVVars.begin(), end = indexVVars.end(); it != end; ++it)
  {
    int i = *it;
    (*errWtVecPtr)[i] = (*relErrTolPtr)[i] * (*errWtVecPtr)[i] + (*absErrTolPtr)[i];
    (*qErrWtVecPtr)[i] = (*relErrTolPtr)[i] * (*qErrWtVecPtr)[i] + (*absErrTolPtr)[i];
  }

  // Current variables
  // if !fastTests, then I vars are treated with V vars above.

  // Masked variables
  for (std::vector<int>::const_iterator it = indexMaskedVars.begin(), end = indexMaskedVars.end(); it != end; ++it)
  {
    int i = *it;
    (*errWtVecPtr)[i] = (*qErrWtVecPtr)[i] = Util::MachineDependentParams::MachineBig();
  }


  if (DEBUG_TIME && isActive(Diag::TIME_ERROR))
  {
    for (int k = 0; k < solutionSize; ++k)
      Xyce::dout() << (*errWtVecPtr)[k] << " "
                   << (*currSolutionPtr)[k] << " "
                   << (*relErrTolPtr)[k] << " "
                   << (*absErrTolPtr)[k] << " " << std::endl;

    Xyce::dout() << ""  << std::endl
                 << Xyce::section_divider << std::endl;
  }
}

//-----------------------------------------------------------------------------
// Function      : DataStore::WRMS_errorNorm
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 5/25/05
//-----------------------------------------------------------------------------
double DataStore::WRMS_errorNorm()
{
  double errorNorm = 0.0, qErrorNorm = 0.0;
  newtonCorrectionPtr->wRMSNorm(*errWtVecPtr, &errorNorm);
  qNewtonCorrectionPtr->wRMSNorm(*qErrWtVecPtr, &qErrorNorm);

  if (DEBUG_TIME && isActive(Diag::TIME_ERROR))
  {
    Xyce::dout() << "DataStore::errorNorm = " << errorNorm << std::endl;
    Xyce::dout() << "DataStore::qErrorNorm = " << qErrorNorm << std::endl;
  }

  // This is for handling the total errorNorm for 2-level solves.
  //
  // Note:  This version of the function assumes that the error norm
  // and q-error norm are the same size.  (they have to be...)
  if ( !(innerErrorInfoVec.empty()) )
  {
    double upperSize = newtonCorrectionPtr->globalLength();
    int sumSize = innerErrorInfoVec.size();

    double totalSize = upperSize;
    double totalSum = errorNorm*errorNorm*upperSize;
    double totalQSum = qErrorNorm*qErrorNorm*upperSize;

    for (int i=0;i<sumSize;++i)
    {
      double innerSum = innerErrorInfoVec[i].xErrorSum;
      double innerQSum = innerErrorInfoVec[i].qErrorSum;
      double innerSize = innerErrorInfoVec[i].innerSize;

      if (DEBUG_TIME && isActive(Diag::TIME_ERROR))
      {
        Xyce::dout() << "DSdae:innerSum["<<i<<"] = " << innerSum <<std::endl;
        Xyce::dout() << "DSdae:innerQSum["<<i<<"] = " << innerQSum <<std::endl;
        Xyce::dout() << "DSdae:innerSize["<<i<<"] = " << innerSize <<std::endl;
      }

      totalSize += innerSize;
      totalSum += innerSum;
      totalQSum += innerQSum;
    }

    double recip = 1.0/totalSize;
    errorNorm = sqrt(recip*totalSum);
    qErrorNorm = sqrt(recip*totalQSum);

    if (DEBUG_TIME && isActive(Diag::TIME_ERROR))
    {
      Xyce::dout() << "DSdae:upperSize = " << upperSize << std::endl;
      Xyce::dout() << "DSdae:totalSum = " << totalSum << std::endl;
      Xyce::dout() << "DSdae:totalQSum = " << totalQSum << std::endl;
      Xyce::dout() << "DSdae:totalSize = " << totalSize << std::endl;
      Xyce::dout() << "DSdae:2-level errorNorm = " << errorNorm << std::endl;
      Xyce::dout() << "DSdae:2-level qErrorNorm = " << qErrorNorm << std::endl;
    }
  }

#ifndef Xyce_USE_Q_NORM
  //errorNorm = errorNorm;
#else
  errorNorm = sqrt(0.5*errorNorm*errorNorm+0.5*qErrorNorm*qErrorNorm);
#endif
  return errorNorm;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::partialQErrorNormSum
// Purpose       : Needed by 2-level solves.  This is the Q-vector version
//                 of this function.
//
// Special Notes : A weighted RMS norm is this:
//
//                  norm = sqrt ( 1/n * sum (x/w)^2 )
//
//                  What this function returns is: sum (x/w)^2
//
//                  It will later be summed with other patial sums
//                  to get the complete WRMS value.
//
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 03/15/06
//-----------------------------------------------------------------------------
double DataStore::partialQErrorNormSum ()
{
  double qErrorNorm = 0.0;
  qNewtonCorrectionPtr->wRMSNorm(*qErrWtVecPtr, &qErrorNorm);
  double sum = qErrorNorm*qErrorNorm;
  double length = qNewtonCorrectionPtr->globalLength();
  sum *= length;

  return sum;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::partialSum_m1
// Purpose       : Needed by 2-level solves.
//
// Special Notes : A weighted RMS norm is this:
//
//                  norm = sqrt ( 1/n * sum (x/w)^2 )
//
//                  What this function returns is: sum (x/w)^2
//
//                  It will later be summed with other patial sums
//                  to get the complete WRMS value.
//
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 03/15/07
//-----------------------------------------------------------------------------
double DataStore::partialSum_m1 (int currentOrder)
{
  double sum = 0.0;

  if (currentOrder>1)
  {
    delta_x->linearCombo(1.0,*(xHistory[currentOrder]),1.0,*newtonCorrectionPtr);
    double norm = 0.0;
    delta_x->wRMSNorm(*errWtVecPtr, &norm);
    sum = norm*norm;
    double length = newtonCorrectionPtr->globalLength();
    sum *= length;
  }

  return sum;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::partialSum_m2
// Purpose       : Needed by 2-level solves.
//
// Special Notes : A weighted RMS norm is this:
//
//                  norm = sqrt ( 1/n * sum (x/w)^2 )
//
//                  What this function returns is: sum (x/w)^2
//
//                  It will later be summed with other patial sums
//                  to get the complete WRMS value.
//
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 03/15/07
//-----------------------------------------------------------------------------
double DataStore::partialSum_m2 (int currentOrder)
{
  double sum = 0.0;

  if (currentOrder>2)
  {
    delta_x->linearCombo(1.0,*(xHistory[currentOrder]),1.0,*newtonCorrectionPtr);
    delta_x->linearCombo(1.0,*(xHistory[currentOrder-1]),1.0,*delta_x);
    double norm = 0.0;
    delta_x->wRMSNorm(*errWtVecPtr, &norm);
    sum = norm*norm;
    double length = newtonCorrectionPtr->globalLength();
    sum *= length;
  }

  return sum;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::partialSum_p1
// Purpose       : Needed by 2-level solves.
//
// Special Notes : A weighted RMS norm is this:
//
//                  norm = sqrt ( 1/n * sum (x/w)^2 )
//
//                  What this function returns is: sum (x/w)^2
//
//                  It will later be summed with other patial sums
//                  to get the complete WRMS value.
//
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 03/15/07
//-----------------------------------------------------------------------------
double DataStore::partialSum_p1 (int currentOrder, int maxOrder)
{
  double sum = 0.0;

  if (currentOrder<maxOrder)
  {
    delta_x->linearCombo(1.0,*newtonCorrectionPtr,-1.0,*(xHistory[currentOrder+1]));
    double norm = 0.0;
    delta_x->wRMSNorm(*errWtVecPtr, &norm);
    sum = norm*norm;
    double length = newtonCorrectionPtr->globalLength();
    sum *= length;
  }

  return sum;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::partialSum_q1
// Purpose       : Needed by 2-level solves.
//
// Special Notes : A weighted RMS norm is this:
//
//                  norm = sqrt ( 1/n * sum (x/w)^2 )
//
//                  What this function returns is: sum (x/w)^2
//
//                  It will later be summed with other patial sums
//                  to get the complete WRMS value.
//
// Scope         : public
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 03/15/07
//-----------------------------------------------------------------------------
double DataStore::partialSum_q1 ()
{
  double sum = 0.0;

  double norm = 0.0;
  (qHistory[1])->wRMSNorm(*qErrWtVecPtr, &norm);

  sum = norm*norm;
  double length = qNewtonCorrectionPtr->globalLength();
  sum *= length;

  return sum;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::delta_x_errorNorm_m1
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 03/15/07
//-----------------------------------------------------------------------------
double DataStore::delta_x_errorNorm_m1()
{
  double errorNorm = 0.0;
  delta_x->wRMSNorm(*errWtVecPtr, &errorNorm);

  // This is for handling the total errorNorm for 2-level solves.
  // It really should get a separate overloaded function.
  if ( !(innerErrorInfoVec.empty()) )
  {
    double upperSize = newtonCorrectionPtr->globalLength();
    int sumSize = innerErrorInfoVec.size();

    double totalSize = upperSize;
    double totalSum = errorNorm*errorNorm*upperSize;

    for (int i=0;i<sumSize;++i)
    {
      double innerSum = innerErrorInfoVec[i].xErrorSum_m1;
      double innerSize = innerErrorInfoVec[i].innerSize;

      totalSize += innerSize;
      totalSum += innerSum;
    }

    double recip = 1.0/totalSize;
    errorNorm = sqrt(recip*totalSum);
  }

  return errorNorm;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::delta_x_errorNorm_m2
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 03/15/07
//-----------------------------------------------------------------------------
double DataStore::delta_x_errorNorm_m2()
{
  double errorNorm = 0.0;
  delta_x->wRMSNorm(*errWtVecPtr, &errorNorm);

  // This is for handling the total errorNorm for 2-level solves.
  // It really should get a separate overloaded function.
  if ( !(innerErrorInfoVec.empty()) )
  {
    double upperSize = newtonCorrectionPtr->globalLength();
    int sumSize = innerErrorInfoVec.size();

    double totalSize = upperSize;
    double totalSum = errorNorm*errorNorm*upperSize;

    for (int i=0;i<sumSize;++i)
    {
      double innerSum = innerErrorInfoVec[i].xErrorSum_m2;
      double innerSize = innerErrorInfoVec[i].innerSize;

      totalSize += innerSize;
      totalSum += innerSum;
    }

    double recip = 1.0/totalSize;
    errorNorm = sqrt(recip*totalSum);
  }

  return errorNorm;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::delta_x_errorNorm_p1
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 03/15/07
//-----------------------------------------------------------------------------
double DataStore::delta_x_errorNorm_p1()
{
  double errorNorm = 0.0;
  delta_x->wRMSNorm(*errWtVecPtr, &errorNorm);

  // This is for handling the total errorNorm for 2-level solves.
  // It really should get a separate overloaded function.
  if ( !(innerErrorInfoVec.empty()) )
  {
    double upperSize = newtonCorrectionPtr->globalLength();
    int sumSize = innerErrorInfoVec.size();

    double totalSize = upperSize;
    double totalSum = errorNorm*errorNorm*upperSize;

    for (int i=0;i<sumSize;++i)
    {
      double innerSum = innerErrorInfoVec[i].xErrorSum_p1;
      double innerSize = innerErrorInfoVec[i].innerSize;

      totalSize += innerSize;
      totalSum += innerSum;
    }

    double recip = 1.0/totalSize;
    errorNorm = sqrt(recip*totalSum);
  }

  return errorNorm;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::delta_x_errorNorm_q1
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 03/15/07
//-----------------------------------------------------------------------------
double DataStore::delta_x_errorNorm_q1()
{
  double errorNorm = 0.0;
  (qHistory[1])->wRMSNorm(*qErrWtVecPtr, &errorNorm);

  // This is for handling the total errorNorm for 2-level solves.
  // It really should get a separate overloaded function.
  if ( !(innerErrorInfoVec.empty()) )
  {
    double upperSize = newtonCorrectionPtr->globalLength();
    int sumSize = innerErrorInfoVec.size();

    double totalSize = upperSize;
    double totalSum = errorNorm*errorNorm*upperSize;

    for (int i=0;i<sumSize;++i)
    {
      double innerSum = innerErrorInfoVec[i].q1HistorySum;
      double innerSize = innerErrorInfoVec[i].innerSize;

      totalSize += innerSize;
      totalSum += innerSum;
    }

    double recip = 1.0/totalSize;
    errorNorm = sqrt(recip*totalSum);
  }

  return errorNorm;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::stepLinearCombo
// Purpose       : setup the newtonCorrection vectors.
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter
// Creation Date : 02/20/07
//-----------------------------------------------------------------------------
void DataStore::stepLinearCombo()
{
  // 03/16/04 tscoffe:  update the newton correction.  Note:  this should be
  // available from NOX, but for now I'm going to do the difference anyway.
  newtonCorrectionPtr->linearCombo (1.0,*nextSolutionPtr,-1.0,*xn0Ptr);

  // We need to compute the correction in Q here
  // I'm assuming dsDaePtr_->daeQVectorPtr will be fresh from the end of the
  // nonlinear solve.
  qNewtonCorrectionPtr->linearCombo (1.0,*daeQVectorPtr,-1.0,*qn0Ptr);

  // We also need a State correction between the time steps
  sNewtonCorrectionPtr->linearCombo (1.0,*nextStatePtr,-1.0,*sn0Ptr);

  // We also need a Store correction between the time steps
  stoNewtonCorrectionPtr->linearCombo (1.0,*nextStorePtr,-1.0,*ston0Ptr);
  stoLeadCurrQNewtonCorrectionPtr->linearCombo (1.0, *nextStoreLeadCurrQPtr, -1.0, *stoQn0Ptr);

  // This is for the lead current vector
  leadCurrentNewtonCorrectionPtr->linearCombo (1.0,*nextLeadCurrentPtr,-1.0,*leadCurrentn0Ptr);
  leadCurrentQNewtonCorrectionPtr->linearCombo (1.0, *nextLeadCurrentQPtr, -1.0, *leadCurrentQn0Ptr );
  leadDeltaVNewtonCorrectionPtr->linearCombo (1.0, *nextLeadDeltaVPtr, -1.0, *leadDeltaVn0Ptr );

  if (DEBUG_TIME && isActive(Diag::TIME_ERROR))
  {
    Xyce::dout() << "\n newtonCorrection: \n" << std::endl;
    newtonCorrectionPtr->printPetraObject(Xyce::dout());
    Xyce::dout() << std::endl;
    Xyce::dout() << "\n qNewtonCorrection: \n" << std::endl;
    qNewtonCorrectionPtr->printPetraObject(Xyce::dout());
    Xyce::dout() << "\n sNewtonCorrection: \n" << std::endl;
    sNewtonCorrectionPtr->printPetraObject(Xyce::dout());
    Xyce::dout() << std::endl;
  }
}

//-----------------------------------------------------------------------------
// Function      : DataStore::getSolnVarData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 7/27/06
//-----------------------------------------------------------------------------
bool
DataStore::getSolnVarData(
  const int &           gid,
  std::vector<double> & varData )
{
  varData.resize(23);
  int i=0;
  varData[i++] = tmpSolVectorPtr->getElementByGlobalIndex( gid );
  varData[i++] = currSolutionPtr->getElementByGlobalIndex( gid );
  varData[i++] = lastSolutionPtr->getElementByGlobalIndex( gid );
  varData[i++] = nextSolutionPtr->getElementByGlobalIndex( gid );
  varData[i++] = currSolutionDerivPtr->getElementByGlobalIndex( gid );
  varData[i++] = lastSolutionDerivPtr->getElementByGlobalIndex( gid );
  varData[i++] = nextSolutionDerivPtr->getElementByGlobalIndex( gid );
  varData[i++] = currSolutionDivDiffPtr->getElementByGlobalIndex( gid );
  varData[i++] = lastSolutionDivDiffPtr->getElementByGlobalIndex( gid );
  varData[i++] = nextSolutionDivDiffPtr->getElementByGlobalIndex( gid );
  varData[i++] = errWtVecPtr->getElementByGlobalIndex( gid );
  varData[i++] = absErrTolPtr->getElementByGlobalIndex( gid );
  varData[i++] = relErrTolPtr->getElementByGlobalIndex( gid );
  varData[i++] = newtonCorrectionPtr->getElementByGlobalIndex( gid );
  varData[i++] = qErrWtVecPtr->getElementByGlobalIndex ( gid );
  varData[i++] = daeQVectorPtr->getElementByGlobalIndex ( gid );
  varData[i++] = daeFVectorPtr->getElementByGlobalIndex ( gid );
  varData[i++] = daeBVectorPtr->getElementByGlobalIndex ( gid );
  varData[i++] = xn0Ptr->getElementByGlobalIndex ( gid );
  varData[i++] = qn0Ptr->getElementByGlobalIndex ( gid );
  varData[i++] = qpn0Ptr->getElementByGlobalIndex ( gid );
  varData[i++] = dFdxdVpVectorPtr->getElementByGlobalIndex ( gid );
  varData[i++] = dQdxdVpVectorPtr->getElementByGlobalIndex ( gid );
  return true;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::getStateVarData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 7/27/06
//-----------------------------------------------------------------------------
bool DataStore::getStateVarData( const int & gid,
			 	       std::vector<double> & varData )
{
  int i=0;
  varData.resize( 14 );
  varData[i++] = tmpStaVectorPtr->getElementByGlobalIndex( gid );
  varData[i++] = tmpStaDerivPtr->getElementByGlobalIndex( gid );
  varData[i++] = tmpStaDivDiffPtr->getElementByGlobalIndex( gid );
  varData[i++] = currStatePtr->getElementByGlobalIndex( gid );
  varData[i++] = lastStatePtr->getElementByGlobalIndex( gid );
  varData[i++] = nextStatePtr->getElementByGlobalIndex( gid );
  varData[i++] = currStateDerivPtr->getElementByGlobalIndex( gid );
  varData[i++] = lastStateDerivPtr->getElementByGlobalIndex( gid );
  varData[i++] = nextStateDerivPtr->getElementByGlobalIndex( gid );
  varData[i++] = currStateDivDiffPtr->getElementByGlobalIndex( gid );
  varData[i++] = lastStateDivDiffPtr->getElementByGlobalIndex( gid );
  varData[i++] = nextStateDivDiffPtr->getElementByGlobalIndex( gid );
  varData[i++] = sn0Ptr->getElementByGlobalIndex( gid );
  varData[i++] = spn0Ptr->getElementByGlobalIndex( gid );
  return true;
}


//-----------------------------------------------------------------------------
// Function      : DataStore::getStoreVarData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date :
//-----------------------------------------------------------------------------
bool DataStore::getStoreVarData( const int & gid,
			 	       std::vector<double> & varData )
{
  int i=0;
  varData.resize( 6 );
  varData[i++] = tmpStoVectorPtr->getElementByGlobalIndex( gid );
  varData[i++] = currStorePtr->getElementByGlobalIndex( gid );
  varData[i++] = lastStorePtr->getElementByGlobalIndex( gid );
  varData[i++] = nextStorePtr->getElementByGlobalIndex( gid );
  varData[i++] = ston0Ptr->getElementByGlobalIndex( gid );
  varData[i++] = stopn0Ptr->getElementByGlobalIndex( gid );
  return true;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::setSolnVarData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 7/27/06
//-----------------------------------------------------------------------------
bool DataStore::setSolnVarData( const int & gid,
				      const std::vector<double> & varData )
{
  int i=0;
  tmpSolVectorPtr->setElementByGlobalIndex       ( gid, varData[i++] );
  currSolutionPtr->setElementByGlobalIndex       ( gid, varData[i++] );
  lastSolutionPtr->setElementByGlobalIndex       ( gid, varData[i++] );
  nextSolutionPtr->setElementByGlobalIndex       ( gid, varData[i++] );
  currSolutionDerivPtr->setElementByGlobalIndex  ( gid, varData[i++] );
  lastSolutionDerivPtr->setElementByGlobalIndex  ( gid, varData[i++] );
  nextSolutionDerivPtr->setElementByGlobalIndex  ( gid, varData[i++] );
  currSolutionDivDiffPtr->setElementByGlobalIndex( gid, varData[i++] );
  lastSolutionDivDiffPtr->setElementByGlobalIndex( gid, varData[i++] );
  nextSolutionDivDiffPtr->setElementByGlobalIndex( gid, varData[i++] );
  errWtVecPtr->setElementByGlobalIndex           ( gid, varData[i++] );
  absErrTolPtr->setElementByGlobalIndex          ( gid, varData[i++] );
  relErrTolPtr->setElementByGlobalIndex          ( gid, varData[i++] );
  newtonCorrectionPtr->setElementByGlobalIndex   ( gid, varData[i++] );
  qErrWtVecPtr->setElementByGlobalIndex          ( gid, varData[i++] );
  daeQVectorPtr->setElementByGlobalIndex         ( gid, varData[i++] );
  daeFVectorPtr->setElementByGlobalIndex         ( gid, varData[i++] );
  daeBVectorPtr->setElementByGlobalIndex         ( gid, varData[i++] );
  xn0Ptr->setElementByGlobalIndex                ( gid, varData[i++] );
  qn0Ptr->setElementByGlobalIndex                ( gid, varData[i++] );
  qpn0Ptr->setElementByGlobalIndex               ( gid, varData[i++] );
  dFdxdVpVectorPtr->setElementByGlobalIndex      ( gid, varData[i++] );
  dQdxdVpVectorPtr->setElementByGlobalIndex      ( gid, varData[i++] );
  return true;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::setStateVarData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 7/27/06
//-----------------------------------------------------------------------------
bool DataStore::setStateVarData( const int & gid,
			 	       const std::vector<double> & varData )
{
  int i=0;
  tmpStaVectorPtr->setElementByGlobalIndex    ( gid, varData[i++] );
  tmpStaDerivPtr->setElementByGlobalIndex     ( gid, varData[i++] );
  tmpStaDivDiffPtr->setElementByGlobalIndex   ( gid, varData[i++] );
  currStatePtr->setElementByGlobalIndex       ( gid, varData[i++] );
  lastStatePtr->setElementByGlobalIndex       ( gid, varData[i++] );
  nextStatePtr->setElementByGlobalIndex       ( gid, varData[i++] );
  currStateDerivPtr->setElementByGlobalIndex  ( gid, varData[i++] );
  lastStateDerivPtr->setElementByGlobalIndex  ( gid, varData[i++] );
  nextStateDerivPtr->setElementByGlobalIndex  ( gid, varData[i++] );
  currStateDivDiffPtr->setElementByGlobalIndex( gid, varData[i++] );
  lastStateDivDiffPtr->setElementByGlobalIndex( gid, varData[i++] );
  nextStateDivDiffPtr->setElementByGlobalIndex( gid, varData[i++] );

  sn0Ptr->setElementByGlobalIndex             ( gid, varData[i++] );
  spn0Ptr->setElementByGlobalIndex            ( gid, varData[i++] );
  return true;
}

//-----------------------------------------------------------------------------
// Function      : DataStore::setStoreVarData
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 7/27/06
//-----------------------------------------------------------------------------
bool DataStore::setStoreVarData( const int & gid,
			 	       const std::vector<double> & varData )
{
  int i=0;
  tmpStoVectorPtr->setElementByGlobalIndex    ( gid, varData[i++] );
  currStorePtr->setElementByGlobalIndex       ( gid, varData[i++] );
  lastStorePtr->setElementByGlobalIndex       ( gid, varData[i++] );
  nextStorePtr->setElementByGlobalIndex       ( gid, varData[i++] );
  ston0Ptr->setElementByGlobalIndex           ( gid, varData[i++] );
  stopn0Ptr->setElementByGlobalIndex          ( gid, varData[i++] );
  return true;
}

} // namespace TimeIntg
} // namespace Xyce
