//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename      : $RCSfile: N_TIA_TIAParams.C,v $
//
// Purpose       : This file implements the class associated with all user
//                 specified parameters which relate to the time integration
//                 algorithms and problem definition.
//
// Special Notes :
//
// Creator       : Buddy Watts
//
// Creation Date : 6/1/00
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.145.2.2 $
//
// Revision Date  : $Date: 2015/04/02 18:20:18 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#include <Xyce_config.h>

#include <iostream>

#include <N_NLS_fwd.h>

#include <N_TIA_TIAParams.h>
#include <N_IO_CmdParse.h>

#include <N_ERH_ErrorMgr.h>
#include <N_UTL_Diagnostic.h>
#include <N_UTL_ExtendedString.h>
#include <N_UTL_FeatureTest.h>

namespace Xyce {
namespace TimeIntg {

int
maxOrder(
  const IO::CmdParse &  command_line)
{
  return command_line.getArgumentIntValue("-maxord", 5);
}


//-----------------------------------------------------------------------------
// Function      : TIAParams::TIAParams
// Purpose       : constructor
// Special Notes :
// Scope         : public
// Creator       : Buddy Watts, SNL
// Creation Date : 6/01/00
//-----------------------------------------------------------------------------

TIAParams::TIAParams()
  : initialOutputTime(0.0),
    initialOutputTimeGiven(false),
    initialTime(0.0),
    finalTime(0.0),
    initialTimeStep(1.0e-10),
    maxTimeStep(1.0e+99),
    maxTimeStepGiven(false),
    minTimeStep(0.0),
    minTimeStepGiven(false),
    constantTimeStepFlag(false),
    useDeviceTimeStepMaxFlag(true),
    newBPStepping(true),
    minTimeStepsBP(10),
    minTimeStepsBPGiven(false),
    newLte(true),
    errorAnalysisOption(LOCAL_TRUNCATED_ESTIMATES),
    NLmin(3),
    NLmax(8),
    delmax(1.0e+99),
    delmaxGiven(false),
    errorAnalysisOptionResetCount(0),
    timestepsReversal(false),
    testFirstStep(false),
    relErrorTol(1.0e-3),
    absErrorTol(1.0e-6),
    relErrorTolGiven(false),
    errTolAcceptance(1.0),
    bpEnable(true),
    restartTimeStepScale(0.005),
    jacLimitFlag(false),
    jacLimit(1.0e+17),
    maxOrder(5),
    minOrder(1),
    interpOutputFlag(true),
    minTimeStepRecoveryCounter(0)
{}

//-----------------------------------------------------------------------------
// Function      : TIAParams::TIAParams
// Purpose       : destructor
// Special Notes :
// Scope         : public
// Creator       : Buddy Watts, SNL
// Creation Date : 6/01/00
//-----------------------------------------------------------------------------
TIAParams::~TIAParams()
{}

//-----------------------------------------------------------------------------
// Function      : TIAParams::printParams
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Eric Keiter, 9233.
// Creation Date : 7/12/01
//-----------------------------------------------------------------------------
void TIAParams::printParams(std::ostream &os, int analysis) const
{
  os << "\n" << std::endl;
  os << Xyce::section_divider << std::endl;
  os << "\n***** Time Integration solver options:\n" << std::endl;

  if (analysis == Xyce::Nonlinear::TRANSIENT)
  {
    os << "\tAnalysis:\t\t\tTRANSIENT" << std::endl
       << "\tInitial Time (sec):\t\t" <<  initialTime << std::endl
       << "\tFinal Time (sec):\t\t" <<  finalTime << std::endl
       << "\tStarting Time Step(sec):\t" <<  initialTimeStep << std::endl
       << "\tRestart Time Step Scale:\t" <<  restartTimeStepScale << std::endl
       << "\tError Analysis option:\t" <<  errorAnalysisOption << std::endl
       << "\tInitial Output Time:\t" <<  initialOutputTime << std::endl
      // << "\tTime Integration method:\t" << integrationMethod << std::endl

       << (constantTimeStepFlag ? "\tUsing Constant Step Size" : "\tUsing Variable Step Size") << std::endl
       << (useDeviceTimeStepMaxFlag ? "\tUsing Device specified maximum stepsize" : "\tNOT using Device specified maximum stepsize") << std::endl;
      // << (nlNearConvFlag ? "\tNL Near Convergence Flag is ON" : "\tNL Near Convergence Flag is OFF") << std::endl
      // << (passNLStall ? "\tNL Pass Non-linear Stalls is ON" : "\tNL Pass Non-linear Stalls is OFF") << std::endl;
  }
  else
  {
    os << "\tAnalysis:\t\t\tDC SWEEP" << std::endl;
  }

  os << "\tabsErrorTol:\t\t\t" <<  absErrorTol  << std::endl
     << "\trelErrorTol:\t\t\t" <<  relErrorTol  << std::endl
     << "\tMaximum Order:\t\t\t" <<  maxOrder << std::endl
     << "\tMinimum Order:\t\t\t" <<  minOrder << std::endl
     << "\tInterpolated Output Flag:\t\t " << (interpOutputFlag ? "true": "false") << std::endl;
  
    // << "\tConductance Test Flag:\t\t" << (condTestFlag  ? "true": "false") << std::endl;

  // if (condTestDeviceNames.empty())
  // {
  //   os << "\tConductance Test Device Name List is:\t\tEMPTY" << std::endl;
  // }
  // else
  // {
  //   os << "\tConductance Test Device Name List contains:  " << std::endl;
  //   for (std::list< std::string >::iterator it = condTestDeviceNames.begin(); it != condTestDeviceNames.end(); ++it)
  //   {
  //     os << " \"" << *it << "\"";
  //   }
  // }
  os << Xyce::section_divider << std::endl;
}

//-----------------------------------------------------------------------------
// Function      : TIAParams::operator=
// Purpose       : "=" operator.
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Computational Sciences
// Creation Date : 12/17/05
//-----------------------------------------------------------------------------
TIAParams::TIAParams(const TIAParams &right)
  : initialOutputTime(right.initialOutputTime),
    initialOutputTimeGiven(right.initialOutputTimeGiven),
    initialTime(right.initialTime),
    finalTime(right.finalTime),
    initialTimeStep(right.initialTimeStep),
    maxTimeStep(right.maxTimeStep),
    maxTimeStepGiven(right.maxTimeStepGiven),
    minTimeStep(right.minTimeStep),
    minTimeStepGiven(right.minTimeStepGiven),
    constantTimeStepFlag(right.constantTimeStepFlag),
    useDeviceTimeStepMaxFlag(right.useDeviceTimeStepMaxFlag),
    errorAnalysisOption(right.errorAnalysisOption),
    errorAnalysisOptionResetCount(right.errorAnalysisOptionResetCount),
    bpEnable(right.bpEnable),
    restartTimeStepScale(right.restartTimeStepScale),
    relErrorTol(right.relErrorTol),
    absErrorTol(right.absErrorTol),
    errTolAcceptance(right.errTolAcceptance),
    jacLimitFlag(right.jacLimitFlag),
    jacLimit(right.jacLimit),
    maxOrder(right.maxOrder),
    minOrder(right.minOrder),
    interpOutputFlag(right.interpOutputFlag)
{}

//-----------------------------------------------------------------------------
// Function      : TIAParams::operator=
// Purpose       : "=" operator.
// Special Notes :
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Computational Sciences
// Creation Date : 12/17/05
//-----------------------------------------------------------------------------
TIAParams & TIAParams::operator=(const TIAParams &right)
{
  if (this != &right)
  {
    initialOutputTime = right.initialOutputTime;
    initialOutputTimeGiven = right.initialOutputTimeGiven;
    initialTime = right.initialTime;
    finalTime = right.finalTime;
    initialTimeStep = right.initialTimeStep;
    maxTimeStep = right.maxTimeStep;
    maxTimeStepGiven = right.maxTimeStepGiven;
    minTimeStep = right.minTimeStep;
    minTimeStepGiven = right.minTimeStepGiven;
    constantTimeStepFlag = right.constantTimeStepFlag;
    useDeviceTimeStepMaxFlag = right.useDeviceTimeStepMaxFlag;
    errorAnalysisOption = right.errorAnalysisOption;
    errorAnalysisOptionResetCount = right.errorAnalysisOptionResetCount;
    bpEnable = right.bpEnable;
    restartTimeStepScale = right.restartTimeStepScale;
    relErrorTol = right.relErrorTol;
    absErrorTol = right.absErrorTol;
    errTolAcceptance = right.errTolAcceptance;
    jacLimitFlag = right.jacLimitFlag;
    jacLimit = right.jacLimit;
    maxOrder = right.maxOrder;
    minOrder = right.minOrder;
    interpOutputFlag = right.interpOutputFlag;
  }

  return *this;
}

void
TIAParams::setMaxOrder(
  int           max_order)
{
  if (max_order < 1)
    max_order = 1;
  else if (max_order > 5)
    max_order = 5;

  maxOrder = max_order;
}

//-----------------------------------------------------------------------------
// Function      : setTimeIntegratorOptions
// Purpose       :
// Special Notes : These are from '.options timeint'
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 04/18/02
//-----------------------------------------------------------------------------
bool
TIAParams::setTimeIntegratorOption(
  const Util::Param &           param)
{
  if (DEBUG_ANALYSIS && param.uTag() == "CONSTSTEP")
  {
    constantTimeStepFlag =  static_cast<bool> (param.getImmutableValue<int>());
  }
  else if (param.uTag() == "USEDEVICEMAX")
  {
    useDeviceTimeStepMaxFlag =  static_cast<bool> (param.getImmutableValue<int>());
  }
  else if (param.uTag() == "RELTOL")
  {
    relErrorTol = param.getImmutableValue<double>();
    relErrorTolGiven = true;
  }
  else if (param.uTag() == "ABSTOL")
  {
    absErrorTol = param.getImmutableValue<double>();
  }
  else if (param.uTag() == "BPENABLE" )
  {
    bpEnable =  static_cast<bool> (param.getImmutableValue<int>());
  }
  else if (param.uTag() == "RESTARTSTEPSCALE" )
  {
    restartTimeStepScale = param.getImmutableValue<double>();
  }
  else if (param.uTag() == "MINTIMESTEPSBP")
  {
    minTimeStepsBP = param.getImmutableValue<int>();
    minTimeStepsBPGiven = true;
  }
  else if (param.uTag() == "ERROPTION" )
  {
    errorAnalysisOption = param.getImmutableValue<int>();
    // tscoffe/tmei 12/7/07:  If error option = 1 (NO LTE) then make sure minTimeStepBP is enabled.
    if (errorAnalysisOption == TimeIntg::NO_LOCAL_TRUNCATED_ESTIMATES)
    {
      if (!minTimeStepsBPGiven)
      {
        minTimeStepsBPGiven = true;
      }
    }
  }
  else if (param.uTag() == "JACLIMITFLAG" )
  {
    jacLimitFlag= static_cast<bool> (param.getImmutableValue<int>());
  }
  else if (param.uTag() == "JACLIMIT" )
  {
    jacLimit = param.getImmutableValue<double>();
  }
  else if (param.uTag() == "MAXORD" )
  {
    maxOrder = param.getImmutableValue<int>();
  }
  else if (param.uTag() == "MINORD" )
  {
    minOrder = param.getImmutableValue<int>();
  }
  else if (param.uTag() == "TIMESTEPSREVERSAL" )
  {
    timestepsReversal  = param.getImmutableValue<bool>();
  }
  else if (param.uTag() == "TESTFIRSTSTEP" )
  {
    testFirstStep = static_cast<bool> (param.getImmutableValue<int>());
  }
  else if (param.uTag() == "DELMAX" )
  {
    delmax  = param.getImmutableValue<double>();
    delmaxGiven = true;
  }
  else if (param.uTag() == "NLMIN" )
  {
    NLmin = param.getImmutableValue<int>();
    if (NLmin > NLmax)
    {
      Report::UserError() << ".options timeint NLMIN = " << NLmin << " > " << NLmax << " = NLMAX!";
    }
  }
  else if (param.uTag() == "NLMAX" )
  {
    NLmax = param.getImmutableValue<int>();
    if (NLmin > NLmax)
    {
      Report::UserError() << ".options timeint NLMIN = " << NLmin << " > " << NLmax << " = NLMAX!";
    }
  }
  else if (param.uTag() == "NEWLTE")
  {
    newLte = param.getImmutableValue<double>();

    if (newLte && !relErrorTolGiven)
      relErrorTol = 1.0e-3;
  }
  else if (param.uTag() == "NEWBPSTEPPING")
  {
    newBPStepping = param.getImmutableValue<double>();
  }
  else if (param.uTag() == "INTERPOUTPUT")
  {
    interpOutputFlag = static_cast<bool> (param.getImmutableValue<int>());
  }
  else if (param.uTag() == "DTMIN")
  {
    minTimeStep = param.getImmutableValue<double>();
    minTimeStepGiven = true;
  }
  else if (param.uTag() == "MINTIMESTEPRECOVERY")
  {
    minTimeStepRecoveryCounter = param.getImmutableValue<int>();
  }
  else
  {
    return false;
  }

  return true;
}


//-----------------------------------------------------------------------------
// Function      : setAnalysisOptions
// Purpose       :
// Special Notes : These are from '.options timeint'
// Scope         : public
// Creator       : Eric R. Keiter, SNL, Parallel Computational Sciences
// Creation Date : 04/18/02
//-----------------------------------------------------------------------------
bool
TIAParams::setAnalysisOption(
  const Util::Param &           param)
{
  if (param.uTag() == "TSTART")
  {
    initialOutputTime = param.getImmutableValue<double>();
    initialOutputTimeGiven = true;
  }
  else if (param.uTag() == "TSTOP")
  {
    finalTime   = param.getImmutableValue<double>();
  }
  else if (param.uTag() == "TSTEP")
  {
    initialTimeStep = param.getImmutableValue<double>();
  }
  else if (param.uTag() == "DTMAX")
  {
    maxTimeStep = param.getImmutableValue<double>();
    maxTimeStepGiven = true;
    if (DEBUG_ANALYSIS && isActive(Diag::TIME_PARAMETERS))
    {
      dout() << "setting maxTimeStep = " << maxTimeStep << std::endl;
    }
  }
  else
  {
    return false;
  }

  return true;
}

} // namespace TimeIntg
} // namespace Xyce
