//-----------------------------------------------------------------------------
// Copyright Notice
//
//   Copyright 2002 Sandia Corporation. Under the terms
//   of Contract DE-AC04-94AL85000 with Sandia Corporation, the U.S.
//   Government retains certain rights in this software.
//
//    Xyce(TM) Parallel Electrical Simulator
//    Copyright (C) 2002-2015 Sandia Corporation
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Filename       : $RCSfile: N_TOP_Misc.h,v $
//
// Purpose        : Misc. typedefs, defines, and structs for Topology
//
// Special Notes  :
//
// Creator        : Rob Hoekstra
//
// Creation Date  : 7/03/01
//
// Revision Information:
// ---------------------
//
// Revision Number: $Revision: 1.18.2.1 $
//
// Revision Date  : $Date: 2015/04/02 18:20:18 $
//
// Current Owner  : $Author: tvrusso $
//-----------------------------------------------------------------------------

#ifndef Xyce_N_TOP_Misc_h
#define Xyce_N_TOP_Misc_h

#include <string>
#include <iostream>
#include <utility>
#if defined(HAVE_UNORDERED_MAP)
#include <unordered_map>
using std::unordered_map;
#elif defined(HAVE_TR1_UNORDERED_MAP)
#include <tr1/unordered_map>
using std::tr1::unordered_map;
#else
#error neither unordered_map or tr1/unordered_map found
#endif

namespace Xyce {

class NodeID : public std::pair< std::string, int >
{
public:
  NodeID()
    : std::pair<std::string, int>()
  {}

  NodeID(const std::string &node, int id)
    : std::pair<std::string, int>( node, id )
  {}
};

inline std::ostream& operator<< (std::ostream &os, const NodeID& n)
{
  return os << "( " << n.first << " , " << n.second << " )";
}

//-----------------------------------------------------------------------------
// Class         : tagged_param
// Purpose       :
// Special Notes :
// Creator       : Eric Keiter, SNL, Parallel Computational Sciences
// Creation Date : 3/16/00
//-----------------------------------------------------------------------------
struct tagged_param
{
  tagged_param()
    : tag(),
      param(0.0),
      given(false)
  {}

  tagged_param(
    const std::string &t,
    const double & p,
    bool g = false )
    : tag(t),
      param(p),
      given(g)
  {}

  tagged_param(const tagged_param & right)
    : tag(right.tag),
      param(right.param),
      given(right.given)
  {}

  tagged_param &operator=( const tagged_param & right )
  {
    tag = right.tag;
    param = right.param;
    given = right.given;

    return *this;
  }

  bool operator<( const tagged_param & rhs ) const
  {
    return tag < rhs.tag;
  }

  std::string tag;
  double      param;
  bool        given;
};

//-----------------------------------------------------------------------------
// Function      : operator==
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rob Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 3/16/00
//-----------------------------------------------------------------------------
inline bool operator==(const tagged_param & tp1, const tagged_param & tp2)
{
  return tp1.tag == tp2.tag;
}

//-----------------------------------------------------------------------------
// Function      : operator!=
// Purpose       :
// Special Notes :
// Scope         : public
// Creator       : Rob Hoekstra, SNL, Parallel Computational Sciences
// Creation Date : 3/16/00
//-----------------------------------------------------------------------------
inline bool operator!=(const tagged_param & tp1, const tagged_param & tp2)
{
  return !(tp1 == tp2);
}

} // namespace Xyce

namespace std {

template<>
struct equal_to<Xyce::NodeID> : public std::binary_function<Xyce::NodeID, Xyce::NodeID, bool>
{
  bool operator()(const Xyce::NodeID &lhs, const Xyce::NodeID &rhs) const
  {
    equal_to<std::string> x0;
    equal_to<int> x1;

    return x1(lhs.second, rhs.second) && x0(lhs.first, rhs.first);
  }
};

#if defined(HAVE_TR1_UNORDERED_MAP) && !defined(HAVE_UNORDERED_MAP)
namespace tr1 {
#endif

template<>
struct hash<Xyce::NodeID> : public std::unary_function<Xyce::NodeID, size_t>
{
  size_t operator()(const Xyce::NodeID &node_id) const
  {
    hash<std::string> x0;
    hash<int> x1;

    return x0(node_id.first) ^ x1(node_id.second);
  }
};

#if defined(HAVE_TR1_UNORDERED_MAP) && !defined(HAVE_UNORDERED_MAP)
} // namespace tr1
#endif

} // std namespace

#endif
